using System;
using System.IO;			// for StreamWriter I/O to a file
using System.Windows.Forms;	// for RichTextBox

namespace MyNamespace
{
	/// <summary>
	///		Multiple precision floating-point Complex algorithms
	///		in Visual C# .Net
	///
	/// ---------------- *** LEGAL NOTICE: *** ------------------
	///
	/// This program is free software; you can redistribute it and/or modify
	/// it under the terms of the GNU General Public License (GPL) as published
	/// by the Free Software Foundation. cf. the file COPYING.txt.
	///
	/// ------------ *** end of legal notice *** ------------------
	/// </summary>

	// ------- A multiple-precision signed floating complex number
	public class MultiC : MultiF	// real part
	{
		public MultiF I;			// imaginary part

		public MultiC()				// initialize with M = 2
			: base()				// calls MultiSI()
		{
			C = 0;	// constructor logic
			this.I = new MultiF();
		}

		public MultiC(int nMaxI)		// initialize with M = nMaxI
			: base(nMaxI)				// calls MultiF(i)
		{
			this.I = new MultiF(nMaxI);
		}

		public MultiC(int nMaxIr, int nMaxIi)	// initialize with M = nMaxIr, I.M = nMaxIi
			: base(nMaxIr)						// calls MultiF(i)
		{
			this.I = new MultiF(nMaxIi);
		}

		//		public static bool operator = (MultiC left, MultiC right)	// not allowed to overload =
//		{
//			MultiCD.SetToC(right, left);
//			return true;
//		}

		public static MultiC operator & (MultiC left, MultiC right)	// use &= for = assignment
		{
			MultiCD.SetToC(right, left);
			return left;
		}

		public static MultiC operator & (MultiC left, MultiF right)	// use &= for = assignment
		{
			MultiCD.SetToFC(right, left); // clears left.I
			return left;
		}

		public static MultiC operator & (MultiC left, MultiSI right)	// use &= for = assignment
		{
			MultiCD.SetToSIC(right, left);
			return left;
		}

		public static MultiC operator & (MultiC left, int right)	// use &= for = assignment
		{
			MultiCD.SetTo1C(right, left);
			return left;
		}

		public static MultiC operator & (MultiC left, long right)	// use &= for = assignment
		{
			MultiCD.SetToExC(right, left);
			return left;
		}

		public static MultiC operator & (MultiC left, double right)	// use &= for = assignment
		{
			MultiCD.SetToDbC(right, left);
			return left;
		}

		public static MultiC operator & (MultiC left, decimal right)	// use &= for = assignment
		{
			MultiCD.SetToDeC(right, left);
			return left;
		}

		public static MultiC operator + (MultiC left, MultiC right) // can be used for a += b
		{
			MultiC z = new MultiC(Calc.fMC);
			MultiCD.AddC(left, right, z);
			return z;
		}

		public static MultiC operator + (MultiC left, int right) // can be used for a += b
		{
			MultiC z = new MultiC(Calc.fMC);
			MultiCD.Add1C(left, right, z);
			return z;
		}

		public static MultiC operator + (int left, MultiC right) // can be used for a += b
		{
			MultiC z = new MultiC(Calc.fMC);
			MultiCD.Add1C(right, left, z);
			return z;
		}

		public static MultiC operator - (MultiC left, MultiC right) // can be used for a -= b
		{
			MultiC z = new MultiC(Calc.fMC);
			MultiCD.SubC(left, right, z);
			return z;
		}

		public static MultiC operator - (MultiC left, int right) // can be used for a -= b
		{
			MultiC z = new MultiC(Calc.fMC);
			MultiCD.Add1C(left, -right, z);
			return z;
		}

		public static MultiC operator - (int left, MultiC right) // can be used for a -= b
		{
			MultiC z = new MultiC(Calc.fMC);
			MultiCD.Add1C(right, -left, z);
			MultiCD.ChSC(z);
			return z;
		}

		public static MultiC operator - (MultiC x)
		{
			MultiC z = new MultiC(x.N, x.I.N);
			MultiCD.SetToC(x, z);
			MultiCD.ChSC(z);
			return z;
		}
		public static MultiC operator * (MultiC left, MultiC right) // can be used for a *= b
		{
			MultiC z = new MultiC(Calc.fMC);
			MultiCD.MulC(left, right, z);
			return z;
		}

		public static MultiC operator * (MultiC left, MultiF right) // can be used for a *= b
		{
			MultiC z = new MultiC(Calc.fMC);
			MultiCD.MulCF(left, right, z);
			return z;
		}

		public static MultiC operator * (MultiF left, MultiC right) // can be used for a *= b
		{
			MultiC z = new MultiC(Calc.fMC);
			MultiCD.MulCF(right, left, z);
			return z;
		}

		public static MultiC operator * (MultiC left, int right) // can be used for a *= b
		{
			MultiC z = new MultiC(Calc.fMC);
			MultiCD.Mul1C(left, right, z);
			return z;
		}

		public static MultiC operator * (int left, MultiC right)
		{
			MultiC z = new MultiC(Calc.fMC);
			MultiCD.Mul1C(right, left, z);
			return z;
		}

		public static MultiC operator / (MultiC left, MultiC right) // can be used for a /= b
		{
			MultiC q = new MultiC(Calc.fMC);
			MultiCD.DivC(left, right, q);
			return q;
		}

		public static MultiC operator / (MultiC left, MultiF right) // can be used for a /= b
		{
			MultiC q = new MultiC(Calc.fMC);
			MultiCD.DivCF(left, right, q);
			return q;
		}

		public static MultiC operator / (MultiC left, int right) // can be used for a /= b
		{
			MultiC q = new MultiC(Calc.fMC);
			MultiCD.Div1C(left, right, q);
			return q;
		}

		public static MultiC operator / (int left, MultiC right) // can be used for a /= b
		{
			MultiC q = new MultiC(Calc.fMC);
			MultiCD.Div1CC(left, right, q);
			return q;
		}

		// % and ^ will default to Singed Integer versions

		public static bool operator == (int d1, MultiC z)
		{
			MultiC dF = new MultiC(MultiID.sdInInt, 1);
			MultiCD.SetTo1C(d1, dF);
			return (MultiCD.CompCC(dF, z) == 0);
		}

		public static bool operator != (int d1, MultiC z)
		{
			MultiC dF = new MultiC(MultiID.sdInInt, 1);
			MultiCD.SetTo1C(d1, dF);
			return (MultiCD.CompCC(dF, z) != 0);
		}

		public static bool operator == (MultiC z, int d1)
		{
			MultiC dF = new MultiC(MultiID.sdInInt, 1);
			MultiCD.SetTo1C(d1, dF);
			return (MultiCD.CompCC(dF, z) == 0);
		}

		public static bool operator != (MultiC z, int d1)
		{
			MultiC dF = new MultiC(MultiID.sdInInt, 1);
			MultiCD.SetTo1C(d1, dF);
			return (MultiCD.CompCC(dF, z) != 0);
		}

		public static bool operator == (long ex, MultiC z)
		{
			MultiC exF = new MultiC(MultiID.sdInLong, 1);
			MultiCD.SetToExC(ex, exF);
			return (MultiCD.CompCC(exF, z) == 0);
		}

		public static bool operator != (long ex, MultiC z)
		{
			MultiC exF = new MultiC(MultiID.sdInLong, 1);
			MultiCD.SetToExC(ex, exF);
			return (MultiCD.CompCC(exF, z) != 0);
		}

		public static bool operator == (MultiC z, long ex)
		{
			MultiC exF = new MultiC(MultiID.sdInLong, 1);
			MultiCD.SetToExC(ex, exF);
			return (MultiCD.CompCC(exF, z) == 0);
		}

		public static bool operator != (MultiC z, long ex)
		{
			MultiC exF = new MultiC(MultiID.sdInLong, 1);
			MultiCD.SetToExC(ex, exF);
			return (MultiCD.CompCC(exF, z) != 0);
		}

		public static bool operator == (MultiC x, MultiC y)
		{
			return (MultiCD.CompCC(y, x) == 0);
		}

		public static bool operator != (MultiC x, MultiC y)
		{
			return (MultiCD.CompCC(y, x) != 0);
		}

		public static bool operator < (MultiC x, MultiC y)
		{
			return (MultiCD.CompCC(y, x) < 0);
		}

		public static bool operator > (MultiC x, MultiC y)
		{
			return (MultiCD.CompCC(y, x) > 0);
		}

		public static bool operator <= (MultiC x, MultiC y)
		{
			return (MultiCD.CompCC(y, x) <= 0);
		}

		public static bool operator >= (MultiC x, MultiC y)
		{
			return (MultiCD.CompCC(y, x) >= 0);
		}

		public static MultiC operator ++ (MultiC x)
		{
			MultiFD.RAddF(MultiFD.fOne, x);
			return x;
		}

		public static MultiC operator -- (MultiC x)
		{
			MultiFD.RAddF(MultiFD.fMOne, x);
			return x;
		}

		public override bool Equals(object o)	// dummy
		{
			return true;
		}

		public override int GetHashCode()		// dummy
		{
			return 0;
		}
	}

	public class MultiCD
	{
		public MultiCD()	// constructor logic
		// initialize Multiple-precision floating decimal package
		// input: MultiID.muDMax, MultiID.muNMax, MultiID.baseM
		{
			int i;
			ci = new MultiC(1);
			SetTo1C(0, 1, ci);
			cZero = new MultiC(1);
			cOne = new MultiC(1);
			SetTo1C(1, 0, cOne);
			cTwo = new MultiC(1);
			SetTo1C(2, 0, cTwo);
			cFour = new MultiC(1);
			SetTo1C(4, 0, cFour);
			cSix = new MultiC(1);
			SetTo1C(6, 0, cSix);
			cMOne = new MultiC(1);
			SetTo1C(-1, 0, cMOne);
			cHalf = new MultiC(1);
			MultiFD.ValueF("0.5", out i, cHalf);	// cHalf = 0.5 + 0.0 * i
			cMHalf = new MultiC(1);
			MultiFD.ValueF("-0.5", out i, cMHalf);	// cMHalf = -0.5 + 0.0 * i
		}

		public static void DoneC(MultiC z) // free up space
		{
			MultiID.DoneSI((MultiF)z);
			MultiID.DoneSI(z.I);
			z.C = z.I.C = 0;
		}

		public static void ReSizeC(int m, MultiC z) // change M and set to zero
		{
			MultiID.ReSizeSI(m, (MultiF)z);
			MultiID.ReSizeSI(m, z.I);
			z.C = z.I.C = 0;
		}

		public static void ClearC(MultiC z)	// set Value = 0
		{
			MultiID.ClearSI((MultiF)z);
			MultiID.ClearSI(z.I);
			z.C = z.I.C = 0;
		}

		public static void ChSC(MultiC z) // change sign, z = - z
		{
			MultiFD.ChSF((MultiF)z);
			MultiFD.ChSF(z.I);
		}

		public static void ChSCN(int zI) // change sign, n = - n
		{
			ChSC(Common.nList[zI]);
		}

		public static void SetToC(MultiC x, MultiC z)	// z = x
		// can reduce precision and round
		{
			MultiFD.SetToF(x, (MultiF)z);
			MultiFD.SetToF(x.I, z.I);
		}

		public static void SetToSIC(MultiSI x, MultiC z) // z = x
		// can reduce precision and round
		{
			MultiFD.SetToSIF(x, (MultiF)z);
			z.I = MultiFD.fZero;
		}

		public static void SetToCS(MultiC x, MultiC z) // z = x, same size
		// copy everything but .Nm Name
		{

			MultiFD.SetToFS(x, (MultiF)z);
			MultiFD.SetToFS(x.I, z.I);
		}

		public static void SetToFC(MultiF x, MultiC z)	// z = x
			// can reduce precision and round
		{
			MultiFD.SetToF(x, (MultiF)z);
			z.I &= 0;
		}

		public static void ClearSmall(MultiC z, int n) // clear z.r or z.i if very small compared to the other
		{
			if (z == MultiFD.fZero || z.I == MultiFD.fZero)
				return;
			long max = Math.Max(z.N + z.C, z.I.N + z.I.C);
			long next = z.N + z.C;
			if (max - next >= Calc.fMC - n)
				MultiFD.ClearF(z);
			next = z.I.N + z.I.C;
			if (max - next >= Calc.fMC - n)
				MultiFD.ClearF(z.I);
		}

		public static void ClearSmall(MultiC z) // clear z.r or z.i if very small compared to the other
		{
			ClearSmall(z, 0);
		}

		public static void NormC(MultiC z) // normalize
		{
			MultiFD.NormF((MultiF)z);
			MultiFD.NormF(z.I);
			ClearSmall(z);
		}

		public static void RAddC(MultiC x, MultiC z) // z = z + x
		// sets muErr true if overflow
		{
			AddC(z, x, z);
		}

		public static void RSubC(MultiC x, MultiC z) // z = z - x
		// sets muErr true if overflow
		{
			if (Object.ReferenceEquals(x, z))
			{
				ClearC(z);
			}
			else
			{
				MultiFD.RSubF(x, z);
				MultiFD.RSubF(x.I, z.I);
			}
		}

		public static void AddC(MultiC x, MultiC y, MultiC z) // z = x + y
		// sets muErr true if overflow
		{
			MultiFD.AddF((MultiF)x, (MultiF)y, (MultiF)z);
			MultiFD.AddF(x.I, y.I, z.I);
		}

		public static void SubC(MultiC x, MultiC y, MultiC z) // z = x - y
		// sets muErr true if overflow
		{
			if (Object.ReferenceEquals(x, y))
				ClearC(z);
			else if (Object.ReferenceEquals(x, z))
				RSubC(y, z);
			else if (Object.ReferenceEquals(y, z))
			{
				RSubC(x, z);
					ChSC(z);
			}
			else
			{
				ChSC(y);
				AddC(x, y, z);
				ChSC(y);
			}
		}

		//public static void ValueC(string st, out int i, MultiC z)
		//// convert String to z, returns i = # of character used
		//// allows st to be a literal, muErr set true if overflow
		//{
		//		MultiFD.ValueF(st, out i, (MultiF)z);
		//		z.I = MultiFD.fZero;
		//}

		public static bool aDigit(string st, string maxD)	// true if ch is a digit <= maxD
		{
			char ch;
			ch = Char.ToUpper(st[0]);
			return (ch >= '0' && ch <= maxD[0] && (ch <= '9' || ch >= 'A'));
		}

		public static void SetTo1C(int d1, MultiC z)	// z = d1
		// allows a literal d1, d1 an integer, d1 not changed
		{
			SetTo1C(d1, 0, z);
		}

		public static void SetTo1C(int d1, int d2, MultiC z)	// z = (d1, d2)
		// allows a literal d1 and d2, d1 and d2 an integer, d1 and d2 not changed
		{
			MultiFD.SetTo1F(d1, (MultiF)z);
			MultiFD.SetTo1F(d2, z.I);
		}

		public static void SetToExC(Int64 ext, MultiC z)	// z = ext (Int64 = long)
		// allows a literal ext, ext not changed, max Int64 = 9223,37203,68547,75807 = 2^63-1
		{
			MultiID.SetToExSI(ext, (MultiF)z);
			z.C = 0;
			z.I &= 0;
			NormC(z);
		}

		public static void SetToDbC(double db, MultiC z) // z = db % baseM^4
		// allows a literal db, db not changed, muErr set true if overflow
		{
			MultiID.SetToDbSI(db, (MultiF)z);
			z.C = 0;
			z.I &= 0;
			NormC(z);
		}

		public static void SetToDeC(decimal de, MultiC z) // z = db
		// allows a literal de, de not changed, muErr set true if overflow
		{
			MultiID.SetToDeSI(de, (MultiF)z);
			z.C = 0;
			z.I &= 0;
			NormC(z);
		}

		public static void Get1C(out int d1, MultiC z) // d1 = z % baseM
		// z is not changed
		{
			MultiFD.Get1F(out d1, (MultiF)z);
		}

		public static void GetDeC(out decimal ext, MultiC z)	// ext = z
		// z is not changed
		// max integer decimal is 2^96-1 = 7922,81625,14264,33759,35439,50335 29 digits
		{
			MultiFD.GetDeF(out ext, (MultiF)z);
		}

		public static void GetDbC(out double db, MultiC z)	// db = IntegerPart(z) % baseM^4
		// z is not changed, 2^53-1 = 9,00719,92547,40991 16 digits
		{
			MultiFD.GetDbF(out db, (MultiF)z);
		}

		public static void GetDbA(out double db, MultiC z)	// db = IntegerPart(z) + FractionPart(z)
		// z is not changed
		{
			MultiFD.GetDbA(out db, (MultiF)z);
		}

		public static void GetDbInC(out double db, double low, double high, MultiC z)
		// db = z % baseM^4, in a range, + or -
		{
			long i = z.N + z.C;
			if (i >= 5)
			{
				if (z.S)
				{
					db = low;
				}
				else
				{
					db = high;
				}
				return;
			}
			GetDbC(out db, z);
			if (db < low) db = low;
			if (db > high) db = high;
		}

		public static void RAdd1C(int d1, MultiC z) // z = z + d1
		// allows literal d1, d1 is not changed, sets muErr true if overflow
		{
			MultiF x = new MultiF(2);
			MultiFD.SetTo1F(d1, x);
			MultiFD.RAddF(x, (MultiF)z);
			MultiFD.DoneF(x);
		}

		public static void RSub1C(int d1, MultiC z) // z = z - d1
		// allows literal d1, d1 is not changed, sets muErr true if overflow
		{
			MultiF x = new MultiF(2);
			MultiFD.SetTo1F(d1, x);
			MultiFD.RSubF(x, (MultiF)z);
			MultiFD.DoneF(x);
		}

		public static void Add1C(MultiC x, int d1, MultiC z) // z = x + d1
		// allows literal d1, d1 is not changed, sets muErr true if overflow
		{
			SetToC(x, z);
			MultiFD.RAdd1F(d1, (MultiF)z);
		}

		public static void Abs(MultiC x, MultiC z)	// z = SqRt(Real(X)^2 + Imag(X)^2)
		{
			MultiFD.SqF(x, z);
			MultiFD.SqF(x.I, z.I);
			MultiFD.RAddF(z.I, z);
			MultiFD.RSqRtF(z);
			MultiFD.ClearF(z.I);
		}

		public static void SignC(MultiC x, MultiC z) // z = sign(x)
		// z = -1 if x is real < 0, 0 if x = 0, +1 if x is real > 0
		// if x not real, z = x/|x|
		// x.S used but not changed, x not changed
		{
			if (MultiFD.ZTestF(x.I))
				z &= MultiFD.SignF(x);
			else
			{
				MultiC a = new MultiC(Calc.fMC);
				Abs(x, a);
				MultiCD.DivC(x, a, z);
			}
		}

		public static int CompCC(MultiC x, MultiC z) // CompCC = sign(z - x)
		// sign = -1 if z < x, sign = 0 if z = x, sign = +1 if z > x
		// z.S and x.S used but not changed, z, x not changed
		{
			int iCompCC;
			MultiC dif = new MultiC(Math.Max(z.N, x.N) + 1, Math.Max(z.I.N, x.I.N) + 1);
			SubC(z, x, dif);
			if (MultiFD.ZTestF(dif.I))
				iCompCC = MultiFD.SignF(dif);
			else
			{
				bool zRDif = MultiFD.ZTestF(dif); // ZTestF prevents -0
				if (dif.S)
					iCompCC = -1;		// rear part < 0
				else if (!zRDif)
					iCompCC = 1;		// real part > 0
				else if (dif.I.S)
					iCompCC = -1;		// real part 0, imag part < 0
				else
					iCompCC = 1;		// real part 0, imag part > 0
			}
			DoneC(dif);
			return iCompCC;
		} // CompCC

		public static void WriteFixedL(RichTextBox outB, StreamWriter outF, MultiC z, ref int tot, int width)
		{
			if (z.I == 0)
				MultiFD.WriteFixedL(outB, outF, z, ref tot, width);
			else if ((MultiF)z == 0)
			{
				MultiFD.WriteFixedL(outB, outF, z.I, ref tot, width);
				MultiID.WriteT(outB, outF, " * i", ref tot, width);
			}
			else
			{
				MultiFD.WriteFixedL(outB, outF, z, ref tot, width);
				string st = " + ";
				if (z.I.S) st = " - ";
				z.I.S = false;
				MultiID.WriteT(outB, outF, st, ref tot, width);
				MultiFD.WriteFixedL(outB, outF, z.I, ref tot, width);
				MultiID.WriteT(outB, outF, " * i", ref tot, width);
				z.I.S = (st == " - ");
			}
		} // WriteFixedL

		public static void WriteFixed(RichTextBox outB, StreamWriter outF, MultiC z, ref int tot, int width)
		{
			int j;
			int k;
			int did;
			int didZ;
			string st;
			did = 0;
			if (z.S) MultiID.WriteT(outB, outF, "-", ref tot, width);
			if (z.N + z.C > 0)
			{
				st = ((int)(MultiID.baseM + z.V[z.N-1])).ToString();
				j = 1;
				do
					j++;
				while (st[j-1] == '0' && j != MultiID.muDMax + 1);
				did = j - 2 - MultiID.muDMax;
				if (z.N + z.C > 1) did -= MultiID.muDMax;
				if (MultiID.muDpG > 0) did = did % MultiID.muDpG;
				if (did < 0) did = did + MultiID.muDpG;
				didZ = did;
				k = MultiID.muDMax + 1;
				j--;
				Write1C(st, ref j, k, ref did, didZ, outB, outF, ref tot, width);
				if (z.N + z.C > 1)
				{
					if (z.N > 1)
						st = ((int)(MultiID.baseM + z.V[z.N-2])).ToString();
					else
					{
						st = ((int)(MultiID.baseM)).ToString();
					}
					k = MultiID.muDMax + 1;
					j = 1;
					Write1C(st, ref j, k, ref did, didZ, outB, outF, ref tot, width);
				}
			}
			else
			{
				MultiID.WriteT(outB, outF, "0", ref tot, width);
			}
			MultiID.WriteT(outB, outF, ".", ref tot, width);
			didZ = did;
			if (-z.C == 2)
			{
				st = ((int)(MultiID.baseM + z.V[1])).ToString();
				k = MultiID.muDMax + 1;
				j = 1;
				Write1C(st, ref j, k, ref did, didZ, outB, outF, ref tot, width);
			}
			if (-z.C > 0)
			{
				st = ((int)(MultiID.baseM + z.V[0])).ToString();
				j = MultiID.muDMax + 2;
				do
					j--;
				while (st[j-1] == '0');
				k = j;
				j = 1;
				Write1C(st, ref j, k, ref did, didZ, outB, outF, ref tot, width);
			}
			else
			{
				MultiID.WriteT(outB, outF, "0", ref tot, width);
			}
			return;
		} // WriteFixed

		private static void Write1C(string st, ref int j, int k, ref int did, int didZ, RichTextBox outB, StreamWriter outF, ref int tot, int width)
		// part of WriteFixed
		{
			while (j < k)
			{
				j++;
				if (did != didZ && MultiID.muDpG > 0)
				{
					if (did % MultiID.muDpG == 0) MultiID.WriteT(outB, outF, ",", ref tot, width);
				}
				MultiID.WriteT(outB, outF, st.Substring(j - 1, 1), ref tot, width);
				did++;
			}
			return;
		}

		public static void WriteC(MultiC z, ref int tot)
		{
			WriteC(MultiID.outS, null, z, ref tot, 0);
		}

		public static void WriteC(RichTextBox outB, StreamWriter outF, MultiC x, ref int tot, int width)
		// output x as a line of text, tot = characters on line so far, x is not modified
		// good for any base = baseO [2, 36], base 10 much faster
		{
			if (MultiFD.ZTestF(x.I))
			{
				MultiFD.WriteF(outB, outF, x, ref tot, width);
			}
			else
			{
				bool bs = x.I.S;
				bool wB = MultiID.writeBool;
				MultiID.writeBool = false;
				if (!MultiFD.ZTestF(x))
				{
					MultiFD.WriteF(outB, outF, x, ref tot, width);
					string s = " + ";
					if (bs) s = " - ";
					x.I.S = false;
					MultiID.WriteT(outB, outF, s, ref tot, width);
				}
				MultiFD.WriteF(outB, outF, x.I, ref tot, width);
				MultiID.WriteT(outB, outF, " * i", ref tot, width);
				MultiID.writeBool = wB;
				x.I.S = bs;
			}
		}

		public static void WriteFLn(RichTextBox outB, StreamWriter outF, MultiC z, ref int tot)
		// output z as a line of text, tot = characters on line so far
		{
			WriteC(outB, outF, z, ref tot, 0);
			MultiID.Write(outB, outF, "\n");
			tot = 0;
		}

		public static void WriteFLn(MultiC z, ref int tot)
		{
			WriteFLn(MultiID.outS, null, z, ref tot);
		}

		public static void WriteFLn(MultiC z)
		{
			int tot = 0;
			WriteFLn(MultiID.outS, null, z, ref tot);
		}

		public static void RMul1C(int d1, MultiC z) // z = z * d1 (d1 <= maxDigit)
		// Multi-precision one super digit multiply d1 is not changed
		// muErr set true if z overflows
		{
			MultiFD.RMul1F(d1, (MultiF)z);
			MultiFD.RMul1F(d1, z.I);
		}

		public static void Mul1C(MultiC x, int d1, MultiC z) // z = x * d1 (d1 <= maxDigit)
		// Multi-precision one super digit multiply x and d1 are not changed
		// muErr set true if z overflows
		{
			SetToC(x, z);
			RMul1C(d1, z);
		}

		public static void MulC(MultiC x, MultiC y, MultiC z) // z = x * y
		// x or y or both may have the same address in memory as z
		// x, y are not changed unless they share memory with z
		// sets muErr true if overflow
		//	x*y = ((x.r*y.r - x.i*y.i), (x.r*y.i + x.i*y.r))
		{
			if (MultiFD.ZTestF(x.I) && MultiFD.ZTestF(y.I)) // not actually needed
			{
				MultiFD.MulF(x, y, z);
				z.I &= 0;
				return;
			}
			MultiF tI = new MultiF(Calc.fMC);
			if (MultiFD.ZTestF(x.I)) // not actually needed
			{
				MultiFD.MulF(x, y.I, tI);
				MultiFD.MulF(x, y, z);
			}
			else if (MultiFD.ZTestF(y.I)) // not actually needed
			{
				MultiFD.MulF(x.I, y, tI);
				MultiFD.MulF(x, y, z);
			}
			else
			{
				MultiF t = new MultiF(Calc.fMC);
				t &= (MultiF)x * (MultiF)y;
				t -= x.I * y.I;
				tI &= (MultiF)x * y.I;
				tI += x.I * (MultiF)y;
				z &= t;
				MultiFD.DoneF(t);
			}
			z.I &= tI;
			MultiFD.DoneF(tI);
		}

		public static void MulCF(MultiC x, MultiF y, MultiC z) // z = x * y
		// x or y or both may have the same address in memory as z
		// x, y are not changed unless they share memory with z
		// sets muErr true if overflow
		//	x*y = (x.r*y.r, x.r*y.i)
		{
			MultiF tI = new MultiF(Calc.fMC);
			MultiFD.MulF(x.I, y, tI);
			MultiFD.MulF(x, y, z);
			z.I &= tI;
			MultiFD.DoneF(tI);
		}

		public static void RMulC(MultiC x, MultiC z) // z = z * x
		// x may have the same address in memory as z
		// x is not changed unless it shares memory with z
		// sets muErr true if overflow
		{
			MulC(z, x, z);
		}

		public static void SqC(MultiC x, MultiC z) // z = x * x
		// x may have the same address in memory as z
		// x is not changed unless it shares memory with z
		// sets muErr true if overflow
		{
			MulC(x, x, z);
		}

		public static void RSqC(MultiC z) // z = z * z
		// sets muErr true if overflow
		{
			MulC(z, z, z);
		}

		public static void RDivCF(MultiF x, MultiC z) // z = z / x
		// sets muErr true if overflow
		{
			MultiFD.RDivF(x, (MultiF)z);
			MultiFD.RDivF(x, z.I);
		}

		public static void RDiv1C(int d1, MultiC z) // z = z / d1 (d1 <= maxDigit)
		// Multi-precision one super digit divide, d1 is not changed
		// allows literal d1, muErr set true if d1 = 0
		{
			MultiF x = new MultiF(2); // temp
			MultiFD.SetTo1F(d1, x);
			RDivCF(x, z);
			MultiFD.DoneF(x);
		}

		public static void Div1C(MultiC u, int d1, MultiC z) // z = u / d1 (d1 <= maxDigit)
		// Multi-precision one super digit divide, d1 and u not changed
		// u and z may have same address in memory, then u is changed
		// allows literal d1, muErr set true if d1 = 0
		{
			SetToC(u, z);
			RDiv1C(d1, z);
		}

		public static void Div1CC(int u1, MultiC d, MultiC z) // z = u1 / d (u1 <= maxDigit)
			// Multi-precision one super digit divide, d1 and u not changed
			// u and z may have same address in memory, then u is changed
			// allows literal d1, muErr set true if d1 = 0
		{
			SetToC(d, z);
			RDiv1C(u1, z);
			DivC(MultiCD.cOne, z, z);
		}

		public static void DivC(MultiC u, MultiC d, MultiC z) // z = u / d
		// u or d or both may have the same address in memory as z
		// u, d are not changed unless they share memory with z
		// sets muErr true if d = 0
		//	u/d = ((ur*dr + ui*di)/(dr^2 + di^2) , (ui*dr - ur*di)/(dr^2 + di^2))
		{
			if (MultiFD.ZTestF(u.I) && MultiFD.ZTestF(d.I))
			{
				MultiFD.DivF(u, d, z);
				z.I &= 0;
				return;
			}
			MultiF t = new MultiF(Calc.fMC);
			if (MultiFD.ZTestF(d.I))		// not actually needed
			{
				MultiFD.DivF(u.I, d, t);
				MultiFD.DivF(u, d, z);
				z.I &= t;
			}
			else
			{
				MultiF s = new MultiF(Calc.fMC);
				s &= (MultiF)d * (MultiF)d + d.I * d.I;
				t &= (u.I * (MultiF)d - (MultiF)u * d.I) / s;
				z &= ((MultiF)u * (MultiF)d + u.I * d.I) / s;
				z.I &= t;
				MultiFD.DoneF(s);
			}
			MultiFD.DoneF(t);
		}

		public static void DivCF(MultiC u, MultiF d, MultiC z) // z = u / d
		// u or d or both may have the same address in memory as z
		// u, d are not changed unless they share memory with z
		// sets muErr true if d = 0
		//	u/d = (ur/dr , ui/dr)
		{
			bool saveExp = MultiFD.expanded;
			int saveFMC = Calc.fMC;
			int saveMMax = MultiID.mMax;
			XPCalLst.Expand();
			z &= (MultiF)u / d;
			z.I &= u.I / d;
			MultiFD.expanded = saveExp;
			Calc.fMC = saveFMC;
			MultiID.mMax = saveMMax;
			NormC(z);
		}

		public static void RDivC(MultiC d, MultiC z) // z = z / d
		// d may have the same address in memory as z
		// d is not changed unless it shares memory with z
		// sets muErr true if d = 0
		{
			DivC(z, d, z);
		}

		public static void SqRootC(MultiC x, MultiC z) // z = SqRt(x)
		// x is not changed, it is ok for x and this to have the same location in memory
		// but then x will be changed. Sets muErr True if error
		{	
			if (MultiFD.ZTestF(x.I) && !x.S)
			{
				MultiFD.SqRootF(x, z);
				z.I &= 0;
				return;
			}
			XCCalLst.PowNC(x, cHalf, z);		// z = x^0.5
		}

		public static void RSqRtC(MultiC z) // z = SqRt(z)
		// sets muErr true if error
		{
			SqRootC(z, z);
		}

		public static void IntC(MultiC x, MultiC z) // z = Integer part of x
		{
			MultiFD.IntF((MultiF)x, (MultiF)z);
			z.I = MultiFD.fZero;
		}

		public static void RIntC(MultiC z) // z = Integer part of z
		{
			IntC(z, z);
		}

		public static void FracC(MultiC x, MultiC z) // z = Fractional part of x
		{
			MultiFD.FracF((MultiF)x, (MultiF)z);
			MultiFD.FracF(x.I, z.I);
		}

		public static void RFracC(MultiC z) // z = Fractional part of z
		{
			FracC(z, z);
		}

		public static void FloorC(MultiC x, MultiC z) // z = integer <= and nearest to x
		{
			MultiFD.FloorF((MultiF)x, (MultiF)z);
			MultiFD.FloorF(x.I, z.I);
		}

		public static void LopC(MultiC x, MultiC z) // z = x with one less super digit
		{
			if (MultiFD.ZTestF(x.I))
				MultiFD.LopF(x, z);
			else if (MultiFD.ZTestF(x))
				MultiFD.LopF(x.I, z.I);
			else
			{
				bool lopI = (x.C >= x.I.C);		// x and z may have the same address in memory
				if (x.C <= x.I.C)
					MultiFD.LopF(x, z);
				else
					MultiFD.SetToF(x, z);		// copies real part
				if (lopI)
					MultiFD.LopF(x.I, z.I);
				else
					MultiFD.SetToF(x.I, z.I);	// copies imaginary part
			}
		}

		public static void RLopC(MultiC z)	// z = z with one less least significant "digit"
		{
			LopC(z, z);
		}

		public static bool ZTestC(MultiC z) // true if z is Zero
		{
			return (MultiFD.ZTestF((MultiF)z) && MultiFD.ZTestF(z.I));
		}

		public static bool ZTestN(int zI) // true if z is Zero
		{
			return ZTestC(Common.nList[zI]);
		}

		public static void ReducePrecisionC(int ps, MultiC d, MultiC dL)
		// remove lower ps super digits, d and dL may be the same object
		{
			MultiFD.ReducePrecisionF(ps, d, dL);
			MultiFD.ReducePrecisionF(ps, d.I, dL.I);
		}

		public static MultiC ci;		// ci = complex i = |SqRt(-1)| = (0, 1)
		public static MultiC cMOne;		// cMOne  = -1.0 + 0.0 * i
		public static MultiC cMHalf;	// cMHalf = -0.5 + 0.0 * i
		public static MultiC cZero;		// cZeto  =  0.0 + 0.0 * i
		public static MultiC cHalf;		// cHalf  =  0.5 + 0.0 * i
		public static MultiC cOne;		// cOne   =  1.0 + 0.0 * i
		public static MultiC cTwo;		// cTwo   =  2.0 + 0.0 * i
		public static MultiC cFour;		// cFour  =  4.0 + 0.0 * i
		public static MultiC cSix;		// cSix   =  6.0 + 0.0 * i
	}
}
