using System;
using System.IO;			// for StreamWriter I/O to a file
using System.Windows.Forms;	// for RichTextBox

namespace MyNamespace
{
	/// <summary>
	///		Extra precision floating-point complex List algorithms
	///		Developed in TURBO Pascal 5.5 and 6.0, converted to Visual Basic 6
	///		and now in Visual C# .Net
	///
	/// ---------------- *** LEGAL NOTICE: *** ------------------
	///
	/// This program is free software; you can redistribute it and/or modify
	/// it under the terms of the GNU General Public License (GPL) as published
	/// by the Free Software Foundation. cf. the file COPYING.txt.
	///
	/// ------------ *** end of legal notice *** ------------------
	/// </summary>

	public class XCCalLst
	{
		public XCCalLst()
		{
			Common.saveI = 2;
		}

		public static void RAddN(MultiF x, int zI) // z = z + x
			// sets muErr true if overflow
		{
			AddN((MultiF)Common.nList[zI], x, zI);
		}

		public static void RSubN(MultiF x, int zI) // z = z - x
			// sets muErr true if overflow
		{
			SubN((MultiF)Common.nList[zI], x, zI);
		}

		public static void AddN(MultiF x, MultiF y, int zI) // z = x + y
		// sets muErr true if overflow
		{
			MultiF t = new MultiF(Calc.fMC);
			MultiFD.AddF(x, y, t);
			Calc.SetToN(t, zI); // clears xI.i
			MultiFD.DoneF(t);
		}

		public static void SubN(MultiF x, MultiF y, int zI) // z = x - y
		{
			// sets muErr true if overflow
			MultiF t = new MultiF(Calc.fMC);
			MultiFD.SubF(x, y, t);
			Calc.SetToN(t, zI);
			MultiFD.DoneF(t);
		}

		public static void RAddN(MultiC x, int zI)	// z = z + x
		// sets muErr true if overflow
		{
			AddN(Common.nList[zI], x, zI);
		}

		public static void RSubN(MultiC x, int zI) // z = z - x
		// sets muErr true if overflow
		{
			SubN(Common.nList[zI], x, zI);
		}

		public static void AddN(MultiC x, MultiC y, int zI) // z = x + y
		// sets muErr true if overflow
		{
			MultiC t = new MultiC(Calc.fMC);
			MultiCD.AddC(x, y, t);
			Calc.SetToN(t, zI);
			MultiCD.DoneC(t);
		}

		public static void SubN(MultiC x, MultiC y, int zI) // z = x - y
		{
			// sets muErr true if overflow
			MultiC t = new MultiC(Calc.fMC);
			MultiCD.SubC(x, y, t);
			Calc.SetToN(t, zI);
			MultiCD.DoneC(t);
		}

		public static void SetTo1N(int d1, int zI) // z = d1 % baseM
		// allows a literal d1, d1 an integer, |d1| < baseM, d1 not changed
		{
			MultiF t = new MultiF(2);
			MultiFD.SetTo1F(d1, t);
			Calc.SetToN(t, zI);	// clears zI.i
			MultiFD.DoneF(t);
		}

		public static void SetToDbN(double db, int zI) // z = db % baseM^4
		// allows a literal db, db not changed, muErr set true if overflow
		{
			MultiF t = new MultiF(4);
			MultiFD.SetToDbF(db, t);
			Calc.SetToN(t, zI);
			MultiFD.DoneF(t);
		}

		public static void RAdd1N(int d1, int zI) // z = z + d1
		// allows literal d1, d1 is not changed, sets muErr true if overflow
		{
			MultiF t = new MultiF(2); // temp
			MultiFD.SetTo1F(d1, t);
			XPCalLst.RAddN(t, zI);
			MultiFD.DoneF(t);
		}

		public static void Add1N(MultiC x, int d1, int zI) // z = x + d1
		// allows literal d1, d1 is not changed, sets muErr true if overflow
		{
			Calc.SetToN(x, zI);
			RAdd1N(d1, zI);
		}

		public static void FormatNB(StreamWriter outF, int zI)
		// output z as a formatted set of lines in base baseO
		{
			int tot = 0;
			MultiID.WriteLn(null, outF);
			MultiID.WriteTLn2(null, outF, "Base(" + MultiID.baseO + ")", ref tot, 0);
			MultiID.WriteT(null, outF, Common.nList[zI].Nm + " =", ref tot, 0);
			MultiCD.WriteC(null, outF, Common.nList[zI], ref tot, 0); // WriteF will call WriteFB
			MultiID.WriteLn(null, outF);
		}

		public static void AndN(MultiC x, MultiC y, int zI) // z = x & y
		{
			if (x == MultiCD.cZero || y == MultiCD.cZero)
				Calc.SetToN(MultiCD.cZero, zI);
			else
				Calc.SetToN(MultiCD.cOne, zI);
		} // AndN

		public static void OrN(MultiC x, MultiC y, int zI) // z = x | y
		{
			if (x == MultiCD.cZero && y == MultiCD.cZero)
				Calc.SetToN(MultiCD.cZero, zI);
			else
				Calc.SetToN(MultiCD.cOne, zI);
		} // OrN

		public static void NotN(MultiC x, int zI)	// z = !x
		{
			bool zip;
			zip = MultiCD.ZTestC(x);
			Calc.ClearN(zI);
			if (zip) Common.nList[zI].V[0] = 1;
		} // NotN

		public static void MulNFF(MultiF x, MultiF y, int zI) // z.r = x * y, z.i unchanged
			// x or y or both may have the same address in memory as z
			// x, y are not changed unless they share memory with z
			// sets muErr true if overflow
		{
			MultiF t = new MultiF(Calc.fMC);
			MultiFD.MulF(x, y, t);
			MultiFD.SetToFS(t, Common.nList[zI]);
			MultiFD.DoneF(t);
		}

		public static void MulNFI(MultiF x, MultiF y, int zI) // z.i = x * y, z.r unchanged
			// x or y or both may have the same address in memory as z
			// x, y are not changed unless they share memory with z
			// sets muErr true if overflow
		{
			MultiF t = new MultiF(Calc.fMC);
			MultiFD.MulF(x, y, t);
			MultiFD.SetToFS(t, Common.nList[zI].I);
			MultiFD.DoneF(t);
		}

		public static void MulN(MultiC x, MultiF y, int zI) // z = x * y
			// x or y or both may have the same address in memory as z
			// x, y are not changed unless they share memory with z
			// sets muErr true if overflow
		{
			MulNFF((MultiF)x, y, zI);
			MulNFI(x.I, y, zI);
		}

		public static void MulN(MultiF x, MultiF y, int zI) // z = x * y
			// x or y or both may have the same address in memory as z
			// x, y are not changed unless they share memory with z
			// sets muErr true if overflow
		{
			MulNFF(x, y, zI);
			MultiFD.ClearF(Common.nList[zI].I);
			Calc.NormN(zI);
		}

		public static void RMul1N(int d1, int zI) // z = z * d1
		// Multi-precision one or two super digit multiply d1 is not changed
		// muErr set true if z overflows
		{
			MultiC t = new MultiC(2, 1);
			MultiFD.SetTo1F(d1, t);
			XCCalLst.RMulN(t, zI);
			MultiFD.DoneF(t);
		}

		public static void Mul1N(MultiC x, int d1, int zI) // z = x * d1
		// Multi-precision one or two super digit multiply x and d1 are not changed
		// muErr set true if z overflows
		{
			Calc.SetToN(x, zI);
			RMul1N(d1, zI);
		}

		public static void MulN(MultiC x, MultiC y, int zI) // z = x * y
		// x or y or both may have the same address in memory as z
		// x, y are not changed unless they share memory with z
		// sets muErr true if overflow
		{
			MultiC t = new MultiC(Calc.fMC);
			MultiCD.MulC(x, y, t);
			Calc.SetToN(t, zI);
			MultiCD.DoneC(t);
		}

		public static void RMulN(MultiC x, int zI) // z = z * x
		// x may have the same address in memory as z
		// x is not changed unless it shares memory with z
		// sets muErr true if overflow
		{
			MulN(Common.nList[zI], x, zI);
		}

		public static void MulNF(MultiC x, MultiF y, int zI) // z = x * y
		// x or y or both may have the same address in memory as z
		// x, y are not changed unless they share memory with z
		// sets muErr true if overflow
		{
			MultiC t = new MultiC(Calc.fMC);
			MultiCD.MulCF(x, y, t);
			Calc.SetToN(t, zI);
			MultiCD.DoneC(t);
		}

		public static void RMulNF(MultiF x, int zI) // z = z * x
		// x may have the same address in memory as z
		// x is not changed unless it shares memory with z
		// sets muErr true if overflow
		{
			MulNF(Common.nList[zI], x, zI);
		}

		public static void ConjN(MultiC x, int zI) // z = Complex Conjugate(x)
			// x may have the same address -in memory as z
			// x is not changed unless it shares memory with z
		{
			Calc.SetToN(x, zI);
			MultiFD.ChSF(Common.nList[zI].I);
		} // ConjN

		public static void RConjN(int zI) // z = Complex Conjugate(z)
		{
			MultiFD.ChSF(Common.nList[zI].I);
		}

		public static void SignN(MultiC x, int zI) // z = Sign(x)
			// x may have the same address -in memory as z
			// x is not changed unless it shares memory with z
		{
			if (MultiFD.ZTestF(x.I))
			{
				MultiFD.SetTo1F(MultiFD.SignF(x), Common.nList[zI]);
				MultiFD.ClearF(Common.nList[zI].I);
			}
			else
			{
				bool saveExp = MultiFD.expanded;
				int saveFMC = Calc.fMC;
				int saveMMax = MultiID.mMax;
				XPCalLst.Expand();
				MultiC s = new MultiC(Calc.fMC);
				MultiCD.SignC(x, s);
				Calc.SetToN(s, zI);
				MultiCD.DoneC(s);
				MultiFD.expanded = saveExp;
				Calc.fMC = saveFMC;
				MultiID.mMax = saveMMax;
			}
			Calc.NormN(zI);
		} // SignN

		public static void RSignN(int zI) // z = Sign(z)
		{
			SignN(Common.nList[zI], zI);
		}

		public static void SqN(MultiC x, int zI) // z = x * x
		// x may have the same address -in memory as z
		// x is not changed unless it shares memory with z
		// sets muErr true if overflow
		{
			MulN(x, x, zI);
		}

		public static void RSqN(int zI) // z = z * z
		// sets muErr true if overflow
		{
			MulN(Common.nList[zI], Common.nList[zI], zI);
		}

		public static void PowN(MultiC b, MultiC p, int zI)
		// g = b^p = Exp(p * Ln(b))
		{
			if (MultiFD.ZTestF(b.I) && MultiFD.ZTestF(p.I))
			{
				if (!b.S || (p.C >= 0)) // b >= 0 or (p an integer)
				{
					XPCalLst.PowN(b, p, zI);
					MultiFD.ClearF(Common.nList[zI].I);
					Calc.NormN(zI);
					return;
				}
			}
			if (MultiCD.ZTestC(b))
			{
				if ((MultiF)p > MultiFD.fZero)		// if positive exponent
					Calc.ClearN(zI);				// 0^(real > 0) = 0
				else								// anything else is undefined
					MultiID.MuWriteErr("Cannot divide by zero, continuing...");
				return;
			}
			bool saveExp = MultiFD.expanded;
			int saveFMC = Calc.fMC;
			int saveMMax = MultiID.mMax;
			XPCalLst.Expand();
			int hI;
			Common.InitNAdd(1, "PowM-hI", out hI);
			XCCalLst.LnN(b, hI);
			XCCalLst.RMulN(p, hI);
			XCCalLst.ExpN(Common.nList[hI], zI);
			Common.LDelete(hI);
			MultiFD.expanded = saveExp;
			Calc.fMC = saveFMC;
			MultiID.mMax = saveMMax;
			Calc.NormN(zI);
		} // PowN

		public static void PowNC(MultiC b, MultiC p, MultiC z) // z = b^p
		{
			int zI;
			Common.InitNAdd(1, "PowNC-zI", out zI);
			PowN(b, p, zI);
			MultiCD.SetToC(Common.nList[zI], z);
			Common.LDelete(zI);
		}

		public static void RPowN(MultiC p, int zI) // z = z^p
		{
			PowN(Common.nList[zI], p, zI);
		}

		public static void PowMN(MultiC b, MultiC p, int zI) // z = b^p % FMB
		{
			int fMBI;
			if (!Common.FMBIok(zI, out fMBI))
				return;
			if (MultiFD.ZTestF(b.I) && MultiFD.ZTestF(p.I) && MultiFD.ZTestF(Common.nList[fMBI].I))
			{
				XPCalLst.PowMN(b, p, zI);
				MultiFD.ClearF(Common.nList[zI].I);
				Calc.NormN(zI);
				return;
			}
			PowN(b, p, zI);
			RModFMBN(zI);
		}

		public static void RPowMN(MultiC p, int zI) // z = z^p % FMB
		{
			PowMN(Common.nList[zI], p, zI);
		}

		public static void RDiv1N(int d1, int zI) // z = z / d1
		// Multi-precision one or two super digit divide, d1 is not changed
		// allows literal d1, muErr set true if d1 = 0
		{
			MultiF t = new MultiF(2);
			MultiFD.SetTo1F(d1, t);
			RDivNF(t, zI);
			MultiFD.DoneF(t);
		}

		public static void Div1N(MultiC u, int d1, int zI) // z = u / d1
		// Multi-precision one or two super digit divide, d1 and u not changed
		// u and z may have same address in memory, then u is changed
		// allows literal d1, muErr set true if d1 = 0
		{
			Calc.SetToN(u, zI);
			RDiv1N(d1, zI);
		}

		public static void DivN(MultiC u, MultiC d, int zI) // z = u / d
		// u or d or both may have the same address in memory as z
		// u, d are not changed unless they share memory with z
		// sets muErr true if d = 0
		{
			MultiC t = new MultiC(Calc.fMC);
			MultiCD.DivC(u, d, t);
			Calc.SetToN(t, zI);
			MultiCD.DoneC(t);
		}

		public static void DivRemN(MultiC u, MultiC d, int zI) // z = Floor(u / d), Re = remainder
		// u or d or both may have the same address in memory as z
		// u, d are not changed unless they share memory with z
		// sets muErr true if d = 0
		{
			int rI;
			Common.LFindAdd("Re", out rI);
			if (rI == zI || rI == Common.iTop)
			{
				MultiID.MuWriteErr(
				"Cannot set same location to both quotient and remainder, continuing...");
				return;
			}
			MultiC t = new MultiC(Calc.fMC);
			MultiC r = new MultiC(Calc.fMC);
			Mod(u, d, r);
			MultiCD.DivC(u-r, d, t);
			if (MultiID.muErr)
			{
				Calc.SetToN(MultiCD.cZero, rI);
				Calc.SetToN(MultiCD.cZero, zI);
				return;
			}
			Calc.SetToN(t, zI);		// z = t = integer quotient
			Calc.SetToN(r, rI);		// Re = r = remainder
		} // DivRemN

		public static void DivIntN(MultiC u, MultiC d, int zI) // z = Floor(u / d)
		// u or d or both may have the same address in memory as z
		// u, d are not changed unless they share memory with z
		// sets muErr true if d = 0
		{
			MultiC t = new MultiC(Calc.fMC);
			Mod(u, d, t);
			MultiCD.DivC(u-t, d, t);
			Calc.SetToN(t, zI);		// z = t = integer quotient
		} // DivIntN

		public static void RDivN(MultiC d, int zI) // z = z / d
		// d may have the same address in memory as z
		// d is not changed unless it shares memory with z
		// sets muErr true if d = 0
		{
			DivN(Common.nList[zI], d, zI);
		}

		public static void DivNF(MultiC u, MultiF d, int zI) // z = u / d
		// u or d or both may have the same address in memory as z
		// u, d are not changed unless they share memory with z
		// sets muErr true if d = 0
		{
			MultiC t = new MultiC(Calc.fMC);
			MultiCD.DivCF(u, d, t);
			Calc.SetToN(t, zI);
			MultiCD.DoneC(t);
		}

		public static void RDivNF(MultiF d, int zI) // z = z / d
		// d may have the same address in memory as z
		// d is not changed unless it shares memory with z
		// sets muErr true if d = 0
		{
			DivNF(Common.nList[zI], d, zI);
		}

		public static bool ZTestNC(int zI) // is z = 0
		{
			return (MultiFD.ZTestF(Common.nList[zI]) && MultiFD.ZTestF(Common.nList[zI].I));
		}
		
		// Function that computes van Wijngaarden's A_j for a given j. // used by LerchT and Lerch
		private static int Aj(MultiC x, MultiC s, MultiC a, int j, MultiC result, string name)
		{
			MultiC bjk = new MultiC(Calc.fMC);
			MultiC ind = new MultiC(Calc.fMC);
			MultiC jFp = new MultiC(Calc.fMC);
			MultiC p = new MultiC(Calc.fMC);		// p is a power
			MultiC sum = new MultiC(Calc.fMC);
			MultiC sumP = new MultiC(Calc.fMC);
			MultiC t = new MultiC(Calc.fMC);		// t is temp
			MultiC two2k = new MultiC(Calc.fMC);
			MultiC z2ind = new MultiC(Calc.fMC);
			two2k &= MultiCD.cHalf;
			jFp &= j + 1;
			int flag = 0;
			for (int k = 0; true; k++)	// Sum b^j_k's over k.
			{
				two2k *= 2;	// Index for the term of the original series.
				ind &= two2k * jFp - 1;	// ind = two2k * (j + 1) - 1
				// If long integer overflow occurs, variables become zero.
				// Not relevant in v1.0 because two2k and ind are double type.	
				if (MultiID.muErr)
					flag = 4;
				// Increment the sum.
				PowNC(x, ind, z2ind);					// z2ind = Pow(x, ind)
				if (MultiID.muAbort || MultiID.muErr)
					break;
				MultiCD.AddC(a, ind, t);				// t = a + ind
				PowNC(t, s, p);							// t = Math.Pow(a + ind, s)
				bjk &= two2k * z2ind / p;				// bjk = two2k * z2ind / Math.Pow(a + ind, s)
				sumP &= sum;
				sum += bjk;
				// Stop summation if either sum is zero or sum == sumP
				if (MultiCD.ZTestC(sum) || (sum == sumP)) break;
				MultiID.MuInterrupt();
				if (MultiID.statReq3)
				{
					Common.SingleSpace();
					MultiID.WriteT(name +": Aj Iteration = " + k);
					MultiID.TimeDel();
					MultiID.statReq3 = false;
				}
				if (k > 1000)
					MultiID.muErr = true;
				if (MultiID.muAbort || MultiID.muErr)
					break;
			}
			if (MultiID.muAbort || MultiID.muErr)
				sum &= 0;
			if (MultiID.muErr)
				flag = 4;
			result &= sum;
			MultiCD.DoneC(z2ind);
			MultiCD.DoneC(two2k);
			MultiCD.DoneC(t);
			MultiCD.DoneC(sumP);
			MultiCD.DoneC(sum);
			MultiCD.DoneC(p);
			MultiCD.DoneC(jFp);
			MultiCD.DoneC(ind);
			MultiCD.DoneC(bjk);
			return flag;
		} // Aj

		public static void Lerch(MultiC x, MultiC s, MultiC a, int zI) // z = Lerch transcendent capitol Phi(x, s, a)
		// Function that computes approximation to Lerch Phi as a
		// converging sequence of CNC transforms S^n_k.
		// x, s, a may have the same address in memory as z				// traditional
		// x, s, a are not changed unless it shares memory with z
		// sets muErr true if overflow or error
		//
		// The LerchT(x, s, a) function is Phi(x, s, a) = Sum {k = 0, 1, ...} of x^k / (a + k)^s,
		// where |x| <= 1 and a != 0, -1, -2, ... .
		// Lerch(x, s, a) = sum[k=0, infinity]{x^k / |a + k|^s}, terms with a + k == 0 are ignored, |x| <= 1
		{
			if (MultiFD.ZTestF(x.I) && MultiFD.ZTestF(s.I) && MultiFD.ZTestF(a.I))
			{
				XPCalLst.Lerch(x, s, a, false, zI);
				MultiFD.ClearF(Common.nList[zI].I);
				Calc.NormN(zI);
				return;
			}
			MultiC saveX = new MultiC(x.N, x.I.N);
			saveX &= x;
			if (s == MultiCD.cTwo && a == MultiCD.cOne && DilogSN(x, zI))	// if special dilog case
			{

				RDivN(saveX, zI);
				MultiCD.DoneC(saveX);
				return;
			}
			if (s == MultiCD.cTwo && a == MultiCD.cOne && x < MultiFD.fMOne)	// if Dilog of x < -1
			{
				DilogN(x, zI);
				RDivN(saveX, zI);
				MultiCD.DoneC(saveX);
				return;
			}
			MultiCD.DoneC(saveX);
			int flag = 0, flagL = 0;
			// Special cases.
			string name = "Lerch";
			// 1 < x. (Return error, Lerch Phi diverges.)	// was <= but LerchPhi(1, 2, 1) = Dilog(1) = Pi^2 / 6
			if (x > MultiCD.cOne)	// if x > 1
			{
				Calc.ClearN(zI);
				flag = 1;
				MultiID.MuWriteErr(name + ": Cannot always compute LerchPhi(x, s, a) for x > 1");
				return;
			}
			bool saveExp = MultiFD.expanded;
			int saveFMC = Calc.fMC;
			int saveMMax = MultiID.mMax;
			XPCalLst.Expand();
			MultiC aL = new MultiC(Calc.fMC);
			MultiC n = new MultiC();
			aL &= a;									// Local copy of a.
			n = new MultiC(Calc.fMC);
			if (a.I == 0 && a <= MultiFD.fZero && a.C >= 0)		// if a <= 0 and an integer
			{
				flagL = 2;
				n &= -aL;
				aL &= 1;
			}
			MultiC t = new MultiC(Calc.fMC);			// t = temp
			double digits = Calc.fMC * MultiID.muDMax;
			int imax = (int)(4.0 * digits);
			int i, j, sign, iter, imin = 0;				// m = 0; const int beta = 1, n = 0;
			bool sEq2 = (s == MultiCD.cTwo);			// does s equal 2?
			if (sEq2 && a == MultiCD.cOne && x.I == 0 && x.N == 1 && x.V[0] == 1) // and a equal 1?
			{											// lerch(+/-1, 2, 1)
				int pI = XPCalLst.GetPiPtr();
				t &= Common.nList[pI] * Common.nList[pI];
				if (x.S)
					t /= 12;			// Lerch(-1, 2, 1) = Pi^2 / 12
				else
					t /= 6;				// Lerch(1, 2, 1) = Pi^2 / 6
				Calc.SetToN(t, zI);
				iter = 0;
				goto EndItEarly;
			}
			MultiC cacc = new MultiC(Calc.fMC);
			MultiC eps = new MultiC(Calc.fMC);
			MultiC eps0 = new MultiC(Calc.fMC);
			MultiC factor = new MultiC(Calc.fMC);
			MultiC factor1 = new MultiC(Calc.fMC);
			MultiC iF = new MultiC(4, 1);				// iF is i in floating-point
			MultiC iom = new MultiC(Calc.fMC);
			MultiC mF = new MultiC(Calc.fMC);
			MultiC omega = new MultiC(Calc.fMC);
			MultiC p1 = new MultiC(Calc.fMC);				// p1 and p2 are powers
			MultiC p2 = new MultiC(Calc.fMC);
			MultiC skn = new MultiC(Calc.fMC);
			MultiC skn0 = new MultiC(Calc.fMC);
			MultiC sn = new MultiC(Calc.fMC);
			MultiC sum1 = new MultiC(Calc.fMC);
			MultiC best = new MultiC(Calc.fMC);
			MultiC minEps = new MultiC(Calc.fMC);
			minEps.V[0] = MultiID.maxDigit;			// initialize min eps arbitrary large
			minEps.C = long.MaxValue;
			MultiC[] num;
			MultiC[] den; 
			MultiC[] StoreAj;
			if (a < MultiCD.cZero && x != 0)	// a < 0 is not integer or zero and x != 0 (x == 0 considered below) ...
			{
				// s is an integer. (Transform a to positive).
				MultiFD.FloorF(a, mF);			// m = -Floor(a.r)
				mF.S = false;
				mF.I &= 0;
				MultiFD.RAddF(mF, aL);			// aL += m
				MultiCD.ClearC(sum1);
				for (iF &= 0; iF < mF; iF++)
				{
					PowNC(x, iF, p1);			// p1 = x^i
					t &= a + iF;
					t &= t * t;
					PowNC(t, s/2, p2);			// p2 = ((a+i)^2)^s/2
					sum1 += p1/p2;
				}
			}
			if (x == 0)							// x == 0 and a not integer <= 0
			{
				iter = 1;						// (Return first term of series.)
				t &= a * a;
				PowNC(t, s/2, p1);				// p1 = Pow(a^2 s/2)
				MultiCD.DivC(MultiCD.cOne, p1, t);	// t = 1 / p1
				Calc.SetToN(t, zI);					// result = 1 / p1
				goto EndIt;
			}
			// General case.
			//
			// For x <= 0.5 van Wijngaarden transformation is not used
			// [hence no calls to Aj()].
			bool zAbsLEH = MultiCD.cMHalf <= x && x <= MultiCD.cHalf;	// Abs(x) <= 0.5
			bool xValLEH = x <= MultiCD.cHalf;							// x <= 0.5
			if (xValLEH)			// if (x <= 0.5)	// Direct summation and CNCT (x < -0.5) case.
			{
				PowNC(aL, s, p1);						// p1 = Pow(aL, s)
				MultiCD.DivC(MultiCD.cOne, p1, omega);	// omega = 1 / p1
			}
			else										// CNCT (x > 0.5) case.
			{
				flag = XCCalLst.Aj(x, s, aL, 0, omega, name);
				if (flag != 0)
				{
					MultiID.MuWriteErr(name + ": Overflow in Aj()");
					Calc.ClearN(zI);		// result = 0.0
					iter = 0;
					goto EndIt;
				}
			}
			num = new MultiC[imax];			// Allocate memory for working arrays.
			den = new MultiC[imax];
			for (i = 0; i < imax; i++)
			{
				num[i] = new MultiC(Calc.fMC);
				den[i] = new MultiC(Calc.fMC);
				MultiID.MuInterrupt();
			}
			if (!xValLEH)					// x > 0.5
			{
				StoreAj = new MultiC[imax];	// StoreAj is used only in CNCT
				for (i = 0; i < imax; i++)
				{
					StoreAj[i] = new MultiC(Calc.fMC);
					MultiID.MuInterrupt();
				}
			}
			else StoreAj = new MultiC[0];
			flag = 0;
			sign = -1;
			for (i = 0; true; i++)	// Main loop: iterations for S^k_n.
			{						// i points to current iterate.
				iF &= i;
				sign = -sign;		// Increment the sum.
				sn += omega;		// Next term: omega.
				if (xValLEH)		// if (x <= 0.5) // "Direct summation".
				{
					t &= aL + iF;
					PowNC(t / (t+1), s, p2);
					omega &= x * p2 * omega;
				}
				else				// CNCT (x > 0.5) case.
				{
					StoreAj[i] = sign * omega;
					if ((i & 1) == 0)			// !Odd
					{
						PowNC(x, iF/2, p1);		// Recurrence for odd pointer i.
						PowNC(aL + iF/2, s, p2);
						omega &= MultiCD.cHalf * (StoreAj[i/2] - p1 / p2);
						if (sign > 0) MultiCD.ChSC(omega);	// * -sign
					}
					else
					{
						flag = XCCalLst.Aj(x, s, aL, i+1, omega, name);
						if (flag != 0)			// flag = 4
						{
							MultiID.MuWriteErr(name + ": Overflow in Aj()");
							Calc.ClearN(zI);	// result = 0.0
							iter = 0;
							goto EndIt;
						}
						else  omega &= -sign * omega;
					}
				}
				// Direct summation case: store current sum and remainder estimate.
				if (zAbsLEH)						// Abs(x) <= 0.5
				{
					skn &= sn;
				}
				else	// CNCT case.
				{			
					if (MultiCD.ZTestC(omega))		// Make sure omega is representable machine number.
					{
						flag = 5;
						if (!MultiID.muAbort && !MultiID.muErr)
							MultiID.MuWriteErr(name + ": Underflow in remainder estimate omega in LerchPhi()");
						Calc.ClearN(zI);			// result = 0.0
						iter = i+1;
						goto EndIt;
					}
					else
						MultiCD.DivC(MultiCD.cOne, omega, iom);	// iom = 1.0 / omega;
					num[i] &= sn * iom;				// Last terms in sums of numerator and denominator of i-th partial sum.
					den[i] &= iom;
					if (i > 0)						// Recurrence computation of numerator and denominator of a S_k^n.
					{
						num[i-1] &= num[i] - num[i-1];
						den[i-1] &= den[i] - den[i-1];
					}
					factor1 &= iF * (iF - 1);		// (beta+n+i-1) * (beta+n+i-2)
					for(j = 2; j <= i; j++)
					{
						t &= i + j - 1;							// beta = 1, n = 0
						MultiCD.DivC(factor1, t, factor);		// factor = factor1 / (beta+n+i+j-2)
						t &= i + j - 2;
						MultiCD.RDivC(t, factor);				// factor = factor1 / (beta+n+i+j-2) / (beta+n+i+j-3)
						num[i-j] &= num[i-j+1] - factor * num[i-j];
						den[i-j] &= den[i-j+1] - factor * den[i-j];
					}
					skn &= num[0] / den[0];			// Current approximation of the sum S_k^n.
				} // else CNCT case.
				MultiCD.SubC(skn, skn0, eps);		// eps = Abs(skn - skn0)
				MultiCD.Abs(eps, eps);
				if (i > 2)
				{
					if ((skn.N + skn.C) - (eps.N + eps.C) >= Calc.fMC - 2) break;
					if (-(eps.N + eps.C) >= Calc.fMC - 3) break;
					// Check the two termination criteria,
					if (MultiFD.ZTestF(eps)) break;		// Successive iterates skn are the same.
				}
				if (eps < minEps)					// save best answer so far
				{
					imin = i + 1;
					minEps &= eps;
					best &= skn;
				}
				if (i > imax-2 || i >= imin + 99)	// Maximum number of iterations is exceeded.
				{
					flag = 6;
					MultiID.WriteLn(name + ": No convergence within the maximum number of iterations = " + imax);
					MultiID.WriteLn(name + ": Results at iteration " + imin + " had a relative error approximately =");
					int sFMC = Calc.fMC;
					Calc.fMC = 2;
					MultiCD.WriteFLn(minEps / best);
					Calc.fMC = sFMC;
					skn &= best;					// result = best;
					break;
				}
				skn0 &= skn;			// Go on to the next iteration.
				eps0 &= eps;
				MultiID.MuInterrupt();
				if (MultiID.statReq2)
				{
					Common.SingleSpace();
					MultiID.WriteT(name + ": Iteration = " + i + " of " + (imax-2));
					MultiID.TimeDel();
					MultiID.statReq2 = false;
				}
				if (MultiID.muAbort || MultiID.muErr)
				{
					Calc.ClearN(zI);
					iter = i+1;
					goto EndIt;
				}
			} // for
			if (a.S)					// Store the resulting sum.
			{
				PowNC(x, mF, p1);		// p1 = x^m (m is a integer > 0)
				skn *= p1;				// skn = skn * p1
				skn += sum1;			// result = skn = sum1 + skn * p1
			}
			iter = i + 1;				// Store the number of iterations.
			if (flag == 0 && flagL == 2)
			{
				PowNC(x, n+1, p1);
				skn *= p1;
				for (MultiCD.ClearC(iF); iF < n; iF++)
				{
					PowNC(x, iF, p1);
					PowNC(n-iF, s, p2);
					skn += p1 / p2;			// result += x^k / (n - k)^s
				}
			}
			Calc.SetToN(skn, zI);		// result = skn;
EndIt:
			if (flag == 0 && flagL == 2)
				MultiID.WriteLn(name + ": Integer a <= 0, one infinite term ignored");
			StoreAj = new MultiC[0];
			den = new MultiC[0];
			num = new MultiC[0];
			MultiCD.DoneC(minEps);
			MultiCD.DoneC(best);
			MultiCD.DoneC(sum1);
			MultiCD.DoneC(sn);
			MultiCD.DoneC(skn0);
			MultiCD.DoneC(skn);
			MultiCD.DoneC(p2);
			MultiCD.DoneC(p1);
			MultiCD.DoneC(omega);
			MultiCD.DoneC(mF);
			MultiCD.DoneC(iom);
			MultiCD.DoneC(iF);
			MultiCD.DoneC(factor1);
			MultiCD.DoneC(factor);
			MultiCD.DoneC(eps0);
			MultiCD.DoneC(eps);
			MultiCD.DoneC(cacc);
			MultiCD.DoneC(aL);
			EndItEarly:
				MultiCD.DoneC(t);
			MultiCD.DoneC(n);
			MultiCD.DoneC(aL);
			MultiID.ADiag(name + ": Iteration = " + iter);
			MultiFD.expanded = saveExp;
			Calc.fMC = saveFMC;
			MultiID.mMax = saveMMax;
			Calc.NormN(zI);
		} // Lerch

		public static void GamN(MultiC x, int zI) // z = gamma(x)
		// x may have the same address in memory as z
		// x is not changed unless it shares memory with z
		// sets muErr true if overflow or error
		// z must be large enough ... (See FacN)
		{
			if (MultiFD.ZTestF(x.I))
			{
				XPCalLst.GamN(x, zI);
				MultiFD.ClearF(Common.nList[zI].I);
				Calc.NormN(zI);
				return;
			}
			MultiF lim = new MultiF(4);
			MultiF limM = new MultiF(4);
			int j;
			MultiFD.ValueF10("+4.05997,83000,01705,88908,81E+18", out j, lim);
			MultiFD.ValueF10("-4.05997,83000,01705,88908,81E+18", out j, limM);
			if (x > lim || x < limM)
			{
				MultiID.MuWriteErr("Number too large for gamma function");
				Calc.ClearN(zI);
				MultiFD.DoneF(limM);
				MultiFD.DoneF(lim);
				return;
			}
			bool saveExp = MultiFD.expanded;
			int saveFMC = Calc.fMC;
			int saveMMax = MultiID.mMax;
			XPCalLst.Expand();
			bool degModeSv = Calc.degMode;
			Calc.degMode = false;
			lim &= -5;
			bool reflection = (x < lim);				// if x < -5, use reflection
			MultiFD.DoneF(limM);
			MultiFD.DoneF(lim);
			MultiC xr = new MultiC(Calc.fMC);
			MultiC xx = new MultiC(Calc.fMC);
			if (reflection)
				xr &= 1 - x;							// x = 1 - x, use reflection formula
			else
				xr &= x;
			double digits = Calc.fMC * MultiID.muDMax;
			int n = (int)(Math.Pow(digits, 1.5) / 13.0) + 1;
			int iMax = (int)(0.6 * digits) + 20;
			xx &= xr;
			xx -= n;
			MultiFD.ChSF(xx);						// xx = n - xr
			if (xx.S)
				n = 0;
			else
				MultiFD.Get1F(out n, xx);
			xx &= xr + n;								// xx = xr + n
			MultiCD.DoneC(xr);
			MultiC lnG = new MultiC(Calc.fMC);			// lnG = (xx-0.5) * ln(xx) - xx + 0.5 * ln(2*pi)
			lnG &= xx;									// lnG = xx
			lnG -= MultiCD.cHalf;						// lnG = xx - 0.5
			int tI;
			Common.InitNAdd(1, "GamN-tI", out tI);		// t = temp on the list
			Calc.SetToN(xx, tI);
			RLnN(tI);									// t = ln(xx)
			MultiCD.MulC(lnG, Common.nList[tI], lnG);	// lnG = (xx-0.5) * ln(xx)
			lnG -= xx;									// - xx
			int pI = XPCalLst.GetPiPtr();
			XCCalLst.MulN(MultiFD.fTwo, Common.nList[pI], tI);	// t = 2 * Pi
			RLnN(tI);									// t = ln(2*Pi)
			XCCalLst.MulN(MultiFD.fHalf, Common.nList[tI], tI);	// t = 0.5 * ln(2*Pi)
			lnG += Common.nList[tI];
			MultiF j2 = new MultiF(1);
			j2.V[0] = 2;								// j2 = 2
			MultiF j2M = new MultiF(1);
			j2M.V[0] = 1;								// j2M = j2 - 1 = 1
			MultiC xP = new MultiC(Calc.fMC);
			xP &= xx;									// xP = xx
			MultiC xSq = new MultiC(Calc.fMC);
			MultiCD.MulC(xx, xx, xSq);					// xSq = xx^2
			int i = Math.Min(iMax, Calc.limBern);
			Common.GenBernI(i);
			MultiC div = new MultiC(Calc.fMC);
			MultiC lnGn = new MultiC(Calc.fMC);
			bool done = false;
			do
			{
				MultiID.MuInterrupt();
				Calc.Bern(j2, tI);				// t = B(j2)
				MultiFD.MulF(j2, j2M, div);		// div = j2 * (j2 - 1)
				MultiFD.ClearF(div.I);			// div.i = 0
				MultiCD.MulC(xP, div, div);		// div = j2 * (j2 - 1) * xP
				MultiCD.DivC(Common.nList[tI], div, div);	// div = t / div
				MultiCD.AddC(lnG, div, lnGn);	// lnG new = lnG + new term
				MultiCD.SubC(lnGn, lnG, div);	// div = amount lnG changed
				if (MultiCD.ZTestC(div))
					done = true;
				else
				{
					MultiCD.SetToC(lnGn, lnG);
					MultiCD.MulC(xP, xSq, xP);
					MultiFD.AddF(j2, MultiFD.fTwo, j2);
					MultiFD.AddF(j2M, MultiFD.fTwo, j2M);
				}
			}
			while (!done && !MultiID.muAbort && !MultiID.muErr);
			if (MultiID.diagOn)
			{
				int j2i;
				MultiFD.Get1F(out j2i, j2);
				MultiID.WriteLn("Used Bernoulli number B(" + j2i + ")");
			}
			XCCalLst.ExpN(lnG, tI);					// t = Exp(lnG) = gamma(xx)
			div &= MultiCD.cOne;
			for (i = 0; i < n; i++)
			{
				MultiFD.Add1F(xx, -1, xx);	// // xx = xx - 1
				MultiCD.MulC(div, xx, div);	// div = div * xx
				MultiID.MuInterrupt();
				if (MultiID.muAbort || MultiID.muErr)
				{
					Calc.ClearN(tI);
					div &= 1;
					break; // exit for
				}
			}
			MultiFD.DoneF(lnGn);
			XCCalLst.RDivN(div, tI);	// t = gamma(xr) = gamma(xx) / div
			MultiCD.DoneC(div);
			MultiCD.DoneC(xSq);
			MultiCD.DoneC(xP);
			MultiFD.DoneF(j2M);
			MultiFD.DoneF(j2);
			if (reflection)
			{
				int t2I;										// x < 0
				Common.InitNAdd(1, "GamN-t2I", out t2I);		// t2 = temp on the list
				MulN(x, Common.nList[pI], t2I);				// t2 = x * Pi
				RSinN(t2I);									// t2 = sin(x * Pi)
				RMulN(Common.nList[t2I], tI);				// t = gamma(xr) * sin(x * Pi)
				DivN(Common.nList[pI], Common.nList[tI], tI);	// t = Pi / t = gamma(x)
				Common.LDelete(t2I);
			}
			Calc.SetToN(Common.nList[tI], zI);
			Common.LDelete(tI);
			MultiFD.DoneF(lnG);
			MultiFD.DoneF(xx);
			Calc.degMode = degModeSv;
			MultiFD.expanded = saveExp;
			Calc.fMC = saveFMC;
			MultiID.mMax = saveMMax;
			Calc.NormN(zI);
		} // GamN

		public static void BetaN(MultiC x, int zI) // z = beta(x)
		// x may have the same address in memory as z
		// x is not changed unless it shares memory with z
		// sets muErr true if overflow or error
		// z must be large enough ... (See FacN)
		{
			if (MultiFD.ZTestF(x.I))
			{
				XPCalLst.BetaN(x, zI);
				MultiFD.ClearF(Common.nList[zI].I);
				Calc.NormN(zI);
				return;
			}
			bool saveExp = MultiFD.expanded;
			int saveFMC = Calc.fMC;
			int saveMMax = MultiID.mMax;
			XPCalLst.Expand();
			XPCalLst.ExpandMore(MultiFD.fGuard);
			bool degModeSv = Calc.degMode;
			Calc.degMode = false;
			MultiC xx = new MultiC(Calc.fMC);
			MultiF lim = new MultiF(1);
			// lim &= -90000000; // diag
			// lim &= -5; // diag
			// lim &= -313;								// for 400 m -313 is best, for 800 m -477 is best ... (for time)
			lim &= MultiFD.fHalf;
			bool reflection = (x < lim);				// if x < lim, use reflection
			// reflection = ((x.V[0] & 3) == 0);		// diag
			MultiFD.DoneF(lim);
			if (reflection)
				xx &= 1 - x;							// x = 1 - x, use reflection formula
			else
				xx &= x;
			// if (reflection) MultiID.WriteLn("Reflection"); // diag
			MultiC d = new MultiC(Calc.fMC);
			MultiC sum = new MultiC(Calc.fMC);				// beta = Sum ...
			MultiC sumN = new MultiC(Calc.fMC);
			double digits = Calc.fMC * MultiID.muDMax;
			int n = (int)((digits / 300.0) * digits + 5.0);
			if (n < 300) n = 300;
			MultiF nF = new MultiF(1);						// n floating
			nF &= n;
			int j = 4 * n + 1;
			MultiC iF = new MultiC(1);						// i complex floating
			MultiF jF = new MultiF(1);						// j floating
			jF &= j;
			sum &= 0;
			MultiID.ADiag("BetaN: n = " + n);
			MultiCD.ChSC(xx);								// x = -x for now
			bool sign = false;
			int tI;
			Common.InitNAdd(1, "BetaN-tI", out tI);			// t = temp on the list
			for (int i = 3; i < j; i += 2)
			{
				MultiID.MuInterrupt();
				if (MultiID.muAbort || MultiID.muErr) break;
				sign = !sign;
				iF &= i;
				PowN(iF, xx, tI);							// t = i ^ -x
				if (sign)
					MultiCD.RSubC(Common.nList[tI], sum);		// sum = sum - t
				else
					MultiCD.RAddC(Common.nList[tI], sum);		// sum = sum + t
				if (MultiID.statReq2)
				{
					MultiID.nDone = i - 2;
					MultiID.nToGo = j - i - 1;
					MultiID.WritenDone("BetaN:");
				}
			}
			iF &= j;
			PowN(iF, xx, tI);								// t = i ^ -x
			RMulN(MultiCD.cHalf, tI);						// 1/2J^x
			MultiCD.RAddC(Common.nList[tI], sum);			// sum = sum +/- t
			MultiCD.ChSC(xx);								// x = +x now
			MultiC twoToK = new MultiC(Calc.fMC);
			MultiC xK = new MultiC(Calc.fMC);
			MultiC xKP = new MultiC(Calc.fMC);
			MultiC kF = new MultiC(Calc.fMC);
			MultiC term = new MultiC(Calc.fMC);
			twoToK &= 4;
			xK &= xx;
			xK += 1;
			PowN(iF, xK, tI);				// t = j ^ (x + 1)
			MultiCD.Mul1C(Common.nList[tI], 4, d);			// d = 4 * (j ^ (x + 1)), denominator 2*k!*(J^(k+x-1)) with k = 2
			xK &= xx;										// xK = x + k - 2
			xKP &= xK;										// xKP = x * (x+1) * (x+2) * ... * (x+k-2)
			Common.GenBernI(Calc.limBern);					// generate Bernoulli numbers
			for (int k = 2; ; k += 2)
			{
				MultiID.MuInterrupt();
				if (MultiID.muAbort || MultiID.muErr) break;
				kF &= k;
				Calc.Bern(kF, tI);							// t = B(k) = k-th Bernoulli number
				MultiCD.SubC(twoToK, MultiCD.cOne, term);
				MultiCD.MulC(term, twoToK, term);
				MultiCD.MulC(term, xKP, term);
				MultiCD.AddC(xK, MultiCD.cOne, xK);
				MultiCD.MulC(xK, xKP, xKP);
				if (k > 2)
				{
					MultiCD.MulC(term, xK, term);
					MultiCD.AddC(xK, MultiCD.cOne, xK);
					MultiCD.MulC(xK, xKP, xKP);
				}
				MultiCD.MulC(term, Common.nList[tI], term);
				MultiCD.DivC(term, d, term);
				MultiCD.AddC(sum, term, sumN);				// sumN = sum + t = new sum
				if (sum == sumN)
				{
					MultiID.ADiag("Used Bernoulli number B(" + k + ")");
					k++; // diag
					break;
				}
				MultiCD.SetToC(sumN, sum);
				MultiCD.MulC(twoToK, MultiCD.cFour, twoToK);	// twoToK = 2^k
				MultiCD.MulC(d, iF, d);							// d = d * j
				MultiCD.MulC(d, iF, d);							// d = d * j
				kF &= k + 1;
				MultiCD.MulC(d, kF, d);							// d = d * (k+1)
				kF &= k + 2;
				MultiCD.MulC(d, kF, d);							// d = d * (k+2)
			}
			MultiCD.DoneC(term);
			MultiCD.DoneC(kF);
			MultiCD.DoneC(xKP);
			MultiCD.DoneC(xK);
			MultiCD.DoneC(twoToK);
			MultiFD.DoneF(jF);
			MultiCD.DoneC(iF);
			MultiFD.DoneF(nF);
			sum += 1;
			if (reflection)
			{
				int pI = XPCalLst.GetPiPtr();
				MulN(xx, MultiCD.cHalf, tI);
				RMulN(Common.nList[pI], tI);						// t = Pi * x / 2
				RSinN(tI);											// t = sin(Pi * x / 2)
				MultiCD.MulC(Common.nList[tI], sum, sum);			// sum = t * sum
				MultiFD.DivF(MultiFD.fTwo, Common.nList[pI], d);	// 2 / Pi
				MultiFD.ClearF(d.I);								// d.i = 0
				PowN(d, xx, tI);									// t = (2 / Pi) ^ x
				MultiCD.MulC(Common.nList[tI], sum, sum);			// sum = t * sum
				GamN(xx, tI);										// t = gamma(x)
				MultiCD.MulC(Common.nList[tI], sum, sum);			// sum = t * sum
			}
			Calc.SetToN(sum, zI);
			Common.LDelete(tI);
			MultiCD.DoneC(sumN);
			MultiCD.DoneC(sum);
			MultiCD.DoneC(d);
			MultiCD.DoneC(xx);
			if (MultiID.muAbort || MultiID.muErr) Calc.ClearN(zI);
			Calc.degMode = degModeSv;
			MultiFD.expanded = saveExp;
			Calc.fMC = saveFMC;
			MultiID.mMax = saveMMax;
			Calc.NormN(zI);
		} // BetaN

		public static void BetaL(MultiC x, MultiC z)
		// z = Beta(x) by prime numbers for large |x|
		{
			if (MultiFD.ZTestF(x.I))
			{
				XPCalLst.BetaL(x, z);
				MultiFD.ClearF(z.I);
				Calc.NormNM(z);
				return;
			}
			int tI, t2I;
			MultiC xL = new MultiC(x.N, x.I.N);
			MultiCD.SetToC(x, xL);
			MultiF lim = new MultiF(1);
			lim &= 100;
			Common.InitNAdd(1, "BetaL-tI", out tI);
			if (x <= lim && x >= -lim)
			{
				MultiID.WriteLn("BetaL: |X.r| <= 100, using Beta(X)");
				BetaN(x, tI);
				Calc.SetToNM(Common.nList[tI], z);
				Common.LDelete(tI);
				Calc.NormNM(z);
				return;
			}
			bool saveExp = MultiFD.expanded;
			int saveFMC = Calc.fMC;
			int saveMMax = MultiID.mMax;
			XPCalLst.Expand();
			bool reflection = (x < MultiFD.fHalf);
			if (reflection)								// if x.r < 1/2, use reflection
			{
				MultiC xx = new MultiC(Calc.fMC, Calc.fMC);
				MultiC be = new MultiC(Calc.fMC, Calc.fMC);
				MultiC zL = new MultiC(Calc.fMC, Calc.fMC);
				xx &= 1 - x;							// xx = 1 - x, use reflection formula
				BetaL(xx, be);							// be can have a reduced .M
				MultiCD.SetToC(be, zL);
				MultiCD.DoneC(be);
				if (!MultiID.muAbort && !MultiID.muErr)
				{
					Calc.SetToN(xx, tI);
					XPCalLst.Mod(xx, MultiFD.fFour, Common.nList[tI]);	// xx.r > 0
					bool degModeSave = Calc.degMode;
					Calc.degMode = true;
					RMul1N(90, tI);										// t = Pi * x / 2
					RSinN(tI);											// t = sin(Pi * x / 2)
					MultiCD.MulC(Common.nList[tI], zL, zL);				// z = t * z
					Calc.degMode = degModeSave;
					int pI = XPCalLst.GetPiPtr();
					MultiC d = new MultiC(Calc.fMC, Calc.fMC);
					MultiCD.DivC(MultiCD.cTwo, Common.nList[pI], d);	// 2 / Pi
					PowN(d, xx, tI);									// t = (2 / Pi) ^ x
					MultiCD.MulC(Common.nList[tI], zL, zL);				// z = t * zm
					GamN(xx, tI);										// t = gamma(x)
					MultiCD.MulC(Common.nList[tI], zL, zL);				// z = t * z
				}
				Common.LDelete(tI);
				Calc.SetToNM(zL, z);
				MultiFD.expanded = saveExp;
				Calc.fMC = saveFMC;
				MultiID.mMax = saveMMax;
				Calc.NormNM(z);
				return;
			}
			Common.InitNAdd(1, "BetaL-t2I", out t2I);
			double p;				// current prime
			int i;					// index for table of primes
			int mC = 105000;		// number of primes each call of function NPrime
			int iQSize = 5500000;	// size of IQ and JQ arrays, must be large enough
			new NPrimes();			// initialize class NPrimes
			new NPrimes(mC, iQSize);	// initialize NPrimes package
			MultiC pF = new MultiC(2);
			MultiC pPri = new MultiC(Calc.fMC, Calc.fMC);
			MultiC pPow = new MultiC(Calc.fMC, Calc.fMC);
			pPri &= 1;
			pPow &= 1;
			p = 2.0;
			bool done = false;
			for (; ; )
			{
				NPrimes.NPrime(mC, false);		// generate mC more primes in iP[1] thru iP[mC]
				for (i = 1; i <= mC; i++)
				{
					p = NPrimes.iP[i];
					if (p == 2.0)
						continue;
					pF &= p;
					PowN(pF, xL, tI);							// t = p^n
					long rem;
					Math.DivRem((long)p, 4, out rem);
					if (rem == 1)
						SubN(Common.nList[tI], MultiCD.cOne, t2I);	// t2 = p^n - 1
					else
						AddN(Common.nList[tI], MultiCD.cOne, t2I);	// t2 = p^n + 1
					if (MultiCD.CompCC(Common.nList[tI], Common.nList[t2I]) == 0) // if t == t2
					{
						done = true;
						break;
					}
					DivN(Common.nList[tI], Common.nList[t2I], tI);	// t = p^n / p^n +/- 1
					MultiCD.MulC(Common.nList[tI], pPow, pPow);			// pPow = Prod[p^n / p^n +/- 1]
					if (MultiID.muAbort || MultiID.muErr)
						break;
					if (MultiID.diagOn || MultiID.statReq2)
					{
						MultiID.WriteLn("BetaL: Using prime " + p);
						//MultiID.WriteLn("BetaL: Using prime " + p + " <= " + nn + ", " +
						//	Math.Floor(100.0 * p / nn).ToString() + "% done");
						MultiID.statReq2 = false;
					}
				}
				MultiID.MuInterrupt();	// allow for abort
				if (MultiID.statReq3)
				{
					long totalMemory = GC.GetTotalMemory(false);
					//string st = "BetaL: Prime = " + p + ", NN = " + nn + ", ";
					//st += Math.Floor(100.0 * p / nn).ToString() + "% done, Memory in use = " + totalMemory;
					string st = "BetaL: Prime = " + p + ", Memory in use = " + totalMemory;
					MultiID.WriteLn(st);
					st = "";
				}
				if (MultiID.muAbort || MultiID.muErr)
				{
					new NPrimes(0, 0);			// release memory
					Calc.SetToNM(MultiCD.cZero, z);
					MultiID.statReq3 = false;
					Common.LDelete(t2I);
					Common.LDelete(tI);
					MultiFD.expanded = saveExp;
					Calc.fMC = saveFMC;
					MultiID.mMax = saveMMax;
					return;
				}
				if (done)
					break;
				if (MultiID.diagOn || MultiID.statReq3)
				{
					MultiID.WriteLn("BetaL: Getting more primes");
					MultiID.statReq3 = false;
				}
			}
			Common.LDelete(t2I);
			Common.LDelete(tI);
			new NPrimes(0, 0);				// release memory
			Calc.SetToNM(pPow, z);
			MultiFD.expanded = saveExp;
			Calc.fMC = saveFMC;
			MultiID.mMax = saveMMax;
			Calc.NormNM(z);
		} // BetaL

		public static void EulerL(MultiC x, MultiC z)
		// z = E(x), the generalized Euler number of x, |x| "large"
		{
			if (x.I == MultiFD.fZero)	// if x is real and x >= 0 and x is an integer, use Euler(n)
			{
				MultiFD.ClearF(z.I);
				XPCalLst.EulerL(x, z);
				Calc.NormNM(z);
				return;
			}
			bool saveExp = MultiFD.expanded;
			int saveFMC = Calc.fMC;
			int saveMMax = MultiID.mMax;
			XPCalLst.Expand();
			XCCalLst.BetaL(-x, z);						// z can have a reduced .M
			MultiC zL = new MultiC(z.N + 1, z.I.N + 1);
			MultiCD.MulC(MultiCD.cTwo, z, zL);			// E(x) = 2 * Beta(-x)
			Calc.SetToNM(zL, z);
			MultiFD.expanded = saveExp;
			Calc.fMC = saveFMC;
			MultiID.mMax = saveMMax;
			Calc.NormNM(z);
		} // EulerL

		public static void EtaN(MultiC x, int zI) // z = eta(x), eta(4) = 7*Pi^4/720
		// x may have the same address in memory as z
		// x is not changed unless it shares memory with z
		// sets muErr true if overflow or error
		// z must be large enough ... (See FacN)
		{
			if (MultiFD.ZTestF(x.I))
			{
				XPCalLst.EtaN(x, zI);
				MultiFD.ClearF(Common.nList[zI].I);
				Calc.NormN(zI);
				return;
			}
			bool saveExp = MultiFD.expanded;
			int saveFMC = Calc.fMC;
			int saveMMax = MultiID.mMax;
			XPCalLst.Expand();
			bool degModeSv = Calc.degMode;
			Calc.degMode = false;
			MultiC xx = new MultiC(Calc.fMC);
			MultiF lim = new MultiF(1);
			lim &= -4;
			bool reflection = (x < lim);					// if x < -4, use reflection
			MultiFD.DoneF(lim);
			if (reflection)
				xx &= 1 - x;								// x = 1 - x, use reflection formula
			else
				xx &= x;
			double digits = Calc.fMC * MultiID.muDMax;
			int n = (int)(1.3 * digits) + 5;
			MultiF nF = new MultiF(2);						// n floating
			nF &= n;
			MultiC sum = new MultiC(Calc.fMC);				// eta = Sum ...
			MultiC kF = new MultiC(2, 1);						// k floating = 0
			int k;
			bool sign = false;								// (-1)^(k-1)
			int tI;
			Common.InitNAdd(1, "EtaN-tI", out tI);			// t = temp on the list
			MultiCD.ChSC(xx);								// x = -x
			MultiF c = new MultiF(Calc.fMC);				// c = C(n, j)
			c &= 1;
			MultiF e = new MultiF(Calc.fMC);				// e = Sum C(n, j)
			e &= 1;
			MultiF nM = new MultiF(2);						// nM = n - j + 1
			nM &= n;
			MultiF jF = new MultiF(2);						// j floating
			jF &= 1;
			MultiC d = new MultiC(Calc.fMC);				// d = delta for sum
			for (k = n+n; k > n; k--)
			{
				MultiID.MuInterrupt();
				if (MultiID.muAbort || MultiID.muErr) break;
				if (MultiID.statReq2)
				{
					MultiID.nDone = n+n-k;
					MultiID.nToGo = k;
					MultiID.WritenDone("EtaN:");
				}
				sign = !sign;
				MultiCD.SetTo1C(k, kF);
				PowN(kF, xx, tI);							// t = k ^ -x
				MultiCD.MulCF(Common.nList[tI], e, d);		// d = e * t
				if (sign)
					MultiCD.RSubC(d, sum);					// sum = sum - d	(d * (-1) ^ (k-1))
				else
					MultiCD.RAddC(d, sum);					// sum = sum + d
				MultiFD.MulF(c, nM, c);						// c = c * (n - j + 1)
				MultiFD.DivF(c, jF, c);						// c = c / j
				MultiFD.AddF(e, c, e);						// e = sum C(n, j)
				MultiFD.SubF(nM, MultiFD.fOne, nM);			// nM = nM - 1
				MultiFD.AddF(jF, MultiFD.fOne, jF);			// jF = jF + 1
			}
			MultiCD.DoneC(d);
			int tpI;
			Common.InitNAdd(1, "EtaN-tpI", out tpI);		// tp = temp on the list
			XPCalLst.PowN(MultiFD.fTwo, nF, tpI);			// tp = 2 ^ n
			MultiFD.ClearF(Common.nList[tpI].I);			// [tpI].i = 0
			MultiCD.DivC(sum, Common.nList[tpI], sum);		// sum = sum / (2 ^ n)
			Common.LDelete(tpI);
			for (k = n; k > 0; k--)
			{
				MultiID.MuInterrupt();
				if (MultiID.muAbort || MultiID.muErr) break;
				if (MultiID.statReq2)
				{
					MultiID.nDone = n+n-k;
					MultiID.nToGo = k;
					MultiID.WritenDone("EtaN:");
				}
				sign = !sign;
				MultiFD.SetTo1F(k, kF);
				PowN(kF, xx, tI);							// t = k ^ -x
				if (sign)
					MultiCD.RSubC(Common.nList[tI], sum);	// sum = sum - (k ^ -x)	(* (-1) ^ (k-1))
				else
					MultiCD.RAddC(Common.nList[tI], sum);	// sum = sum + (k ^ -x)
			}
			MultiFD.DoneF(jF);
			MultiFD.DoneF(nM);
			MultiFD.DoneF(e);
			MultiCD.DoneC(kF);
			if (reflection)
			{
				int pI = XPCalLst.GetPiPtr();
				MulN(x, MultiCD.cHalf, tI);
				RMulN(Common.nList[pI], tI);					// t = Pi * x / 2
				RSinN(tI);										// t = sin(Pi * x / 2)
				MultiCD.MulC(Common.nList[tI], sum, sum);		// sum = t * sum
				MultiC f = new MultiC(Calc.fMC);				// f = 2 * pi
				MultiFD.MulF(MultiFD.fTwo, Common.nList[pI], f);// f = 2 * Pi
				MultiFD.ClearF(f.I);
				PowN(f, xx, tI);								// t = (2 * Pi) ^ (x-1)
				MultiCD.MulC(Common.nList[tI], sum, sum);		// sum = t * sum
				MultiCD.ChSC(xx);								// xx = +(1-x)
				GamN(xx, tI);									// t = Gamma(1-x)
				MultiCD.MulC(Common.nList[tI], sum, sum);		// sum = t * sum
				PowN(MultiCD.cTwo, xx, tI);						// t = 2 ^ (1-x)
				MultiCD.SubC(MultiCD.cOne, Common.nList[tI], f);// f = 1 - 2^(1-x)
				MultiCD.MulC(f, sum, sum);						// sum = f * sum
				RInvN(tI);										// t = 2 ^ (x-1)
				MultiCD.SubC(MultiCD.cHalf, Common.nList[tI], f);// f = .5 - 2^(x-1)
				MultiCD.DivC(sum, f, sum);						// sum = sum / f
				MultiCD.DoneC(f);
			}
			Calc.SetToN(sum, zI);
			Common.LDelete(tI);
			MultiFD.DoneF(nF);
			MultiCD.DoneC(xx);
			Calc.degMode = degModeSv;
			MultiFD.expanded = saveExp;
			Calc.fMC = saveFMC;
			MultiID.mMax = saveMMax;
			Calc.NormN(zI);
		} // EtaN

		public static void LambdaN(MultiC x, int zI) // z = lambda(x)
		// x may have the same address in memory as z
		// x is not changed unless it shares memory with z
		// sets muErr true if overflow or error
		// z must be large enough ... (See FacN)
		{
			if (MultiFD.ZTestF(x.I))
			{
				XPCalLst.LambdaN(x, zI);
				MultiFD.ClearF(Common.nList[zI].I);
				Calc.NormN(zI);
				return;
			}
			bool saveExp = MultiFD.expanded;
			int saveFMC = Calc.fMC;
			int saveMMax = MultiID.mMax;
			XPCalLst.Expand();
			int zetaI, tI;
			Common.InitNAdd(1, "LambdaN-zetaI", out zetaI);	// zeta = temp on the list
			ZetaN(x, zetaI);									// lambda(x) = Zeta(x) * (1 - 2^(-x))
			Common.InitNAdd(1, "LambdaN-t2I", out tI);			// t = temp on the list
			MultiCD.ChSC(x);									// x = -x for now
			PowN(MultiCD.cTwo, x, tI);							// t = 2 ^ (-x)
			MultiCD.ChSC(x);									// x = +x now
			MultiC f = new MultiC(Calc.fMC);
			MultiCD.SubC(MultiCD.cOne, Common.nList[tI], f);	// f = 1 - 2^(-x)
			MultiCD.MulC(Common.nList[zetaI], f, f);			// f = zeta * f
			Calc.SetToN(f, zI);									// lambda(x) = f
			MultiFD.DoneF(f);
			Common.LDelete(tI);
			Common.LDelete(zetaI);
			MultiFD.expanded = saveExp;
			Calc.fMC = saveFMC;
			MultiID.mMax = saveMMax;
			Calc.NormN(zI);
		} // LambdaN

		public static void ZetaN(MultiC x, int zI) // z = Zeta(x)
		// x may have the same address in memory as z
		// x is not changed unless it shares memory with z
		// sets muErr true if overflow or error
		// z must be large enough ... (See FacN)
		{
			if (MultiFD.ZTestF(x.I))
			{
				XPCalLst.ZetaN(x, zI);
				MultiFD.ClearF(Common.nList[zI].I);
				Calc.NormN(zI);
				return;
			}
			bool saveExp = MultiFD.expanded;
			int saveFMC = Calc.fMC;
			int saveMMax = MultiID.mMax;
			XPCalLst.Expand();
			int etaI, tI;
			Common.InitNAdd(1, "ZetaN-etaI", out etaI);		// eta = temp on the list
			EtaN(x, etaI);									// Zeta(x) = eta(x) / (1 - 2^(1-x))
			Common.InitNAdd(1, "ZetaN-t2I", out tI);		// t = temp on the list
			SubN(MultiCD.cOne, x, tI);						// t = 1 - x
			PowN(MultiCD.cTwo, Common.nList[tI], tI);		// t = 2 ^ (1-x)
			MultiC f = new MultiC(Calc.fMC);
			MultiCD.SubC(MultiCD.cOne, Common.nList[tI], f);	// f = 1 - 2^(1-x)
			MultiCD.DivC(Common.nList[etaI], f, f);				// f = eta / f
			Calc.SetToN(f, zI);									// Zeta(x) = f
			MultiCD.DoneC(f);
			Common.LDelete(tI);
			Common.LDelete(etaI);
			MultiFD.expanded = saveExp;
			Calc.fMC = saveFMC;
			MultiID.mMax = saveMMax;
			Calc.NormN(zI);
		} // ZetaN

		public static void PsiN(MultiC x, int zI) // z = digamma(x)
		// x may have the same address in memory as z
		// x is not changed unless it shares memory with z
		// sets muErr true if overflow or error
		// z must be large enough ... (See FacN)
		{
			if (MultiFD.ZTestF(x.I))
			{
				XPCalLst.PsiN(x, zI);
				MultiFD.ClearF(Common.nList[zI].I);
				Calc.NormN(zI);
				return;
			}
			// numbers are (almost) never too large for Psi
			bool saveExp = MultiFD.expanded;
			int saveFMC = Calc.fMC;
			int saveMMax = MultiID.mMax;
			XPCalLst.Expand();
			MultiF lim = new MultiF(1);
			lim &= -5;
			bool reflection = (x < lim);				// if x < -5, use reflection
			MultiFD.DoneF(lim);
			MultiC xr = new MultiC(Calc.fMC);
			MultiC xx = new MultiC(Calc.fMC);
			if (reflection)
				xr &= 1 - x;							// x = 1 - x, use reflection formula
			else
				xr &= x;
			double digits = Calc.fMC * MultiID.muDMax;
			double dN = (Math.Pow(digits, 1.5) / 13.0) + 1;
			if (dN > Int32.MaxValue)
				dN = Int32.MaxValue;
			int n = (int)dN;
			int iMax = (int)(0.6 * digits) + 20;
			xx &= xr;
			xx -= n;
			MultiFD.ChSF(xx);						// xx = n - xr
			if (xx.S)
				n = 0;
			else
				MultiFD.Get1F(out n, xx);
			xx &= xr + n;								// xx = xr + n
			MultiCD.DoneC(xr);
			MultiC psi = new MultiC(Calc.fMC);			// psi = ln(xx) - 1 / (2 * xx)
			MultiCD.DivC(MultiCD.cMHalf, xx, psi);		// -0.5 / xx
			int tI;
			Common.InitNAdd(1, "PsiN-tI", out tI);		// t = temp on the list
			XCCalLst.LnN(xx, tI);								// t = ln(xx)
			MultiCD.AddC(psi, Common.nList[tI], psi);	// psi = ln(xx) - 1 / (2 * xx)
			MultiF j2 = new MultiF(1);
			j2.V[0] = 2;								// j2 = 2
			MultiC xP = new MultiC(Calc.fMC);
			MultiC xSq = new MultiC(Calc.fMC);
			MultiCD.MulC(xx, xx, xSq);					// xSq = xx^2
			xP &= xSq;									// xP = xSq
			int i = Math.Min(iMax, Calc.limBern);
			Common.GenBernI(i);
			MultiC div = new MultiC(Calc.fMC);
			MultiC psiN = new MultiC(Calc.fMC);
			bool done = false;
			int k = 0;
			do
			{
				k++;
				MultiID.MuInterrupt();
				if (MultiID.statReq2)
				{
					MultiID.WriteT("PsiN: k = " + k);
					MultiID.TimeDel();
					MultiID.statReq2 = false;
				}
				Calc.Bern(j2, tI);							// t = B(j2)
				MultiCD.MulCF(xP, j2, div);					// div = j2 * xP
				MultiCD.DivC(Common.nList[tI], div, div);	// div = t / div
				MultiCD.SubC(psi, div, psiN);				// psi new = psi + new term
				MultiCD.SubC(psiN, psi, div);				// div = amount psi changed
				if (MultiCD.ZTestC(div))
					done = true;
				else
				{
					MultiCD.SetToC(psiN, psi);
					MultiCD.MulC(xP, xSq, xP);
					MultiFD.AddF(j2, MultiFD.fTwo, j2);
				}
			}
			while (!done && !MultiID.muAbort && !MultiID.muErr);
			if (MultiID.diagOn)
			{
				int j2i;
				MultiFD.Get1F(out j2i, j2);
				MultiID.WriteLn("Used Bernoulli number B(" + j2i + ")");
			}
			Calc.SetToN(psi, tI);							// t = psi
			div &= MultiCD.cOne;
			for (i = 0; i < n; i++)
			{
				MultiCD.Add1C(xx, -1, xx);					// xx = xx - 1
				MultiCD.DivC(MultiCD.cMOne, xx, div);		// div = -1 / xx
				MultiCD.AddC(psi, div, psi);
				MultiID.MuInterrupt();
				if (MultiID.muAbort || MultiID.muErr)
				{
					Calc.ClearN(tI);
					div &= 1;
					break; // exit for
				}
			}
			MultiCD.DoneC(psiN);
			MultiCD.DoneC(div);
			MultiCD.DoneC(xSq);
			MultiCD.DoneC(xP);
			MultiFD.DoneF(j2);
			if (reflection && !MultiID.muAbort && !MultiID.muErr)
			{
				int pI = XPCalLst.GetPiPtr();
				int t2I;
				Common.InitNAdd(1, "PsiN-t2I", out t2I);			// t2 = temp on the list
				bool degModeSave = Calc.degMode;
				Calc.degMode = false;
				MulN(x, Common.nList[pI], t2I);
				RTanN(t2I);										// t2 = tan(x * Pi)
				DivN(Common.nList[pI], Common.nList[t2I], tI);	// t = Pi / tan(x * Pi)
				Calc.degMode = degModeSave;
				MultiCD.AddC(psi, Common.nList[tI], psi);			// psi = psi + Pi / tan(x * Pi)
				Common.LDelete(t2I);
			}
			Calc.SetToN(psi, zI);
			if (MultiID.muAbort || MultiID.muErr) Calc.ClearN(zI);
			Common.LDelete(tI);
			MultiCD.DoneC(psi);
			MultiCD.DoneC(xx);
			MultiFD.expanded = saveExp;
			Calc.fMC = saveFMC;
			MultiID.mMax = saveMMax;
			Calc.NormN(zI);
		} // PsiN

		public static void RFacN(int zI) // "FAC" z = z! = 1*2*3*...*x
		// sets muErr true if overflow or error
		{
			XPCalLst.RFacN(zI);
			MultiFD.ClearF(Common.nList[zI].I);
			Calc.NormN(zI);
		}

		public static void RFac2N(int zI, int t2I) // "FAC2" z = z! = 1*2*3*...*x
		// z = Fac(a) / Fac(b) by binary splitting
		// a or b or both may have the same address in memory as z
		// a, b are not changed unless they share memory with z
		// sets muErr true if overflow or error
		{
			XPCalLst.Fac2N(Common.nList[zI], Common.nList[t2I], zI, false);
			MultiFD.ClearF(Common.nList[zI].I);
			Calc.NormN(zI);
		}

		public static void RFacMBN(int zI) // "FACM" z = z! = 1*2*3*...*x % FMB
		// sets muErr true if overflow or error
		{
			XPCalLst.RFacMBN(zI);
			MultiFD.ClearF(Common.nList[zI].I);
			Calc.NormN(zI);
		}

		public static void RFacMB2N(int zI, int t2I) // "FACM2"
		// z = Fac(a) / Fac(b) % fMB by binary splitting
		// a or b or both may have the same address in memory as z
		// a, b are not changed unless they share memory with z
		// sets muErr true if overflow or error
		{
			XPCalLst.Fac2N(Common.nList[zI], Common.nList[t2I], zI, true);
			MultiFD.ClearF(Common.nList[zI].I);
			Calc.NormN(zI);
		}

		public static void RFacMBSN(int zI) // "FACMS" z = z! = 1*2*3*...*x % FMB
		// sets muErr true if overflow or error
		{
			XPCalLst.RFacMBSN(zI);
			MultiFD.ClearF(Common.nList[zI].I);
			Calc.NormN(zI);
		}

		public static void RFacSN(int zI) // "FACS" z = z! = 1*2*3*...*x
		// sets muErr true if overflow or error
		{
			XPCalLst.RFacSN(zI);
			MultiFD.ClearF(Common.nList[zI].I);
			Calc.NormN(zI);
		}

		public static void RBetaN(int zI) // z = beta(z)
		// sets muErr true if overflow or error
		{
			BetaN(Common.nList[zI], zI);
		}
		
		public static void RErfN(int zI) // z = erf(z)
		// sets muErr true if overflow or error
		{
			ErfN(Common.nList[zI], zI);
		}

		public static void RErfCN(int zI) // z = erfc(z)
		// sets muErr true if overflow or error
		{
			ErfCN(Common.nList[zI], zI);
		}

		public static void REtaN(int zI) // z = eta(z)
		// sets muErr true if overflow or error
		{
			EtaN(Common.nList[zI], zI);
		}

		public static void ErfN(MultiC x, int zI)
		// z = error function erf(x)
		// x may have the same address in memory as zI
		// x is not changed unless it shares memory with zI
		// sets muErr true if overflow or error
		{
			bool saveExp = MultiFD.expanded;
			int saveFMC = Calc.fMC;
			int saveMMax = MultiID.mMax;
			XPCalLst.Expand();
			if (MultiFD.ZTestF(x.I))
			{
				XPCalLst.ErfNF(x, zI);
				MultiFD.ClearF(Common.nList[zI].I);
			}
			else
			{
				ErfCN(x, zI);
				RSubN(MultiCD.cOne, zI);
				Common.nList[zI].S = !Common.nList[zI].S;
				Common.nList[zI].I.S = !Common.nList[zI].I.S;
			}
			MultiFD.expanded = saveExp;
			Calc.fMC = saveFMC;
			MultiID.mMax = saveMMax;
			Calc.NormN(zI);
		} // ErfN

		public static void ErfCN(MultiC x, int zI)
		// z = complementary error function erfC(x)
		// x may have the same address in memory as zI
		// x is not changed unless it shares memory with zI
		// sets muErr true if overflow or error
		{
			// If x >= 0, erfc(x) = gamq(1/2, x^2), where gamq(x, y) is the upper regularized
			// incomplete gamma function (see the GamQ(X, Y) function). If x < 0, erfc(x) = 2 - erfc(-x).
			bool saveExp = MultiFD.expanded;
			int saveFMC = Calc.fMC;
			int saveMMax = MultiID.mMax;
			XPCalLst.Expand();
			if (MultiFD.ZTestF(x.I))
			{
				XPCalLst.ErfCNF(x, zI);
				MultiFD.ClearF(Common.nList[zI].I);
			}
			else
			{
				if (x > MultiCD.cZero)			// if x > 0, erfc(x) = gamq(1/2, x^2)
					GamQN(MultiCD.cHalf, x*x, zI);
				else								// if x < 0, erfc = 2 - erfc(-x)
				{									// note, i > 0, -i < 0
					ErfCN(-x, zI);
					RSubN(MultiCD.cTwo, zI);
					Common.nList[zI].S = !Common.nList[zI].S;
					Common.nList[zI].I.S = !Common.nList[zI].I.S;
				}
			}
			MultiFD.expanded = saveExp;
			Calc.fMC = saveFMC;
			MultiID.mMax = saveMMax;
			Calc.NormN(zI);
		} // ErfCN

		public static void RLambdaN(int zI) // z = lambda(z)
		// sets muErr true if overflow or error
		{
			LambdaN(Common.nList[zI], zI);
		}

		public static void RZetaN(int zI) // z = Zeta(z)
		// sets muErr true if overflow or error
		{
			ZetaN(Common.nList[zI], zI);
		}
		
		public static void RGamN(int zI) // z = gamma(z)
		// sets muErr true if overflow or error
		{
			GamN(Common.nList[zI], zI);
		}
		
		public static void RPsiN(int zI) // z = psi(z)
		// sets muErr true if overflow or error
		{
			PsiN(Common.nList[zI], zI);
		}
		
		public static void CuRootN(MultiC x, int zI) // z = CuRt(x)
		// x is not changed, it is ok for x and z to have the same location in memory
		// but then x will be changed. Sets muErr true if error
		{
			if (MultiFD.ZTestF(x.I))
			{
				XPCalLst.CuRootN(x, zI);
				MultiFD.ClearF(Common.nList[zI].I);
				Calc.NormN(zI);
				return;
			}
			int tI;
			bool saveExp = MultiFD.expanded;
			int saveFMC = Calc.fMC;
			int saveMMax = MultiID.mMax;
			XPCalLst.Expand();
			Common.InitNAdd(1, "CuRootN-tI", out tI);
			XPCalLst.DivN(MultiFD.fOne, MultiFD.fThree, tI);
			PowN(x, Common.nList[tI], zI);
			Common.LDelete(tI);
			MultiFD.expanded = saveExp;
			Calc.fMC = saveFMC;
			MultiID.mMax = saveMMax;
			Calc.NormN(zI);
		}

		public static void SqRootN(MultiC x, int zI) // z = SqRt(x)
		// x is not changed, it is ok for x and z to have the same location in memory
		// but then x will be changed. Sets muErr true if error
		{
			MultiC t = new MultiC(Calc.fMC); // expand done by SqRootC
			MultiCD.SqRootC(x, t);
			Calc.SetToN(t, zI);
			MultiCD.DoneC(t);
		}
		
		public static void RCuRtN(int zI) // z = SqRt(z)
		// sets muErr true if error
		{
			CuRootN(Common.nList[zI], zI);
		}

		public static void RSqRtN(int zI) // z = SqRt(z)
		// sets muErr true if error
		{
			SqRootN(Common.nList[zI], zI);
		}

		public static void RSqRtRemN(int zI) // z = Floor(SquareRoot(z)) and set Re to remainder
		// sets muErr true if error
		{
			int rI;
			Common.LFindAdd("Re", out rI);
			if (rI == zI || rI == Common.iTop)
			{
				MultiID.MuWriteErr(
				"Cannot set same location to both SqRt and remainder, continuing...");
				return;
			}
			MultiC t = new MultiC(Calc.fMC);
			MultiCD.SqRootC(Common.nList[zI], t);
			MultiCD.FloorC(t, t);
			MultiC r = new MultiC(Calc.fMC);
			r &= Common.nList[zI] - t * t;
			Calc.SetToN(t, zI);     // z = t = integer square root
			Calc.SetToN(r, rI);     // Re = r = remainder
		} // RSqRtRemN

		public static void InvN(MultiC x, int zI) // z = 1 / x
		{
			DivN(MultiCD.cOne, x, zI);			// z will be resized if not large enough
		}

		public static void RInvN(int zI)		// z = 1 / z
		{
			InvN(Common.nList[zI], zI);
		}

		public static void SetFMBN(int zI) // set Modulo base FMB on Common.nList = value of zI
		{
			int fMBI;
			Common.LFindAdd("FMB", out fMBI);
			if (fMBI != zI)
				Calc.SetToII(zI, fMBI);
		}

		public static void ModFMBN(MultiF x, int zI) // z = x % FMB
		{
			int fMBI;
			if (Common.FMBIok(zI, out fMBI) && !ZTestNC(fMBI))
				XCCalLst.RModN(Common.nList[fMBI], zI);
		}

		public static void RModFMBN(int zI) // z = z % FMB
		{
			ModFMBN(Common.nList[zI], zI);
		}

		public static void SinQuadN(MultiC x, out int qu, int zI)
		// z = Sin(x), qu = Quadrant (1..4), returned
		{
			if (MultiFD.ZTestF(x.I))
			{
				XPCalLst.SinQuadN(x, out qu, zI);
				MultiFD.ClearF(Common.nList[zI].I);
				Calc.NormN(zI);
				return;
			}
			int xI;
			int tI;
			int t2I;
			bool saveExp = MultiFD.expanded;
			int saveFMC = Calc.fMC;
			int saveMMax = MultiID.mMax;
			XPCalLst.Expand();
			Common.InitNAdd(1, "SinQuadN-xI", out xI);
			Common.InitNAdd(1, "SinQuadN-tI", out tI);
			Calc.SetToN(x, xI);
			if (Calc.degMode) RToRadN(xI);
			bool degModeSave = Calc.degMode;
			Calc.degMode = false;
			if (MultiFD.ZTestF(x))
			{
				// sin(i*y) = i*sinh(y);
				XPCalLst.SinHN(Common.nList[xI].I, tI);
				Calc.ClearNF(zI);
				MultiFD.SetToFS(Common.nList[tI], Common.nList[zI].I);
			}

			else
			{	
				// sin(x) = sin(x.r)* cosh(x.i) + i*cos(x.r) * sinh(x.i)
				Common.InitNAdd(1, "SinQuadN-t2I", out t2I);
				XPCalLst.SinN(Common.nList[xI], tI);
				XPCalLst.CosHN(Common.nList[xI].I, t2I);
				XPCalLst.RMulN(Common.nList[t2I], tI);
				MultiFD.SetToFS(Common.nList[tI], Common.nList[zI]);
				XPCalLst.CosN(Common.nList[xI], tI);
				XPCalLst.SinHN(Common.nList[xI].I, t2I);
				XPCalLst.RMulN(Common.nList[t2I], tI);
				MultiFD.SetToFS(Common.nList[tI], Common.nList[zI].I);
				Common.LDelete(t2I);
			}
			Common.LDelete(tI);
			Common.LDelete(xI);
			MultiFD.expanded = saveExp;
			Calc.fMC = saveFMC;
			MultiID.mMax = saveMMax;
			Calc.degMode = degModeSave;
			Calc.NormN(zI);
			qu = 0;
		} // SinQuadN

		public static void SinN(MultiC x, int zI)	// z = Sin(x)
		{
			int qu;
			SinQuadN(x, out qu, zI);
		}

		public static void RSinN(int zI)	// z = Sin(z)
		{
			SinN(Common.nList[zI], zI);
		}

		public static void CosN(MultiC x, int zI)	// z = Cos(x)
		{
			if (MultiFD.ZTestF(x.I))
			{
				XPCalLst.CosN(x, zI);
				MultiFD.ClearF(Common.nList[zI].I);
				Calc.NormN(zI);
				return;
			}
			int xI;
			int tI;
			int t2I;
			bool saveExp = MultiFD.expanded;
			int saveFMC = Calc.fMC;
			int saveMMax = MultiID.mMax;
			XPCalLst.Expand();
			Common.InitNAdd(1, "CosN-xI", out xI);
			Common.InitNAdd(1, "CosN-tI", out tI);
			Calc.SetToN(x, xI);
			if (Calc.degMode) RToRadN(xI);
			bool degModeSave = Calc.degMode;
			Calc.degMode = false;
			if (MultiFD.ZTestF(x))
			{
				// cos(i*y) = i*cosh(y);
				XPCalLst.CosHN(Common.nList[xI].I, tI);
				MultiFD.ClearF(Common.nList[zI].I);
				MultiFD.SetToFS(Common.nList[tI], Common.nList[zI]);
			}
			else
			{
				// cos(x) = cos(x.r)* cosh(x.i) - i*sin(x.r) * sinh(x.i)
				Common.InitNAdd(1, "CosN-t2I", out t2I);
				XPCalLst.CosN(Common.nList[xI], tI);
				XPCalLst.CosHN(Common.nList[xI].I, t2I);
				XPCalLst.RMulN(Common.nList[t2I], tI);
				MultiFD.SetToFS(Common.nList[tI], Common.nList[zI]);
				XPCalLst.SinN(Common.nList[xI], tI);
				XPCalLst.SinHN(Common.nList[xI].I, t2I);
				XPCalLst.RMulN(Common.nList[t2I], tI);
				MultiFD.ChSF(Common.nList[tI]);
				MultiFD.SetToFS(Common.nList[tI], Common.nList[zI].I);
				Common.LDelete(t2I);
			}
			Common.LDelete(tI);
			Common.LDelete(xI);
			MultiFD.expanded = saveExp;
			Calc.fMC = saveFMC;
			MultiID.mMax = saveMMax;
			Calc.degMode = degModeSave;
			Calc.NormN(zI);
		} // CosN

		public static void RCosN(int zI) // z = Cos(z)
		{
			CosN(Common.nList[zI], zI);
		}

		public static void TanN(MultiC x, int zI)	// z = Tan(x)
		// Tan(x) = Sin(x) / SqRt(1 - Sq(Sin(x))
		// and change sign of Tan(x) if in 2nd or 3rd quadrant
		{
			if (MultiFD.ZTestF(x.I))
			{
				XPCalLst.TanN(x, zI);
				MultiFD.ClearF(Common.nList[zI].I);
				Calc.NormN(zI);
				return;
			}
			int xI;
			int tI;
			int t2I;
			bool saveExp = MultiFD.expanded;
			int saveFMC = Calc.fMC;
			int saveMMax = MultiID.mMax;
			XPCalLst.Expand();
			Common.InitNAdd(1, "TanN-xI", out xI);
			Common.InitNAdd(1, "TanN-tI", out tI);
			Calc.SetToN(x, xI);
			if (Calc.degMode) RToRadN(xI);
			bool degModeSave = Calc.degMode;
			Calc.degMode = false;
			if (MultiFD.ZTestF(x))
			{
				// tan(i*y) = i*tanh(y);
				XPCalLst.TanHN(Common.nList[xI].I, tI);
				Calc.ClearNF(zI);
				MultiFD.SetToFS(Common.nList[tI], Common.nList[zI].I);
			}
			else
			{
				// tan(x) = [sin(2*x.r) + i*sinh(2*x.i)] / [cos(2*x.r) + cosh(2*x.i)]
				Common.InitNAdd(1, "TanN-t2I", out t2I);
				RAddN(Common.nList[xI], xI);
				XPCalLst.CosN(Common.nList[xI], tI);
				XPCalLst.CosHN(Common.nList[xI].I, t2I);
				XPCalLst.RAddN(Common.nList[tI], t2I);
				XPCalLst.SinN(Common.nList[xI], tI);
				XPCalLst.DivN(Common.nList[tI], Common.nList[t2I], zI);
				XPCalLst.SinHN(Common.nList[xI].I, tI);
				XPCalLst.RDivN(Common.nList[t2I], tI);
				MultiFD.SetToFS(Common.nList[tI], Common.nList[zI].I);
				Common.LDelete(t2I);
			}
			Common.LDelete(tI);
			Common.LDelete(xI);
			MultiFD.expanded = saveExp;
			Calc.fMC = saveFMC;
			MultiID.mMax = saveMMax;
			Calc.degMode = degModeSave;
			Calc.NormN(zI);
		} // TanN

		public static void RTanN(int zI) // z = Tan(z)
		{
			TanN(Common.nList[zI], zI);
		}

		public static void ATanN(MultiC x, int zI)	// z = ATan(x)
		{
			if (MultiFD.ZTestF(x.I))
			{
				XPCalLst.ATanN(x, zI);
				MultiFD.ClearF(Common.nList[zI].I);
				Calc.NormN(zI);
				return;
			}
			int xI;
			int tI;
			int t2I;
			int t3I;
			int pI;
			bool saveExp = MultiFD.expanded;
			int saveFMC = Calc.fMC;
			int saveMMax = MultiID.mMax;
			XPCalLst.Expand();
			bool degModeSave = Calc.degMode;
			Calc.degMode = false;
			Common.InitNAdd(1, "ATanN-xI", out xI);
			Common.InitNAdd(1, "ATanN-tI", out tI);
			if (MultiFD.ZTestF(x))
			{
				// atan(i*y) = i*ln(|(y+1)/(y-1)|) / 2;
				XPCalLst.AddN(x.I, MultiFD.fMOne, tI);
				XPCalLst.AddN(x.I, MultiFD.fOne, xI);
				XPCalLst.RDivN(Common.nList[tI], xI);
				if (!MultiID.muErr)
				{
					Common.nList[xI].S = false;
					XPCalLst.RLnN(xI);
					if (!MultiID.muErr)
					{
						XPCalLst.RDivN(MultiFD.fTwo, xI);
						Calc.ClearNF(zI);
						if (x.I > MultiFD.fOne || x.I < MultiFD.fMOne)
						{
							pI = XPCalLst.GetPiPtr();
							XPCalLst.DivN(Common.nList[pI], MultiFD.fTwo, zI);
						}
						MultiFD.SetToFS(Common.nList[xI], Common.nList[zI].I);
					}
				}
			}
			else
			{
				// atan(x) = k*Pi + atan(2*x.r / (1 - x.r^2 - x.i^2)) / 2 +
				//	i*ln((x.r^2 + (x.i + 1)^2) / (x.r^2 + (x.i - 1)^2)) / 4
				Common.InitNAdd(1, "ATanN-t2I", out t2I);
				Common.InitNAdd(1, "ATanN-t3I", out t3I);
				XPCalLst.SqN(x, xI);
				XPCalLst.SqN(x.I, tI);
				XPCalLst.RAddN(Common.nList[xI], tI);
				XPCalLst.SubN(MultiFD.fOne, Common.nList[tI], tI);
				int bAdd = 0;
				if (!x.S && Common.nList[tI].S) bAdd = 1;
				if (x.S && Common.nList[tI].S) bAdd = -1;
				XPCalLst.DivN(x, Common.nList[tI], tI);
				XPCalLst.RAddN(Common.nList[tI], tI);
				XPCalLst.ATanN(Common.nList[tI], t3I);
				if (bAdd == 1)
				{
					pI = XPCalLst.GetPiPtr();
					XPCalLst.RAddN(Common.nList[pI], t3I);
				}
				if (bAdd == -1)
				{
					pI = XPCalLst.GetPiPtr();
					XPCalLst.RSubN(Common.nList[pI], t3I);
				}
				XPCalLst.RDivN(MultiFD.fTwo, t3I);
				XPCalLst.AddN(x.I, MultiFD.fMOne, t2I);
				XPCalLst.RSqN(t2I);
				XPCalLst.RAddN(Common.nList[xI], t2I);
				XPCalLst.AddN(x.I, MultiFD.fOne, tI);
				XPCalLst.RSqN(tI);
				XPCalLst.RAddN(Common.nList[xI], tI);
				XPCalLst.RDivN(Common.nList[t2I], tI);
				XPCalLst.RLnN(tI);
				XPCalLst.RDivN(MultiFD.fFour, tI);
				MultiFD.SetToFS(Common.nList[t3I], Common.nList[zI]);
				MultiFD.SetToFS(Common.nList[tI], Common.nList[zI].I);
				Common.LDelete(t2I);
			}
			Common.LDelete(tI);
			Common.LDelete(xI);
			Calc.degMode = degModeSave;
			if (Calc.degMode) XCCalLst.RToDegN(zI);
			MultiFD.expanded = saveExp;
			Calc.fMC = saveFMC;
			MultiID.mMax = saveMMax;
			Calc.NormN(zI);
		} // ATanN

		public static void RATanN(int zI) // z = ATan(z)
		{
			ATanN(Common.nList[zI], zI);
		}

		public static void ATan2N(MultiC y, MultiC x, int zI)	// z = ATan(y / x)
		// Compute Arctan(y / x)
		{
			if (MultiFD.ZTestF(x.I) && MultiFD.ZTestF(y.I))
			{
				XPCalLst.ATan2N(y, x, zI);
				MultiFD.ClearF(Common.nList[zI].I);
				Calc.NormN(zI);
				return;
			}
			// atan2(y, x) = -i*Ln((x + i*y)/SqRt(x^2 + y^2))
			// when x^2 + y^2 = 1, ATan2(y, x) gives a such that x = Cos(a) and y = Sin(a)
			int xI;
			int yI;
			int tI;
			bool saveExp = MultiFD.expanded;
			int saveFMC = Calc.fMC;
			int saveMMax = MultiID.mMax;
			XPCalLst.Expand();
			bool degModeSave = Calc.degMode;
			Calc.degMode = false;
			Common.InitNAdd(1, "ATan2N-xI", out xI);
			Common.InitNAdd(1, "ATan2N-yI", out yI);
			Common.InitNAdd(1, "ATan2N-tI", out tI);

			XCCalLst.SqN(x, xI);
			XCCalLst.SqN(y, yI);
			XCCalLst.RAddN(Common.nList[yI], xI);	// xI = SqRt(x^2 + y^2)
			XCCalLst.RSqRtN(xI);

			XPCalLst.AddN(x.I, y, tI);								// yI = x + i*y
			XPCalLst.SubN(x, y.I, yI);								// = x.r-y.i, x.i+y.r
			MultiFD.SetToFS(Common.nList[tI], Common.nList[yI].I);

			XCCalLst.RDivN(Common.nList[xI], yI);
			XCCalLst.RLnN(yI);

			MultiFD.SetToFS(Common.nList[yI].I, Common.nList[zI]);	// zI = -i*yI
			MultiFD.SetToFS(Common.nList[yI], Common.nList[zI].I);
			MultiFD.ChSF(Common.nList[zI].I);

			Common.LDelete(tI);
			Common.LDelete(yI);
			Common.LDelete(xI);
			Calc.degMode = degModeSave;
			if (Calc.degMode) XCCalLst.RToDegN(zI);
			MultiFD.expanded = saveExp;
			Calc.fMC = saveFMC;
			MultiID.mMax = saveMMax;
			Calc.NormN(zI);
		} // ATan2N
			
		public static void AbsN(MultiC x, int zI)	// z = SqRt(Real(X)^2 + Imag(X)^2)
		{
			XPCalLst.MagN(x, x.I, zI);
			MultiFD.ClearF(Common.nList[zI].I);
			Calc.NormN(zI);
		}

		public static void MagN(MultiC x, MultiC y, int zI)	// z = SqRt(Sq(x) + Sq(y))
		{
			MultiF t = new MultiF(Calc.fMC);
			t &= (MultiF)x*(MultiF)x + (MultiF)y*(MultiF)y + x.I*x.I + y.I*y.I;
			MultiFD.RSqRtF(t);
			Calc.SetToNFF(t, zI);
			MultiFD.ClearF(Common.nList[zI].I);
			Calc.NormN(zI);
			MultiFD.DoneF(t);
		}

		public static void MeanCN(MultiC x, MultiC y, int zI)	// z = common mean of x and y
		// same as PARI GP agm(x, y)
		{
			if (MultiFD.ZTestF(x.I) && MultiFD.ZTestF(y.I) && !x.S && !y.S)
			{
				XPCalLst.MeanCN(x, y, zI);
				MultiFD.ClearF(Common.nList[zI].I);
				Calc.NormN(zI);
				return;
			}
			if (MultiCD.ZTestC(x) || MultiCD.ZTestC(y))
			{
				Calc.ClearN(zI); // common mean of zero and a is zero
				return;
			}
			bool saveExp = MultiFD.expanded;
			int saveFMC = Calc.fMC;
			int saveMMax = MultiID.mMax;
			XPCalLst.Expand();
			MultiC a = new MultiC(Calc.fMC);
			MultiC g = new MultiC(Calc.fMC);
			MultiC s = new MultiC(Calc.fMC);
			MultiC d1 = new MultiC(Calc.fMC);
			MultiC d2 = new MultiC(Calc.fMC);
			MultiC ar = new MultiC(Calc.fMC - 2);
			MultiC gr = new MultiC(Calc.fMC - 2);
			a &= x;
			g &= y;
			for (int n = 1; true; n++)
			{
				MultiID.MuInterrupt();
				if (MultiID.statReq2)
				{
					Common.SingleSpace();
					MultiID.WriteT("MeanCN: Iteration = " + n);
					MultiID.TimeDel();
					MultiID.statReq2 = false;
				}
				if (MultiID.muAbort || MultiID.muErr)
				{
					Calc.ClearN(zI);
					break;					// exit for
				}
				s &= a + g;					// s = sum of a & g
				g &= a * g;					// g = product of a & g
				MultiCD.RSqRtC(g);			// g = geometric mean of a & g
				a &= s / 2;					// a = arithmetic mean of a & g
				s &= (MultiF)a - (MultiF)g;
				d1 &= s*s;
				s &= a.I - g.I;
				d1 += s*s;
				s &= (MultiF)a + (MultiF)g;
				d2 &= s*s;
				s &= a.I + g.I;
				d2 += s*s;
				if ((MultiF)d2 < (MultiF)d1)
					MultiCD.ChSC(g);
				MultiCD.SetToC(g, gr);		// gr = g reduced in precision
				MultiCD.SetToC(a, ar);		// ar = a reduced in precision
				if (ar == gr || MultiCD.ZTestC(g))	// if a = g or g = 0
				{
					Calc.SetToN(g, zI);
					break;					// exit for
				}
			}
			MultiFD.DoneF(gr);
			MultiFD.DoneF(ar);
			MultiFD.DoneF(d2);
			MultiFD.DoneF(d1);
			MultiFD.DoneF(g);
			MultiFD.DoneF(a);
			MultiFD.expanded = saveExp;
			Calc.fMC = saveFMC;
			MultiID.mMax = saveMMax;
			Calc.NormN(zI);
		} // MeanCN

		public static void MaxN(MultiC x, MultiC y, int zI)	// z = max of x and y
		{
			if (x >= y)
				Calc.SetToN(x, zI);
			else
				Calc.SetToN(y, zI);
		} //MaxN

		public static void MinN(MultiC x, MultiC y, int zI)	// z = min of x and y
		{
			if (x >= y)
				Calc.SetToN(y, zI);
			else
				Calc.SetToN(x, zI);
		} //MinN

		public static void NormRemC(MultiC z)	// z = normalized complex remainder of z
		// put Rem in first quadrant by rotating it by n * 90 degrees
		{
			if (z > MultiFD.fZero && z.I >= MultiFD.fZero)
				return;
			if (z <= MultiFD.fZero && z.I <= MultiFD.fZero)
			{
				MultiCD.ChSC(z);
				return;
			}
			MultiF rp = new MultiF(z.N);
			rp &= z;
			XPCalLst.SetToNM(z.I, z);
			XPCalLst.SetToNM(rp, z.I);
			z.S = false;
			z.I.S = false;
		} // NormRemC

		public static void ModCheck(MultiC x, MultiC y, MultiC z)	// z = Mod(x, y))
		// same as FORTRAN definition: z = Mod(x, y) = x - (Math.Floor(x/y) * y)
		{
			if (MultiFD.ZTestF(x.I) && MultiFD.ZTestF(y.I))
			{
				XPCalLst.Mod(x, y, z);
				MultiFD.ClearF(z.I);
			}
			else
			{
				int tI;
				Common.InitNAdd(1, "ModCheck-tI", out tI);
				FloorN(x/y, tI);
				RMulN(-y, tI);
				RAddN(x, tI);
				Calc.SetToNM(Common.nList[tI], z);
				Common.LDelete(tI);
				MultiCD.ClearSmall(z);
			}
		} // ModCheck

		public static void Mod(MultiC x, MultiC y, MultiC z)	// z = Mod(x, y))
		// same as FORTRAN definition: z = Mod(x, y) = x - (Math.Floor(x/y) * y)
		{
			if (MultiFD.ZTestF(x.I) && MultiFD.ZTestF(y.I))
			{
				XPCalLst.Mod(x, y, z);
				MultiFD.ClearF(z.I);
			}
			else
			{
				bool saveExp = MultiFD.expanded;
				int saveFMC = Calc.fMC;
				int saveMMax = MultiID.mMax;
				XPCalLst.Expand();
				int tI;
				Common.InitNAdd(1, "Mod-tI", out tI);
				FloorN(x/y, tI);
				RMulN(-y, tI);
				RAddN(x, tI);
				Calc.SetToNM(Common.nList[tI], z);
				Common.LDelete(tI);
				MultiFD.expanded = saveExp;
				Calc.fMC = saveFMC;
				MultiID.mMax = saveMMax;
				MultiCD.ClearSmall(z);
				Calc.NormNM(z);
				ModCheck(z, y, z);
			}
		} // Mod

		public static void ModN(MultiC x, MultiC y, int zI)	// z = Mod(x, y))
		// same as FORTRAN definition: z = Mod(x, y) = x - (Math.Floor(x/y) * y)
		{
			MultiC zz = new MultiC();
			Mod(x, y, zz);
			Calc.SetToN(zz, zI);
		} // ModN

		public static void RModN(MultiC x, int zI) // z = Mod(z, x))
		{
			ModN(Common.nList[zI], x, zI);
		}

		public static void FibN(MultiC x, int zI) // z = Fib(x))
		{
			if (x.I == MultiCD.cZero)
			{
				XPCalLst.FibN(x, zI);
				MultiFD.ClearF(Common.nList[zI].I);
			}
			else
			{
				bool saveExp = MultiFD.expanded;
				int saveFMC = Calc.fMC;
				int saveMMax = MultiID.mMax;
				XPCalLst.Expand();
				// Fib(x) = (Phi^x - cos(x*Pi)/Phi^x) / Sqrt(5)
				MultiF a = new MultiF(Calc.fMC);
				MultiC b = new MultiC(Calc.fMC, 1);
				MultiC c = new MultiC(Calc.fMC);	
				int tI;
				Common.InitNAdd(1, "FibN-tI", out tI);	// t = temp on the list
				a &= 5;									// a = 5
				MultiFD.RSqRtF(a);						// a = SqRt(5)
				b &= a;									// b = SqRt(5)
				b += 1;									// b = 1 + SqRt(5)
				Calc.NormNM(b);
				MultiFD.MulF(b, MultiFD.fHalf, b);		// b = (1 + SqRt(5)) / 2
				PowNC(b, x, c);							// c = b ^ x
				bool degModeSave = Calc.degMode;
				Calc.degMode = true;
				CosN(x * 180, tI);						// t = cos(x*Pi)
				Calc.degMode = degModeSave;
				Calc.SetToN((c - Common.nList[tI] / c) / a, zI);
				Common.LDelete(tI);
				MultiCD.DoneC(c);
				MultiFD.DoneF(b);
				MultiFD.DoneF(a);
				MultiFD.expanded = saveExp;
				Calc.fMC = saveFMC;
				MultiID.mMax = saveMMax;
			}
			Calc.NormN(zI);
		} // FibN

		public static void GamPN(MultiC a, MultiC x, int zI) // z = GamP(a, x)
		// GamP(a, x) is the lower regularized incomplete gamma function with parameter a and argument x
		{
			if (MultiFD.ZTestF(a) && MultiFD.ZTestF(a.I))	// a = 0
			{
				if (MultiFD.ZTestF(x) && MultiFD.ZTestF(x.I))
				{
					Calc.ClearN(zI);	// GamP(0, 0) = 0
					return;
				}
				Calc.SetToN(MultiFD.fOne, zI);	// GamP(0, x!=0) = 1;
				return;
			}
			if (MultiFD.ZTestF(a.I) && (a.C >= 0 && a.S))	// a = -int
			{
				if (MultiFD.ZTestF(x) && MultiFD.ZTestF(x.I))
				{
					MultiID.MuWriteErr("Cannot take GamP(a, 0) if a is an integer < zero");
					Calc.ClearN(zI);	// GamP(-int, 0) not defined
					return;
				}
				Calc.SetToN(MultiFD.fOne, zI);	// GamP(-int, x!=0) = 1;
				return;
			}
			if (MultiFD.ZTestF(a.I) && MultiFD.ZTestF(x.I))
			{
				bool tryF = true;
				if (x.S || MultiFD.ZTestF(x))	// x <= 0.0)
					tryF = false;
				if (tryF)
				{
					XPCalLst.GamPN(a, x, zI);
					MultiFD.ClearF(Common.nList[zI].I);
					Calc.NormN(zI);
					return;
				}
			}
			bool saveExp = MultiFD.expanded;
			int saveFMC = Calc.fMC;
			int saveMMax = MultiID.mMax;
			XPCalLst.Expand();
			int tI, t2I;
			Common.InitNAdd(1, "GamPN-tI", out tI);
			Common.InitNAdd(1, "GamPN-t2I", out t2I);
			GamLN(a, x, tI);
			GamN(a, t2I);
			MultiC b = new MultiC(Calc.fMC);
			MultiCD.DivC(Common.nList[tI], Common.nList[t2I], b);	// GamP(a, x) = GamL(a, x) / Gam(a)
			Calc.NormNM(b);
			Calc.SetToN(b, zI);
			MultiCD.DoneC(b);
			Common.LDelete(t2I);
			Common.LDelete(tI);
			MultiFD.expanded = saveExp;
			Calc.fMC = saveFMC;
			MultiID.mMax = saveMMax;
			Calc.NormN(zI);
		} // GamPN

		public static void GamQN(MultiC a, MultiC x, int zI) // z = GamQ(a, x)
		// GamQ(a, x) is the upper regularized incomplete gamma function with parameter a and argument x
		{
			if (MultiFD.ZTestF(a) && MultiFD.ZTestF(a.I))	// a = 0
			{
				if (MultiFD.ZTestF(x) && MultiFD.ZTestF(x.I))
				{
					Calc.SetToN(MultiFD.fOne, zI);	// GamQ(0, 0) = 1
					return;
				}
				Calc.ClearN(zI);	// GamQ(0, x!=0) = 0;
				return;
			}
			if (MultiFD.ZTestF(a.I) && (a.C >= 0 && a.S))	// a = -int
			{
				if (MultiFD.ZTestF(x) && MultiFD.ZTestF(x.I))
				{
					MultiID.MuWriteErr("Cannot take GamQ(a, 0) if a is an integer < zero");
					Calc.ClearN(zI);	// GamQ(-int, 0) not defined
					return;
				}
				Calc.ClearN(zI);	// GamQ(-int, x!=0) = 0;
				return;
			}
			if (MultiFD.ZTestF(a.I) && MultiFD.ZTestF(x.I))
			{
				bool tryF = true;
				if (a.C >= 0 && a.S)			// a is an integer < zero
					tryF = false;
				if (x.S || MultiFD.ZTestF(x))	// x <= 0.0)
					tryF = false;
				if (tryF)
				{
					XPCalLst.GamQN(a, x, zI);
					MultiFD.ClearF(Common.nList[zI].I);
					Calc.NormN(zI);
					return;
				}
			}
			bool saveExp = MultiFD.expanded;
			int saveFMC = Calc.fMC;
			int saveMMax = MultiID.mMax;
			XPCalLst.Expand();
			int tI, t2I;
			Common.InitNAdd(1, "GamQN-tI", out tI);
			Common.InitNAdd(1, "GamQN-t2I", out t2I);
			GamUN(a, x, tI);
			GamN(a, t2I);
			MultiC b = new MultiC(Calc.fMC);
			MultiCD.DivC(Common.nList[tI], Common.nList[t2I], b);	// GamQ(a, x) = GamU(a, x) / Gam(a)
			Calc.NormNM(b);
			Calc.SetToN(b, zI);
			MultiCD.DoneC(b);
			Common.LDelete(t2I);
			Common.LDelete(tI);
			MultiFD.expanded = saveExp;
			Calc.fMC = saveFMC;
			MultiID.mMax = saveMMax;
			Calc.NormN(zI);
		} // GamQN

		public static void GamLN(MultiC a, MultiC x, int zI) // z = GamL(a, x)
		// GamL(a, x) is the lower incomplete gamma function with parameter a and argument x
		{
			if (MultiFD.ZTestF(a) && MultiFD.ZTestF(a.I) && MultiFD.ZTestF(x) && MultiFD.ZTestF(x.I))
			{
				Calc.ClearN(zI);	// GamL(0, 0) = 0
				return;
			}
			if (MultiFD.ZTestF(a.I) && ((a.C >= 0 && a.S) || MultiFD.ZTestF(a)))
			{
				MultiID.MuWriteErr("Cannot take GamL(a, x) if a is an integer <= zero, except GamL(0, 0) = 0");
				Calc.ClearN(zI);
				return;
			}
			if (MultiFD.ZTestF(x) && MultiFD.ZTestF(x.I))
			{
				Calc.ClearN(zI);	// GamL(a, 0) = 0 if a is not a neg integer
				return;
			}
			bool tryF = true;
			if (MultiFD.ZTestF(a.I) && MultiFD.ZTestF(x.I))
			{
				if (x.S)				// x < 0.0)
					tryF = false;
				if (tryF)
				{
					XPCalLst.GamLN(a, x, zI);
					MultiFD.ClearF(Common.nList[zI].I);
					Calc.NormN(zI);
					return;
				}
			}
			if (MultiFD.ZTestF(a.I) && ((a.C >= 0 && a.S) || MultiFD.ZTestF(a)))
			{
				MultiID.MuWriteErr("Cannot take GamL(a, x) if a is an integer <= zero, except GamL(0, 0) = 0");
				Calc.ClearN(zI);
				return;
			}
			bool saveExp = MultiFD.expanded;
			int saveFMC = Calc.fMC;
			int saveMMax = MultiID.mMax;
			XPCalLst.Expand();
			MultiC b = new MultiC(Calc.fMC);
			b &= a + 1;
			MultiF lim = new MultiF(1);
			lim &= 100;
			if (x < b || x < lim)			// if x < a + 1 or x < 100
			{
				GamLS(a, x, b);				// b = GamL(a, x) by series
			}
			else
			{
				int tI;
				Common.InitNAdd(1, "GamLN-tI", out tI);
				GamUC(a, x, b);				// b = GamU(a, x) by continued fraction
				GamN(a, tI);				// t = Gamma(a)
				b &= Common.nList[tI] - b;	// b = t - b = Gamma(a) - GamU(a, x)
				Common.LDelete(tI);
			}
			Calc.NormNM(b);
			Calc.SetToN(b, zI);
			MultiCD.DoneC(b);
			MultiFD.expanded = saveExp;
			Calc.fMC = saveFMC;
			MultiID.mMax = saveMMax;
			Calc.NormN(zI);
		} // GamLN

		public static void GamUN(MultiC a, MultiC x, int zI) // z = GamU(a, x)
		// GamU(a, x) is the upper incomplete gamma function with parameter a and argument x
		{
			bool aNPInt = MultiFD.ZTestF(a.I) && (MultiFD.ZTestF(a) || (a.C >= 0 && a.S));
			bool tryF = true;
			if (MultiFD.ZTestF(a.I) && MultiFD.ZTestF(x.I))
			{
				if (a.C >= 0 && a.S)	// a is an integer < zero
					tryF = false;
				if (x.S)				// x < 0.0)
					tryF = false;
				if (tryF)
				{
					XPCalLst.GamUN(a, x, zI);
					MultiFD.ClearF(Common.nList[zI].I);
					Calc.NormN(zI);
					return;
				}
			}
			if (MultiFD.ZTestF(x) && MultiFD.ZTestF(x.I))
			{
				if (aNPInt)
				{
					MultiID.MuWriteErr("Cannot take GamU(a, 0) if a is an integer <= zero");
					Calc.ClearN(zI);
					return;
				}
				GamN(a, zI);	// GamU(a, 0) = Gam(a)
				return;
			}
			bool saveExp = MultiFD.expanded;
			int saveFMC = Calc.fMC;
			int saveMMax = MultiID.mMax;
			XPCalLst.Expand();
			MultiC b = new MultiC(Calc.fMC);
			b &= a + 1;
			MultiF lim = new MultiF(1);
			lim &= 100;
			if (!aNPInt && (x < b || x < lim))	// if !aNPInt && (x < a + 1 or x < 100)
//			if (x < b || (x < lim && tryF))		// if x < a + 1 or x < 100
			{
				int tI;
				Common.InitNAdd(1, "GamUN-tI", out tI);
				GamLS(a, x, b);					// b = GamL(a, x) by series
				if (MultiID.muAbort || MultiID.muErr)
					b &= 0;
				else
				{
					GamN(a, tI);					// t = Gamma(a)
					b &= Common.nList[tI] - b;		// b = t - b = Gamma(a) - GamL(a, x)
				}
				Common.LDelete(tI);
			}
			else
			{
				GamUC(a, x, b);					// b = GamU(a, x) by continued fraction
			}
			Calc.NormNM(b);
			Calc.SetToN(b, zI);
			MultiCD.DoneC(b);
			MultiFD.expanded = saveExp;
			Calc.fMC = saveFMC;
			MultiID.mMax = saveMMax;
			Calc.NormN(zI);
		} // GamUN

		public static void GamUC(MultiC a, MultiC x, MultiC z) // z = GamU(a, x)
		// evaluated by its continued fraction representation
		{
			MultiC g = new MultiC(Calc.fMC);
			MultiC gr = new MultiC(Calc.fMC - 2);
			MultiC gOld = new MultiC(Calc.fMC - 2);
			MultiC a0 = new MultiC(Calc.fMC);
			MultiC a1 = new MultiC(Calc.fMC);
			MultiC b0 = new MultiC(Calc.fMC);
			MultiC b1 = new MultiC(Calc.fMC);
			MultiC fac = new MultiC(Calc.fMC);
			MultiC anf = new MultiC(Calc.fMC);
			MultiC an = new MultiC(Calc.fMC);
			MultiC ana = new MultiC(Calc.fMC);
			MultiC t = new MultiC(Calc.fMC);
			gOld &= 0;
			a0 &= 1;
			a1 &= x;
			b0 &= 0;
			b1 &= 1;
			fac &= 1;
			an &= 0;
			for (int n = 1; true; n++)
			{
				MultiID.MuInterrupt();
				if (MultiID.statReq2)
				{
					Common.SingleSpace();
					MultiID.WriteT("GamUC: n = " + n);
					MultiID.TimeDel();
					MultiID.statReq2 = false;
				}
				if (MultiID.muAbort || MultiID.muErr)
				{
					z &= 0;
					break;								// exit for
				}
				MultiCD.AddC(an, MultiCD.cOne, an);		// an = n
				MultiCD.SubC(an, a, ana);				// ana = an - a;
				MultiCD.MulC(a0, ana, a0);				// a0 = (a1 + a0 * ana) * fac
				MultiCD.AddC(a0, a1, a0);
				MultiCD.MulC(a0, fac, a0);
				MultiCD.MulC(b0, ana, b0);				// b0 = (b1 + b0 * ana) * fac
				MultiCD.AddC(b0, b1, b0);
				MultiCD.MulC(b0, fac, b0);
				MultiCD.MulC(an, fac, anf);				// anf = an * fac
				MultiCD.MulC(x, a0, t);					// a1 = x * a0 + anf * a1
				MultiCD.MulC(anf, a1, a1);
				MultiCD.AddC(t, a1, a1);
				MultiCD.MulC(x, b0, t);					// b1 | x * b0 + anf * b1
				MultiCD.MulC(anf, b1, b1);
				MultiCD.AddC(t, b1, b1);
				if (!MultiCD.ZTestC(a1))
				{
					MultiCD.DivC(MultiCD.cOne, a1, fac);	// fac = 1 / a1
					MultiCD.MulC(b1, fac, g);				// g = b1 * fac
					MultiCD.SetToC(g, gr);					// gr = g reduced in precision
					if (gr == gOld && n > 1)						// if gr == gOldn
					{
						int tI;
						Common.InitNAdd(1, "GamUC-tI", out tI);
						LnN(x, tI);									// t = ln(x)
						an &= a * Common.nList[tI] - x;				// an = a * ln(x) - x
						ExpN(an, tI);								// t = exp(a * ln(x) - x)
						MultiCD.MulC(g, Common.nList[tI], z);		// z = g * t
						Common.LDelete(tI);
						break;
					}
					else if (n > 1000)
					{
						MultiCD.RDivC(gr, gOld);
						MultiFD.RSubF(MultiFD.fOne, gOld);
						if (gOld.C + gOld.N > -2 && !MultiFD.ZTestF(gOld))
						{
							MultiID.MuWriteErr("GamU(a, x): Continued fraction diverged");
							MultiCD.ClearC(z);
							break;
						}
					}
					gOld &= gr;										// save old gr
				}
			}
			MultiCD.DoneC(t);
			MultiCD.DoneC(ana);
			MultiCD.DoneC(an);
			MultiCD.DoneC(anf);
			MultiCD.DoneC(fac);
			MultiCD.DoneC(b1);
			MultiCD.DoneC(b0);
			MultiCD.DoneC(a1);
			MultiCD.DoneC(a0);
			MultiCD.DoneC(gOld);
			MultiCD.DoneC(gr);
			MultiCD.DoneC(g);
		} // GamUC

		public static void GamLS(MultiC a, MultiC x, MultiC z) // z = GamL(a, x)
		// evaluated by its series representation
		{
			int n;
			MultiC sum = new MultiC(Calc.fMC);
			MultiC del = new MultiC(Calc.fMC);
			MultiC aP = new MultiC(Calc.fMC);
			MultiC b = new MultiC(Calc.fMC);
			if (MultiFD.ZTestF(x.I) && (x.S || MultiFD.ZTestF(x))) // x <= 0.0)
			if (MultiFD.ZTestF(x.I) && MultiFD.ZTestF(x)) // x <= 0.0)
			{
				if (x.S) MultiID.MuWriteErr("GamLS: x <= 0");
				MultiID.MuWriteErr("GamLS: x = 0");
				z &= 0;
				return;
			}
			aP &= a;
			MultiCD.DivC(MultiCD.cOne, a, sum);		// sum = 1 / a
			del &= sum;								// del = sum
			for (n = 1; ; n++)
			{
				MultiID.MuInterrupt();
				if (MultiID.statReq2)
				{
					Common.SingleSpace();
					MultiID.Write("GamLS: n = " + n);
					MultiID.TimeDel();
					MultiID.statReq2 = false;
				}
				aP += 1;
				MultiCD.DivC(x, aP, b);			// b = x / aP
				if (MultiID.muAbort || MultiID.muErr)
				{
					z &= 0;
					break;						// exit for
				}
				MultiCD.MulC(del, b, del);		// del = del * x / aP
				b &= sum;						// b = previous sum
				sum += del;
				del &= sum - b;
				if (MultiCD.ZTestC(del))
				{
					int tI;
					Common.InitNAdd(1, "GamLS-tI", out tI);
					LnN(x, tI);									// t = ln(x)
					aP &= a * Common.nList[tI] - x;				// aP = a * ln(x) - x
					ExpN(aP, tI);								// t = exp(a * ln(x) - x)
					MultiCD.MulC(sum, Common.nList[tI], z);		// z = sum * t
					Common.LDelete(tI);
					break;
				}
				if (del.C > 1000 || del.I.C > 1000)
				{
					MultiID.MuWriteErr("GamLS: Did not converge");
					MultiCD.ClearC(z);
					break;
				}
			}
			MultiCD.DoneC(b);
			MultiCD.DoneC(aP);
			MultiCD.DoneC(del);
			MultiCD.DoneC(sum);
		} // GamLS

		public static void BinoN(MultiC x, MultiC y, int zI)
		// z = Binomial coefficient (x, y), uses the integer part of y
		{
			MultiC yL = new MultiC((int)Math.Min(y.N, y.N + y.C), 1);
			MultiCD.IntC(y, yL);						// 1: yL = integer part of y
			if (yL.S)									// 2: (x, y) = 0 for all x if y < 0
			{
				MultiCD.DoneC(yL);
				Calc.ClearN(zI);
				return;
			}
			bool sign = false;							// set sign +
			bool xInt = false;
			MultiC xL = new MultiC(Calc.fMC);
			xL &= x;
			if (MultiFD.ZTestF(xL.I) && xL.C >= 0)		// 3: if x is an integer
			{
				xInt = true;
				if (xL.S)								// 3.1: if x < 0
				{
					MultiCD.ChSC(xL);
					xL &= yL - 1;					// (x, y) = (-1)^y * (y-x-1, y)
					sign = yL.C == 0 && (yL.V[0] & 1) == 1; // set sign true if y is odd
				}
				if (yL > xL)							// 3.2: if y > x
				{
					MultiCD.DoneC(xL);
					MultiCD.DoneC(yL);
					Calc.ClearN(zI);					// (x, y) = 0
					return;
				}
				MultiC h = new MultiC(xL.N+1, xL.I.N+1);
				h &= xL / 2;
				if (yL > h)								// 3.3: y > x / 2
				{
					yL &= xL - yL;						// y = x - y
				}
				MultiCD.DoneC(h);
			}
			if (MultiCD.ZTestC(yL))						// 4: (x, 0) = 1 for all x
			{
				Calc.SetToN(MultiFD.fOne, zI);
				Common.nList[zI].S = sign;
				MultiCD.DoneC(xL);
				MultiCD.DoneC(yL);
				return;
			}
			bool saveExp = MultiFD.expanded;
			int saveFMC = Calc.fMC;
			int saveMMax = MultiID.mMax;
			XPCalLst.Expand();
			MultiC v = new MultiC(Calc.fMC);
			v &= Int32.MaxValue;
			if (yL > v)									// 5: y > 2^31 -1
			{
				Calc.ClearN(zI);						// (x, y) = 0 with error message
				MultiID.MuWriteErr("BinoN: numbers too large for binomial coefficient");
			}
			else
			{
				MultiC nf = new MultiC(Calc.fMC);		// 6: (x, y) = x * (x-1) * (x-2) * ... * (x-y+1)
				MultiC df = new MultiC(Calc.fMC);		//			/ 1 * 2 * ... * n
				nf &= xL;
				df &= 1;
				v &= nf;
				double yd;
				MultiFD.GetDbF(out yd, yL);
				int yi = (int)yd;
				for (int i = 1; i < yi; i++)
				{
					MultiID.MuInterrupt();
					if (MultiID.muAbort)
					{
						MultiCD.ClearC(v);
						sign = false;
						break;
					}
					MultiFD.AddF(nf, MultiFD.fMOne, nf);	// nf = nf - 1
					MultiFD.AddF(df, MultiFD.fOne, df);		// df = df + 1
					MultiCD.MulC(v, nf, v);					// v = (v * nf) / df
					MultiCD.DivC(v, df, v);
					if (MultiID.statReq2)
					{
						MultiID.nDone = i;
						MultiID.nToGo = yi - i - 1;
						MultiID.WritenDone("BinoN:");
					}
				}
				Calc.SetToN(v, zI);
				if (xInt) Common.nList[zI].S = sign;
				MultiCD.DoneC(df);
				MultiCD.DoneC(nf);
			}
			MultiCD.DoneC(v);
			MultiCD.DoneC(xL);
			MultiCD.DoneC(yL);
			MultiFD.expanded = saveExp;
			Calc.fMC = saveFMC;
			MultiID.mMax = saveMMax;
			Calc.NormN(zI);
		} // BinoN

		public static void BetaCN(MultiC x, MultiC y, int zI)
		// z = the complete beta function B(x, y) = gamma(x) * gamma(y) / gamma(x+y)
		// also equal to 1 / (x * Bino(x+y-1, y-1)
		{
			if (MultiFD.ZTestF(x.I) && MultiFD.ZTestF(y.I))
			{
				XPCalLst.BetaCN(x, y, zI);
				MultiFD.ClearF(Common.nList[zI].I);
				Calc.NormN(zI);
				return;
			}
			bool saveExp = MultiFD.expanded;
			int saveFMC = Calc.fMC;
			int saveMMax = MultiID.mMax;
			XPCalLst.Expand();
			MultiC xL = new MultiC(Calc.fMC);
			xL &= x + y;
			bool gxInf = MultiFD.ZTestF(x.I) && ((x.C >= 0 && x.S) || MultiFD.ZTestF(x));		// gamma(x) is infinite
			bool gyInf = MultiFD.ZTestF(y.I) && ((y.C >= 0 && y.S) || MultiFD.ZTestF(y));		// gamma(y) is infinite
			bool gxPyInf = MultiFD.ZTestF(xL.I) && ((xL.C >= 0 && xL.S) || MultiFD.ZTestF(xL));	// gamma(x+y) is infinite
			if (!gxInf && !gyInf && gxPyInf)
			{
				Calc.ClearN(zI);					// finite / infinity = 0
			}
			else
			{
				if ( ((gxInf || gyInf) && !gxPyInf) || (gxInf && gyInf) )	// infinity / finite = infinity
				{															// infinity * infinity / v = infinity
					MultiID.MuWriteErr("BetaC(x, y) error: BetaC is infinite");
					Calc.ClearN(zI);
				}
				else
				{
					xL -= 1;
					MultiC yL = new MultiC(Calc.fMC);
					yL &= y - 1;
					int tI;
					Common.InitNAdd(1, "BetaCN-tI", out tI);			// t = temp on the list
					BinoGN(xL, yL, tI);
					xL &= x * Common.nList[tI];
					InvN(xL, zI);
					Common.LDelete(tI);
					MultiCD.DoneC(yL);
				}
			}
			MultiCD.DoneC(xL);
			MultiFD.expanded = saveExp;
			Calc.fMC = saveFMC;
			MultiID.mMax = saveMMax;
			Calc.NormN(zI);
		} // BetaCN

		public static void BinoGN(MultiC x, MultiC y, int zI)
		// z = Binomial coefficient (x, y) using the gamma function
		{
			if (MultiFD.ZTestF(x.I) && MultiFD.ZTestF(y.I))
			{
				XPCalLst.BinoGN(x, y, zI);
				MultiFD.ClearF(Common.nList[zI].I);
				Calc.NormN(zI);
				return;
			}
			if (MultiCD.ZTestC(y) || x == y)
			{
				Calc.SetToN(MultiCD.cOne, zI);	// (x, 0) = (x, x) = 1 for all x
				return;
			}
			if (y == MultiCD.cOne)
			{
				Calc.SetToN(x, zI);				// (x, 1) = x for all x
				return;
			}
			bool sign = false;
			MultiC xL = new MultiC(Calc.fMC);
			xL &= x;
			if (MultiFD.ZTestF(y.I) && y.C >= 0)	// if y is an integer
			{
				if (y.S)						// if y is a negative integer, (x, y) = 0
				{
					MultiCD.DoneC(xL);
					Calc.ClearN(zI);
					return;
				}
				MultiF lim = new MultiF(1);
				XPCalLst.limFac = 2000;
				lim &= XPCalLst.limFac + 1;
				if (MultiFD.ZTestF(y.I) && y <= lim)				// if y is a small integers
				{
					MultiFD.DoneF(lim);
					BinoN(xL, y, zI);
					MultiCD.DoneC(xL);
					Common.nList[zI].S = Common.nList[zI].S ^ sign; // ^ is xor
					return;
				}
				MultiFD.DoneF(lim);
			}
			bool saveExp = MultiFD.expanded;
			int saveFMC = Calc.fMC;
			int saveMMax = MultiID.mMax;
			XPCalLst.Expand();
			MultiC yL = new MultiC(Calc.fMC);
			yL &= xL - y;				// yL = x - y
			if (MultiFD.ZTestF(yL.I) && yL.C >= 0 && yL.S)	// if gamma of x - y + 1 would be infinite
			{
				Calc.ClearN(zI);		// divide by infinity gives zero
			}
			else
			{
				if (yL == MultiCD.cOne)
				{
					Calc.SetToN(xL, zI);	// (a, a-1) = a for all a
				}
				else
				{
					yL &= yL + 1;
					xL &= xL + 1;
					int tI;
					Common.InitNAdd(1, "BinoGN-tI", out tI); // t = temp on the list
					GamN(xL, tI);
					if (MultiID.muErr || MultiID.muAbort)
					{
						Calc.ClearN(zI);
					}
					else
					{
						int t2I;
						Common.InitNAdd(1, "BinoGN-t2I", out t2I); // t2 = temp on the list
						GamN(yL, t2I);
						MultiC v = new MultiC(Calc.fMC);
						v &= Common.nList[tI] / Common.nList[t2I];
						yL &= y + 1;
						GamN(yL, t2I);
						v /= Common.nList[t2I];
						Calc.SetToN(v, zI);
						Common.nList[zI].S = Common.nList[zI].S ^ sign; // ^ is xor
						MultiCD.DoneC(v);
						Common.LDelete(t2I);
					}
					Common.LDelete(tI);
				}
			}
			MultiCD.DoneC(yL);
			MultiCD.DoneC(xL);
			MultiFD.expanded = saveExp;
			Calc.fMC = saveFMC;
			MultiID.mMax = saveMMax;
			Calc.NormN(zI);
		} // BinoGN

		public static bool AlmostEqual(MultiC a, MultiC b) // is a almost = b
		{
			int fMCO2 = Calc.fMC / 2;
			MultiC t = new MultiC();
			Calc.SetToNM(a - b, t);
			if (MultiFD.ZTestF(t) || ((a.N + a.C - t.N - t.C >= Calc.fMC - fMCO2)))
				if (MultiFD.ZTestF(t.I) || ((a.I.N + a.I.C - t.I.N - t.I.C >= Calc.fMC - fMCO2)))
					return true;
			return false;
		} // AlmostEqual

		public static void GCD(MultiC a, MultiC b, MultiC z) // z = GCD(a, b))
		{
			if (MultiCD.ZTestC(a))
				Calc.SetToNM(b, z);
			else if (MultiCD.ZTestC(b))
				Calc.SetToNM(a, z);
			else
			{
				bool fini = false;
				int tot = 0;
				int nn = 0;
				int match = 0;
				MultiC aL = new MultiC();
				MultiC bL = new MultiC();
				MultiC t = new MultiC();
				Calc.SetToNM(a, aL);
				Calc.SetToNM(b, bL);
				if (MultiID.debug) Common.SingleSpace();
				int fMCO2 = Calc.fMC / 2;
				while (!fini)
				{
					nn++;
					Mod(aL, bL, t);
					MultiCD.ClearSmall(t, fMCO2);
					NormRemC(t);
					if (nn > 1 && AlmostEqual(t, aL))
					match++;
					else
						match = 0;
					if (match > 1)
					{
						Calc.SetToNM(t, z);
						return;
					}
					Calc.SetToNM(bL, aL);
					Calc.SetToNM(t, bL);
					if (MultiID.debug)
					{
						MultiID.WriteT(nn + " ", ref tot);
						MultiCD.WriteC(t, ref tot);
						MultiID.WriteLn();
						tot = 0;
					}
					if (MultiFD.ZTestF(bL) || ((aL.N + aL.C - bL.N - bL.C >= Calc.fMC - fMCO2)))
						if (MultiFD.ZTestF(bL.I) || ((aL.I.N + aL.I.C - bL.I.N - bL.I.C >= Calc.fMC - fMCO2)))
						{
							Calc.SetToNM(aL, z);
							fini = true;
						}				
				}
			}
			NormRemC(z);
		} // GCD

		public static void GCDN(MultiC a, MultiC b, int zI) // z = GCD(a, b))
		{
			MultiC zz = new MultiC();
			GCD(a, b, zz);
			Calc.SetToN(zz, zI);
		} // GCDN

		public static void LCM(MultiC a, MultiC b, MultiC z) // z = LCM(a, b))
		// LCM(a, b) = a * b / GCD(a, b)
		{
			MultiC zz = new MultiC();
			GCD(a, b, zz);
			bool saveExp = MultiFD.expanded;
			int saveFMC = Calc.fMC;
			int saveMMax = MultiID.mMax;
			XPCalLst.Expand();
			Calc.SetToNM((a * b) / zz, z);
			MultiFD.expanded = saveExp;
			Calc.fMC = saveFMC;
			MultiID.mMax = saveMMax;
			Calc.NormNM(z);
		} // LCM

		public static void LCMN(MultiC a, MultiC b, int zI) // z = LCM(a, b))
		// LCM(a, b) = a * b / GCD(a, b)
		{
			MultiC zz = new MultiC();
			LCM(a, b, zz);
			Calc.SetToN(zz, zI);
		} // LCMN

		public static void ASinN(MultiC x, int zI)	// z = ASin(x)
		// ASin(S) = ATan2(S, SqRt(1 - Sq(S)), but if Sq(S) > 1 then Error
		{
			if (MultiFD.ZTestF(x.I))
			{
				if (MultiFD.fMOne <= x && x <= MultiFD.fOne)
				{
					XPCalLst.ASinN(x, zI);
					MultiFD.ClearF(Common.nList[zI].I);
					Calc.NormN(zI);
					return;
				}
			}
			int xI;
			int tI;
			int t2I;
			bool saveExp = MultiFD.expanded;
			int saveFMC = Calc.fMC;
			int saveMMax = MultiID.mMax;
			XPCalLst.Expand();
			Common.InitNAdd(1, "ASinN-xI", out xI);
			Common.InitNAdd(1, "ASinN-tI", out tI);
			bool degModeSave = Calc.degMode;
			Calc.degMode = false;
			if (MultiFD.ZTestF(x))
			{
				// asin(i*y) = +/- i * ln(|y| + sqrt(1+y^2))
				Calc.SetToN(x.I, xI);
				Common.nList[xI].S = false;
				XPCalLst.SqN(x.I, tI);
				XPCalLst.RAddN(MultiFD.fOne, tI);
				XPCalLst.RSqRtN(tI);
				XPCalLst.RAddN(Common.nList[tI], xI);
				XPCalLst.RLnN(xI);
				Common.nList[xI].S = x.I.S;
				Calc.ClearNF(zI);
				MultiFD.SetToFS(Common.nList[xI], Common.nList[zI].I);
			}
			else if (MultiFD.ZTestF(x.I))
			{
				// asin(x) = +/- pI/2 +/- i * ln(|x| + sqrt(x^2 - 1))
				Calc.SetToN(x, xI);
				Common.nList[xI].S = false;
				XPCalLst.SqN(x, tI);
				XPCalLst.RAddN(MultiFD.fMOne, tI);
				XPCalLst.RSqRtN(tI);
				XPCalLst.RAddN(Common.nList[tI], xI);
				XPCalLst.RLnN(xI);
				Common.nList[xI].S = x.S;
				int pI = XPCalLst.GetPiPtr();
				XPCalLst.DivN(Common.nList[pI], MultiFD.fTwo, zI);
				Common.nList[zI].S = Common.nList[xI].S;
				MultiFD.SetToFS(Common.nList[xI], Common.nList[zI].I);
			}
			else
			{
				// asin(x) = k*Pi + (-1)^k * [ asin(v) + i * ln(u + sqrt(u^2 - 1)) ]
				// where u = a + b and v = a - b
				// and a = sqrt((x.r+1)^2 + x.i^2) / 2 and b = sqrt((x.r-1)^2 + x.i^2) / 2
				Common.InitNAdd(1, "ASinN-t2I", out t2I);
				XPCalLst.SqN(x.I, xI);
				XPCalLst.AddN(x, MultiFD.fOne, tI);
				XPCalLst.RSqN(tI);
				XPCalLst.RAddN(Common.nList[xI], tI);
				XPCalLst.RSqRtN(tI);
				XPCalLst.RDivN(MultiFD.fTwo, tI);		// tI = a
				XPCalLst.AddN(x, MultiFD.fMOne, t2I);
				XPCalLst.RSqN(t2I);
				XPCalLst.RAddN(Common.nList[xI], t2I);
				XPCalLst.RSqRtN(t2I);
				XPCalLst.RDivN(MultiFD.fTwo, t2I);		// t2I = b
				XPCalLst.AddN(Common.nList[tI], Common.nList[t2I], xI);	// xI = X
				XPCalLst.RSubN(Common.nList[t2I], tI);	// tI = Y
				XPCalLst.RASinN(tI);					// tI = real part of z
				XPCalLst.SqN(Common.nList[xI], t2I);
				XPCalLst.RAddN(MultiFD.fMOne, t2I);
				XPCalLst.RSqRtN(t2I);
				XPCalLst.RAddN(Common.nList[t2I], xI);
				XPCalLst.RLnN(xI);
				Common.LDelete(t2I);
				Common.nList[xI].S = x.I.S;
				if (MultiFD.ZTestF(x.I)) Common.nList[xI].S = x.S;
				MultiFD.SetToFS(Common.nList[xI], Common.nList[zI].I);
				MultiFD.SetToFS(Common.nList[tI], Common.nList[zI]);
			}
			Calc.degMode = degModeSave;
			if (Calc.degMode) XCCalLst.RToDegN(zI);
			Common.LDelete(tI);
			Common.LDelete(xI);
			MultiFD.expanded = saveExp;
			Calc.fMC = saveFMC;
			MultiID.mMax = saveMMax;
			Calc.NormN(zI);
		} // ASinN

		public static void RASinN(int zI) // z = ASin(z)
		{
			ASinN(Common.nList[zI], zI);
		}

		public static void ACosN(MultiC x, int zI)	// z = ACos(x)
		// ACos(C) = ATan2(SqRt(1 - Sq(C), C)
		{
			if (MultiFD.ZTestF(x.I))
			{
				if (MultiFD.fMOne <= x && x <= MultiFD.fOne)
				{
					XPCalLst.ACosN(x, zI);
					MultiFD.ClearF(Common.nList[zI].I);
					Calc.NormN(zI);
					return;
				}
			}
			int xI;
			int tI;
			int t2I;
			bool saveExp = MultiFD.expanded;
			int saveFMC = Calc.fMC;
			int saveMMax = MultiID.mMax;
			XPCalLst.Expand();
			Common.InitNAdd(1, "ACosN-xI", out xI);
			Common.InitNAdd(1, "ACosN-tI", out tI);
			bool degModeSave = Calc.degMode;
			Calc.degMode = false;
			if (MultiFD.ZTestF(x))
			{
				// acos(i*y) = Pi/2 - i * ln(|y| + sqrt(1+y^2))
				Calc.SetToN(x.I, xI);
				Common.nList[xI].S = false;
				XPCalLst.SqN(x.I, tI);
				XPCalLst.RAddN(MultiFD.fOne, tI);
				XPCalLst.RSqRtN(tI);
				XPCalLst.RAddN(Common.nList[tI], xI);
				XPCalLst.RLnN(xI);
				if (!MultiFD.ZTestF(Common.nList[xI])) Common.nList[xI].S = !x.I.S;
				int pI = XPCalLst.GetPiPtr();
				XPCalLst.DivN(Common.nList[pI], MultiFD.fTwo, zI);
				MultiFD.SetToFS(Common.nList[xI], Common.nList[zI].I);
			}
			else if (MultiFD.ZTestF(x.I))
			{
				// acos(x) = (0 or pI/2) + i * ln(|x| + sqrt(x^2 - 1))
				Calc.SetToN(x, xI);
				Common.nList[xI].S = false;
				XPCalLst.SqN(x, tI);
				XPCalLst.RAddN(MultiFD.fMOne, tI);
				XPCalLst.RSqRtN(tI);
				XPCalLst.RAddN(Common.nList[tI], xI);
				XPCalLst.RLnN(xI);
				if (x.S)
				{
					int pI = XPCalLst.GetPiPtr();
					MultiFD.SetToFS(Common.nList[pI], Common.nList[zI]);
				}
				else
				{
					Calc.ClearNF(zI);
				}
				MultiFD.SetToFS(Common.nList[xI], Common.nList[zI].I);
			}
			else
			{
				// acos(x) = 2*k*Pi +/- [ acos(v) - i*ln(u + sqrt(u^2 - 1)) ]
				// where u = a + b and v = a - b
				// and a = sqrt((x.r+1)^2 + x.i^2) / 2 and b = sqrt((x.r-1)^2 + x.i^2) / 2
				Common.InitNAdd(1, "ACosN-t2I", out t2I);
				XPCalLst.SqN(x.I, xI);
				XPCalLst.AddN(x, MultiFD.fOne, tI);
				XPCalLst.RSqN(tI);
				XPCalLst.RAddN(Common.nList[xI], tI);
				XPCalLst.RSqRtN(tI);
				XPCalLst.RDivN(MultiFD.fTwo, tI);		// tI = a
				XPCalLst.AddN(x, MultiFD.fMOne, t2I);
				XPCalLst.RSqN(t2I);
				XPCalLst.RAddN(Common.nList[xI], t2I);
				XPCalLst.RSqRtN(t2I);
				XPCalLst.RDivN(MultiFD.fTwo, t2I);		// t2I = b
				XPCalLst.AddN(Common.nList[tI], Common.nList[t2I], xI);	// xI = X
				XPCalLst.RSubN(Common.nList[t2I], tI);	// tI = Y
				XPCalLst.RACosN(tI);					// tI = real part of z
				XPCalLst.SqN(Common.nList[xI], t2I);
				XPCalLst.RAddN(MultiFD.fMOne, t2I);
				XPCalLst.RSqRtN(t2I);
				XPCalLst.RAddN(Common.nList[t2I], xI);
				XPCalLst.RLnN(xI);
				Common.LDelete(t2I);
				if (!MultiFD.ZTestF(Common.nList[xI])) Common.nList[xI].S = !x.I.S;
				if (MultiFD.ZTestF(x.I) && !MultiFD.ZTestF(Common.nList[xI])) Common.nList[xI].S = !x.S;
				MultiFD.SetToFS(Common.nList[xI], Common.nList[zI].I);
				MultiFD.SetToFS(Common.nList[tI], Common.nList[zI]);
			}
			Calc.degMode = degModeSave;
			if (Calc.degMode) XCCalLst.RToDegN(zI);
			Common.LDelete(tI);
			Common.LDelete(xI);
			MultiFD.expanded = saveExp;
			Calc.fMC = saveFMC;
			MultiID.mMax = saveMMax;
			Calc.NormN(zI);
		} // ACosN

		public static void RACosN(int zI) // z = ACos(z)
		{
			ACosN(Common.nList[zI], zI);
		}

		public static void ExpLN(MultiC x, bool bL1, int zI)
		// z = (e^x) - 1, accurate for small x, (- 1 only if bL1 is true)
		// n = Math.Floor(x / Ln(10))
		// g = Frac(x / Ln(10)) * Ln(10)
		// answer = Exp(g) * 10^n
		{
			// exp(x) = exp(x.r) * (cos(x.i) + i*sin(x.i))
			if (MultiFD.ZTestF(x.I))
			{
				XPCalLst.ExpLN(x, bL1, zI);
				MultiFD.ClearF(Common.nList[zI].I);
				Calc.NormN(zI);
				return;
			}
			int xI;
			int tI;
			int t2I;
			bool saveExp = MultiFD.expanded;
			int saveFMC = Calc.fMC;
			int saveMMax = MultiID.mMax;
			XPCalLst.Expand();
			bool degModeSv = Calc.degMode;
			Calc.degMode = false;
			int pI = XPCalLst.GetPiPtr();
			Common.InitNAdd(1, "ExpLN-xI", out xI);
			Common.InitNAdd(1, "ExpLN-tI", out tI);
			Common.InitNAdd(1, "ExpLN-t2I", out t2I);
			XPCalLst.ExpN(x, xI);
			MultiF PiR = new MultiF(saveFMC);
			PiR &= Common.nList[pI];
			if (x.I == Common.nList[pI] || x.I == PiR)
			{
				MultiFD.SetToFS(MultiFD.fMOne, Common.nList[tI]);
				MultiFD.SetToFS(MultiFD.fZero, Common.nList[t2I]);
			}
			else
			{
				XPCalLst.CosN(x.I, tI);
				XPCalLst.SinN(x.I, t2I);
			}
			MultiFD.DoneF(PiR);
			XPCalLst.RMulN(Common.nList[xI], tI);
			XPCalLst.RMulN(Common.nList[xI], t2I);
			if (bL1) XPCalLst.RAddN(MultiFD.fMOne, tI);
			MultiFD.SetToFS(Common.nList[tI], Common.nList[zI]);
			MultiFD.SetToFS(Common.nList[t2I], Common.nList[zI].I);
			Calc.degMode = degModeSv;
			Common.LDelete(t2I);
			Common.LDelete(tI);
			Common.LDelete(xI);
			MultiFD.expanded = saveExp;
			Calc.fMC = saveFMC;
			MultiID.mMax = saveMMax;
			Calc.NormN(zI);
		} // ExpLN

		public static void RExpLN(int zI) // z = (e^z) - 1, accurate for small x
		{
			ExpLN(Common.nList[zI], true, zI);
		}

		public static void ExpN(MultiC x, int zI)	// z = e^x
		{
			ExpLN(x, false, zI);
		}

		public static void RExpN(int zI) // z = e^z
		{
			ExpN(Common.nList[zI], zI);
		}

		public static void LnLN(MultiC y, bool bL1, int zI)
		// z = Ln(y + 1), accurate for small y
		// x = y + 1 = Integer * MultiID.baseM^C
		// Multiply by 10 until MSD of i >= MultiID.baseM / 10, save count
		// change C of x to -N so x < 1, save original C+N
		// answer = Ln(x) + Ln(10) * NT, NT = MultiID.muDMax * C+N - count
		{
			// ln(x) = ln(x.r^2 + x.i^2) / 2 + i*atan2(x.i, x.r)
			if (MultiFD.ZTestF(y.I) && (y >= MultiFD.fZero || (bL1 && y >= MultiFD.fMOne)))
			{
				XPCalLst.LnLN(y, bL1, zI);
				MultiFD.ClearF(Common.nList[zI].I);
				Calc.NormN(zI);
				return;
			}
			int xI;
			int tI;
			int t2I;
			bool saveExp = MultiFD.expanded;
			int saveFMC = Calc.fMC;
			int saveMMax = MultiID.mMax;
			XPCalLst.Expand();
			bool degModeSv = Calc.degMode;
			Calc.degMode = false;
			Common.InitNAdd(1, "LnLN-xI", out xI);
			Common.InitNAdd(1, "LnLN-tI", out tI);
			Common.InitNAdd(1, "LnLN-t2I", out t2I);
			Calc.SetToN(y, t2I);
			if (bL1) XPCalLst.RAddN(MultiFD.fOne, t2I);
			if (MultiFD.ZTestF(Common.nList[t2I]))
			{
				Calc.SetToN(y.I, xI);
				Common.nList[xI].S = false;
				XPCalLst.RLnN(xI);
			}
			else if (MultiFD.ZTestF(y.I))
			{
				Calc.SetToN(Common.nList[t2I], xI);
				Common.nList[xI].S = false;
				XPCalLst.RLnN(xI);
			}
			else
			{
				XPCalLst.SqN(Common.nList[t2I], xI);
				XPCalLst.SqN(y.I, tI);
				XPCalLst.RAddN(Common.nList[tI], xI);
				XPCalLst.RLnN(xI);
				XPCalLst.RDivN(MultiFD.fTwo, xI);
			}
			XPCalLst.ATan2N(y.I, Common.nList[t2I], tI);
			MultiFD.SetToFS(Common.nList[xI], Common.nList[zI]);
			MultiFD.SetToFS(Common.nList[tI], Common.nList[zI].I);
			Calc.degMode = degModeSv;
			Common.LDelete(t2I);
			Common.LDelete(tI);
			Common.LDelete(xI);
			MultiFD.expanded = saveExp;
			Calc.fMC = saveFMC;
			MultiID.mMax = saveMMax;
			Calc.NormN(zI);
		} // LnLN

		public static void RLnLN(int zI) // z = Ln(z + 1), accurate for small z
		{
			LnLN(Common.nList[zI], true, zI);
		}

		public static void LnN(MultiC x, int zI)	// z = Ln(x)
		{
			LnLN(x, false, zI);
		}

		public static void RLnN(int zI) // z = Ln(z)
		{
			LnN(Common.nList[zI], zI);
		}

		public static void LogN(MultiC x, int zI)	// z = Log(x)
		// Log(x) = Ln(x) / Ln(10), Log base 10 of x
		{
			bool saveExp = MultiFD.expanded;
			int saveFMC = Calc.fMC;
			int saveMMax = MultiID.mMax;
			XPCalLst.Expand();
			LnN(x, zI);
			if (!MultiID.muErr)
			{
				int ln10 = XPCalLst.GetLn10Ptr();
				RDivNF(Common.nList[ln10], zI);
			}
			MultiFD.expanded = saveExp;
			Calc.fMC = saveFMC;
			MultiID.mMax = saveMMax;
			Calc.NormN(zI);
		}

		public static void RLogN(int zI) // z = Log(z)
		{
			LogN(Common.nList[zI], zI);
		}

		public static void SinHN(MultiC x, int zI)	// z = SinH(x)
		// for (x >= 0.1 SinH(x) = (y - 1/y) / 2, where y = Exp(x)
		// for (x <  0.1 SinH(x) = y / (2 * SqRt(y + 1)), where y = Exp(2 * x) - 1
		// SinH(-x) = -SinH(x)
		{
			if (MultiFD.ZTestF(x.I)) 
			{
				// sinh(x + i*0) = sinh(x) + i*0
				XPCalLst.SinHN(x, zI);
				MultiFD.ClearF(Common.nList[zI].I);
				Calc.NormN(zI);
				return;
			}
			bool degModeSave = Calc.degMode;
			Calc.degMode = false;
			if (MultiFD.ZTestF(x))
			{
				// sinh(i*y) = 0 + i*sin(y))
				XPCalLst.SinN(x.I, zI);
				MultiFD.SetToFS(Common.nList[zI], Common.nList[zI].I);
				Calc.ClearNF(zI);
				Calc.degMode = degModeSave;
				return;
			}
			// sinh(x) = sinh(x.r) * cos(x.i) + i*cosh(x.r) * sin(x.i)
			int xI;
			int tI;
			int t2I;
			bool saveExp = MultiFD.expanded;
			int saveFMC = Calc.fMC;
			int saveMMax = MultiID.mMax;
			XPCalLst.Expand();
			Common.InitNAdd(1, "SinHN-xI", out xI);
			Common.InitNAdd(1, "SinHN-tI", out tI);
			Common.InitNAdd(1, "SinHN-t2I", out t2I);
			XPCalLst.SinHN(x, xI);
			XPCalLst.CosN(x.I, tI);
			XPCalLst.RMulN(Common.nList[xI], tI);
			XPCalLst.CosHN(x, xI);
			XPCalLst.SinN(x.I, t2I);
			XPCalLst.RMulN(Common.nList[t2I], xI);
			MultiFD.SetToFS(Common.nList[xI], Common.nList[zI].I);
			MultiFD.SetToFS(Common.nList[tI], Common.nList[zI]);
			Calc.degMode = degModeSave;
			Common.LDelete(t2I);
			Common.LDelete(tI);
			Common.LDelete(xI);
			MultiFD.expanded = saveExp;
			Calc.fMC = saveFMC;
			MultiID.mMax = saveMMax;
			Calc.NormN(zI);
		} // SinHN

		public static void RSinHN(int zI) // z = SinH(z)
		{
			SinHN(Common.nList[zI], zI);
		}

		public static void LiN(MultiC x, int zI)	// z = Logarithmic integral of x
		// Li(x) = Ei(Ln(x))
		{
			if (MultiFD.ZTestF(x.I)&& !x.S)
			{
				XPCalLst.LiN(x, zI);
				MultiFD.ClearF(Common.nList[zI].I);
				Calc.NormN(zI);
				return;
			}
			bool saveExp = MultiFD.expanded;
			int saveFMC = Calc.fMC;
			int saveMMax = MultiID.mMax;
			XPCalLst.Expand();
			int tI;
			Common.InitNAdd(1, "LiN-tI", out tI);
			LnN(x, tI);							// tI = ln(x)
			EiN(Common.nList[tI], zI);			// z = Ei(ln(x))
			Common.LDelete(tI);
			MultiFD.expanded = saveExp;
			Calc.fMC = saveFMC;
			MultiID.mMax = saveMMax;
			Calc.NormN(zI);
		} // LiN

		public static void EiN(MultiC x, int zI)	// z = entire exponential integral of x
		// Ei(x) = -E1(-x)
		{
			if (MultiFD.ZTestF(x.I))
			{
				XPCalLst.EiN(x, zI);
				MultiFD.ClearF(Common.nList[zI].I);
				Calc.NormN(zI);
				return;
			}
			MultiC mx = new MultiC(x.N, x.I.N);	// mx = -x
			mx &= x;
			MultiCD.ChSC(mx);
			E1N(mx, zI);
			MultiCD.ChSC(Common.nList[zI]);	// E1 = -Ei(-x)
		} // EiN

		public static void E1N(MultiC x, int zI)	// z = exponential integral 1 of x
		// E1(x) = integral{x to inf}(e^-t / t)dt = -Ei(-x)
		{
			if (MultiFD.ZTestF(x.I))
			{
				XPCalLst.E1N(x, zI);
				MultiFD.ClearF(Common.nList[zI].I);
				Calc.NormN(zI);
				return;
			}
			// E1(x) = GamU(0,x) - (ln(-x) - Ln(-1/x))/2 + ln(x)
			bool saveExp = MultiFD.expanded;
			int saveFMC = Calc.fMC;
			int saveMMax = MultiID.mMax;
			XPCalLst.Expand();
			MultiC mx = new MultiC(x.N, x.I.N);		// mx = -x
			mx &= x;
			MultiCD.ChSC(mx);
			MultiC mOx = new MultiC(Calc.fMC);		// mOx = -1/x
			mOx = MultiCD.cOne / mx;
			int tI;
			Common.InitNAdd(1, "E1N-tI", out tI);	// tI = Ln(-x)
			LnN(mx, tI);
			int t1I;
			Common.InitNAdd(1, "E1N-t1I", out t1I);	// t1I = Ln(-1/x)
			LnN(mOx, t1I);
			RSubN(Common.nList[t1I], tI);
			RDivN(MultiCD.cTwo, tI);				// tI = (Ln(-x) - Ln(-1/x))/2
			LnN(x, t1I);							// t1I = ln(x)
			RSubN(Common.nList[tI], t1I);			// t1I = -(ln(-x) - Ln(-1/x))/2 + ln(x)
			GamUN(MultiCD.cZero, x, tI);			// tI = GamU(0, x)
			RAddN(Common.nList[tI], t1I);			// t1I = E1(x)
			Calc.SetToN(Common.nList[t1I], zI);
			Common.LDelete(t1I);					// ln(x) != -ln(1/x) when x is real and < 0
			Common.LDelete(tI);
			MultiCD.DoneC(mOx);
			MultiCD.DoneC(mx);
			MultiFD.expanded = saveExp;
			Calc.fMC = saveFMC;
			MultiID.mMax = saveMMax;
			Calc.NormN(zI);
		} // E1N

		public static bool DilogSN(MultiC x, int zI)	// z = Dilog(x) special case
		// Dilog(x) = Polylog(2, x) = x * Lerch(x, 2, 1)
		{
			if (((MultiF)x >= MultiFD.fTwo && (MultiF)x >= x.I && (MultiF)x >= -x.I) ||
				((MultiF)x > MultiFD.fOne && (MultiF)x >= 2*x.I && (MultiF)x >= -2*x.I))
			{
				bool saveExp = MultiFD.expanded;
				int saveFMC = Calc.fMC;
				int saveMMax = MultiID.mMax;
				XPCalLst.Expand();
				int pI = XPCalLst.GetPiPtr();
				int tI;
				int t2I;
				Common.InitNAdd(1, "DilogSN-tI", out tI);
				Common.InitNAdd(1, "DilogSN-t2I", out t2I);
				SubN(MultiCD.cOne, x, tI);						// t = 1-x
				LnN(x, t2I);									// t2 = Ln(x)
				LnN(Common.nList[tI], zI);						// z = Ln(1-x)
				PolylogN(MultiCD.cTwo, Common.nList[tI], tI);	// t = dilog(1-x)
				RMulN(Common.nList[zI], t2I);					// t2 = Ln(x)*Ln(1-x)
				SqN(Common.nList[pI], zI);						// z = Pi^2
				RDivN(MultiCD.cSix, zI);						// z = Pi^2 / 6
				RSubN(Common.nList[tI], zI);
				RSubN(Common.nList[t2I], zI);					// z = Pi^2 / 6 - dilog(1-x) - Ln(x)*Ln(1-x)
				Common.LDelete(t2I);
				Common.LDelete(tI);
				MultiFD.expanded = saveExp;
				Calc.fMC = saveFMC;
				MultiID.mMax = saveMMax;
				Calc.NormN(zI);
				return true;
			}
			return false;
		} // DilogSN

		public static void DilogN(MultiC x, int zI)	// z = Dilog(x)
		// Dilog(x) = Polylog(2, x) = x * Lerch(x, 2, 1)
		{
			if (DilogSN(x, zI)) // if special dilog case
				return;
			PolylogN(MultiCD.cTwo, x, zI);
		} // DilogN

		public static void PolylogN(MultiC s, MultiC x, int zI)	// z = Polylog(s, x)
		// Polylog(s, x) = x * Lerch(x, s, 1)
		{
			if (MultiFD.ZTestF(s.I) && MultiFD.ZTestF(x.I) && x <= MultiFD.fOne)
			{
				XPCalLst.PolylogN(s, x, zI);
				MultiFD.ClearF(Common.nList[zI].I);
				Calc.NormN(zI);
				return;
			}
			if (s == MultiCD.cTwo && DilogSN(x, zI)) // if special dilog case
				return;
			bool saveExp = MultiFD.expanded;
			int saveFMC = Calc.fMC;
			int saveMMax = MultiID.mMax;
			XPCalLst.Expand();
			if (s == MultiCD.cTwo && x < MultiFD.fMOne)			// if Dilog of x < -1
			{
				MultiC y = new MultiC(Calc.fMC);
				y &= MultiCD.cOne / (1 - x);					// y = 1 / (1 - x)
				int tI;
				Common.InitNAdd(1, "PolylogN-tI", out tI);
				SubN(MultiCD.cOne, y, zI);						// z = 1 - y
				PolylogN(MultiCD.cTwo, Common.nList[zI], zI);	// z = Dilog(1-y)
				LnN(y, tI);										// t = ln(y)
				RSqN(tI);										// t = (ln(y))^2
				RMulN(MultiCD.cHalf, tI);						// t = (ln(y))^2 / 2
				RAddN(Common.nList[tI], zI);					// z = Dilog(1-y) + (ln(y))^2 / 2
				MultiCD.ChSC(Common.nList[zI]);					// z = -Dilog(1-y) - (ln(y))^2 / 2
				Common.LDelete(tI);
				MultiCD.DoneC(y);
			}
			else
			{	
				if (x > MultiFD.fOne && s != MultiCD.cTwo)	// if x > 1 and s != 2
				{
					Calc.ClearN(zI);
					MultiID.MuWriteErr("Cannot always compute Polylog(s, x) for x > 1 and s != 2");
				}
				else
				{
					MultiC saveX = new MultiC(x.N, x.I.N);
					saveX &= x;
					Lerch(x, s, MultiCD.cOne, zI);
					RMulN(saveX, zI);
					MultiCD.DoneC(saveX);
				}
			}
			MultiFD.expanded = saveExp;
			Calc.fMC = saveFMC;
			MultiID.mMax = saveMMax;
			Calc.NormN(zI);
		} // PolylogN

		public static void ZetaHN(MultiC s, MultiC a, int zI)	// z = ZetaH(s, a) Hurwitz Zeta Function
		// ZetaH(s, a) = Lerch(1, s, a)
		{
			Lerch(MultiCD.cOne, s, a, zI);
		} // ZetaHN

		public static void CosHN(MultiC x, int zI)	// z = CosH(x)
		// CosH(x) = (y + 1/y) / 2, where y = Exp(|x|)
		{
			if (MultiFD.ZTestF(x.I)) 
			{
				// cosh(x + i*0) = cosh(x) + i*0
				XPCalLst.CosHN(x, zI);
				MultiFD.ClearF(Common.nList[zI].I);
				Calc.NormN(zI);
				return;
			}
			bool degModeSave = Calc.degMode;
			Calc.degMode = false;
			if (MultiFD.ZTestF(x))
			{
				// cosh(i*y) = cos(y) + i*0
				XPCalLst.CosN(x.I, zI);
				MultiFD.ClearF(Common.nList[zI].I);
				Calc.NormN(zI);
				Calc.degMode = degModeSave;
				return;
			}
			// cosh(x) = cosh(x.r) * cos(x.i) + i*sinh(x.r) * sin(x.i)
			int xI;
			int tI;
			int t2I;
			bool saveExp = MultiFD.expanded;
			int saveFMC = Calc.fMC;
			int saveMMax = MultiID.mMax;
			XPCalLst.Expand();
			Common.InitNAdd(1, "CosHN-xI", out xI);
			Common.InitNAdd(1, "CosHN-tI", out tI);
			Common.InitNAdd(1, "CosHN-t2I", out t2I);
			XPCalLst.CosHN(x, xI);
			XPCalLst.CosN(x.I, tI);
			XPCalLst.RMulN(Common.nList[xI], tI);
			XPCalLst.SinHN(x, xI);
			XPCalLst.SinN(x.I, t2I);
			XPCalLst.RMulN(Common.nList[t2I], xI);
			MultiFD.SetToFS(Common.nList[xI], Common.nList[zI].I);
			MultiFD.SetToFS(Common.nList[tI], Common.nList[zI]);
			Calc.degMode = degModeSave;
			Common.LDelete(t2I);
			Common.LDelete(tI);
			Common.LDelete(xI);
			MultiFD.expanded = saveExp;
			Calc.fMC = saveFMC;
			MultiID.mMax = saveMMax;
			Calc.NormN(zI);
		} // CosHN

		public static void RCosHN(int zI) // z = CosH(z)
		{
			CosHN(Common.nList[zI], zI);
		}

		public static void TanHN(MultiC x, int zI)	// z = TanH(x)
		// TanH(x) = y / (y + 2), where y = Exp(2 * x) - 1
		// TanH(-x) = -TanH(x)
		{
			if (MultiFD.ZTestF(x.I))
			{
				// tanh(x + i*0) = tanh(x) + i*0
				XPCalLst.TanHN(x, zI);
				MultiFD.ClearF(Common.nList[zI].I);
				Calc.NormN(zI);
				return;
			}
			int xI;
			int tI;
			int t2I;
			bool saveExp = MultiFD.expanded;
			int saveFMC = Calc.fMC;
			int saveMMax = MultiID.mMax;
			XPCalLst.Expand();
			Common.InitNAdd(1, "TanHN-xI", out xI);
			Common.InitNAdd(1, "TanHN-tI", out tI);
			Calc.SetToN(x, xI);
			bool degModeSave = Calc.degMode;
			Calc.degMode = false;
			if (MultiFD.ZTestF(x))
			{
				// tanh(i*y) = i*tan(y);
				XPCalLst.TanN(Common.nList[xI].I, tI);
				Calc.ClearNF(zI);
				MultiFD.SetToFS(Common.nList[tI], Common.nList[zI].I);
			}
			else
			{
				// tanh(x) = [sinh(2*x.r) + i*sin(2*x.i)] / [cosh(2*x.r) + cos(2*x.i)]
				Common.InitNAdd(1, "TanHN-t2I", out t2I);
				RAddN(Common.nList[xI], xI);
				XPCalLst.CosHN(Common.nList[xI], tI);
				XPCalLst.CosN(Common.nList[xI].I, t2I);
				XPCalLst.RAddN(Common.nList[tI], t2I);
				XPCalLst.SinHN(Common.nList[xI], tI);
				XPCalLst.DivN(Common.nList[tI], Common.nList[t2I], zI);
				XPCalLst.SinN(Common.nList[xI].I, tI);
				XPCalLst.RDivN(Common.nList[t2I], tI);
				MultiFD.SetToFS(Common.nList[tI], Common.nList[zI].I);
				Common.LDelete(t2I);
			}
			Common.LDelete(tI);
			Common.LDelete(xI);
			MultiFD.expanded = saveExp;
			Calc.fMC = saveFMC;
			MultiID.mMax = saveMMax;
			Calc.degMode = degModeSave;
			Calc.NormN(zI);
		} // TanHN

		public static void RTanHN(int zI) // z = TanH(z)
		{
			TanHN(Common.nList[zI], zI);
		}

		public static void ASinHN(MultiC x, int zI)	// z = ASinH(x)
		// for (x >= 0.1 ASinH(x) = Ln(x + SqRt(1 + Sq(x)))
		// for (x <  0.1 ASinH(x) = LnL(x + Sq(x) / SqRt(1 + Sq(x)))
		// ASinH(-x) = -ASinH(x)
		{
			if (MultiFD.ZTestF(x.I))
			{
				XPCalLst.ASinHN(x, zI);
				MultiFD.ClearF(Common.nList[zI].I);
				Calc.NormN(zI);
				return;
			}
			// asinh(x) = i*asin(x.i - i*x.r)
			int xI;
			bool degModeSave = Calc.degMode;
			Calc.degMode = false;
			Common.InitNAdd(1, "ASinHN-xI", out xI);
			XPCalLst.SetToNM(x.I, (MultiF)Common.nList[xI]);
			XPCalLst.SetToNM((MultiF)x, Common.nList[xI].I);
			MultiFD.ChSF(Common.nList[xI].I);
			RASinN(xI);
			XPCalLst.SetToNM((MultiF)Common.nList[xI], Common.nList[zI].I);
			XPCalLst.SetToNM(Common.nList[xI].I, (MultiF)Common.nList[zI]);
			MultiFD.ChSF(Common.nList[zI]);
			Common.LDelete(xI);
			Calc.degMode = degModeSave;
		} // ASinHN

		public static void RASinHN(int zI) // z = ASinH(z)
		{
			ASinHN(Common.nList[zI], zI);
		}

		public static void ACosHN(MultiC x, int zI)	// z = ACosH(x)
		// ACosH(x) = Ln(x + SqRt(Sq(x) - 1))
		// complex if x < 1
		{
			bool xIZero = MultiFD.ZTestF(x.I);
			if (xIZero)
			{
				if (MultiFD.fOne <= x)
				{
					XPCalLst.ACosHN(x, zI);
					MultiFD.ClearF(Common.nList[zI].I);
					Calc.NormN(zI);
					return;
				}
			}
			// acosh(x) = i*acos(x)
			int xI;
			bool degModeSave = Calc.degMode;
			Calc.degMode = false;
			Common.InitNAdd(1, "ACosHN-xI", out xI);
			ACosN(x, xI);
			XPCalLst.SetToNM((MultiF)Common.nList[xI], Common.nList[zI].I);
			XPCalLst.SetToNM(Common.nList[xI].I, (MultiF)Common.nList[zI]);
			MultiFD.ChSF(Common.nList[zI]);
			Common.LDelete(xI);
			if (Common.nList[zI].S)
			{
				Common.nList[zI].S = false;
				MultiFD.ChSF(Common.nList[zI].I);
			}
			if (MultiFD.ZTestF(Common.nList[zI]) || xIZero)
				Common.nList[zI].I.S = false;
			Calc.degMode = degModeSave;
		} // ACosHN

		public static void RACosHN(int zI) // z = ACosH(z)
		{
			ACosHN(Common.nList[zI], zI);
		}

		public static void RACotN(int zI) // z = ACot(z)
		// ACot(z) = ATan(1/z), error if z = 0
		{
			if (MultiFD.ZTestF(Common.nList[zI].I)) 
			{
				XPCalLst.RACotN(zI);
				MultiFD.ClearF(Common.nList[zI].I);
				Calc.NormN(zI);
				return;
			}
			bool saveExp = MultiFD.expanded;
			int saveFMC = Calc.fMC;
			int saveMMax = MultiID.mMax;
			XPCalLst.Expand();
			ATanN(MultiCD.cOne/Common.nList[zI], zI);
			MultiFD.expanded = saveExp;
			Calc.fMC = saveFMC;
			MultiID.mMax = saveMMax;
			Calc.NormN(zI);
		} // RACotN

		public static void RACotHN(int zI) // z = ACotH(z)
		// ACotH(z) = ATanH(1/z), error if z = 0
		{
			if (MultiFD.ZTestF(Common.nList[zI].I)) 
			{
				if (Common.nList[zI] < MultiFD.fMOne || MultiFD.fOne < Common.nList[zI])
				{
					XPCalLst.RACotHN(zI);
					MultiFD.ClearF(Common.nList[zI].I);
					Calc.NormN(zI);
					return;
				}
			}
			bool saveExp = MultiFD.expanded;
			int saveFMC = Calc.fMC;
			int saveMMax = MultiID.mMax;
			XPCalLst.Expand();
			ATanHN(MultiCD.cOne/Common.nList[zI], zI);
			MultiFD.expanded = saveExp;
			Calc.fMC = saveFMC;
			MultiID.mMax = saveMMax;
			Calc.NormN(zI);
		} // RACotHN

		public static void RACscN(int zI) // z = ACsc(z)
		// ACsc(z) = ASin(1/z), error if z = 0
		{
			if (MultiFD.ZTestF(Common.nList[zI].I)) 
			{
				if (Common.nList[zI] <= MultiFD.fMOne || MultiFD.fOne <= Common.nList[zI])
				{
					XPCalLst.RACscN(zI);
					MultiFD.ClearF(Common.nList[zI].I);
					Calc.NormN(zI);
					return;
				}
			}
			bool saveExp = MultiFD.expanded;
			int saveFMC = Calc.fMC;
			int saveMMax = MultiID.mMax;
			XPCalLst.Expand();
			ASinN(MultiCD.cOne/Common.nList[zI], zI);
			MultiFD.expanded = saveExp;
			Calc.fMC = saveFMC;
			MultiID.mMax = saveMMax;
			Calc.NormN(zI);
		} // RACscN

		public static void RACscHN(int zI) // z = ACscH(z)
		// ACscH(z) = ASinH(1/z), error if z = 0
		{
			if (MultiFD.ZTestF(Common.nList[zI].I)) 
			{
				XPCalLst.RACscHN(zI);
				MultiFD.ClearF(Common.nList[zI].I);
				Calc.NormN(zI);
				return;
			}
			bool saveExp = MultiFD.expanded;
			int saveFMC = Calc.fMC;
			int saveMMax = MultiID.mMax;
			XPCalLst.Expand();
			ASinHN(MultiCD.cOne/Common.nList[zI], zI);
			MultiFD.expanded = saveExp;
			Calc.fMC = saveFMC;
			MultiID.mMax = saveMMax;
			Calc.NormN(zI);
		} // RACscHN

		public static void RASecN(int zI) // z = ASec(z)
		// ASec(z) = ACos(1/z), error if z = 0
		{
			if (MultiFD.ZTestF(Common.nList[zI].I)) 
			{
				if (Common.nList[zI] <= MultiFD.fMOne || MultiFD.fOne <= Common.nList[zI])
				{
					XPCalLst.RASecN(zI);
					MultiFD.ClearF(Common.nList[zI].I);
					Calc.NormN(zI);
					return;
				}
			}
			bool saveExp = MultiFD.expanded;
			int saveFMC = Calc.fMC;
			int saveMMax = MultiID.mMax;
			XPCalLst.Expand();
			ACosN(MultiCD.cOne/Common.nList[zI], zI);
			MultiFD.expanded = saveExp;
			Calc.fMC = saveFMC;
			MultiID.mMax = saveMMax;
			Calc.NormN(zI);
		} // RASecN

		public static void RASecHN(int zI) // z = ASecH(z)
		// ASecH(z) = ACosH(1/z), error if z = 0
		{
			if (MultiFD.ZTestF(Common.nList[zI].I)) 
			{
				if (MultiFD.fZero < Common.nList[zI] && Common.nList[zI] <= MultiFD.fOne)
				{
					XPCalLst.RASecHN(zI);
					MultiFD.ClearF(Common.nList[zI].I);
					Calc.NormN(zI);
					return;
				}
			}
			bool saveExp = MultiFD.expanded;
			int saveFMC = Calc.fMC;
			int saveMMax = MultiID.mMax;
			XPCalLst.Expand();
			ACosHN(MultiCD.cOne/Common.nList[zI], zI);
			MultiFD.expanded = saveExp;
			Calc.fMC = saveFMC;
			MultiID.mMax = saveMMax;
			Calc.NormN(zI);
		} // RASecHN

		public static void RCotN(int zI) // z = Cot(z)
		// Cot(z) = 1/Tan(z), error if Tan(z) = 0
		{
			if (MultiFD.ZTestF(Common.nList[zI].I)) 
			{
				XPCalLst.RCotN(zI);
				MultiFD.ClearF(Common.nList[zI].I);
				Calc.NormN(zI);
				return;
			}
			bool saveExp = MultiFD.expanded;
			int saveFMC = Calc.fMC;
			int saveMMax = MultiID.mMax;
			XPCalLst.Expand();
			TanN(Common.nList[zI], zI);
			RInvN(zI);
			MultiFD.expanded = saveExp;
			Calc.fMC = saveFMC;
			MultiID.mMax = saveMMax;
			Calc.NormN(zI);
		} // RCotN

		public static void RCotHN(int zI) // z = CotH(z)
		// CotH(z) = 1/TanH(z), error if TanH(z) = 0
		{
			if (MultiFD.ZTestF(Common.nList[zI].I)) 
			{
				XPCalLst.RCotHN(zI);
				MultiFD.ClearF(Common.nList[zI].I);
				Calc.NormN(zI);
				return;
			}
			bool saveExp = MultiFD.expanded;
			int saveFMC = Calc.fMC;
			int saveMMax = MultiID.mMax;
			XPCalLst.Expand();
			TanHN(Common.nList[zI], zI);
			RInvN(zI);
			MultiFD.expanded = saveExp;
			Calc.fMC = saveFMC;
			MultiID.mMax = saveMMax;
			Calc.NormN(zI);
		} // RCotHN

		public static void RCscN(int zI) // z = Csc(z)
		// Csc(z) = 1/Sin(z), error if Sin(z) = 0
		{
			if (MultiFD.ZTestF(Common.nList[zI].I)) 
			{
				XPCalLst.RCscN(zI);
				MultiFD.ClearF(Common.nList[zI].I);
				Calc.NormN(zI);
				return;
			}
			bool saveExp = MultiFD.expanded;
			int saveFMC = Calc.fMC;
			int saveMMax = MultiID.mMax;
			XPCalLst.Expand();
			SinN(Common.nList[zI], zI);
			RInvN(zI);
			MultiFD.expanded = saveExp;
			Calc.fMC = saveFMC;
			MultiID.mMax = saveMMax;
			Calc.NormN(zI);
		} // RCscN

		public static void RCscHN(int zI) // z = CscH(z)
		// CscH(z) = 1/SinH(z), error if SinH(z) = 0
		{
			if (MultiFD.ZTestF(Common.nList[zI].I)) 
			{
				XPCalLst.RCscHN(zI);
				MultiFD.ClearF(Common.nList[zI].I);
				Calc.NormN(zI);
				return;
			}
			bool saveExp = MultiFD.expanded;
			int saveFMC = Calc.fMC;
			int saveMMax = MultiID.mMax;
			XPCalLst.Expand();
			SinHN(Common.nList[zI], zI);
			RInvN(zI);
			MultiFD.expanded = saveExp;
			Calc.fMC = saveFMC;
			MultiID.mMax = saveMMax;
			Calc.NormN(zI);
		} // RCscHN

		public static void RSecN(int zI) // z = Sec(z)
		// Sec(z) = 1/Cos(z), error if Cos(z) = 0
		{
			if (MultiFD.ZTestF(Common.nList[zI].I)) 
			{
				XPCalLst.RSecN(zI);
				MultiFD.ClearF(Common.nList[zI].I);
				Calc.NormN(zI);
				return;
			}
			bool saveExp = MultiFD.expanded;
			int saveFMC = Calc.fMC;
			int saveMMax = MultiID.mMax;
			XPCalLst.Expand();
			CosN(Common.nList[zI], zI);
			RInvN(zI);
			MultiFD.expanded = saveExp;
			Calc.fMC = saveFMC;
			MultiID.mMax = saveMMax;
			Calc.NormN(zI);
		} // RSecN

		public static void RSecHN(int zI) // z = SecH(z)
		// SecH(z) = 1/CosH(z), error if CosH(z) = 0
		{
			if (MultiFD.ZTestF(Common.nList[zI].I)) 
			{
				XPCalLst.RSecHN(zI);
				MultiFD.ClearF(Common.nList[zI].I);
				Calc.NormN(zI);
				return;
			}
			bool saveExp = MultiFD.expanded;
			int saveFMC = Calc.fMC;
			int saveMMax = MultiID.mMax;
			XPCalLst.Expand();
			CosHN(Common.nList[zI], zI);
			RInvN(zI);
			MultiFD.expanded = saveExp;
			Calc.fMC = saveFMC;
			MultiID.mMax = saveMMax;
			Calc.NormN(zI);
		} // RSecHN

		public static void ATanHN(MultiC x, int zI)	// z = ATanH(x)
		// atanh(x) = i*atan(x.i - i*x.r)
		// error if x = +/-1
		{
			if (x == MultiCD.cOne || x == MultiCD.cMOne)
			{
				MultiID.MuWriteErr("ATanH(x) error: x = +/-1");
				return;
			}
			if (MultiFD.ZTestF(x.I)) 
			{
				if(MultiFD.fMOne < x && x < MultiFD.fOne)
				{
					XPCalLst.ATanHN(x, zI);
					MultiFD.ClearF(Common.nList[zI].I);
					Calc.NormN(zI);
					return;
				}
			}
			int xI;
			bool degModeSave = Calc.degMode;
			Calc.degMode = false;
			Common.InitNAdd(1, "ATanHN-xI", out xI);
			XPCalLst.SetToNM(x.I, (MultiF)Common.nList[xI]);
			XPCalLst.SetToNM((MultiF)x, Common.nList[xI].I);
			MultiFD.ChSF(Common.nList[xI].I);
			RATanN(xI);
			XPCalLst.SetToNM((MultiF)Common.nList[xI], Common.nList[zI].I);
			XPCalLst.SetToNM(Common.nList[xI].I, (MultiF)Common.nList[zI]);
			MultiFD.ChSF(Common.nList[zI]);
			Common.LDelete(xI);
			Calc.degMode = degModeSave;
		} // ATanHN

		public static void RATanHN(int zI) // z = ATanH(z)
		{
			ATanHN(Common.nList[zI], zI);
		}

		public static void CeilingN(MultiC x, int zI)	// z = integer nearest to x
		{
			int xI;
			Common.InitNAdd(1, "CeilingN-xI", out xI);
			XPCalLst.CeilingN(x, xI);
			XPCalLst.SetToNM(Common.nList[xI], (MultiF)Common.nList[zI]);
			XPCalLst.CeilingN(x.I, xI);
			XPCalLst.SetToNM(Common.nList[xI], Common.nList[zI].I);
			Common.LDelete(xI);
		}

		public static void RCeilingN(int zI) // z = largest integer <= z
		{
			CeilingN(Common.nList[zI], zI);
		}

		public static void FloorN(MultiC x, int zI)	// z = integer nearest to x
		{
			int xI;
			Common.InitNAdd(1, "FloorN-xI", out xI);
			XPCalLst.FloorN(x, xI);
			XPCalLst.SetToNM(Common.nList[xI], (MultiF)Common.nList[zI]);
			XPCalLst.FloorN(x.I, xI);
			XPCalLst.SetToNM(Common.nList[xI], Common.nList[zI].I);
			Common.LDelete(xI);
		}

		public static void RFloorN(int zI) // z = largest integer <= z
		{
			FloorN(Common.nList[zI], zI);
		}

		public static void RoundN(MultiC x, int zI)	// z = integer nearest to x
		{
			int xI;
			Common.InitNAdd(1, "RoundN-xI", out xI);
			XPCalLst.RoundN(x, xI);
			XPCalLst.SetToNM(Common.nList[xI], (MultiF)Common.nList[zI]);
			XPCalLst.RoundN(x.I, xI);
			XPCalLst.SetToNM(Common.nList[xI], Common.nList[zI].I);
			Common.LDelete(xI);
		}

		public static void RRoundN(int zI) // z = integer nearest to z
		{
			RoundN(Common.nList[zI], zI);
		}

		public static void IntN(MultiC x, int zI)	// z = Integer part of x
		{
			int xI;
			Common.InitNAdd(1, "IntN-xI", out xI);
			XPCalLst.IntN(x, xI);
			XPCalLst.SetToNM(Common.nList[xI], (MultiF)Common.nList[zI]);
			XPCalLst.IntN(x.I, xI);
			XPCalLst.SetToNM(Common.nList[xI], Common.nList[zI].I);
			Common.LDelete(xI);
		}

		public static void RIntN(int zI) // z = Integer part of z
		{
			IntN(Common.nList[zI], zI);
		}

		public static void FracN(MultiC x, int zI)	// z = Fractional part of x
		{
			int xI;
			Common.InitNAdd(1, "FracN-xI", out xI);
			XPCalLst.FracN(x, xI);
			XPCalLst.SetToNM(Common.nList[xI], (MultiF)Common.nList[zI]);
			XPCalLst.FracN(x.I, xI);
			XPCalLst.SetToNM(Common.nList[xI], Common.nList[zI].I);
			Common.LDelete(xI);
		}

		public static void RFracN(int zI) // z = Fractional part of z
		{
			FracN(Common.nList[zI], zI);
		}

		public static void LopN(MultiC x, int zI) // z = x with one less super digit
		{
			MultiCD.LopC(x, Common.nList[zI]);
			Calc.NormN(zI);
		}

		public static void RLopN(int zI) // z = z with one less least significant "digit"
		{
			MultiCD.RLopC(Common.nList[zI]);
			Calc.NormN(zI);
		}

		public static void ToDegN(MultiC x, int zI) // z = x converted Rad to Degrees
		{
			bool saveExp = MultiFD.expanded;
			int saveFMC = Calc.fMC;
			int saveMMax = MultiID.mMax;
			XPCalLst.Expand();
			int pI = XPCalLst.GetPiPtr();
			Mul1N(x, 180, zI);
			RDivNF((MultiF)Common.nList[pI], zI);
			MultiFD.expanded = saveExp;
			Calc.fMC = saveFMC;
			MultiID.mMax = saveMMax;
			Calc.NormN(zI);
		}

		public static void RToDegN(int zI) // z = z converted Radians to Degrees
		{
			ToDegN(Common.nList[zI], zI);
		}

		public static void ToRadN(MultiC x, int zI)	// z = x converted Deg to Radians
		{
			bool saveExp = MultiFD.expanded;
			int saveFMC = Calc.fMC;
			int saveMMax = MultiID.mMax;
			XPCalLst.Expand();
			int pI = XPCalLst.GetPiPtr();
			MulNF(x, (MultiF)Common.nList[pI], zI);
			RDiv1N(180, zI);
			MultiFD.expanded = saveExp;
			Calc.fMC = saveFMC;
			MultiID.mMax = saveMMax;
			Calc.NormN(zI);
		}

		public static void RToRadN(int zI) // z = z converted Deg to Radians
		{
			ToRadN(Common.nList[zI], zI);
		}

		public static void CalSFTn(int zI) // set number of digits in truncate in display
		{
			double db;
			MultiCD.GetDbInC(out db, 0, Int32.MaxValue, Common.nList[zI]);
			db = 1 + Math.Floor((db - 1) / MultiID.muDMax); // set for at least N decimal digits
			MultiFD.fTn = (int)db;
			XPCalLst.CheckNMax();
		}

		public static void NormC(MultiC x, MultiC m, MultiC s, int zI)
		// z = Cumulative distribution function (cdf) of normal distribution
		// m = mean, s = standard deviation
		// x, m, s may have the same address in memory as z
		// x, m, s are not changed unless it shares memory with z
		// sets muErr true if overflow or error
		{
			// NormC(x, m, s) = 0.5 * ErfC((m - x)/(s*SqRt(2))
			if (s == MultiCD.cZero)
			{
				if (x < m)
					Calc.SetToN(MultiCD.cZero, zI);
				else if (x > m)
					Calc.SetToN(MultiCD.cOne, zI);
				else
					Calc.SetToN(MultiCD.cHalf, zI);
			}
			else
			{
				bool saveExp = MultiFD.expanded;
				int saveFMC = Calc.fMC;
				int saveMMax = MultiID.mMax;
				XPCalLst.Expand();
				MultiC t = new MultiC(Calc.fMC);
				MultiCD.SqRootC(MultiCD.cTwo, t);
				t &= (m - x)/(s*t);
				ErfCN(t, zI);
				RDiv1N(2, zI);
				MultiFD.expanded = saveExp;
				Calc.fMC = saveFMC;
				MultiID.mMax = saveMMax;
				Calc.NormN(zI);
			}
		} // NormC

		public static void NormS(int zI)
		// z = Cumulative distribution function (cdf) of standard normal distribution
		// mean = 0, standard deviation = 1
		// sets muErr true if overflow or error
		{
			NormC(Common.nList[zI], MultiCD.cZero, MultiCD.cOne, zI);
		} // NormS

		public static void QuadRoot(MultiC a, MultiC b, MultiC c, int zI)
		// z = a root of quadratic equation 0 = a*x^2 + b*x + c, roots also stored in X1 and X2
		// a, b, c may have the same address in memory as z
		// a, b, c are not changed unless it shares memory with z
		// sets muErr true if overflow or error
		{
			bool saveExp = MultiFD.expanded;
			int saveFMC = Calc.fMC;
			int saveMMax = MultiID.mMax;
			XPCalLst.Expand();
			int x1I;	// "X1"
			int x2I;	// "X2"
			Common.LFindAdd("X1", out x1I);
			Common.LFindAdd("X2", out x2I);
			MultiC d = new MultiC(Calc.fMC);
			d &= b*b - 4*a*c;
			if (a.I == 0 && b.I == 0 && c.I == 0)
			{
				if (d >= MultiCD.cZero)
				{
					XPCalLst.QuadRoot(a, b, c, zI);
				}
				else
				{
					d.S = false;
					MultiFD.RSqRtF(d);
					XPCalLst.DivN(b, -2*a, x1I);
					XPCalLst.DivN(d, 2*a, x2I);
					XPCalLst.SetToNM(Common.nList[x2I], Common.nList[x1I].I);
					Calc.SetToII(x1I, x2I);
					Common.nList[x2I].I.S = !Common.nList[x2I].I.S;
				}
			}
			else if (a == 0 && b == 0)
			{
				MultiID.MuWriteErr("QuadR: No solutions");
			}
			else if (x1I == zI || x2I == zI || x1I == Common.iTop || x2I == Common.iTop)
			{
				MultiID.MuWriteErr("QuadR: Cannot set X1 or X2");
			}
			else if (a == 0)
			{
				XCCalLst.DivN(c, b, x1I);
				MultiCD.ChSC(Common.nList[x1I]);
				Calc.SetToII(x1I, x2I);
			}
			else
			{
				MultiCD.RSqRtC(d);
				d += b;
				if (d == 0)
				{
					Calc.SetToN(MultiCD.cZero, x1I);
					Calc.SetToN(MultiCD.cZero, x2I);
				}
				else
				{
					MultiC r1 = new MultiC(Calc.fMC);
					MultiC r2 = new MultiC(Calc.fMC);
					d *= MultiCD.cMHalf;
					MultiCD.DivC(d, a, r1);
					QuadNewton(a, b, c, r1, "X1");
					MultiCD.DivC(c, d, r2);
					QuadNewton(a, b, c, r2, "X2");
					Calc.SetToN(r1, x1I);
					Calc.SetToN(r2, x2I);
				}
			}
			MultiFD.expanded = saveExp;
			Calc.fMC = saveFMC;
			MultiID.mMax = saveMMax;
			Calc.NormN(x1I);
			Calc.NormN(x2I);
			Calc.SetToII(x2I, zI);
		} // QuadRoot

		private static void QuadNewton(MultiC a, MultiC b, MultiC c, MultiC r0, string root)
		{
			MultiC r1 = new MultiC(Calc.fMC);
			MultiC r0r = new MultiC(Calc.fMC - 2);
			MultiC r1r = new MultiC(Calc.fMC - 2);
			for (int k = 1; true; k++)
			{
				r1 &= r0 - (a*r0*r0 + b*r0 + c) / (2*a*r0 + b);
				QuadZero(r1);
				MultiCD.SetToC(r0, r0r);		// r0r = r0 reduced in precision
				MultiCD.SetToC(r1, r1r);		// r1r = r1 reduced in precision
				MultiCD.SetToC(r1, r0);
				if (r1r == r0r)
				{
					MultiID.ADiag("QuadNewton: " + root + " converged in " + k + " iterations");
					break;
				}
				MultiID.MuInterrupt();
				if (MultiID.statReq3 || k >= 1000)
				{
					MultiID.WriteT("QuadNewton: Iteration = " + k);
					MultiID.TimeDel();
					MultiID.statReq3 = false;
					if (k >= 1000)
					{
						MultiID.WriteT("QuadNewton: " + root + " Did not converge");
						break;
					}
				}
				if (MultiID.muAbort || MultiID.muErr)
					break;
			}
		} // QuadNewton

		public static void QuadZero(MultiC r)
		{
			if (r.N + r.C > r.I.N + r.I.C + Calc.fMC && !MultiFD.ZTestF(r))
				MultiFD.SetToF(MultiFD.fZero, r.I);
			else if (r.I.N + r.I.C > r.N + r.C + Calc.fMC && !MultiFD.ZTestF(r.I))
				MultiFD.SetToF(MultiFD.fZero, r);
		} // QuadZero

		public static void WriteFixedB(RichTextBox outB, StreamWriter outF, MultiC z, int tot, int width)
		{
			string ch;
			int d;
			int did;		// number of fractional digits output
			int dToD;		// digits to display
			int i;
			int k;
			int nOutF;		// number of significant digits in fraction output
			int nOutI;		// number of significant digits in integer output
			string st;
			string stD;		// formatted NOut
			MultiC t2 = new MultiC();
			k = 'A' - 10;
			nOutI = 0;
			nOutF = 0;
			did = 0;
			if (outB == null) MultiID.WriteTLn2(outB, outF, "", ref tot, width);
			MultiC t1 = new MultiC(z.N+1, z.I.N+1);
			MultiCD.SetToC(z, t1); // t1 = z with one extra digit
			if (z.S) MultiID.WriteT(outB, outF, "-", ref tot, width);
			t1.S = false;
			if (z.N + z.C > 0)
			{
				MultiCD.ReSizeC((int)(z.N + z.C), t2);
				MultiCD.IntC(t1, t2);
				MultiCD.RFracC(t1);
				if (MultiID.EQ1SI(1, t2))
				{
					MultiID.WriteT(outB, outF, "1", ref tot, width);
					nOutI = 1;
				}
				else
				{
					MultiID.WriteSI(outB, outF, t2, ref tot, width, false);
					nOutI = MultiID.nOutSI;
				}
				MultiCD.DoneC(t2);
			}
			else
			{
				MultiID.WriteT(outB, outF, "0", ref tot, width);
			}
			MultiID.WriteT(outB, outF, ".", ref tot, width);
			MultiCD.ReSizeC(1, t2);
			if (outB == null)
				dToD = 1 + (int)Math.Round((Calc.fMC * MultiID.muDMax) / Math.Log10(MultiID.baseO));
			else
				dToD = 1 + (int)Math.Round(((Calc.fMC - MultiFD.fTn) * MultiID.muDMax) / Math.Log10(MultiID.baseO));
			dToD = Math.Min(dToD, MultiID.maxDis);
			dToD = Math.Max(dToD, 2); // an extra digit is output, but not rounded
			do
			{
				MultiCD.RMul1C(MultiID.baseO, t1);
				MultiCD.IntC(t1, t2);
				MultiCD.RFracC(t1);
				MultiCD.Get1C(out d, t2);
				if (d < 10)
					ch = d.ToString();
				else
					ch = ((char)(d + k)).ToString();
				if (MultiID.muDpG > 0)
				{
					if (did > 0 && did % MultiID.muDpG == 0)
						MultiID.WriteT(outB, outF, ",", ref tot, width);
				}
				MultiID.WriteT(outB, outF, ch, ref tot, width);
				did++;
				if (ch != "0" || nOutI + nOutF > 0) nOutF = nOutF + 1;
			}
			while (!MultiCD.ZTestC(t1) && !(nOutI + nOutF >= dToD));
			did = nOutI + nOutF;
			if (outF != null)
			{
				MultiID.Write(outB, outF, "\n");
				tot = 0;
				ch = "";
			}
			else
			{
				ch = " ";
			}
			stD = "";
			if (did > MultiID.muLenD)
				stD = nOutI.ToString() + "." + nOutF.ToString() + " Digits ";
			MultiID.WriteT(outB, outF, ch + "(" + stD + "Base " + MultiID.baseO + ")", ref tot, width);
			if (outF == null)
			{
				i = z.V[z.N-1].ToString().Length;
				st = (MultiID.baseM + z.V[0]).ToString();
				while (st[st.Length - 1] == '0') st = st.Remove(st.Length - 1, 1);
				i += st.Length - 1 - MultiID.muDMax;
				st = (i + (z.N - 1) * MultiID.muDMax).ToString();
				MultiID.WriteT(outB, outF, " [" + st + "]", ref tot, width);
			}
			MultiCD.DoneC(t2);
			MultiCD.DoneC(t1);
			return;
		} // WriteFixedB

		public static void WriteFB(RichTextBox outB, StreamWriter outF, MultiC z, ref int tot, int width)
		// output z as a line of text, tot = characters on line so far, z is not modified
		{
			int bI;				// natural log of baseO
			string ch;
			int d;
			decimal dE;			// exponent
			double db;
			int did;			// digits output
			int dToD;			// digits to display
			bool bFixed;
			int i;
			int k;
			int nI;
			string st;			// formatted number
			string stE;			// formatted exponent "E+xxx,xxx...x"
			MultiC t1 = new MultiC();	// temp
			MultiC t2 = new MultiC(1);
			int trunc;
			int min = 2;
			bFixed = (!MultiFD.scieN) && (z.N + z.C <= min) && (z.C >= -min);
			if (bFixed)
			{
				WriteFixedB(outB, outF, z, tot, 0); // format <= 99999999999999.0
				return;
			}
			if (outF != null) // always do full precision to file
				dToD = (int)Math.Round((Calc.fMC * MultiID.muDMax) / Math.Log10(MultiID.baseO)); // convert to baseO digits
			else
			{
				dToD = (int)Math.Round(((Calc.fMC - MultiFD.fTn) * MultiID.muDMax) / Math.Log10(MultiID.baseO));
				dToD = Math.Min(dToD, MultiID.maxDis);
				dToD = Math.Max(dToD, 2);	// needed or rounding factor will be wrong
			}
			bool saveExp = MultiFD.expanded;
			int saveFMC = Calc.fMC;
			int saveMMax = MultiID.mMax;
			XPCalLst.Expand();
			db = MultiID.maxDis * Math.Log10(MultiID.baseO); // convert maxDis to Decimal digits
			trunc = (int)Math.Round(Math.Max(Calc.fMC - (db + 12) / 7, MultiFD.fTn));
			trunc = Math.Min(trunc, Calc.fMC - 2);
			if (trunc > 0 && outF == null)
				MultiCD.ReSizeC(Calc.fMC - trunc, t1);
			else
				MultiCD.ReSizeC(Calc.fMC, t1);
			MultiCD.SetToC(z, t1);	// rounds to desired digits
			t1.S = false;
			MultiC bF = new MultiC(Calc.fMC);	// baseO as a MultiC
			MultiCD.SetTo1C(MultiID.baseO, bF);
			Common.InitNAdd(6, "WriteFB-nI", out nI);
			Common.InitNAdd(1, "WriteFB-bI", out bI);
			i = t1.V[t1.N-1].ToString().Length;
			dE = Decimal.Floor((( (decimal)t1.C + t1.N - 1) * MultiID.muDMax + i) / (decimal)Math.Log10(MultiID.baseO));
			MultiCD.SetToDeC(dE, Common.nList[nI]);
			MultiCD.SetTo1C(MultiID.baseO, Common.nList[bI]);
			if (dE < 0)
			{
				Common.nList[nI].S = false;
				RPowN(Common.nList[nI], bI);
				if (MultiID.muErr) goto WriteFBEnd;
				MultiCD.MulC(t1, Common.nList[bI], t1);
			}
			else if (dE > 0)
			{
				RPowN(Common.nList[nI], bI);
				if (MultiID.muErr) goto WriteFBEnd;
				MultiCD.DivC(t1, Common.nList[bI], t1);
			}
			while (t1 < MultiFD.fOne)				// while t1 < 1
			{
				MultiCD.RMul1C(MultiID.baseO, t1);
				dE--;
			}
			while (t1 >= bF)						// t1 >= baseO
			{
				MultiCD.RDiv1C(MultiID.baseO, t1);
				dE++;
			}
			MultiCD.ReSizeC(1, t2);
			if (MultiFD.fRound > 0)
			{
				SetTo1N(MultiID.baseO, bI);
				MultiCD.SetTo1C(-dToD + 1, t2);		// compute rounding factor (+1)
				RPowN(t2, bI);
				RDiv1N(2, bI);
				MultiCD.RAddC(Common.nList[bI], t1);
			}
			MultiCD.IntC(t1, t2);					// t2 = Integer part of t1
			MultiCD.Get1C(out d, t2);
			if (d == MultiID.baseO)
			{
				MultiCD.RDiv1C(MultiID.baseO, t1);
				dE++;
				MultiCD.IntC(t1, t2);				// t2 = Integer part of t1
				MultiCD.Get1C(out d, t2);
			}
			FormatStE(out stE, dE);
			if (outF != null)
				MultiID.WriteLn2(null, outF, " m.n " + stE + ", m.n ="); // line may > 80 ch
			MultiCD.RFracC(t1);						// t1 = Fractional part
			did = 0;
			k = 'A' - 10;
			st = "";
			Save1(ref st, ref did, d, k);			// first digit
			do
			{
				MultiCD.RMul1C(MultiID.baseO, t1);	// one digit at a time
				MultiCD.IntC(t1, t2);
				MultiCD.RFracC(t1);
				MultiCD.Get1C(out d, t2);
				Save1(ref st, ref did, d, k);
			}
			while (did < dToD && !MultiCD.ZTestC(t1));
			i = st.Length - 1;
			while (i > 1)				// built for speed, don't call WriteT too often
			{
				ch = st.Substring(i, 1);
				if (ch == "0")			// delete trailing zeros and don't count them
				{
					st = st.Substring(0, i);
					did--;
				}
				else if (ch == ",")		// delete trailing commas
					st = st.Substring(0, i);
				else
					break;				// exit while
				i--;
			}
			if (z.S) MultiID.WriteT(outB, outF, "-", ref tot, width);
			if (st[0] > '9') MultiID.WriteT(outB, outF, "0", ref tot, width);
			MultiID.WriteT(outB, outF, st.Substring(0, 1) + ".", ref tot, width);
			if (outF != null) MultiID.WriteTLn(null, outF, "", ref tot);
			MultiID.WriteT(outB, outF, st.Substring(1), ref tot, width);
			if (outF != null) MultiID.WriteTLn(null, outF, "", ref tot);
			st = "";
			if (did > MultiID.muLenD)
				st = did.ToString() + " Digits ";
			MultiID.WriteT(outB, outF, stE + " (" + st + "Base " + MultiID.baseO + ")", ref tot, width);
			if (outF == null)
			{
				i = z.V[z.N-1].ToString().Length;
				st = (MultiID.baseM + z.V[0]).ToString();
				while (st[st.Length - 1] == '0') st = st.Remove(st.Length - 1, 1);
				i += st.Length - 1 - MultiID.muDMax;
				st = (i + (z.N - 1) * MultiID.muDMax).ToString();
				MultiID.WriteT(outB, outF, " [" + st + "]", ref tot, width);
			}
WriteFBEnd:
			MultiCD.DoneC(t2);
			MultiCD.DoneC(bF);
			MultiCD.DoneC(t1);
			Common.LDelete(bI);
			Common.LDelete(nI);
			MultiFD.expanded = saveExp;
			Calc.fMC = saveFMC;
			MultiID.mMax = saveMMax;
			return;
		} // WriteFB

		private static void FormatStE(out string stE, decimal dE)
		// part of WriteFB
		{
			if (MultiID.baseO > 10) { stE = "`"; } else { stE = "E"; }
			if (dE >= 0) { stE = stE + "+"; } else { stE = stE + "-"; }
			string st = (Math.Abs(dE)).ToString();
			int k = st.Length;
			if (k > MultiID.muDpG && MultiID.muDpG != 0)
			{
				int j = k + 1;
				int iL = (k - 1) / MultiID.muDpG;
				for (int i = 1; i <= iL; i++)
				{
					j -= MultiID.muDpG;
					st = st.Substring(0, j - 1) + "," + st.Substring(j - 1);
				}
			}
			stE += st;
			st = "";
			return;
		}

		private static void Save1(ref string st, ref int did, int d, int k)
		// part of WriteFB
		{
			string ch;
			if (MultiID.muDpG > 0)
			{
				if ((did - 1) % MultiID.muDpG == 0)
				{
					if (did > 1) st += ",";
				}
			}
			if (d < 10)
				ch = d.ToString();
			else
				ch = ((char)(d + k)).ToString();
			st += ch;
			did++;
			return;
		}
	}
}

