using System;
using System.IO;			// for StreamWriter I/O to a file
using System.Windows.Forms;	// for RichTextBox

namespace MyNamespace
{
	/// <summary>
	///		Extra precision complex matrix List algorithms
	///		in Visual C# .Net
	///
	/// ---------------- *** LEGAL NOTICE: *** ------------------
	///
	/// This program is free software; you can redistribute it and/or modify
	/// it under the terms of the GNU General Public License (GPL) as published
	/// by the Free Software Foundation. cf. the file COPYING.txt.
	///
	/// ------------ *** end of legal notice *** ------------------
	/// </summary>

	public class XZCalLst
	{
		public XZCalLst()
		{
			uFirst = 1;
			Common.saveI = 2;
		}

		public static void Trace(MultiZ x, MultiZ z) // z = sum of elements on the principal diagonal
		{
			if (x.Rs == 0)
				MultiZD.ErrNotMatrix();
			else
			{
				bool saveExp = MultiFD.expanded;
				int saveFMC = Calc.fMC;
				int saveMMax = MultiID.mMax;
				XPCalLst.Expand();
				MultiC sum = new MultiC(Calc.fMC);
				int n = Math.Min(x.Rs, x.Cs);
				for (int i = 0; i < n; i++)
					sum += x.E[i, i];
				MultiFD.expanded = saveExp;
				Calc.fMC = saveFMC;
				MultiID.mMax = saveMMax;
				Calc.SetToNM(sum, z);
			}
		} // Trace

		public static void TraceN(MultiZ x, int zI) // z = sum of elements on the principal diagonal
		{
			if (x.Rs == 0)
				MultiZD.ErrNotMatrix();
			else
			{
				MultiZ z = new MultiZ();
				Trace(x, z);
				Calc.SetToN(z, zI);			}
		} // TraceN

		public static void IsDiagN(MultiZ x, int zI) // is x a diagonal matrix 1 (True), 0 (False
		{
			if (x.Rs == 0 || x.Rs != x.Cs)
				Calc.ClearN(zI);
			else
			{
				int i = 1;
				for (int r = 0; r < x.Rs; r++)
				{
					for (int c = 0; c < x.Cs; c++)
					{
						if (r != c && x.E[r, c] != MultiCD.cZero)
						{
							i = 0;
							break;
						}
					}
					if (i == 0)
						break;
				}
				Calc.ClearN(zI);
				Common.nList[zI].V[0] = i;
			}
		} // IsDiagN

		public static void MatIdN(MultiZ x, int zI) // z = x by x identity matrix
		{
			int n;
			MultiFD.Get1F(out n, x);
			if (x.Rs != 0)
				MultiZD.ErrNotScalar();
			else if (n < 1)
				MultiZD.ErrIncompatible();
			else
			{
				int tI;
				Common.InitNAdd(1, "MatIdN-tI", out tI);
				MatrixN(n, n, tI);
				for (int r = 0; r < n; r++)
					for (int c = 0; c < n; c++)
						Common.nList[tI].E[r, c] &= MultiCD.cZero;
				for (int i = 0; i < n; i++)
					Common.nList[tI].E[i, i] &= MultiCD.cOne;
				Calc.SetToII(tI, zI);
				Common.LDelete(tI);
			}
		} // MatIdN

		public static void HilbertN(MultiZ x, int zI) // z = x by x Hilbert matrix
		// zij = 1/(i+j-1)
		{
			int n;
			MultiFD.Get1F(out n, x);
			if (x.Rs != 0)
				MultiZD.ErrNotScalar();
			else if (n < 1)
				MultiZD.ErrIncompatible();
			else
			{
				bool saveExp = MultiFD.expanded;
				int saveFMC = Calc.fMC;
				int saveMMax = MultiID.mMax;
				XPCalLst.Expand();
				MultiZ z = new MultiZ(Calc.fMC, n, n);
				for (int i = 0; i < n; i++)
					for (int j = 0; j < n; j++)
						MultiCD.Div1C(MultiCD.cOne, i + j + 1, z.E[i, j]);
				Calc.SetToN(z, zI);
				MultiFD.expanded = saveExp;
				Calc.fMC = saveFMC;
				MultiID.mMax = saveMMax;
				Calc.NormN(zI);
			}
		} // HilbertN

		public static void PascalN(MultiZ x, int zI) // z = x by x Pascal triangle
		// zij = Bino(i, j)
		{
			int n;
			MultiFD.Get1F(out n, x);
			if (x.Rs != 0)
				MultiZD.ErrNotScalar();
			else if (n < 1)
				MultiZD.ErrIncompatible();
			else
			{
				int tI;
				Common.InitNAdd(1, "PascalN-tI", out tI);
				MultiF iF = new MultiF(1);
				MultiF jF = new MultiF(1);
				MatrixN(n, n, zI);
				for (int i = 0; i < n; i++)
				{
					iF &= i;
					for (int j = 0; j < n; j++)
					{
						jF &= j;
						XPCalLst.BinoN(iF, jF, tI);
						if (MultiID.muErr)
						{
							Common.LDelete(tI);
							return;
						}
						Calc.SetToNM(Common.nList[tI], Common.nList[zI].E[i, j]);
					}
				}
				Common.LDelete(tI);
			}
		} // PascalN

		public static void DiagonalN(MultiZ x, int zI) // z = diagonal matrix from vector x
		{
			Calc.SetToN(MultiZD.Diagonal(x), zI);
		} // DiagonalN

		public static void Del(MultiZ x, int c, MultiZ z) // delete column c from matrix x
		{
			if (c < 0 || c >= x.Cs)
				Calc.SetToNM(x, z);
			else if (x.Cs == 1)
				Calc.SetToNM(MultiCD.cZero, z);
			else
			{
				MultiZ zz = new MultiZ(1, x.Rs, x.Cs - 1);
				for (int r = 0; r < x.Rs; r++)
					for (int ci = 0; ci < c; ci++)
						Calc.SetToNM(x.E[r, ci], zz.E[r, ci]);
				for (int r = 0; r < x.Rs; r++)
					for (int ci = c + 1; ci < x.Cs; ci++)
						Calc.SetToNM(x.E[r, ci], zz.E[r, ci - 1]);
				Calc.SetToNM(zz, z);
			}
		} // Del

		public static void DelN(MultiZ x, MultiZ y, int zI) // delete column y from matrix x
		{
			int c;
			MultiFD.Get1F(out c, y);
			if (x.Rs == 0)
				MultiZD.ErrNotMatrix();
			else if (y.Rs != 0)
				MultiZD.ErrNotScalar();
			else if (c < 1 || c > x.Cs)
				MultiZD.ErrIncompatible();
			else
			{
				c -= uFirst;
				MultiZ zz = new MultiZ();
				Del(x, c, zz);
				Calc.SetToN(zz, zI);
			}
		} // DelN

		public static void DelR(MultiZ x, int r, MultiZ z) // delete row r from matrix x
		{
			if (r < 0 || r >= x.Rs)
				Calc.SetToNM(x, z);
			else if (x.Rs == 1)
				Calc.SetToNM(MultiCD.cZero, z);
			else
			{
				MultiZ zz = new MultiZ(1, x.Rs - 1, x.Cs);
				for (int ri = 0; ri < r; ri++)
					for (int c = 0; c < x.Cs; c++)
						Calc.SetToNM(x.E[ri, c], zz.E[ri, c]);
				for (int ri = r + 1; ri < x.Rs; ri++)
					for (int c = 0; c < x.Cs; c++)
						Calc.SetToNM(x.E[ri, c], zz.E[ri - 1, c]);
				Calc.SetToNM(zz, z);
			}
		} // DelR

		public static void DelRN(MultiZ x, MultiZ y, int zI) // delete row y from matrix x
		{
			int r;
			MultiFD.Get1F(out r, y);
			if (x.Rs == 0)
				MultiZD.ErrNotMatrix();
			else if (y.Rs != 0)
				MultiZD.ErrNotScalar();
			else if (r < 1 || r > x.Rs)
				MultiZD.ErrIncompatible();
			else
			{
				r -= uFirst;
				MultiZ zz = new MultiZ();
				DelR(x, r, zz);
				Calc.SetToN(zz, zI);
			}
		} // DelRN

		public static void ExtractN(MultiZ x, MultiZ y, int zI) // extract column y from matrix x
		{
			int c;
			MultiFD.Get1F(out c, y);
			if (x.Rs == 0)
				MultiZD.ErrNotMatrix();
			else if (y.Rs != 0)
				MultiZD.ErrNotScalar();
			else if (c < 1 || c > x.Cs)
				MultiZD.ErrIncompatible();
			else
			{
				c -= uFirst;
				int tI;
				Common.InitNAdd(1, "ExtractN-tI", out tI);
				MatrixN(x.Rs, 1, tI);
				for (int r = 0; r < x.Rs; r++)
					Calc.SetToNM(x.E[r, c], Common.nList[tI].E[r, 0]);
				Calc.SetToII(tI, zI);
				Common.LDelete(tI);
			}
		} // ExtractN

		public static void ExtractRN(MultiZ x, MultiZ y, int zI) // extract row y from matrix x
		{
			int r;
			MultiFD.Get1F(out r, y);
			if (x.Rs == 0)
				MultiZD.ErrNotMatrix();
			else if (y.Rs != 0)
				MultiZD.ErrNotScalar();
			else if (r < 1 || r > x.Rs)
				MultiZD.ErrIncompatible();
			else
			{
				r -= uFirst;
				int tI;
				Common.InitNAdd(1, "ExtractRN-tI", out tI);
				MatrixN(1, x.Cs, tI);
				for (int c = 0; c < x.Cs; c++)
					Calc.SetToNM(x.E[r, c], Common.nList[tI].E[0, c]);
				Calc.SetToII(tI, zI);
				Common.LDelete(tI);
			}
		} // ExtractRN

		public static void ConcatN(MultiZ x, MultiZ y, int zI) // z = [x y]
		{
			if (x.Rs == 0 || y.Rs == 0)
				MultiZD.ErrNotMatrix();
			else if (x.Rs != y.Rs)
				MultiZD.ErrIncompatible();
			else
			{
				int tI;
				Common.InitNAdd(1, "ConcatN-tI", out tI);
				MatrixN(x.Rs, x.Cs + y.Cs, tI);
				for (int r = 0; r < x.Rs; r++)
					for (int c = 0; c < x.Cs; c++)
						Calc.SetToNM(x.E[r, c], Common.nList[tI].E[r, c]);
				for (int r = 0; r < y.Rs; r++)
					for (int c = 0; c < y.Cs; c++)
						Calc.SetToNM(y.E[r, c], Common.nList[tI].E[r, c+x.Cs]);
				Calc.SetToII(tI, zI);
				Common.LDelete(tI);
			}
		} // ConcatN

		public static void ConcatRN(MultiZ x, MultiZ y, int zI) // z =	[x]
		{														//		[y]
			if (x.Rs == 0 || y.Rs == 0)
				MultiZD.ErrNotMatrix();
			else if (x.Cs != y.Cs)
				MultiZD.ErrIncompatible();
			else
			{
				int tI;
				Common.InitNAdd(1, "ConcatRN-tI", out tI);
				MatrixN(x.Rs + y.Rs, x.Cs, tI);
				for (int r = 0; r < x.Rs; r++)
					for (int c = 0; c < x.Cs; c++)
						Calc.SetToNM(x.E[r, c], Common.nList[tI].E[r, c]);
				for (int r = 0; r < y.Rs; r++)
					for (int c = 0; c < y.Cs; c++)
						Calc.SetToNM(y.E[r, c], Common.nList[tI].E[r+x.Rs, c]);
				Calc.SetToII(tI, zI);
				Common.LDelete(tI);
			}
		} // ConcatRN

		public static void ContFracN(MultiZ x, MultiZ y, int zI)
		{ // z = x as continuous fraction row vector [b0, b1, b2,..., by-1]
			if (x.Rs != 0 || y.Rs != 0)
				MultiZD.ErrNotScalar();
			else
			{
				MultiF xL = new MultiF(Calc.fMC);
				xL &= x;
				int n;
				MultiFD.Get1F(out n, y);
				if (n < 1)
					n = Int16.MaxValue;
				int tI;
				Common.InitNAdd(1, "ContFracN-tI", out tI);
				MatrixN(1, n, tI);
				MultiF temp = new MultiF(Calc.fMC);
				MultiFD.FloorF(xL, temp);
				XPCalLst.SetToNM(temp, Common.nList[tI].E[0, 0]);
				MultiF z2 = new MultiF(Calc.fMC);
				MultiF z1 = new MultiF(Calc.fMC);
				MultiF aM2 = new MultiF(Calc.fMC);
				MultiF aM1 = new MultiF(Calc.fMC);
				MultiF bM2 = new MultiF(Calc.fMC);
				MultiF bM1 = new MultiF(Calc.fMC);
				MultiF aK = new MultiF(Calc.fMC);
				MultiF bK = new MultiF(Calc.fMC);
				int c;
				for (c = 1; c < n; c++)
				{
					if (c == 1)
					{
						aM1 &= MultiFD.fOne;
						bM1 &= MultiFD.fZero;
						aK &= temp;
						bK &= MultiFD.fOne;
					}
					xL -= temp;
					if (xL == MultiFD.fZero)
						break;
					MultiFD.DivF(MultiFD.fOne, xL, xL); // x = 1 / x
					MultiFD.FloorF(xL, temp);
					if (temp.M - temp.N <= 2)
						break;
					XPCalLst.SetToNM(temp, Common.nList[tI].E[0, c]);
					aM2 &= aM1;
					bM2 &= bM1;
					aM1 &= aK;
					bM1 &= bK;
					aK &= temp * aM1 + aM2;
					bK &= temp * bM1 + bM2;
					z2 &= aK / bK;
					if (z1 == z2)
						break;
					z1 &= z2;
				}
				if (c < n)
				{
					Common.nList[tI].Cs = c;
				}
				n = Common.nList[tI].Cs;
				if (n > 1 && Common.nList[tI].E[0, n - 1] == MultiZD.mOne)
				{
					Common.nList[tI].E[0, n - 2] += MultiZD.mOne;
					Common.nList[tI].Cs--;
				}
				Calc.SetToII(tI, zI);
				Common.LDelete(tI);
			}
		} // ContFracN

		public static void ContFracVN(MultiZ x, int zI)
		// z = Value of continuous fraction x
		{
			if (x.Rs != 1 && x.Cs != 1)
				MultiZD.ErrNotVector();
			else
			{
				bool saveExp = MultiFD.expanded;
				int saveFMC = Calc.fMC;
				int saveMMax = MultiID.mMax;
				XPCalLst.Expand();
				int n = Math.Max(x.Rs, x.Cs);
				MultiF aM2 = new MultiF(Calc.fMC);
				MultiF aM1 = new MultiF(Calc.fMC);
				MultiF bM2 = new MultiF(Calc.fMC);
				MultiF bM1 = new MultiF(Calc.fMC);
				MultiF aK = new MultiF(Calc.fMC);
				MultiF bK = new MultiF(Calc.fMC);
				aM1 &= MultiFD.fOne;
				bM1 &= MultiFD.fZero;
				aK &= x.E[0, 0];
				bK &= MultiFD.fOne;
				for (int i = 1; i < n; i++)
				{
					aM2 &= aM1;
					bM2 &= bM1;
					aM1 &= aK;
					bM1 &= bK;
					if (x.Rs == 1)
					{
						aK &= x.E[0, i] * aM1 + aM2;
						bK &= x.E[0, i] * bM1 + bM2;
					}
					else
					{
						aK &= x.E[i, 0] * aM1 + aM2;
						bK &= x.E[i, 0] * bM1 + bM2;
					}
				}
				aM1 &= aK / bK;
				Calc.SetToN(aM1, zI);
				MultiFD.expanded = saveExp;
				Calc.fMC = saveFMC;
				MultiID.mMax = saveMMax;
				Calc.NormN(zI);
			}
		} // ContFracVN

		public static void InsertN(MultiZ x, MultiZ y, MultiZ cF, int zI)
		// z = x with y inserted at column c
		{
			int c;
			MultiFD.Get1F(out c, cF);
			if (x.Rs == 0 || y.Rs == 0)
				MultiZD.ErrNotMatrix();
			else if (cF.Rs != 0)
				MultiZD.ErrNotScalar();
			else if (x.Rs != y.Rs)
				MultiZD.ErrIncompatible();
			else if (c < 1 || c > x.Cs + 1)
				MultiZD.ErrIncompatible();
			else
			{
				c -= uFirst;
				int tI;
				Common.InitNAdd(1, "InsertN-tI", out tI);
				MatrixN(x.Rs, x.Cs + y.Cs, tI);
				for (int r = 0; r < x.Rs; r++)
					for (int ci = 0; ci < c; ci++)
						Calc.SetToNM(x.E[r, ci], Common.nList[tI].E[r, ci]);
				for (int r = 0; r < y.Rs; r++)
					for (int ci = 0; ci < y.Cs; ci++)
						Calc.SetToNM(y.E[r, ci], Common.nList[tI].E[r, ci + c]);
				for (int r = 0; r < y.Rs; r++)
					for (int ci = c; ci < x.Cs; ci++)
						Calc.SetToNM(x.E[r, ci], Common.nList[tI].E[r, ci + y.Cs]);
				Calc.SetToII(tI, zI);
				Common.LDelete(tI);
			}
		} // InsertN

		public static void InsertRN(MultiZ x, MultiZ y, MultiZ rF, int zI)
		// z = x with y inserted at row r
		{
			int r;
			MultiFD.Get1F(out r, rF);
			if (x.Rs == 0 || y.Rs == 0)
				MultiZD.ErrNotMatrix();
			else if (rF.Rs != 0)
				MultiZD.ErrNotScalar();
			else if (x.Cs != y.Cs)
				MultiZD.ErrIncompatible();
			else if (r < 1 || r > x.Rs + 1)
				MultiZD.ErrIncompatible();
			else
			{
				r -= uFirst;
				int tI;
				Common.InitNAdd(1, "InsertRN-tI", out tI);
				MatrixN(x.Rs + y.Rs, x.Cs, tI);
				for (int ri = 0; ri < r; ri++)
					for (int c = 0; c < x.Cs; c++)
						Calc.SetToNM(x.E[ri, c], Common.nList[tI].E[ri, c]);
				for (int ri = 0; ri < y.Rs; ri++)
					for (int c = 0; c < y.Cs; c++)
						Calc.SetToNM(y.E[ri, c], Common.nList[tI].E[ri + r, c]);
				for (int ri = r; ri < x.Rs; ri++)
					for (int c = 0; c < y.Cs; c++)
						Calc.SetToNM(x.E[ri, c], Common.nList[tI].E[ri + y.Rs, c]);
				Calc.SetToII(tI, zI);
				Common.LDelete(tI);
			}
		} // InsertRN

		public static void RSizeN(int zI) // z = size of z
		{
			int rs = Common.nList[zI].Rs;
			if (rs == 0)
				Calc.ClearN(zI);
			else
			{
				int cs = Common.nList[zI].Cs;
				MatrixN(1, 2, zI);
				Common.nList[zI].E[0, 0] &= rs;
				Common.nList[zI].E[0, 1] &= cs;
			}
		}

		public static void RRIntN(int yI, int zI) // z = Random Integer between y and z inclusive
		{
			int yrs = Common.nList[yI].Rs;
			int zrs = Common.nList[zI].Rs;
			int ycs = Common.nList[yI].Cs;
			int zcs = Common.nList[zI].Cs;
			if (yrs == 0 && zrs == 0)
				XPCalLst.RRIntN(yI, zI);
			else if (yrs != zrs || ycs != zcs)
				MultiZD.ErrIncompatible();
			else
			{
				int tyI;
				int tzI;
				Common.InitNAdd(1, "RRIntN-tyI", out tyI);
				Common.InitNAdd(1, "RRIntN-tzI", out tzI);
				for (int r = 0; r < yrs; r++)
					for (int c = 0; c < ycs; c++)
					{
						XPCalLst.SetToNM(Common.nList[yI].E[r, c], (MultiF)Common.nList[tyI]);
						XPCalLst.SetToNM(Common.nList[zI].E[r, c], (MultiF)Common.nList[tzI]);
						XPCalLst.RRIntN(tyI, tzI);
						if (MultiID.muErr)
						{
							Common.LDelete(tzI);
							Common.LDelete(tyI);
							return;
						}
						Calc.SetToNM(Common.nList[tzI], Common.nList[zI].E[r, c]);
					}
				Common.LDelete(tzI);
				Common.LDelete(tyI);
			}		
		} // RRIntN

		public static void RAddN(MultiZ x, int zI)	// z = z + x
		// sets muErr true if overflow
		{
			AddN(Common.nList[zI], x, zI);
		}

		public static void RSubN(MultiZ x, int zI) // z = z - x
		// sets muErr true if overflow
		{
			SubN(Common.nList[zI], x, zI);
		}

		public static void AddN(MultiZ x, MultiZ y, int zI) // z = x + y
		// sets muErr true if overflow
		{
			if (x.Rs == 0 && y.Rs == 0)
				XCCalLst.AddN(x, y, zI);
			else if (x.Rs != y.Rs || x.Cs != y.Cs)
				MultiZD.ErrIncompatible();
			else
			{
				if (Common.nList[zI].Rs != x.Rs || Common.nList[zI].Cs != x.Cs)
					MatrixN(x.Rs, x.Cs, zI);
				for (int r = 0; r < x.Rs; r++)
					for (int c = 0; c < x.Cs; c++)
						Calc.SetToNM(x.E[r, c] + y.E[r, c], Common.nList[zI].E[r, c]);
			}
		}

		public static void SubN(MultiZ x, MultiZ y, int zI) // z = x - y
		{
			if (x.Rs == 0 && y.Rs == 0)
				XCCalLst.SubN(x, y, zI);
			else if (x.Rs != y.Rs || x.Cs != y.Cs)
				MultiZD.ErrIncompatible();
			else
			{
				if (Common.nList[zI].Rs != x.Rs || Common.nList[zI].Cs != x.Cs)
					MatrixN(x.Rs, x.Cs, zI);
				for (int r = 0; r < x.Rs; r++)
					for (int c = 0; c < x.Cs; c++)
						Calc.SetToNM(x.E[r, c] - y.E[r, c], Common.nList[zI].E[r, c]);
			}
		}

		public static void FormatN(StreamWriter outF, string name, MultiZ z)
		// output z as a formatted set of lines and pages
		// z is not modified
		{
			if (z.Rs == 0)
			{
				string nmR = name;
				if (!MultiFD.ZTestF(z.I)) nmR = name + " .r";
				XPCalLst.FormatN(outF, nmR, z); // output x to file #outF
				if (!MultiFD.ZTestF(z.I))
				{
					MultiID.WriteLn(null, outF, "End of real part");
					XPCalLst.FormatN(outF, name + " .i", z.I); // output x to file #outF
				}
				return;
			}
			bool wB = MultiID.writeBool;
			MultiID.writeBool = false;
			int bO = MultiID.baseO;
			MultiID.baseO = 10;
			int fTn = MultiFD.fTn;
			MultiFD.fTn = 0;
			bool wE = MultiFD.writeExtra;
			MultiFD.writeExtra = false;
			string nLc = MultiID.nLc;
			MultiID.nLc = "-" + MultiID.nL;
			MultiID.WriteLn(null, outF, "@Echo Off");
			MultiID.WriteLn(null, outF, "Rem  Code file for matrix " + name);
			int rs = z.Rs;
			int cs = z.Cs;
			MultiID.WriteLn(null, outF);
			MultiID.WriteLn(null, outF, name + " = Mat(" + rs + ", " + cs + ");");
			for (int r = 0; r < rs; r++)
				for (int c = 0; c < cs; c++)
				{
					int tot = 0;
					MultiID.WriteLn(null, outF);
					MultiID.WriteT(null, outF, name + " = Set(" + name + ", " + (r+uFirst) + ", " + (c+uFirst) + ", ", ref tot);
					MultiCD.WriteC(null, outF, z.E[r, c], ref tot, 0);
					MultiID.WriteTLn(null, outF, ");", ref tot);
				}
			MultiID.writeBool = wB;
			MultiID.baseO = bO;
			MultiFD.fTn = fTn;
			MultiFD.writeExtra = wE;
			MultiID.nLc = nLc;
		} // FormatN

		public static void RepUnitTCN(MultiC x, MultiC y, int zI) // z = x -- y
		{
			int tI;
			Common.InitNAdd(1, "RepUnitTCN-tI", out tI);
			bool saveExp = MultiFD.expanded;
			int saveFMC = Calc.fMC;
			int saveMMax = MultiID.mMax;
			XPCalLst.Expand();
			Calc.SetToN(y, tI);
			Calc.RepUnitN(tI);
			XCCalLst.MulN(x, Common.nList[tI], zI);
			MultiFD.expanded = saveExp;
			Calc.fMC = saveFMC;
			MultiID.mMax = saveMMax;
			Calc.NormN(zI);
			Common.LDelete(tI);
		} // RepUnitTN

		public static void RepUnitTN(MultiZ x, MultiZ y, int zI) // z = x -- y
		{
			if (x.Rs == 0 && y.Rs == 0)
				RepUnitTCN(x, y, zI);
			else if (x.Rs != y.Rs || x.Cs != y.Cs)
				MultiZD.ErrIncompatible();
			else
			{
				int tI;
				Common.InitNAdd(1, "RepUnitTN-tI", out tI);
				if (Common.nList[zI].Rs != x.Rs || Common.nList[zI].Cs != x.Cs)
					MatrixN(x.Rs, x.Cs, zI);
				for (int r = 0; r < x.Rs; r++)
					for (int c = 0; c < x.Cs; c++)
					{
						RepUnitTCN(x.E[r, c], y.E[r, c], tI);
						if (MultiID.muErr)
						{
							Common.LDelete(tI);
							return;
						}
						Calc.SetToNM(Common.nList[tI], Common.nList[zI].E[r, c]);
					}
				Common.LDelete(tI);
			}
		} // RepUnitTN

		public static void CompCCN(MultiC x, MultiC y, string op, string op2, int zI) // z = x op y
		{
			int sign;
			bool met;
			sign = MultiCD.CompCC(x, y);
			if (op2 == "<=")
				met = (sign <= 0);
			else if (op2 == "<>" || op2 == "!=")
				met = (sign != 0);
			else if (op2 == ">=")
				met = (sign >= 0);
			else if (op2 == "==")
				met = (sign == 0);
			else if (op == "<")
				met = (sign < 0);
			else if (op == "=")
				met = (sign == 0);
			else if (op == ">")
				met = (sign > 0);
			else
				met = false; // should not get here
			Calc.ClearN(zI);
			if (met) Common.nList[zI].V[0] = 1; // n is initially zero
		} // CompCCN

		public static void CompN(MultiZ x, MultiZ y, string op, string op2, int zI) // z = x op y
		{
			int tI;
			if (x.Rs == 0 && y.Rs == 0)
			{
				Common.InitNAdd(1, "CompN-tI", out tI);
				CompCCN(x, Common.nList[zI], op, op2, tI);
				Calc.SetToNM(Common.nList[tI], Common.nList[zI]);
				Common.LDelete(tI);
			}
			else if (x.Rs != y.Rs || x.Cs != y.Cs)
				MultiZD.ErrIncompatible();
			else
			{
				Common.InitNAdd(1, "CompN-tI", out tI);
				if (Common.nList[zI].Rs != x.Rs || Common.nList[zI].Cs != x.Cs)
					MatrixN(x.Rs, x.Cs, zI);
				for (int r = 0; r < x.Rs; r++)
					for (int c = 0; c < x.Cs; c++)
					{
						CompCCN(x.E[r, c], y.E[r, c], op, op2, tI);
						if (MultiID.muErr)
						{
							Common.LDelete(tI);
							return;
						}
						Calc.SetToNM(Common.nList[tI], Common.nList[zI].E[r, c]);
					}
				Common.LDelete(tI);
			}
		} // CompN

		public static void AndN(MultiZ x, MultiZ y, int zI) // z = x & y
		{
			if (x.Rs == 0 && y.Rs == 0)
				XCCalLst.AndN((MultiC)x, (MultiC)y, zI);
			else if (x.Rs != y.Rs || x.Cs != y.Cs)
				MultiZD.ErrIncompatible();
			else
			{
				int tI;
				Common.InitNAdd(1, "AndN-tI", out tI);
				if (Common.nList[zI].Rs != x.Rs || Common.nList[zI].Cs != x.Cs)
					MatrixN(x.Rs, x.Cs, zI);
				for (int r = 0; r < x.Rs; r++)
					for (int c = 0; c < x.Cs; c++)
					{
						XCCalLst.AndN(x.E[r, c], y.E[r, c], tI);
						Calc.SetToNM(Common.nList[tI], Common.nList[zI].E[r, c]);
					}
				Common.LDelete(tI);
			}
		} // AndN

		public static void OrN(MultiZ x, MultiZ y, int zI) // z = x & y
		{
			if (x.Rs == 0 && y.Rs == 0)
				XCCalLst.OrN(x, y, zI);
			else if (x.Rs != y.Rs || x.Cs != y.Cs)
				MultiZD.ErrIncompatible();
			else
			{
				int tI;
				Common.InitNAdd(1, "OrN-tI", out tI);
				if (Common.nList[zI].Rs != x.Rs || Common.nList[zI].Cs != x.Cs)
					MatrixN(x.Rs, x.Cs, zI);
				for (int r = 0; r < x.Rs; r++)
					for (int c = 0; c < x.Cs; c++)
					{
						XCCalLst.OrN(x.E[r, c], y.E[r, c], tI);
						if (MultiID.muErr)
						{
							Common.LDelete(tI);
							return;
						}
						Calc.SetToNM(Common.nList[tI], Common.nList[zI].E[r, c]);
					}
				Common.LDelete(tI);
			}
		} // OrN

		public static void MulN(MultiZ x, MultiZ y, int zI) // z = x * y
		// x or y or both may have the same address in memory as z
		// x, y are not changed unless they share memory with z
		// sets muErr true if overflow
		{
			if (x.Rs == 0 && y.Rs == 0)
				XCCalLst.MulN(x, y, zI);
			else if (x.Rs != 0 && y.Rs != 0 && x.Cs != y.Rs)
				MultiZD.ErrIncompatible();
			else if (x.Rs == 0)
			{
				MultiZ z = new MultiZ(Calc.fMC, y.Rs, y.Cs);
				MultiZD.MulZ(x, y, z);
				Calc.SetToN(z, zI);
			}
			else if (y.Rs == 0)
			{
				MultiZ z = new MultiZ(Calc.fMC, x.Rs, x.Cs);
				MultiZD.MulZ(x, y, z);
				Calc.SetToN(z, zI);
			}
			else
			{
				bool saveExp = MultiFD.expanded;
				int saveFMC = Calc.fMC;
				int saveMMax = MultiID.mMax;
				XPCalLst.Expand();
				MultiZ z = new MultiZ(Calc.fMC, x.Rs, y.Cs);
				MultiZD.MulZ(x, y, z);
				Calc.SetToN(z, zI);
				MultiFD.expanded = saveExp;
				Calc.fMC = saveFMC;
				MultiID.mMax = saveMMax;
				Calc.NormN(zI);
			}
		} // MulN

		public static void MulQN(MultiZ x, MultiZ y, int zI) // z = x * y, all quaternions
		// x or y or both may have the same address in memory as z
		// x, y are not changed unless they share memory with z
		// sets muErr true if overflow
		{
				bool saveExp = MultiFD.expanded;
				int saveFMC = Calc.fMC;
				int saveMMax = MultiID.mMax;
				XPCalLst.Expand();
				MultiZ z = new MultiZ(Calc.fMC, 4, 1);
				MultiZD.MulQ(x, y, z);
				Calc.SetToN(z, zI);
				MultiFD.expanded = saveExp;
				Calc.fMC = saveFMC;
				MultiID.mMax = saveMMax;
				Calc.NormN(zI);
		} // MulN

		public static void RSqN(int zI) // z = z * z
		// sets muErr true if overflow
		{
			int rs = Common.nList[zI].Rs;
			int cs = Common.nList[zI].Cs;
			if (rs == 0)
				XCCalLst.RSqN(zI);
			else if (rs != cs)
				MultiZD.ErrNotSquare();
			else
			{
				bool saveExp = MultiFD.expanded;
				int saveFMC = Calc.fMC;
				int saveMMax = MultiID.mMax;
				XPCalLst.Expand();
				MultiZ z = new MultiZ(Calc.fMC, rs, cs);
				z &= Common.nList[zI];
				MultiZD.RSqZ(z);
				Calc.SetToN(z, zI);
				MultiFD.expanded = saveExp;
				Calc.fMC = saveFMC;
				MultiID.mMax = saveMMax;
				Calc.NormN(zI);
			}		
		} // RSqN

		public static void PowN(MultiZ x, MultiZ y, int zI) // z = x^y
		{
			if (x.Rs == 0 && y.Rs == 0)
				XCCalLst.PowN(x, y, zI);
			else if (x.Rs != y.Rs || x.Cs != y.Cs)
				MultiZD.ErrIncompatible();
			else
			{
				int tI;
				Common.InitNAdd(1, "PowN-tI", out tI);
				if (Common.nList[zI].Rs != x.Rs || Common.nList[zI].Cs != x.Cs)
					MatrixN(x.Rs, x.Cs, zI);
				for (int r = 0; r < x.Rs; r++)
					for (int c = 0; c < x.Cs; c++)
					{
						XCCalLst.PowN(x.E[r, c], y.E[r, c], tI);
						if (MultiID.muErr)
						{
							Common.LDelete(tI);
							return;
						}
						Calc.SetToNM(Common.nList[tI], Common.nList[zI].E[r, c]);
					}
				Common.LDelete(tI);
			}		
		} // PowN

		public static void RPowN(MultiZ p, int zI) // z = z^p
		{
			PowN(Common.nList[zI], p, zI);
		}

		public static void PowMN(MultiZ x, MultiZ y, int zI) // z = x^y % FMB
		{
			if (x.Rs == 0 && y.Rs == 0)
				XCCalLst.PowMN(x, y, zI);
			else if (x.Rs != y.Rs || x.Cs != y.Cs)
				MultiZD.ErrIncompatible();
			else
			{
				int tI;
				Common.InitNAdd(1, "PowMN-tI", out tI);
				if (Common.nList[zI].Rs != x.Rs || Common.nList[zI].Cs != x.Cs)
					MatrixN(x.Rs, x.Cs, zI);
				for (int r = 0; r < x.Rs; r++)
					for (int c = 0; c < x.Cs; c++)
					{
						XCCalLst.PowMN(x.E[r, c], y.E[r, c], tI);
						if (MultiID.muErr)
						{
							Common.LDelete(tI);
							return;
						}
						Calc.SetToNM(Common.nList[tI], Common.nList[zI].E[r, c]);
					}
				Common.LDelete(tI);
			}		
		} // PowMN

		public static void RPowMN(MultiZ p, int zI) // z = z^p % FMB
		{
			PowMN(Common.nList[zI], p, zI);
		}

		public static void DivN(MultiZ u, MultiZ d, int zI) // z = u / d
		// u or d or both may have the same address in memory as z
		// u, d are not changed unless they share memory with z
		// sets muErr true if d = 0
		{
			MultiZ z;
			int n = 1;
			if(d.Rs == 0)
			{
				z = new MultiZ(Calc.fMC, u.Rs, u.Cs);
			}
			else
			{
				n = Math.Max(u.Rs, d.Rs);
				z = new MultiZ(Calc.fMC, n, n);
			}
			MultiZD.DivZ(u, d, z);
			Calc.SetToN(z, zI);
		} // DivN

		public static void DivRemN(MultiZ x, MultiZ y, int zI) // zij = Floor(xij / yij), Reij = remainder
		// u or d or both may have the same address in memory as z
		// u, d are not changed unless they share memory with z
		// sets muErr true if d = 0
		{
			if (x.Rs == 0 && y.Rs == 0)
				XCCalLst.DivRemN(x, y, zI);
			else if (x.Rs != y.Rs || x.Cs != y.Cs)
				MultiZD.ErrIncompatible();
			else
			{
				int tI;
				int reMI;
				Common.InitNAdd(1, "DivRemN-tI", out tI);
				Common.InitNAdd(1, "DivRemN-x2MI", out reMI);
				MatrixN(x.Rs, x.Cs, reMI);
				int reI;
				Common.LFindAdd("Re", out reI);
				if (Common.nList[zI].Rs != x.Rs || Common.nList[zI].Cs != x.Cs)
					MatrixN(x.Rs, x.Cs, zI);
				for (int r = 0; r < x.Rs; r++)
					for (int c = 0; c < x.Cs; c++)
					{
						XCCalLst.DivRemN(x.E[r, c], y.E[r, c], tI);
						if (MultiID.muErr)
						{
							Common.LDelete(reMI);
							Common.LDelete(tI);
							Calc.SetToN(MultiZD.mZero, reI);
							Calc.SetToN(MultiZD.mZero, zI);
							return;
						}
						Calc.SetToNM(Common.nList[tI], Common.nList[zI].E[r, c]);
						Calc.SetToNM(Common.nList[reI], Common.nList[reMI].E[r, c]);
					}
				Calc.SetToII(reMI, reI);
				Common.LDelete(reMI);
				Common.LDelete(tI);
			}
		} // DivRemN

		public static void DivIntN(MultiZ x, MultiZ y, int zI) // zij = Floor(xij / yij)
		// u or d or both may have the same address in memory as z
		// u, d are not changed unless they share memory with z
		// sets muErr true if d = 0
		{
			if (x.Rs == 0 && y.Rs == 0)
				XCCalLst.DivIntN(x, y, zI);
			else if (x.Rs != y.Rs || x.Cs != y.Cs)
				MultiZD.ErrIncompatible();
			else
			{
				int tI;
				Common.InitNAdd(1, "DivIntN-tI", out tI);
				if (Common.nList[zI].Rs != x.Rs || Common.nList[zI].Cs != x.Cs)
					MatrixN(x.Rs, x.Cs, zI);
				for (int r = 0; r < x.Rs; r++)
					for (int c = 0; c < x.Cs; c++)
					{
						XCCalLst.DivIntN(x.E[r, c], y.E[r, c], tI);
						if (MultiID.muErr)
						{
							Common.LDelete(tI);
							return;
						}
						Calc.SetToNM(Common.nList[tI], Common.nList[zI].E[r, c]);
					}
				Common.LDelete(tI);
			}
		} // DivIntN

		public static void FacSN(MultiZ x, int zI, bool MFlag, int fMBI)
		// z = x! = 1*2*3*...*x % FMB
		// x may have the same address in memory as zI
		// x is not changed unless it shares memory with zI
		// sets muErr true if overflow or error
		{
			if (x.Rs == 0)
				XPCalLst.FacSN(x, zI, MFlag, fMBI);
			else
			{
				int tI;
				Common.InitNAdd(1, "FacN-tI", out tI);
				if (Common.nList[zI].Rs != x.Rs || Common.nList[zI].Cs != x.Cs)
					MatrixN(x.Rs, x.Cs, zI);
				for (int r = 0; r < x.Rs; r++)
					for (int c = 0; c < x.Cs; c++)
					{
						XPCalLst.FacSN(x.E[r, c], tI, MFlag, fMBI);
						if (MultiID.muErr)
						{
							Common.LDelete(tI);
							return;
						}
						Calc.SetToNM(Common.nList[tI], Common.nList[zI].E[r, c]);
					}
				Common.LDelete(tI);
			}
		} // FacSN

		public static void Fac2N(MultiZ x, MultiZ y, int zI, bool MFlag)
		// z = Fac(x) / Fac(y) % fMB by binary splitting
		// x or y or both may have the same address in memory as z
		// x, y are not changed unless they share memory with z
		// sets muErr true if y = 0
		{
			if (x.Rs == 0 && y.Rs == 0)
				XPCalLst.Fac2N(x, y, zI, MFlag);
			else if (x.Rs != y.Rs || x.Cs != y.Cs)
				MultiZD.ErrIncompatible();
			else
			{
				int tI;
				Common.InitNAdd(1, "Fac2N-tI", out tI);
				if (Common.nList[zI].Rs != x.Rs || Common.nList[zI].Cs != x.Cs)
					MatrixN(x.Rs, x.Cs, zI);
				for (int r = 0; r < x.Rs; r++)
					for (int c = 0; c < x.Cs; c++)
					{
						XPCalLst.Fac2N(x.E[r, c], y.E[r, c], tI, MFlag);
						if (MultiID.muErr)
						{
							Common.LDelete(tI);
							return;
						}
						Calc.SetToNM(Common.nList[tI], Common.nList[zI].E[r, c]);
					}
				Common.LDelete(tI);
			}
		} // Fac2N

		public static void FacN(MultiZ x, int zI, bool MFlag, int fMBI)
		// z = x! = 1*2*3*...*x % FMB
		// x may have the same address in memory as zI
		// x is not changed unless it shares memory with zI
		// sets muErr true if overflow or error
		{
			if (x.Rs == 0)
				XPCalLst.FacN(x, zI, MFlag, fMBI);
			else
			{
				int tI;
				Common.InitNAdd(1, "FacN-tI", out tI);
				if (Common.nList[zI].Rs != x.Rs || Common.nList[zI].Cs != x.Cs)
					MatrixN(x.Rs, x.Cs, zI);
				for (int r = 0; r < x.Rs; r++)
					for (int c = 0; c < x.Cs; c++)
					{
						XPCalLst.FacN(x.E[r, c], tI, MFlag, fMBI);
						if (MultiID.muErr)
						{
							Common.LDelete(tI);
							return;
						}
						Calc.SetToNM(Common.nList[tI], Common.nList[zI].E[r, c]);
					}
				Common.LDelete(tI);
			}
		} // FacN

		public static void AFibN(MultiZ x, int zI)
		// z = Arc Fibonacci number Abs(x), always an integer, accurate if x is a Fibonacci number
		// x may have the same address in memory as zI
		// x is not changed unless it shares memory with zI
		// sets muErr true if overflow or error
		{
			if (x.Rs == 0)
			{
				XPCalLst.AFibN(x, zI);
				MultiFD.ClearF(Common.nList[zI].I);
				Calc.NormN(zI);
			}
			else
			{
				int tI;
				Common.InitNAdd(1, "AFibN-tI", out tI);
				if (Common.nList[zI].Rs != x.Rs || Common.nList[zI].Cs != x.Cs)
					MatrixN(x.Rs, x.Cs, zI);
				for (int r = 0; r < x.Rs; r++)
					for (int c = 0; c < x.Cs; c++)
					{
						XPCalLst.AFibN(x.E[r, c], tI);
						if (MultiID.muErr)
						{
							Common.LDelete(tI);
							return;
						}
						MultiFD.ClearF(Common.nList[tI].I);
						Calc.NormN(tI);
						Calc.SetToNM(Common.nList[tI], Common.nList[zI].E[r, c]);
					}
				Common.LDelete(tI);
			}
		} // AFibN

		public static void FibN(MultiZ x, int zI)
		// z = Fibonacci number of x
		// x may have the same address in memory as zI
		// x is not changed unless it shares memory with zI
		// sets muErr true if overflow or error
		{
			if (x.Rs == 0)
				XCCalLst.FibN(x, zI);
			else
			{
				int tI;
				Common.InitNAdd(1, "FibN-tI", out tI);
				if (Common.nList[zI].Rs != x.Rs || Common.nList[zI].Cs != x.Cs)
					MatrixN(x.Rs, x.Cs, zI);
				for (int r = 0; r < x.Rs; r++)
					for (int c = 0; c < x.Cs; c++)
					{
						XCCalLst.FibN(x.E[r, c], tI);
						if (MultiID.muErr)
						{
							Common.LDelete(tI);
							return;
						}
						Calc.SetToNM(Common.nList[tI], Common.nList[zI].E[r, c]);
					}
				Common.LDelete(tI);
			}		
		} // FibN

		public static void RIsFibN(int zI)
		// z = 1 (True) if |z| is a Fibonacci number, else 0 (False)
		// True iff z is an integer and the closed interval
		// [Phi*z - 1/z, Phi*z + 1/z] contains a positive integer
		{
			if (Common.nList[zI].Rs != 0)
				MultiZD.ErrNotScalar();
			else if (Common.nList[zI].I != MultiFD.fZero)
				Calc.SetToN(MultiFD.fZero, zI);
			else
				XPCalLst.RIsFibN(zI);
		} // RIsFibN

		public static void RIsFib2N(int zI)
		// z = 1 (True) if |z| is a Fibonacci number, else 0 (False)
		// True iff z is an integer and 5z^2+4 or 5z^2-4 is a perfect square
		{
			if (Common.nList[zI].Rs != 0)
				MultiZD.ErrNotScalar();
			else if (Common.nList[zI].I != MultiFD.fZero)
				Calc.SetToN(MultiFD.fZero, zI);
			else
				XPCalLst.RIsFib2N(zI);
		} // RIsFib2N

		public static void Lerch1(MultiZ x, MultiZ s, MultiZ a, bool T, int zI) // z = Lerch transcendent capitol Phi(x, s, a)
		// x, s, a may have the same address in memory as z				// to 16 decimal digits and traditional
		// x, s, a are not changed unless it shares memory with x
		// sets muErr true if overflow or error
		//
		// The LerchT1(x, s, a) function is Psi(x, s, a) = Sum {k = 0, 1, ...} of x^k / (a + k)^s,
		// where |x| <= 1 and a != 0, -1, -2, ... .
		// Lerch1(x, s, a) = sum[k=0, infinity]{x^k / |a + k|^s}, terms with a + k == 0 are ignored, |x| <= 1
		{
			if (x.Rs == 0 && s.Rs == 0 && a.Rs == 0)
				XPCalLst.Lerch1(x, s, a, T, zI);
			else if (x.Rs != s.Rs || x.Rs != a.Rs || x.Cs != s.Cs || x.Cs != a.Cs)
				MultiZD.ErrIncompatible();
			else
			{
				int tI;
				Common.InitNAdd(1, "Lerch1-tI", out tI);
				if (Common.nList[zI].Rs != x.Rs || Common.nList[zI].Cs != x.Cs)
					MatrixN(x.Rs, x.Cs, zI);
				for (int r = 0; r < x.Rs; r++)
					for (int c = 0; c < x.Cs; c++)
					{
						XPCalLst.Lerch1(x.E[r, c], s.E[r, c], a.E[r, c], T, tI);
						if (MultiID.muErr)
						{
							Common.LDelete(tI);
							return;
						}
						Calc.SetToNM(Common.nList[tI], Common.nList[zI].E[r, c]);
					}
				Common.LDelete(tI);
			}
		} // Lerch1

		public static void LerchT(MultiZ x, MultiZ s, MultiZ a, int zI) // z = Lerch transcendent capitol Phi(x, s, a)
		{
			if (x.Rs == 0 && s.Rs == 0 && a.Rs == 0)
				XPCalLst.Lerch(x, s, a, true, zI);
			else if (x.Rs != s.Rs || x.Rs != a.Rs || x.Cs != s.Cs || x.Cs != a.Cs)
				MultiZD.ErrIncompatible();
			else
			{
				int tI;
				Common.InitNAdd(1, "LerchT-tI", out tI);
				if (Common.nList[zI].Rs != x.Rs || Common.nList[zI].Cs != x.Cs)
					MatrixN(x.Rs, x.Cs, zI);
				for (int r = 0; r < x.Rs; r++)
					for (int c = 0; c < x.Cs; c++)
					{
						XPCalLst.Lerch(x.E[r, c], s.E[r, c], a.E[r, c], true, tI);
						if (MultiID.muErr)
						{
							Common.LDelete(tI);
							return;
						}
						Calc.SetToNM(Common.nList[tI], Common.nList[zI].E[r, c]);
					}
				Common.LDelete(tI);
			}
		} // LerchT

		public static void Lerch(MultiZ x, MultiZ s, MultiZ a, bool T, int zI) // z = Lerch transcendent capitol Phi(x, s, a)
		// Function that computes approximation to Lerch Phi as a
		// converging sequence of CNC transforms S^n_k.
		// x, s, a may have the same address in memory as z				// traditional
		// x, s, a are not changed unless it shares memory with z
		// sets muErr true if overflow or error
		//
		// The LerchT(x, s, a) function is Phi(x, s, a) = Sum {k = 0, 1, ...} of x^k / (a + k)^s,
		// where |x| <= 1 and a != 0, -1, -2, ... .
		// Lerch(x, s, a) = sum[k=0, infinity]{x^k / |a + k|^s}, terms with a + k == 0 are ignored, |x| <= 1
		{
			if (T)
				LerchT(x, s, a, zI);
			else if (x.Rs == 0 && s.Rs == 0 && a.Rs == 0)
				XCCalLst.Lerch(x, s, a, zI);
			else if (x.Rs != s.Rs || x.Rs != a.Rs || x.Cs != s.Cs || x.Cs != a.Cs)
				MultiZD.ErrIncompatible();
			else
			{
				int tI;
				Common.InitNAdd(1, "Lerch-tI", out tI);
				if (Common.nList[zI].Rs != x.Rs || Common.nList[zI].Cs != x.Cs)
					MatrixN(x.Rs, x.Cs, zI);
				for (int r = 0; r < x.Rs; r++)
					for (int c = 0; c < x.Cs; c++)
					{
						XCCalLst.Lerch(x.E[r, c], s.E[r, c], a.E[r, c], tI);
						if (MultiID.muErr)
						{
							Common.LDelete(tI);
							return;
						}
						Calc.SetToNM(Common.nList[tI], Common.nList[zI].E[r, c]);
					}
				Common.LDelete(tI);
			}
		} // Lerch

		public static void Lerch2(MultiZ x, MultiZ s, MultiZ a, bool T, int zI) // z = Lerch transcendent capitol Phi(x, s, a)
		// x, s, a may have the same address in memory as z			// by a simple sum and traditional
		// x, s, a are not changed unless it shares memory with z
		// sets muErr true if overflow or error
		//
		// The LerchT2(x, s, a) function is Phi(x, s, a) = Sum {k = 0, 1, ...} of x^k / (a + k)^s,
		// where |x| <= 1 and a != 0, -1, -2, ... .
		// Lerch2(x, s, a) = sum[k=0, infinity]{x^k / |a + k|^s}, terms with a + k == 0 are ignored, |x| <= 1
		{
			if (x.Rs == 0 && s.Rs == 0 && a.Rs == 0)
				XPCalLst.Lerch2(x, s, a, T, zI);
			else if (x.Rs != s.Rs || x.Rs != a.Rs || x.Cs != s.Cs || x.Cs != a.Cs)
				MultiZD.ErrIncompatible();
			else
			{
				int tI;
				Common.InitNAdd(1, "Lerch2-tI", out tI);
				if (Common.nList[zI].Rs != x.Rs || Common.nList[zI].Cs != x.Cs)
					MatrixN(x.Rs, x.Cs, zI);
				for (int r = 0; r < x.Rs; r++)
					for (int c = 0; c < x.Cs; c++)
					{
						XPCalLst.Lerch2(x.E[r, c], s.E[r, c], a.E[r, c], T, tI);
						if (MultiID.muErr)
						{
							Common.LDelete(tI);
							return;
						}
						Calc.SetToNM(Common.nList[tI], Common.nList[zI].E[r, c]);
					}
				Common.LDelete(tI);
			}
		} // Lerch2

		public static void RPYMN(MultiZ r, MultiZ p, MultiZ y, int zI) // z = Rotation matrix
		// r = roll, p = pitch, y = yaw
		// roll matrix = {1, 0, 0; 0, cos(r), sin(r); 0, -sin(r), cos(r)}
		// pitch matrix = {cos(p), 0, -sin(p); 0, 1, 0; sin(p), 0, cos(p)}
		// yaw matrix = {cos(y), sin(y), 0; 0, -sin(y), cos(y), 0; 0, 0, 1}
		// Rotation matrix = yaw * pitch * roll
		{
			int tI;
			Common.InitNAdd(1, "RPYMN-tI", out tI);

			int rI;
			Common.InitNAdd(1, "RPYMN-rI", out rI);
			MatrixN(3, 3, rI);
			CosN(r, tI);
			Calc.SetToNM(Common.nList[tI], Common.nList[rI].E[1, 1]);
			SinN(r, tI);
			Calc.SetToNM(Common.nList[tI], Common.nList[rI].E[1, 2]);
			Calc.SetToNM(-Common.nList[rI].E[1, 2], Common.nList[rI].E[2, 1]);
			Calc.SetToNM(Common.nList[rI].E[1, 1], Common.nList[rI].E[2, 2]);
			Calc.SetToNM(MultiCD.cOne, Common.nList[rI].E[0, 0]);

			int pI;
			Common.InitNAdd(1, "RPYMN-pI", out pI);
			MatrixN(3, 3, pI);
			CosN(p, tI);
			Calc.SetToNM(Common.nList[tI], Common.nList[pI].E[0, 0]);
			SinN(p, tI);
			Calc.SetToNM(Common.nList[tI], Common.nList[pI].E[2, 0]);
			Calc.SetToNM(-Common.nList[pI].E[2, 0], Common.nList[pI].E[0, 2]);
			Calc.SetToNM(Common.nList[pI].E[0, 0], Common.nList[pI].E[2, 2]);
			Calc.SetToNM(MultiCD.cOne, Common.nList[pI].E[1, 1]);

			int yI;
			Common.InitNAdd(1, "RPYMN-yI", out yI);
			MatrixN(3, 3, yI);
			CosN(y, tI);
			Calc.SetToNM(Common.nList[tI], Common.nList[yI].E[0, 0]);
			SinN(y, tI);
			Calc.SetToNM(Common.nList[tI], Common.nList[yI].E[0, 1]);
			Calc.SetToNM(-Common.nList[yI].E[0, 1], Common.nList[yI].E[1, 0]);
			Calc.SetToNM(Common.nList[yI].E[0, 0], Common.nList[yI].E[1, 1]);
			Calc.SetToNM(MultiCD.cOne, Common.nList[yI].E[2, 2]);

			MulN(Common.nList[yI], Common.nList[pI], yI);
			MulN(Common.nList[yI], Common.nList[rI], yI);

			Calc.SetToII(yI, zI);
			Common.LDelete(yI);
			Common.LDelete(pI);
			Common.LDelete(rI);
			Common.LDelete(tI);
		} // RPYMN

		public static void GamN(MultiZ x, int zI)
		// z = gamma(x)
		// x may have the same address in memory as zI
		// x is not changed unless it shares memory with zI
		// sets muErr true if overflow or error
		{
			if (x.Rs == 0)
				XCCalLst.GamN(x, zI);
			else
			{
				int tI;
				Common.InitNAdd(1, "GamN-tI", out tI);
				if (Common.nList[zI].Rs != x.Rs || Common.nList[zI].Cs != x.Cs)
				MatrixN(x.Rs, x.Cs, zI);
				for (int r = 0; r < x.Rs; r++)
				for (int c = 0; c < x.Cs; c++)
				{
					XCCalLst.GamN(x.E[r, c], tI);
					if (MultiID.muErr)
					{
						Common.LDelete(tI);
						return;
					}
					Calc.SetToNM(Common.nList[tI], Common.nList[zI].E[r, c]);
				}
				Common.LDelete(tI);
			}		
		} // GamN

		public static void BetaN(MultiZ x, int zI)
		// z = beta(x)
		// x may have the same address in memory as zI
		// x is not changed unless it shares memory with zI
		// sets muErr true if overflow or error
		{
			if (x.Rs == 0)
				XCCalLst.BetaN(x, zI);
			else
			{
				int tI;
				Common.InitNAdd(1, "BetaN-tI", out tI);
				if (Common.nList[zI].Rs != x.Rs || Common.nList[zI].Cs != x.Cs)
					MatrixN(x.Rs, x.Cs, zI);
				for (int r = 0; r < x.Rs; r++)
					for (int c = 0; c < x.Cs; c++)
					{
						XCCalLst.BetaN(x.E[r, c], tI);
						if (MultiID.muErr)
						{
							Common.LDelete(tI);
							return;
						}
						Calc.SetToNM(Common.nList[tI], Common.nList[zI].E[r, c]);
					}
				Common.LDelete(tI);
			}
		} // BetaN

		public static void BetaL(MultiZ x, int zI)
		// z = beta(x)
		// x may have the same address in memory as zI
		// x is not changed unless it shares memory with zI
		// sets muErr true if overflow or error
		{
			if (x.Rs == 0)
				XCCalLst.BetaL(x, Common.nList[zI]);
			else
			{
				int tI;
				Common.InitNAdd(1, "BetaL-tI", out tI);
				if (Common.nList[zI].Rs != x.Rs || Common.nList[zI].Cs != x.Cs)
					MatrixN(x.Rs, x.Cs, zI);
				for (int r = 0; r < x.Rs; r++)
					for (int c = 0; c < x.Cs; c++)
					{
						XCCalLst.BetaL(x.E[r, c], Common.nList[tI]);
						if (MultiID.muErr)
						{
							Common.LDelete(tI);
							return;
						}
						Calc.SetToNM(Common.nList[tI], Common.nList[zI].E[r, c]);
					}
				Common.LDelete(tI);
			}
		} // BetaL

		public static void ErfN(MultiZ x, int zI)
		// z = error function erf(x)
		// x may have the same address in memory as zI
		// x is not changed unless it shares memory with zI
		// sets muErr true if overflow or error
		{
			if (x.Rs == 0)
				XCCalLst.ErfN(x, zI);
			else
			{
				int tI;
				Common.InitNAdd(1, "ErfN-tI", out tI);
				if (Common.nList[zI].Rs != x.Rs || Common.nList[zI].Cs != x.Cs)
					MatrixN(x.Rs, x.Cs, zI);
				for (int r = 0; r < x.Rs; r++)
					for (int c = 0; c < x.Cs; c++)
					{
						XCCalLst.ErfN(x.E[r, c], tI);
						if (MultiID.muErr)
						{
							Common.LDelete(tI);
							return;
						}
						Calc.SetToNM(Common.nList[tI], Common.nList[zI].E[r, c]);
					}
				Common.LDelete(tI);
			}
		} // ErfN

		public static void ErfCN(MultiZ x, int zI)
		// z = complementary error function erfC(x)
		// x may have the same address in memory as zI
		// x is not changed unless it shares memory with zI
		// sets muErr true if overflow or error
		{
			if (x.Rs == 0)
				XCCalLst.ErfCN(x, zI);
			else
			{
				int tI;
				Common.InitNAdd(1, "ErfCN-tI", out tI);
				if (Common.nList[zI].Rs != x.Rs || Common.nList[zI].Cs != x.Cs)
					MatrixN(x.Rs, x.Cs, zI);
				for (int r = 0; r < x.Rs; r++)
					for (int c = 0; c < x.Cs; c++)
					{
						XCCalLst.ErfCN(x.E[r, c], tI);
						if (MultiID.muErr)
						{
							Common.LDelete(tI);
							return;
						}
						Calc.SetToNM(Common.nList[tI], Common.nList[zI].E[r, c]);
					}
				Common.LDelete(tI);
			}
		} // ErfCN

		public static void EtaN(MultiZ x, int zI)
		// z = eta(x)
		// x may have the same address in memory as zI
		// x is not changed unless it shares memory with zI
		// sets muErr true if overflow or error
		{
			if (x.Rs == 0)
				XCCalLst.EtaN(x, zI);
			else
			{
				int tI;
				Common.InitNAdd(1, "EtaN-tI", out tI);
				if (Common.nList[zI].Rs != x.Rs || Common.nList[zI].Cs != x.Cs)
					MatrixN(x.Rs, x.Cs, zI);
				for (int r = 0; r < x.Rs; r++)
					for (int c = 0; c < x.Cs; c++)
					{
						XCCalLst.EtaN(x.E[r, c], tI);
						if (MultiID.muErr)
						{
							Common.LDelete(tI);
							return;
						}
						Calc.SetToNM(Common.nList[tI], Common.nList[zI].E[r, c]);
					}
				Common.LDelete(tI);
			}
		} // EtaN

		public static void LambdaN(MultiZ x, int zI)
		// z = lambda(x)
		// x may have the same address in memory as zI
		// x is not changed unless it shares memory with zI
		// sets muErr true if overflow or error
		{
			if (x.Rs == 0)
				XCCalLst.LambdaN(x, zI);
			else
			{
				int tI;
				Common.InitNAdd(1, "LambdaN-tI", out tI);
				if (Common.nList[zI].Rs != x.Rs || Common.nList[zI].Cs != x.Cs)
					MatrixN(x.Rs, x.Cs, zI);
				for (int r = 0; r < x.Rs; r++)
					for (int c = 0; c < x.Cs; c++)
					{
						XCCalLst.LambdaN(x.E[r, c], tI);
						if (MultiID.muErr)
						{
							Common.LDelete(tI);
							return;
						}
						Calc.SetToNM(Common.nList[tI], Common.nList[zI].E[r, c]);
					}
				Common.LDelete(tI);
			}
		} // LambdaN

		public static void ZetaN(MultiZ x, int zI)
		// z = Zeta(x)
		// x may have the same address in memory as zI
		// x is not changed unless it shares memory with zI
		// sets muErr true if overflow or error
		{
			if (x.Rs == 0)
				XCCalLst.ZetaN(x, zI);
			else
			{
				int tI;
				Common.InitNAdd(1, "ZetaN-tI", out tI);
				if (Common.nList[zI].Rs != x.Rs || Common.nList[zI].Cs != x.Cs)
					MatrixN(x.Rs, x.Cs, zI);
				for (int r = 0; r < x.Rs; r++)
					for (int c = 0; c < x.Cs; c++)
					{
						XCCalLst.ZetaN(x.E[r, c], tI);
						if (MultiID.muErr)
						{
							Common.LDelete(tI);
							return;
						}
						Calc.SetToNM(Common.nList[tI], Common.nList[zI].E[r, c]);
					}
				Common.LDelete(tI);
			}		
		} // ZetaN

		public static void GamN1(MultiZ x, int zI) // 16-digit gamma function
		// z = gamma(x)
		// x may have the same address in memory as zI
		// x is not changed unless it shares memory with zI
		// sets muErr true if overflow or error
		{
			if (x.Rs == 0)
				XPCalLst.GamN1(x, zI);
			else
			{
				int tI;
				Common.InitNAdd(1, "GamN1-tI", out tI);
				if (Common.nList[zI].Rs != x.Rs || Common.nList[zI].Cs != x.Cs)
					MatrixN(x.Rs, x.Cs, zI);
				for (int r = 0; r < x.Rs; r++)
					for (int c = 0; c < x.Cs; c++)
					{
						XPCalLst.GamN1(x.E[r, c], tI);
						if (MultiID.muErr)
						{
							Common.LDelete(tI);
							return;
						}
						Calc.SetToNM(Common.nList[tI], Common.nList[zI].E[r, c]);
					}
				Common.LDelete(tI);
			}		
		} // GamN1

		public static void PsiN(MultiZ x, int zI)
		// z = psi(x) = gamma'(x) / gamma(x) = digamma function
		// x may have the same address in memory as zI
		// x is not changed unless it shares memory with zI
		// sets muErr true if overflow or error
		{
			if (x.Rs == 0)
				XCCalLst.PsiN(x, zI);
			else
			{
				int tI;
				Common.InitNAdd(1, "PsiN-tI", out tI);
				if (Common.nList[zI].Rs != x.Rs || Common.nList[zI].Cs != x.Cs)
					MatrixN(x.Rs, x.Cs, zI);
				for (int r = 0; r < x.Rs; r++)
					for (int c = 0; c < x.Cs; c++)
					{
						XCCalLst.PsiN(x.E[r, c], tI);
						if (MultiID.muErr)
						{
							Common.LDelete(tI);
							return;
						}
						Calc.SetToNM(Common.nList[tI], Common.nList[zI].E[r, c]);
					}
				Common.LDelete(tI);
			}		
		} // PsiN

		public static void RFacN(int zI) // z = z! = 1*2*3*...*x
		// sets muErr true if overflow or error
		{
			int fMBI = 1;
			FacN(Common.nList[zI], zI, false, fMBI);
			MultiFD.ClearF(Common.nList[zI].I);
			Calc.NormN(zI);
		}

		public static void RFacSN(int zI) // z = z! = 1*2*3*...*x
		// sets muErr true if overflow or error
		{
			int fMBI = 1;
			FacSN(Common.nList[zI], zI, false, fMBI);
		}
		
		public static void RFacMBN(int zI) // z = z! = 1*2*3*...*x % FMB
		// sets muErr true if overflow or error
		{
			int fMBI;
			if (Common.FMBIok(zI, out fMBI))
				FacN(Common.nList[zI], zI, true, fMBI);
		}

		public static void RFacMBSN(int zI) // z = z! = 1*2*3*...*x % FMB
		// sets muErr true if overflow or error
		{
			int fMBI;
			if (Common.FMBIok(zI, out fMBI))
				FacSN(Common.nList[zI], zI, true, fMBI);
		}

		public static void RBetaN(int zI) // z = beta(z)
		// sets muErr true if overflow or error
		{
			BetaN(Common.nList[zI], zI);
		}

		public static void RErfN(int zI) // z = erf(z)
		// sets muErr true if overflow or error
		{
			ErfN(Common.nList[zI], zI);
		}

		public static void RErfCN(int zI) // z = erfc(z)
		// sets muErr true if overflow or error
		{
			ErfCN(Common.nList[zI], zI);
		}

		public static void REtaN(int zI) // z = eta(z)
		// sets muErr true if overflow or error
		{
			EtaN(Common.nList[zI], zI);
		}

		public static void RLambdaN(int zI) // z = lambda(z)
		// sets muErr true if overflow or error
		{
			LambdaN(Common.nList[zI], zI);
		}

		public static void RZetaN(int zI) // z = Zeta(z)
		// sets muErr true if overflow or error
		{
			ZetaN(Common.nList[zI], zI);
		}

		public static void RGamN(int zI) // z = gamma(z)
		// sets muErr true if overflow or error
		{
			GamN(Common.nList[zI], zI);
		}

		public static void RGamN1(int zI) // z = gamma(z), 16-digit gamma function
		// sets muErr true if overflow or error
		{
			GamN1(Common.nList[zI], zI);
			MultiFD.ClearF(Common.nList[zI].I);
			Calc.NormN(zI);
		}

		public static void RPsiN(int zI) // z = psi(z)
		// sets muErr true if overflow or error
		{
			PsiN(Common.nList[zI], zI);
		}

		public static void SqRootN(MultiZ x, int zI) // z = SqRt(x)
		// x is not changed, it is ok for x and z to have the same location in memory
		// but then x will be changed. Sets muErr true if error
		{
			if (x.Rs == 0)
				XCCalLst.SqRootN(x, zI);
			else
			{
				int tI;
				Common.InitNAdd(1, "SqRootN-tI", out tI);
				if (Common.nList[zI].Rs != x.Rs || Common.nList[zI].Cs != x.Cs)
					MatrixN(x.Rs, x.Cs, zI);
				for (int r = 0; r < x.Rs; r++)
					for (int c = 0; c < x.Cs; c++)
					{
						XCCalLst.SqRootN(x.E[r, c], tI);
						if (MultiID.muErr)
						{
							Common.LDelete(tI);
							return;
						}
						Calc.SetToNM(Common.nList[tI], Common.nList[zI].E[r, c]);
					}
				Common.LDelete(tI);
			}
		} // SqRootN

		public static void RCuRtN(int zI) // z = SqRt(z)
		// sets muErr true if error
		{
			int rs = Common.nList[zI].Rs;
			if (rs == 0)
				XCCalLst.RCuRtN(zI);
			else
			{
				int cs = Common.nList[zI].Cs;
				int tI;
				Common.InitNAdd(1, "RCuRtN-tI", out tI);
				for (int r = 0; r < rs; r++)
					for (int c = 0; c < cs; c++)
					{
						Calc.SetToNM(Common.nList[zI].E[r, c], Common.nList[tI]);
						XCCalLst.RCuRtN(tI);
						if (MultiID.muErr)
						{
							Common.LDelete(tI);
							return;
						}
						Calc.SetToNM(Common.nList[tI], Common.nList[zI].E[r, c]);
					}
				Common.LDelete(tI);
			}
		} // RCuRtN

		public static void RSqRtN(int zI) // z = SqRt(z)
		// sets muErr true if error
		{
			SqRootN(Common.nList[zI], zI);
		}

		public static void RSqRtRemN(int zI) // z = Floor(SquareRoot(z)) and set Re to remainder
		// sets muErr true if error
		{
			if (Common.nList[zI].Rs == 0)
				XCCalLst.RSqRtRemN(zI);
			else
			{
				int reI;
				Common.LFindAdd("Re", out reI);
				if (reI == zI || reI == Common.iTop)
				{
					MultiID.MuWriteErr(
					"Cannot set same location to both SqRt and remainder, continuing...");
					return;
				}
				int tI;
				int reMI;
				Common.InitNAdd(1, "RSqRtRemN-tI", out tI);
				Common.InitNAdd(1, "RSqRtRemN-reMI", out reMI);
				MatrixN(Common.nList[zI].Rs, Common.nList[zI].Cs, reMI);
				for (int r = 0; r < Common.nList[zI].Rs; r++)
					for (int c = 0; c < Common.nList[zI].Cs; c++)
					{
						Calc.SetToN(Common.nList[zI].E[r, c], tI);
						XCCalLst.RSqRtRemN(tI);
						if (MultiID.muErr)
						{
							Common.LDelete(reMI);
							Common.LDelete(tI);
							Calc.SetToN(MultiZD.mZero, reI);
							Calc.SetToN(MultiZD.mZero, zI);
							return;
						}
						Calc.SetToNM((MultiC)Common.nList[tI], Common.nList[zI].E[r, c]);
						Calc.SetToNM((MultiC)Common.nList[reI], Common.nList[reMI].E[r, c]);
					}
				Calc.SetToII(reMI, reI);
				Common.LDelete(reMI);
				Common.LDelete(tI);
			}
		} // RSqRtRemN

		public static void SolveN(MultiZ x, MultiZ y, int zI) // solve for z, x * z = y
		// this code is adapted from Numerical Recipes in C
		{
			if (x.Rs != x.Cs || y.Rs != x.Rs || y.Cs != 1)
				MultiZD.ErrIncompatible();
			else
			{
				bool saveExp = MultiFD.expanded;
				int saveFMC = Calc.fMC;
				int saveMMax = MultiID.mMax;
				XPCalLst.Expand();
				int n = x.Rs;
				MultiZ a = new MultiZ(Calc.fMC, n, n);
				a &= x;
				int[] index;
				index = new int[n];
				MultiC d = new MultiC();
				bool sing;
				MultiZD.LUdecomp(a, index, d, out sing); // decompose the matrix
				if (sing)
					MultiZD.ErrSingular();
				else
				{
					MultiZ b = new MultiZ(Calc.fMC, n, 1);
					b &= y;
					MultiZD.LUbackSub(a, index, b);
					Calc.SetToN(b, zI);
				}
				MultiFD.expanded = saveExp;
				Calc.fMC = saveFMC;
				MultiID.mMax = saveMMax;
				Calc.NormN(zI);
			}
		} // SolveN

		public static void SolveSN(MultiZ x, MultiZ y, int zI) // solve for z, x * z = y
		// this code is adapted from Numerical Recipes in C
		{
			if (y.Rs != x.Rs || y.Cs != 1 || x.Rs == 0)
				MultiZD.ErrIncompatible();
			else if (x.Rs < x.Cs)
			{
				MultiZ xx = new MultiZ(1, x.Cs, x.Cs);
				MultiZ yy = new MultiZ(1, x.Cs, 1);
				for (int r = 0; r < x.Rs; r++)
				{
					Calc.SetToNM(y.E[r, 0], yy.E[r, 0]);
					for (int c = 0; c < x.Cs; c++)
						Calc.SetToNM(x.E[r, c], xx.E[r, c]);
				}
				for (int r = x.Rs; r < x.Cs; r++)
				{
					yy.E[r, 0] &= MultiCD.cZero;
					for (int c = 0; c < x.Cs; c++)
						xx.E[r, c] &= MultiCD.cZero;
				}
				SolveSSN(xx, yy, zI);
			}
			else if (x.Rs > x.Cs)
			{
				int rs = x.Rs;
				int cs = x.Cs;
				MultiZ xx = new MultiZ(Calc.fMC, x.Rs, x.Rs);
				for (int r = 0; r < x.Rs; r++)
					for (int c = 0; c < x.Cs; c++)
						Calc.SetToNM(x.E[r, c], xx.E[r, c]);
				for (int r = 0; r < x.Rs; r++)
					for (int c = x.Cs; c < x.Rs; c++)
						xx.E[r, c] &= MultiCD.cZero;
				SolveSSN(xx, y, zI);
				MultiZ yy = new MultiZ(Calc.fMC, cs, 1);
				for (int r = 0; r < cs; r++)
					yy.E[r, 0] &= Common.nList[zI].E[r, 0];
				Calc.SetToN(yy, zI);
			}
			else
				SolveSSN(x, y, zI);
		} // SolveSN

		public static void SolveSSN(MultiZ x, MultiZ y, int zI) // solve for z, x * z = y, x square
		// this code is adapted from Numerical Recipes in C
		{
			if (x.Rs != x.Cs || y.Rs != x.Rs || y.Cs != 1 || x.Rs == 0)
				MultiZD.ErrIncompatible();
			else
			{
				int i;
				bool saveExp = MultiFD.expanded;
				int saveFMC = Calc.fMC;
				int saveMMax = MultiID.mMax;
				XPCalLst.Expand();
				int n = x.Rs;
				MultiZ u = new MultiZ(Calc.fMC, n, n);
				MultiZ w = new MultiZ(Calc.fMC, 1, n);
				MultiZ v = new MultiZ(Calc.fMC, n, n);
				u &= x;
				SVD(u, w, v);
				string sing = "";
				int rank = 0;
				for (i = 0; i < n; i++)
					if (w.E[0, i] != MultiCD.cZero)
					{
						w.E[0, i] &= MultiCD.cOne / w.E[0, i];
						rank++;
					}
					else
						sing = "Singular, ";
				MultiZ z = new MultiZ(Calc.fMC, n, 1);
				MultiZ r = new MultiZ(Calc.fMC, n, 1);
				MultiZD.SetToZ(v * (MultiZD.Diagonal(w) * (ConjTran(u) * y)), z);
				r &= x * z - y;
				long next;
				long max = Int64.MinValue;
				for (i = 0; i < n; i++)
				{
					next = Int64.MinValue;
					if ((MultiF)y.E[i, 0] != MultiFD.fZero)
						next = y.E[i, 0].N + y.E[i, 0].C;
					if ((MultiF)y.E[i, 0].I != MultiFD.fZero)
						next = Math.Max(next, y.E[i, 0].I.N + y.E[i, 0].I.C);
					if (next > max)
						max = next;
				}
				for (i = 0; i < n; i++)
				{
					next = r.E[i, 0].N + r.E[i, 0].C;
					if (max - next >= Calc.fMC - MultiFD.fGuard)
						XPCalLst.SetToNM(MultiFD.fZero, r.E[i, 0]);
					next = r.E[i, 0].I.N + r.E[i, 0].I.C;
					if (max - next >= Calc.fMC - MultiFD.fGuard)
						XPCalLst.SetToNM(MultiFD.fZero, r.E[i, 0].I);
				}
				r.E[0, 0] &= Abs2(r.E[0, 0]);
				for (i = 1; i < n; i++)
					r.E[0, 0] += Abs2(r.E[i, 0]);
				MultiCD.SqRootC(r.E[0, 0], r.E[0, 0]);
				max = Int64.MinValue;
				for (i = 0; i < n; i++)
				{
					next = Int64.MinValue;
					if ((MultiF)z.E[i, 0] != MultiFD.fZero)
						next = z.E[i, 0].N + z.E[i, 0].C;
					if ((MultiF)z.E[i, 0].I != MultiFD.fZero)
						next = Math.Max(next, z.E[i, 0].I.N + z.E[i, 0].I.C);
					if (next > max)
						max = next;
				}
				for (i = 0; i < n; i++)
				{
					next = z.E[i, 0].N + z.E[i, 0].C;
					if (max - next >= Calc.fMC - MultiFD.fGuard)
						XPCalLst.SetToNM(MultiFD.fZero, z.E[i, 0]);
					next = z.E[i, 0].I.N + z.E[i, 0].I.C;
					if (max - next >= Calc.fMC - MultiFD.fGuard)
						XPCalLst.SetToNM(MultiFD.fZero, z.E[i, 0].I);
				}
				next = r.E[0, 0].N + r.E[0, 0].C;
				if (max - next >= Calc.fMC - MultiFD.fGuard)
					XPCalLst.SetToNM(MultiFD.fZero, r.E[0, 0]);
				next = r.E[0, 0].I.N + r.E[0, 0].I.C;
				if (max - next >= Calc.fMC - MultiFD.fGuard)
					XPCalLst.SetToNM(MultiFD.fZero, r.E[0, 0].I);
				Calc.SetToN(z, zI);
				MultiFD.expanded = saveExp;
				Calc.fMC = saveFMC;
				MultiID.mMax = saveMMax;
				MultiCD.NormC(r.E[0, 0]);
				int tot = 0;
				MultiID.WriteT(sing + "Rank = " + rank + ", Residual = ", ref tot);
				MultiCD.WriteC(MultiID.outS, null, r.E[0, 0], ref tot, 0);
				Calc.NormN(zI);
			}
		} // SolveSSN

		public static void SolveH(MultiZ x, MultiZ z) // solves homogeneous equation x * z = 0
		// z = all non trivial solutions of matrix x * z = 0
		{
			int m = x.Rs;
			int n = x.Cs;
			if (m == 0)
				MultiZD.ErrNotMatrix();
			else
			{
				bool saveExp = MultiFD.expanded;
				int saveFMC = Calc.fMC;
				int saveMMax = MultiID.mMax;
				XPCalLst.Expand();
				int j, ri, ci;
				MultiC di = new MultiC();
				if (m < n)
					m = n;
				MultiZ u = new MultiZ(Calc.fMC, m, n);
				MultiZ w = new MultiZ(Calc.fMC, 1, n);
				MultiZ v = new MultiZ(Calc.fMC, n, n);
				for (int r = 0; r < x.Rs; r++)
					for (int c = 0; c < n; c++)
						MultiCD.SetToC(x.E[r, c], u.E[r, c]);
				SVD(u, w, v);
				MultiFD.expanded = saveExp;
				Calc.fMC = saveFMC;
				MultiID.mMax = saveMMax;
				Calc.NormNM(w);
				XPCalLst.Expand();
				for (j = n - 1; j >= 0; j--)			// delete columns of v where w != 0
				{
					if (w.E[0, j] != MultiCD.cZero)
						Del(v, j, v);
				}
				MultiZ zz = new MultiZ(1, n, v.Cs);
				for (ci = 0; ci < v.Cs; ci++)			// copy v to zz
					for (ri = 0; ri < n; ri++)
						Calc.SetToNM(v.E[ri, ci], zz.E[ri, ci]);
				Calc.fMC -= MultiFD.fGuard;
				Calc.NormNMCol(zz);
				Calc.fMC += MultiFD.fGuard;
				bool err = false;
				for (ci = zz.Cs -1; ci >= 0; ci--)		// for all columns of zz
				{
					Calc.SetToNM(MultiCD.cZero, di);
					for (ri = n - 1; ri >= 0; ri--)		// find last non-zero element in this column
						if (zz.E[ri, ci] != MultiCD.cZero)
						{
							Calc.SetToNM(zz.E[ri, ci], di);
							break;
						}
					if (di == MultiCD.cZero)			// if column is all zeros
					{
						Del(zz, ci, zz);				// Delete columns of zz with all zeros
						err = true;
					}
					else
					{
						for (ri = 0; ri < n; ri++)		// Make last non-zero element equal to one
							Calc.SetToNM(zz.E[ri, ci] / di, zz.E[ri, ci]);
					}
				}
				MultiFD.expanded = saveExp;
				Calc.fMC = saveFMC;
				MultiID.mMax = saveMMax;
				Calc.NormNM(zz);
				for (ci = zz.Cs - 1; ci > 0; ci--)
				{
					m = 0;
					for (ri = 0; ri < n; ri++)
					{
						if (zz.E[ri, ci] == zz.E[ri, ci - 1])
							m++;
						else
							break;
					}
					if (m == n)
					{
						Del(zz, ci, zz);
						err = true;
					}
				}
				Calc.SetToNM(zz, z);
				if (err)
					MultiID.WriteLn("Not all vectors found");
			}
		} // SolveH

		public static void SolveHN(MultiZ x, int zI) // solves homogeneous equation x * z = 0
		// z = all non trivial solutions of matrix x * z = 0
		{
			if (x.Rs == 0)
				MultiZD.ErrNotMatrix();
			else
			{
				MultiZ z = new MultiZ();
				SolveH(x, z);
				Calc.SetToN(z, zI);
			}
		} // SolveHN

		public static int Get1InM(int low, int high, MultiZ z)
		{
			if (z.Rs != 0) return low;
			int d1;
			if (z.N + z.C > 1)
			{
				d1 = 100000000;
				if (z.S) d1 = -d1;
			}
			else
				MultiFD.Get1F(out d1, z);
			if (d1 < low) return low;
			if (d1 > high) return high;
			return d1;
		} // Get1InM

		public static void SortC(MultiZ x) // x = sorted columns of x
		{
			MultiZ rightM = new MultiZ(2);
			MultiFD.SetTo1F(x.Rs, rightM);
			SortCS(x, MultiZD.mOne, rightM);
		} // SortC

		public static void SortCS(MultiZ x, MultiZ leftM, MultiZ rightM) // x = sorted columns of x
		{
			swapped = false;
			if (x.Rs == 0)
				return;
			int left = Get1InM(1, x.Rs, leftM);
			left -= uFirst;
			int right = Get1InM(1, x.Rs, rightM);
			if (rightM.Rs != 0)
				right = x.Rs;
			right -= uFirst;
			if (left > right)
			{
				int swap = left;
				left = right;
				right = swap;
			}
			for (int c = 0; c < x.Cs; c++)
				QuickSortC(x, left, right, c);
			if (swapped && !Common.quiet) MultiID.WriteLn("SortC: Some elements were swapped");
			if (!swapped && !Common.quiet) MultiID.WriteLn("SortC: No elements were swapped");
		} // SortCS

		public static void QuickSortC(MultiZ x, int left, int right, int i) // sort column i of x
		{
			if (right <= left)
				return;
			int leftIndex = left;
			int rightIndex = right;
			// Arbitrarily establishing partition element as the midpoint of the array.
			int j = (left + right) / 2;
			MultiZ partionElement = new MultiZ(x.E[j, i].M);
			partionElement &= x.E[j, i];
			// loop through the array until indices cross
			while (leftIndex <= rightIndex)
			{
				// find the first element that is greater than or equal to
				// the partionElement starting from the leftIndex.
				while ((leftIndex < right) && (x.E[(leftIndex), i] < partionElement))
					leftIndex = leftIndex + 1;
				// find an element that is smaller than or equal to
				// the partionElement starting from the rightIndex.
				while ((rightIndex > left) && (x.E[(rightIndex), i] > partionElement))
					rightIndex = rightIndex - 1;
				// if the indexes have not crossed, swap
				if (leftIndex <= rightIndex)
				{
					if (leftIndex != rightIndex)
					{
						SwapC(ref x.E[(leftIndex), i], ref x.E[(rightIndex), i]);
						swapped = true;
					}
					leftIndex = leftIndex + 1;
					rightIndex = rightIndex - 1;
				}
			}
			// If the right index has not reached the left side of array
			// must now sort the left partition.
			if (left < rightIndex)
				QuickSortC(x, left, rightIndex, i);
			// If the left index has not reached the right side of array
			// must now sort the right partition.
			if (leftIndex < right)
				QuickSortC(x, leftIndex, right, i);
		} // QuickSortC

		public static void SortR(MultiZ x) // x = sorted rows of x
		{
			MultiZ rightM = new MultiZ(2);
			MultiFD.SetTo1F(x.Cs, rightM);
			SortRS(x, MultiZD.mOne, rightM);
		} // SortR

		public static void SortRS(MultiZ x, MultiZ leftM, MultiZ rightM) // x = sorted rows of x
		{
			swapped = false;
			if (x.Rs == 0)
				return;
			int left = Get1InM(1, x.Cs, leftM);
			left -= uFirst;
			int right = Get1InM(1, x.Cs, rightM);
			if (rightM.Rs != 0)
				right = x.Cs;
			right -= uFirst;
			if (left > right)
			{
				int swap = left;
				left = right;
				right = swap;
			}
			for (int r = 0; r < x.Rs; r++)
				QuickSortR(x, left, right, r);
			if (swapped && !Common.quiet) MultiID.WriteLn("SortR: Some elements were swapped");
			if (!swapped && !Common.quiet) MultiID.WriteLn("SortR: No elements were swapped");
		} // SortRS

		public static void QuickSortR(MultiZ x, int left, int right, int i) // sort row i of x
		{
			if (right <= left)
				return;
			int leftIndex = left;
			int rightIndex = right;
			// Arbitrarily establishing partition element as the midpoint of the array.
			int j = (left + right) / 2;
			MultiZ partionElement = new MultiZ(x.E[i, j].M);
			partionElement &= x.E[i, j];
			// loop through the array until indices cross
			while (leftIndex <= rightIndex)
			{
				// find the first element that is greater than or equal to
				// the partionElement starting from the leftIndex.
				while ((leftIndex < right) && (x.E[i, (leftIndex)] < partionElement))
					leftIndex = leftIndex + 1;
				// find an element that is smaller than or equal to
				// the partionElement starting from the rightIndex.
				while ((rightIndex > left) && (x.E[i, (rightIndex)] > partionElement))
					rightIndex = rightIndex - 1;
				// if the indexes have not crossed, swap
				if (leftIndex <= rightIndex)
				{
					if (leftIndex != rightIndex)
					{
						SwapC(ref x.E[i, (leftIndex)], ref x.E[i, (rightIndex)]);
						swapped = true;
					}
					leftIndex = leftIndex + 1;
					rightIndex = rightIndex - 1;
				}
			}
			// If the right index has not reached the left side of array
			// must now sort the left partition.
			if (left < rightIndex)
				QuickSortR(x, left, rightIndex, i);
			// If the left index has not reached the right side of array
			// must now sort the right partition.
			if (leftIndex < right)
				QuickSortR(x, leftIndex, right, i);
		} // QuickSortR

		public static void Image(MultiZ x, MultiZ z) // z = a basis for the image/range of matrix x
		{
			int m = x.Rs;
			int n = x.Cs;
			if (m == 0)
				MultiZD.ErrNotMatrix();
			else
			{
				MultiZ xx = new MultiZ();
				MultiZ zz = new MultiZ(1, m, n);
				RREF(x, xx);
				int c = 0;
				int ci = 0;
				for (int r = 0; r < m; r++)
				{
					for ( ; c < n; c++)
					{
						if (xx.E[r, c] == MultiCD.cOne)
						{
							// insert column c of x into column ci of zz
							for (int ri = 0; ri < m; ri++)
								Calc.SetToNM(x.E[ri, c], zz.E[ri, ci]);
							ci++;
							c++;
							break;
						}
					}
				}
				zz.Cs = ci;
				Calc.SetToNM(zz, z);
			}
		} // Image

		public static void ImageN(MultiZ x, int zI) // z = a basis for the image/range of matrix x
		{
			if (x.Rs == 0)
				MultiZD.ErrNotMatrix();
			else
			{
				MultiZ z = new MultiZ();
				Image(x, z);
				Calc.SetToN(z, zI);
			}
		} // ImageN

		public static void Rank(MultiZ x, MultiZ z) // z = rank(x)
		{
			int m = x.Rs;
			int n = x.Cs;
			if (m == 0)
				MultiZD.ErrNotMatrix();
			else
			{
				int rank = m;
				bool zero = true;
				MultiZ zz = new MultiZ();
				RREF(x, zz);
				for (int r = m - 1; r >= 0; r--)
				{
					for (int c = 0; c < n; c++)
					{
						if (zz.E[r, c] != MultiCD.cZero)
						{
							zero = false;
							break;
						}
					}
					if (zero)
						rank--;
					else
						break;
				}
				MultiZD.SetTo1Z(rank, z);
			} 
		} // Rank

		public static void RankN(MultiZ x, int zI) // z = rank(x)
		{
			if (x.Rs == 0)
				MultiZD.ErrNotMatrix();
			else
			{
				MultiZ z = new MultiZ();
				Rank(x, z);
				Calc.SetToN(z, zI);
			}
		} // RankN

		public static void RealN(MultiZ x, int zI) // z = real(x)
		{
			if (x.Rs == 0)
				Calc.SetToN((MultiF)x, zI);
			else
			{
				int tI;
				Common.InitNAdd(1, "RealN-tI", out tI);
				if (Common.nList[zI].Rs != x.Rs || Common.nList[zI].Cs != x.Cs)
					MatrixN(x.Rs, x.Cs, zI);
				for (int r = 0; r < x.Rs; r++)
					for (int c = 0; c < x.Cs; c++)
					{
						Calc.SetToN((MultiF)x.E[r, c], tI);
						if (MultiID.muErr)
						{
							Common.LDelete(tI);
							return;
						}
						Calc.SetToNM(Common.nList[tI], Common.nList[zI].E[r, c]);
					}
				Common.LDelete(tI);
			} 
		} // RealN

		public static void ImagN(MultiZ x, int zI) // z = Imag(x)
		{
			if (x.Rs == 0)
				Calc.SetToN(x.I, zI);
			else
			{
				int tI;
				Common.InitNAdd(1, "RealN-tI", out tI);
				if (Common.nList[zI].Rs != x.Rs || Common.nList[zI].Cs != x.Cs)
					MatrixN(x.Rs, x.Cs, zI);
				for (int r = 0; r < x.Rs; r++)
					for (int c = 0; c < x.Cs; c++)
					{
						Calc.SetToN(x.E[r, c].I, tI);
						if (MultiID.muErr)
						{
							Common.LDelete(tI);
							return;
						}
						Calc.SetToNM(Common.nList[tI], Common.nList[zI].E[r, c]);
					}
				Common.LDelete(tI);
			} 
		} // ImagN

		public static MultiC Conj(MultiC x)	// return x.r - x.r
		{
			MultiC t = new MultiC(x.M, x.I.M);
			t &= (MultiF)x;
			t.I &= -x.I;
			return t;
		}
		
		public static MultiC Abs2(MultiC x)	// return |x|^2 = x.r^2 + x.i^^2
		{
			MultiC t = new MultiC(Calc.fMC);
			MultiFD.SqF(x, t);
			MultiFD.SqF(x.I, t.I);
			MultiFD.RAddF(t.I, t);
			MultiFD.ClearF(t.I);
			return t;
		}
		
		public static MultiC Abs(MultiC x)	// return |x|
		{
			MultiC t = new MultiC(Calc.fMC);
			MultiCD.Abs(x, t);
			return t;
		}
		
//		public static MultiF AbsF(MultiC x)	// return |x| // not used
//		{
//			MultiC t = new MultiC(Calc.fMC);
//			MultiCD.Abs(x, t);
//			return (MultiF)t;
//		}
//		
		public static MultiF Abs(MultiF x)	// return |x|
		{
			if (x < MultiFD.fZero)
				return -x;
			else
				return x;
		}

		public static MultiF Max(MultiF x, MultiF y)	// return larger of x and y
		{
			if (x >= y)
				return x;
			else
				return y;
		}

		public static MultiF Sign(MultiF x, MultiF y)	// return x with sign of y.r
		// actually, change sign of x if y < 0
		{
			if (x == MultiFD.fZero || y >= MultiFD.fZero)
				return x;
			else
				return -x;
		}
		
		public static MultiF Pythag(MultiF x, MultiF y)	// return SqRt(x.r^2 + y.r^2)
		{
			if (x == MultiFD.fZero)
				return Abs(y);
			if (y == MultiFD.fZero)
				return Abs(x);
			MultiF t = new MultiF(Calc.fMC);
			t = x * x + y * y;
			MultiFD.SqRootF(t, t);
			return t;
		}

		public static void SortUWV(MultiZ u, MultiZ w, MultiZ v)
		{
			int m = u.Rs;
			int n = u.Cs;
			int i, j, k;
			MultiC x = new MultiC(1);
			MultiZ uc = new MultiZ(1, m, 1);
			MultiZ vc = new MultiZ(1, n, 1);
			for (j = 1; j < n; j++) // Sort w by straight insertion
			{
				Calc.SetToNM(w.E[0, j], x);
				for (k = 0; k < m; k++)
					Calc.SetToNM(u.E[k, j], uc.E[k, 0]);
				for (k = 0; k < n; k++)
					Calc.SetToNM(v.E[k, j], vc.E[k, 0]);
				for (i = j - 1; i >= 0; i--)
				{
					if (w.E[0, i] <= x) // least first
						break;
					Calc.SetToNM(w.E[0, i], w.E[0, i + 1]);
					for (k = 0; k < m; k++)
						Calc.SetToNM(u.E[k, i], u.E[k, i + 1]);
					for (k = 0; k < n; k++)
						Calc.SetToNM(v.E[k, i], v.E[k, i + 1]);
				}
				Calc.SetToNM(x, w.E[0, i + 1]);
				for (k = 0; k < m; k++)
					Calc.SetToNM(uc.E[k, 0], u.E[k, i + 1]);
				for (k = 0; k < n; k++)
					Calc.SetToNM(vc.E[k, 0], v.E[k, i + 1]);
			}
		} // SortUWV

		public static void CSVD(MultiZ a, int p, MultiZ u, MultiZ s, MultiZ v)
		// Singular Value Decomposition, a = u * s * ConjTran(v), a is destroyed by CSVD
		// the diagonal matrix s is output as a row vector, a.Rs must be >= a.Cs
		// if smaller, a should be filled with zero rows
		// this code is adapted from Collected Algorithms from ACM, Algorithm 358
		// The transformation ConjTran(u) is applied to the p vectors given in columns
		// n, n+1, ..., n+p-1 of matrix a
		// See: http://www.scs.fsu.edu/~burkardt/f77_src/toms358/toms358.f 
		// and: http://www.scs.fsu.edu/~burkardt/f77_src/toms358/toms358_prb.f 
		{
			int m = a.Rs;
			int n = a.Cs - p;
			int nu = m;
			int nv = n;
			int mMax = m;
			int nMax = n;
			int np = n + p;
			int n1 = n + 1;
			int j;
			int bctn = n;
			MultiF[] b = new MultiF[bctn];
			MultiF[] c = new MultiF[bctn];
			MultiF[] t = new MultiF[bctn];
			for (j = 0; j < bctn; j++)
			{
				b[j] = new MultiF(Calc.fMC);
				c[j] = new MultiF(Calc.fMC);
				t[j] = new MultiF(Calc.fMC);
			}
			MultiF cs = new MultiF(Calc.fMC);
			MultiF eps = new MultiF(Calc.fMC);
			MultiF eta = new MultiF(1);
			MultiF f = new MultiF(Calc.fMC);
			MultiF g = new MultiF(Calc.fMC);
			MultiF h = new MultiF(Calc.fMC);
			int i;
			int k;
			int k1;
			int L = 0;
			int L1;
			int nM1 = n - 1;
			MultiC q = new MultiC(Calc.fMC);
			MultiC r = new MultiC(Calc.fMC);
			MultiF sn = new MultiF(Calc.fMC);
			MultiF tol = new MultiF(1);
			MultiF w = new MultiF(Calc.fMC);
			MultiF x = new MultiF(Calc.fMC);
			MultiF y = new MultiF(Calc.fMC);
			MultiF z = new MultiF(Calc.fMC);
			eta &= MultiFD.fOne;
			eta.C = 1 - Calc.fMC;			// eta = the relative machine precision
			tol &= MultiFD.fOne;
			tol.C = Int64.MinValue - eta.C;	// tol = the smallest normalized positive number, divided by eta
			MultiZ aT = new MultiZ(Calc.fMC, a.Rs, a.Cs);
			aT &= a;
			a = aT; // copies pointer only, original a not changed
			//
			// HOUSEHOLDER REDUCTION
			c[0] &= MultiFD.fZero;
			k = 0;
			while (true)
			{
				k1 = k + 1;
				//
				// ELIMINATION OF a(i, k), i = k, ..., m-1.
				z &= MultiFD.fZero;
				for (i = k; i < m; i++)
					z += (MultiF)a.E[i, k] * (MultiF)a.E[i, k] + a.E[i, k].I * a.E[i, k].I;
				b[k] &= MultiFD.fZero;
				if (z > tol)
				{
					MultiFD.RSqRtF(z);
					b[k] &= z;
					w &= Abs(a.E[k, k]);
					q &= MultiCD.cOne;
					if (w != MultiFD.fZero) q &= a.E[k, k] / w;
					a.E[k, k] &= q * (z + w);
					if (k != np - 1)
					{
						for (j = k1; j < np; j++)
						{
							q &= MultiCD.cZero;
							for (i = k; i < m; i++)
								q += Conj(a.E[i, k]) * a.E[i, j];
							q /= z * (z + w);
							for (i = k; i < m; i++)
								a.E[i, j] -= q * a.E[i, k];
						}
					}
					//
					// PHASE TRANSFORMATION
					q &= -Conj(a.E[k, k]) / Abs(a.E[k, k]);
					for (j = k1; j < np; j++)
						a.E[k, j] *= q;
				}
				//
				// ELIMINATION OF a[k, j], j = k+2, ..., n-1
				if (k == nM1) break;
				z &= MultiFD.fZero;
				for (j = k1; j < n; j++)
					z += (MultiF)a.E[k, j] * (MultiF)a.E[k, j] + a.E[k, j].I * a.E[k, j].I;
				c[k1] &= MultiFD.fZero;
				if (z > tol)
				{
					MultiFD.RSqRtF(z);
					c[k1] &= z;
					w &= Abs(a.E[k, k1]);
					q &= MultiCD.cOne;
					if (w != MultiFD.fZero) q &= a.E[k, k1] / w;
					a.E[k, k1] &= q * (z + w);
					for (i = k1; i < m; i++)
					{
						q &= MultiCD.cZero;
						for (j = k1; j < n; j++)
							q += Conj(a.E[k, j]) * a.E[i, j];
						q /= z * (z + w);
						for (j = k1; j < n; j++) 
							a.E[i, j] -= q * a.E[k, j];
					}
					//
					// PHASE TRANSFORMATION
					q &= -Conj(a.E[k, k1]) / Abs(a.E[k, k1]);
					for (i = k1; i < m; i++)
						a.E[i, k1] *= q;
				}
				k = k1;
			}
			//
			// TOLERANCE FOR NEGLIGIBLE ELEMENTS
			eps &= MultiFD.fZero;
			for (k = 0; k < n; k++)
			{
				s.E[0, k] &= b[k];
				t[k] &= c[k];
				if ((MultiF)s.E[0, k] + t[k] > eps)
					eps &= (MultiF)s.E[0, k] + t[k];
			}
			eps *= eta;
			//
			// INITIALIZATION OF u AND v
			if (nu > 0)
			{
				for (j = 0; j < nu; j++)
				{
					for (i = 0; i < m; i++)
						u.E[i, j] &= MultiCD.cZero;
					u.E[j, j] &= MultiCD.cOne;
				}
			}
			if (nv > 0)
			{
				for (j = 0; j < nv; j++)
				{
					for (i = 0; i < n; i++)
						v.E[i, j] &= MultiCD.cZero;
					v.E[j, j] &= MultiCD.cOne;
				}
			}
			//
			// QR DIAGONALIZATION
			for (k = nM1; k >= 0; k--)
			{
				//
				// TEST FOR SPLIT
				while (true)
				{
					for (L = k; L >= 0; L--)
					{
						if (Abs(t[L]) <= eps) goto Test;
						if (Abs(s.E[0, L - 1]) <= eps) break;
					}
					//
					// CANCELLATION OF E(L)
					cs &= MultiFD.fZero;
					sn &= MultiFD.fOne;
					L1 = L - 1;
					for (i = L; i <= k; i++)
					{
						f &= sn * t[i];
						t[i] *= cs;
						if (Abs(f) <= eps) goto Test;
						h &= s.E[0, i];
						MultiFD.SqRootF(f * f + h * h, w);
						s.E[0, i] &= w;
						cs &= h / w;
						sn &= -f / w;
						if (nu > 0)
						{
							for (j = 0; j < n; j++)
							{
								x &= (MultiF)u.E[j, L1];
								y &= (MultiF)u.E[j, i];
								u.E[j, L1] &= x * cs + y * sn;
								u.E[j, i] &= y * cs - x * sn;
							}
						}
						if (np == n) continue;
						for (j = n; j < np; j++)
						{
							q &= a.E[L1, j];
							r &= a.E[i, j];
							a.E[L1, j] &= q * cs + r * sn;
							a.E[i, j] &= r * cs - q * sn;
						}
					}
					//
					// TEST FOR CONVERGENCE
				Test:	w &= s.E[0, k];
					if (L == k) break;
					//
					// ORIGIN SHIFT
					x &= s.E[0, L];
					y &= s.E[0, k - 1];
					g &= t[k - 1];
					h &= t[k];
					f &= ((y - w) * (y + w) + (g - h) * (g + h)) / (MultiFD.fTwo * h * y);
					MultiFD.SqRootF(f * f + MultiFD.fOne, g);
					if (f < MultiFD.fZero) g &= -g;
					f &= ((x - w) * (x + w) + (y / (f + g) - h) * h) / x;
					//
					// QR STEP
					cs &= MultiFD.fOne;
					sn &= MultiFD.fOne;
					L1 = L + 1;
					for (i = L1; i <= k; i++)
					{
						g &= t[i];
						y &= s.E[0, i];
						h &= sn * g;
						g &= cs * g;
						MultiFD.SqRootF(h * h + f * f, w);
						t[i - 1] &= w;
						cs &= f / w;
						sn &= h / w;
						f &= x * cs + g * sn;
						g &= g * cs - x * sn;
						h &= y * sn;
						y &= y * cs;
						if (nv > 0)
						{
							for (j = 0; j < n; j++)
							{
								x &= (MultiF)v.E[j, i - 1];
								w &= (MultiF)v.E[j, i];
								v.E[j, i - 1] &= x * cs + w * sn;
								v.E[j, i] &= w * cs - x * sn;
							}
						}
						MultiFD.SqRootF(h * h + f * f, w);
						s.E[0, i - 1] &= w;
						cs &= f / w;
						sn &= h / w;
						f &= cs * g + sn * y;
						x &= cs * y - sn * g;
						if (nu > 0)
						{
							for (j = 0; j < n; j++)
							{
								y &= (MultiF)u.E[j, i - 1];
								w &= (MultiF)u.E[j, i];
								u.E[j, i - 1] &= y * cs + w * sn;
								u.E[j, i] &= w * cs - y * sn;
							}
						}
						if (n == np) continue;
						for (j = n1; j < np; j++)
						{
							q &= a.E[i - 1, j];
							r &= a.E[i, j];
							a.E[i - 1, j] &= q * cs + r * sn;
							a.E[i, j] &= r * cs - q * sn;
						}
					}
					t[L] &= MultiFD.fZero;
					t[k] &= f;
					s.E[0, k] &= x;
				}
				//
				// CONVERGENCE
				if (w >= MultiFD.fZero) continue;
				s.E[0, k] &= -w;
				if (nv == 0) continue;
				for (j = 0; j < n; j++)
					v.E[j, k] &= -v.E[j, k];
			}
			//
			// SORT SINGULAR VALUES
//			for (k = 0; k < n; k++)			// sort descending
			for (k = n - 1; k >= 0; k--)	// sort ascending	
			{
				g &= MultiFD.fMOne;
				j = k;
//				for (i = k; i < n; i++)		// sort descending
				for (i = k; i >= 0; i--)	// sort ascending
				{
					if (s.E[0, i] <= g) continue;
					g &= s.E[0, i];
					j = i;
				}
				if (j == k) continue;
				s.E[0, j] &= s.E[0, k];
				s.E[0, k] &= g;
				if (nv > 0)
				{
					for (i = 0; i < n; i++)
					{
						q &= v.E[i, j];
						v.E[i, j] &= v.E[i, k];
						v.E[i, k] &= q;
					}
				}
				if (nu > 0)
				{
					for (i = 0; i < n; i++)
					{
						q &= u.E[i, j];
						u.E[i, j] &= u.E[i, k];
						u.E[i, k] &= q;
					}
				}
				if (n == np) continue;
				for (i = n; i < np; i++)
				{
					q &= a.E[j, i];
					a.E[j, i] &= a.E[k, i];
					a.E[k, i] &= q;
				}
			}
			//
			// BACK TRANSFORMATION
			if (nu > 0)
			{
				for (k = nM1; k >= 0; k--)
				{
					if (b[k] == MultiFD.fZero) continue;
					q &= -a.E[k, k] / Abs(a.E[k, k]);
					for (j = 0; j < nu; j++)
						u.E[k, j] *= q;
					for (j = 0; j < nu; j++)
					{
						q &= MultiCD.cZero;
						for (i = k; i < m; i++)
							q += Conj(a.E[i, k]) * u.E[i, j];
						q /= Abs(a.E[k, k]) * b[k];
						for (i = k; i < m; i++)
							u.E[i, j] -= q * a.E[i ,k];
					}
				}
			}
			if (nv > 0 && n > 1)
			{
				for (k = n - 2; k >= 0; k--)
				{
					k1 = k + 1;
					if (c[k1] == MultiFD.fZero) continue;
					q &= -Conj(a.E[k, k1]) / Abs(a.E[k, k1]);
					for (j = 0; j < nv; j++)
						v.E[k1, j] *= q;
					for (j = 0; j < nv; j++)
					{
						q &= MultiCD.cZero;
						for (i = k1; i < n; i++)
							q += a.E[k, i] * v.E[i, j];
						q &= q / (Abs(a.E[k, k1]) * c[k1]);
						for (i = k1; i < n; i++)
							v.E[i, j] -= q * Conj(a.E[k, i]);
					}
				}
			}
		} // CSVD

		public static void SVD(MultiZ a, MultiZ w, MultiZ v)
		// Singular Value Decomposition, a = u * w * Tran(v), a replaced with u
		// the diagonal matrix is output as a row vector, a.Rs must be >= a.Cs
		// if smaller, a should be filled with zero rows
		// this code is adapted from Numerical Recipes in C
		{
			int m = a.Rs;
			int n = a.Cs;
			if (m == 0)
				MultiZD.ErrNotMatrix();
			else
			{
				bool complex = false;
				int flag, i, its, j, jj, k, L = 0, nm = 0;
				for (i = 0; i < m; i++)
					for (j = 0; j < n; j++)
						if(a.E[i, j].I != MultiCD.cZero)
						{
							complex = true;
							i = m;
							j = n;
						}
				if (complex)
				{
					MultiZ u = new MultiZ(Calc.fMC, m, m);
					CSVD(a, 0, u, w, v);
					for (i = 0; i < m; i++)
						for (j = 0; j < n; j++)
							a.E[i, j] &= u.E[i, j];
					return;
				}
				MultiC anorm = new MultiC(Calc.fMC);
				MultiC c = new MultiC(Calc.fMC);
				MultiC f = new MultiC(Calc.fMC);
				MultiC g = new MultiC(Calc.fMC);
				MultiC h = new MultiC(Calc.fMC);
				MultiC s = new MultiC(Calc.fMC);
				MultiC scale = new MultiC(Calc.fMC);
				MultiC x = new MultiC(Calc.fMC);
				MultiC y = new MultiC(Calc.fMC);
				MultiC z = new MultiC(Calc.fMC);
				MultiZ rv1 = new MultiZ(Calc.fMC, 1, n);
				// Householder reduction to bidiagonal form
				g &= scale &= anorm &= MultiCD.cZero;
				for (i = 0; i < n; i++) 
				{
					L = i + 1;
					rv1.E[0, i] &= scale * g;
					g &= s &= scale &= MultiCD.cZero;
					if (i < m)
					{
						for (k = i; k < m; k++)
							scale += Abs(a.E[k, i]);
						if (scale != MultiCD.cZero)
						{
							for (k = i; k < m; k++)
							{
								a.E[k, i] /= scale;
								s += a.E[k, i] * a.E[k, i];
							}
							f &= a.E[i, i];
							MultiCD.SqRootC(s, g);	// g = -SIGN(sqrt(s), f);
							g &= -Sign(g, f);
							h &= f * g - s;
							a.E[i, i] &= f - g;
							for (j = L; j < n; j++)
							{
								for (s &= MultiCD.cZero, k = i; k < m; k++)
									s += a.E[k, i]* a.E[k, j];
								f &= s / h;
								for (k = i; k < m; k++)
									a.E[k, j] += f * a.E[k, i];
							}
							for (k = i; k < m; k++)
								a.E[k, i] *= scale;
						}
					}
					w.E[0, i] &= scale * g;
					g &= s &= scale &= MultiCD.cZero;
					if (i < m && i != n)
					{
						for (k = L; k < n; k++)
							scale += Abs(a.E[i, k]);
						if (scale != MultiCD.cZero)
						{
							for (k = L; k < n; k++)
							{
								a.E[i, k] /= scale;
								s += a.E[i, k] * a.E[i, k];
							}
							f &= a.E[i, L];
							MultiCD.SqRootC(s, g);
							g &= -Sign(g, f);
							h &= f * g - s;
							a.E[i, L] &= f - g;
							for (k = L; k < n; k++)
								rv1.E[0, k] &= a.E[i, k] / h;
							for (j = L; j < m; j++)
							{
								for (s &= MultiCD.cZero, k = L; k < n; k++)
									s += a.E[j, k] * a.E[i, k];
								for (k = L; k < n; k++)
									a.E[j, k] += s * rv1.E[0, k];
							}
							for (k = L; k < n; k++)
								a.E[i, k] *= scale;
						}
					}
					anorm &= Max(anorm,(Abs(w.E[0, i]) + Abs(rv1.E[0, i])));
				}
				// accumulation of right-hand transformations
				for (i = n - 1; i >= 0; i--)
				{
					if (i < n - 1)
					{
						if (g != MultiCD.cZero)
						{
							for (j = L; j < n; j++)	// double division to avoid possible underflow
								v.E[j, i] &= (a.E[i, j] / a.E[i, L]) / g;
							for (j = L; j < n; j++)
							{
								for (s &= MultiCD.cZero, k = L; k < n; k++)
									s += a.E[i, k] * v.E[k, j];
								for (k = L; k < n; k++)
									v.E[k, j] += s * v.E[k, i];
							}
						}
						for (j = L; j < n; j++)
							v.E[i, j] &= v.E[j, i] &= MultiCD.cZero;
					}
					v.E[i, i] &= MultiCD.cOne;
					g &= rv1.E[0, i];
					L = i;
				}
				// accumulation of left-hand transformations
				for (i = Math.Min(m, n) - 1; i >= 0; i--)
				{
					L = i + 1;
					g &= w.E[0, i];
					for (j = L; j < n; j++)
						a.E[i, j] &= MultiCD.cZero;
					if (g != MultiCD.cZero)
					{
						g &= MultiCD.cOne / g;
						for (j = L; j < n; j++)
						{
							for (s &= MultiCD.cZero, k = L; k < m; k++)
								s += a.E[k, i] * a.E[k, j];
							f &= (s / a.E[i, i]) * g;
							for (k = i; k < m; k++)
								a.E[k, j] += f * a.E[k, i];
						}
						for (j = i; j < m; j++)
							a.E[j, i] *= g;
					}
					else
						for (j = i; j <m; j++)
							a.E[j, i] &= MultiCD.cZero;
					a.E[i, i] += MultiCD.cOne;
				}
				// diagonalization of the bidiagonal form
				for (k = n - 1; k >= 0; k--)				// loop over singular values
				{
					for (its = 1; its <= 300; its++)		// loop over allowed iterations
					{
						flag = 1;
						for (L = k; L >= 0; L--)		// test for splitting
						{								// note that rv1.E[0, 0] should always be zero
							nm = L - 1;
							if (L == 0 || (Abs(rv1.E[0, L]) + anorm) == anorm)
							{
								flag = 0;
								break;
							}
							if ((Abs(w.E[0, nm]) + anorm) == anorm)
								break;
						}
						if (flag != 0)
						{
							c &= MultiCD.cZero;			// cancellation of rv1.E[0, L], if L > 1
							s &= MultiCD.cOne;
							for (i = L; i < k + 1; i++)
							{
								f &= s * rv1.E[0, i];
								rv1.E[0, i] *= c;
								if ((Abs(f) + anorm) == anorm)
									break;
								g &= w.E[0, i];
								h &= Pythag(f, g);
								w.E[0, i] &= h;
								h &= MultiCD.cOne / h;
								c &= g * h;
								s &= -f * h;
								for (j = 0; j < m; j++)
								{
									y &= a.E[j, nm];
									z &= a.E[j, i];
									a.E[j, nm] &= y * c + z * s;
									a.E[j, i] &= z * c - y * s;
								}
							}
						}
						z &= w.E[0, k];
						if (L == k)						// convergence
						{
							if (z < MultiCD.cZero)		// singular value is made nonnegative
							{
								w.E[0, k] &= -z;
								for (j = 0; j < n; j++)
									v.E[j, k] &= -v.E[j, k];
							}
							break;
						}
						if (its >= 300)
						{
							MultiID.MuWriteErr("No convergence in 300 SVD iterations");
							return;
						}
						x &= w.E[0, L];				// shift from bottom 2-by-2 minor
						nm = k - 1;
						y &= w.E[0, nm];
						g &= rv1.E[0, nm];
						h &= rv1.E[0, k];
						f &= ((y - z) * (y + z) + (g - h) * (g + h)) / (MultiCD.cTwo * h * y);
						g &= Pythag(f, MultiCD.cOne);
						f &= ((x - z) * (x + z) + h*((y / (f + Sign(g, f))) - h)) / x;
						// next QR transform
						c &= s &= MultiCD.cOne;
						for (j = L; j <= nm; j++)
						{
							i = j + 1;
							g &= rv1.E[0, i];
							y &= w.E[0, i];
							h &= s * g;
							g &= c * g;
							z &= Pythag(f, h);
							rv1.E[0, j] &= z;			// can sets rv1.E[0, 0] > 0
							c &= f / z;
							s &= h / z;
							f &= x * c + g * s;
							g &= g * c - x * s;
							h &= y * s;
							y *= c;
							for (jj = 0; jj < n; jj++)
							{
								x &= v.E[jj, j];
								z &= v.E[jj, i];
								v.E[jj, j] &= x * c + z * s;
								v.E[jj, i] &= z * c - x * s;
							}
							z &= Pythag(f, h);
							w.E[0, j] &= z;			// rotation can be arbitrary if z == 0
							if (z != MultiCD.cZero)
							{
								z &= MultiCD.cOne / z;
								c &= f * z;
								s &= h *z ;
							}
							f &= c * g + s * y;
							x &= c * y - s * g;
							for (jj = 0; jj < m; jj++)
							{
								y &= a.E[jj, j];
								z &= a.E[jj, i];
								a.E[jj, j] &= y * c + z * s;
								a.E[jj, i] &= z * c - y * s;
							}
						}
						rv1.E[0, L] &= MultiCD.cZero;
						rv1.E[0, k] &= f;
						w.E[0, k] &= x;
					}
				}
				long next;
				long max = Int64.MinValue;
				for (i = 0; i < n; i++)
				{
					if (w.E[0, i] != MultiCD.cZero)
					{
						next = w.E[0, i].N + w.E[0, i].C;
						if (next > max)
							max = next;
					}
				}
				for (i = 0; i < n; i++)
				{
					next = w.E[0, i].N + w.E[0, i].C;
					if (max - next > Calc.fMC - MultiFD.fGuard)
						Calc.SetToNM(MultiCD.cZero, w.E[0, i]);
				}
			}
			SortUWV(a, w, v);
		} // SVD ZZZ

		public static void SVD2(MultiZ a, MultiZ u, MultiZ w, MultiZ v)
		// Singular Value Decomposition, a = u * w * Tran(v)
		// the diagonal matrix is output as a row vector, a.Rs must be >= a.Cs
		// if smaller, a should be filled with zero rows
		{
			int m = a.Rs;
			int n = a.Cs;
			if (m == 0)
				MultiZD.ErrNotMatrix();
			else
				CSVD(a, 0, u, w, v);
		} // SVD2 ZZZZ

		public static void SVDN(MultiZ x, int zI) // z = w part of SVD(x)
		{
			int m = x.Rs;
			int n = x.Cs;
			if (m == 0)
				MultiZD.ErrNotMatrix();
			else
			{
				bool saveExp = MultiFD.expanded;
				int saveFMC = Calc.fMC;
				int saveMMax = MultiID.mMax;
				XPCalLst.Expand();
				if (m < n)
					m = n;
				MultiZ u = new MultiZ(Calc.fMC, m, n);
				MultiZ w = new MultiZ(Calc.fMC, 1, n);
				MultiZ v = new MultiZ(Calc.fMC, n, n);
				for (int r = 0; r < x.Rs; r++)
					for (int c = 0; c < n; c++)
						MultiCD.SetToC(x.E[r, c], u.E[r, c]);
				SVD(u, w, v);
				MultiFD.expanded = saveExp;
				Calc.fMC = saveFMC;
				MultiID.mMax = saveMMax;
				int uI, wI, vI;
				Common.LFindAdd("MatU", out uI);
				Common.LFindAdd("MatW", out wI);
				Common.LFindAdd("MatV", out vI);
				Calc.SetToN(u, uI);
				Calc.SetToN(w, wI);
				Calc.SetToN(v, vI);
				Calc.SetToN(w, zI);
			}
		} // SVDN

		public static void SVD2N(MultiZ x, int zI) // z = w part of SVD(x)
		{
			int m = x.Rs;
			int n = x.Cs;
			if (m == 0)
				MultiZD.ErrNotMatrix();
			else
			{
				bool saveExp = MultiFD.expanded;
				int saveFMC = Calc.fMC;
				int saveMMax = MultiID.mMax;
				XPCalLst.Expand();
				if (m < n)
					m = n;
				MultiZ u = new MultiZ(Calc.fMC, m, m);
				MultiZ w = new MultiZ(Calc.fMC, 1, n);
				MultiZ v = new MultiZ(Calc.fMC, n, n);
				SVD2(x, u, w, v);
				MultiFD.expanded = saveExp;
				Calc.fMC = saveFMC;
				MultiID.mMax = saveMMax;
				int uI, wI, vI;
				Common.LFindAdd("MatU", out uI);
				Common.LFindAdd("MatW", out wI);
				Common.LFindAdd("MatV", out vI);
				Calc.SetToN(u, uI);
				Calc.SetToN(w, wI);
				Calc.SetToN(v, vI);
				Calc.SetToN(w, zI);
			}
		} // SVD2N

		public static void SVDInvN(MultiZ u, MultiZ w, MultiZ v, int zI) // z = u * w * v^H
		{
			int m = u.Rs;
			int n = v.Rs;
			if (m == 0 || n == 0 || m < n || u.Cs < n || w.Rs != 1 || w.Cs != n || v.Cs != n)
				MultiZD.ErrIncompatible();
			else
			{
				bool saveExp = MultiFD.expanded;
				int saveFMC = Calc.fMC;
				int saveMMax = MultiID.mMax;
				XPCalLst.Expand();
				MultiZ z = new MultiZ(Calc.fMC, m, n);
				SVDInv(u, w, v, z);
				MultiFD.expanded = saveExp;
				Calc.fMC = saveFMC;
				MultiID.mMax = saveMMax;
				Calc.SetToN(z, zI);
			}
		} // SVDInvN

		public static void SVDInv(MultiZ u, MultiZ w, MultiZ v, MultiZ z)
		// Inverse of Singular Value Decomposition, z = u * w * v^H
		// the diagonal matrix w is input as an n length row vector
		// v is an n by n square matrix, u is an m >= n by nu >= n matrix
		// only the first n columns of u are used
		// z is output as an m by n matrix
		{
			int m = u.Rs;
			int n = v.Rs;
			if (m == 0 || n == 0 || m < n || u.Cs < n || w.Rs != 1 || w.Cs != n || v.Cs != n)
				MultiZD.ErrIncompatible();
			else
			{
				for (int i = 0; i < m; i++)		// Compute z = u * w * v^H
					for (int j = 0; j < n; j++)
					{
						z.E[i, j] &= MultiCD.cZero;
						for (int k = 0; k < m && k < n; k++)
							z.E[i, j] += u.E[i, k] * w.E[0, k] * Conj(v.E[j, k]);
					}
			}
		} // SVDInv

		public static void Cyclo(MultiZ x, MultiZ z) // z = Cyclotonic polynomial of Int(x.r)
		{
			if (x.Rs != 0)
				MultiZD.ErrNotScalar();
			else
			{
				MultiF n = new MultiF();
				MultiFD.IntF(x, n);
				if (n < MultiFD.fOne)
				{
					MultiID.MuWriteErr("Cyclo: argument must be >= 1");
					return;
				}
				int saveFMC = Calc.fMC;
				Calc.fMC = 3;
				MultiF k = new MultiF();
				MultiF gcd = new MultiF();
				MultiZ p = new MultiZ(1, 2, 1);
				MultiZ zz = new MultiZ(1, 1, 1);
				int pI = XPCalLst.GetPiPtr();
				int tI;
				Common.InitNAdd(1, "Cyclo-tI", out tI);		// t = temp on the list
				int eI;
				Common.InitNAdd(1, "Cyclo-eI", out eI);		// e = temp on the list
				p.E[0, 0] &= MultiCD.cOne;
				zz.E[0, 0] &= MultiCD.cOne;
				Calc.SetToN(MultiCD.cTwo, tI);
				XCCalLst.RMulNF(Common.nList[pI], tI);
				XCCalLst.RDivNF(n, tI);
				XCCalLst.RMulN(MultiCD.ci, tI);
				for (k &= MultiFD.fOne; k <= n; k++)
				{
					if (MultiID.muErr || MultiID.muAbort)
						break;
					XPCalLst.GCD(k, n, gcd);
					if (gcd == MultiFD.fOne)
					{
						XCCalLst.MulN(Common.nList[tI], k, eI);
						XCCalLst.RExpN(eI);
						Calc.SetToNM(-Common.nList[eI], p.E[1, 0]);
						PolMul(p, zz, zz);
					}
				}
				Common.LDelete(eI);
				Calc.SetToN(zz, tI);
				RRoundN(tI);
				Calc.SetToNM(Common.nList[tI], z);
				Common.LDelete(tI);
				Calc.fMC = saveFMC;
			}
		} // Cyclo

		public static void CycloN(MultiZ x, int zI) // z = Cyclotonic polynomial of Int(x.r)
		{
			if (x.Rs != 0)
				MultiZD.ErrNotScalar();
			else
			{
				MultiZ z = new MultiZ();
				Cyclo(x, z);
				Calc.SetToN(z, zI);
			}
		} // CycloN

		public static MultiF FacF(MultiF x) // return x!
		{
			int zI;
			Common.InitNAdd(1, "FacF-zI", out zI);
			XPCalLst.FacSN(x, zI, false, 0);
			MultiF z = new MultiF(Calc.fMC);
			XPCalLst.SetToNM(Common.nList[zI], z);
			Common.LDelete(zI);
			return z;
		} // FacF

		public static MultiF PowNF(MultiF b, MultiF p) // return z = b^p
		{
			MultiF z = new MultiF(Calc.fMC);
			XPCalLst.PowNF(b, p, z);
			return z;
		} // PowNF

		public static void Legen(MultiZ x, MultiZ z) // z = Legendre polynomial of Int(x.r)
		{
			if (x.Rs != 0)
				MultiZD.ErrNotScalar();
			else
			{
				MultiF nF = new MultiF();
				MultiF mF = new MultiF();
				MultiFD.IntF(x, nF);
				int n;
				MultiFD.Get1F(out n, nF);
				if (n < 0)
				{
					MultiID.MuWriteErr("Legen: argument must be >= 0");
					return;
				}
				bool saveExp = MultiFD.expanded;
				int saveFMC = Calc.fMC;
				int saveMMax = MultiID.mMax;
				XPCalLst.Expand();
				MultiF t = new MultiF();
				MultiZ zz = new MultiZ(1, n + 1, 1);
				// z(n) = Fac(n+n) / (2^n * Fac(n)^2)
				XPCalLst.SetToNM(FacF(nF + nF) / (PowNF(MultiFD.fTwo, nF) * PowNF(FacF(nF), MultiFD.fTwo)), zz.E[0, 0]);
				for (int m = n - 2; m >= 0; m -= 2)
				{
					// f = (m+2)*(m+1) / ((n-m)*(n+m+1))
					MultiFD.SetTo1F(m, mF);
					Calc.SetToNM(-zz.E[n - m - 2, 0] * (mF + MultiFD.fTwo) * (mF + MultiFD.fOne) /
						((nF - mF) * (nF + mF + MultiFD.fOne)), zz.E[n - m, 0]);
				}
				MultiFD.expanded = saveExp;
				Calc.fMC = saveFMC;
				MultiID.mMax = saveMMax;
				Calc.SetToNM(zz, z);
			}
		} // Legen

		public static void LegenN(MultiZ x, int zI) // z = Legendre polynomial of Int(x.r)
		{
			if (x.Rs != 0)
				MultiZD.ErrNotScalar();
			else
			{
				MultiZ z = new MultiZ();
				Legen(x, z);
				Calc.SetToN(z, zI);
			}
		} // LegenN

		public static void Cheby(MultiZ x, MultiZ z) // z = Chebyshev polynomial of Int(x.r)
		{
			if (x.Rs != 0)
				MultiZD.ErrNotScalar();
			else
			{
				MultiF nF = new MultiF();
				MultiFD.IntF(x, nF);
				nF.S = false;					// Cheby(-n) = Cheby(n)
				int n;
				MultiFD.Get1F(out n, nF);
				MultiZ zz = new MultiZ(1, n + 1, 1);
				if (n < 2)
				{
					zz.E[0, 0] &= MultiFD.fOne;	// Cheby(0) = {1}
					Calc.SetToNM(zz, z);		// Cheby(1) = {1; 0}
					return;
				}
				MultiZ p0 = new MultiZ(1, 1, 1);
				MultiZ p1 = new MultiZ(1, 2, 1);
				MultiZ p2 = new MultiZ(1, 2, 1);
				MultiZ t = new MultiZ();
				p0.E[0, 0] &= MultiFD.fOne;		// p1 = {1}
				p1.E[0, 0] &= MultiFD.fOne;		// p2 = {1; 0}
				p2.E[0, 0] &= MultiFD.fTwo;		// p2 = {2; 0}
				for (int i = 2; i <= n; i++)
				{
					PolMul(p2, p1, t);			// Cheby(n+1) = {2; 0} * Cheby(n)
					PolAdd(t, -p0, zz);			// - Cheby(n-1)
					if (i < n)
					{
						Calc.SetToNM(p1, p0);	// update Cheby(n-1)
						Calc.SetToNM(zz, p1);	// and Cheby(n)
					}
				}
				Calc.SetToNM(zz, z);
			}
		} // Cheby

		public static void ChebyN(MultiZ x, int zI) // z = Chebyshev polynomial of Int(x.r)
		{
			if (x.Rs != 0)
				MultiZD.ErrNotScalar();
			else
			{
				MultiZ z = new MultiZ();
				Cheby(x, z);
				Calc.SetToN(z, zI);
			}
		} // ChebyN

		public static void DecompN(MultiZ x, int zI) // z = LUdecomp(x)
		{
			if (x.Rs == 0)
				MultiZD.ErrNotMatrix();
			else if (x.Cs != x.Rs)
				MultiZD.ErrNotSquare();
			else
			{
				bool saveExp = MultiFD.expanded;
				int saveFMC = Calc.fMC;
				int saveMMax = MultiID.mMax;
				XPCalLst.Expand();
				int n = x.Rs;
				MultiZ a = new MultiZ(Calc.fMC, n, n);
				a &= x;
				int[] index;
				index = new int[n];
				MultiC d = new MultiC();
				bool sing;
				MultiZD.LUdecomp(a, index, d, out sing);	// decompose the matrix
				Calc.SetToN(a, zI);
				MultiFD.expanded = saveExp;
				Calc.fMC = saveFMC;
				MultiID.mMax = saveMMax;
				Calc.NormN(zI);
			}
		} // DecompN

		public static void RREF(MultiZ x, int e, MultiZ z) // z = RREF of x less e right columns
		// RREF = Reduced Row Echelon Form
		{
			int m = x.Rs;
			int ne = x.Cs;
			int n = ne - e;
			int i = 0;
			int j = 0;
			int k;
			int c;
			int u;
			int maxInd;
			MultiC maxVal = new MultiC();
			MultiC val = new MultiC();
			MultiC temp = new MultiC();
			MultiZ a = new MultiZ();
			Calc.SetToNM(x, a);
			bool saveExp = MultiFD.expanded;
			int saveFMC = Calc.fMC;
			int saveMMax = MultiID.mMax;
			XPCalLst.Expand();
			MultiC d = new MultiC(Calc.fMC - MultiFD.fGuard);
			MultiC auc = new MultiC(Calc.fMC - MultiFD.fGuard);
			while (i < m && j < n)
			{
				// Find pivot in column j, starting in row i
				Calc.SetToNM(a.E[i, j], maxVal);
				maxInd = i;
				for (k = i+1; k < m; k++)
				{
					Calc.SetToNM(a.E[k, j], val);
					if (Abs(val) > Abs(maxVal))
					{
						Calc.SetToNM(val, maxVal);
						maxInd = k;
					}
				}
				if (maxVal != MultiCD.cZero)
				{
					// switch rows i and maxInd, but remain the values of i and maxInd
					// Now a[i, j] will contain the same value as maxVal
					for (c = 0; c < ne; c++)
					{
						if (c != j)
							Calc.SetToNM(a.E[maxInd, c], temp);
						Calc.SetToNM(a.E[i, c], a.E[maxInd, c]);
						// divide row i by maxVal
						if (c != j)
							Calc.SetToNM(temp / maxVal, a.E[i, c]);
					}
					// Now A[i, j] will have the value 1
					Calc.SetToNM(MultiCD.cOne, a.E[i, j]);
					for (u = 0; u < m; u++)
					{
						if (u != i)
						{
							// subtract a[u, j] * row i from row u
							for (c = 0; c < ne; c++)
								if (c != j)
								{
									auc &= a.E[u, c];
									Calc.SetToNM(a.E[u, c] - a.E[u, j] * a.E[i, c], a.E[u, c]);
									d &= auc - a.E[u, c];
									if (d == auc)
										Calc.SetToNM(MultiCD.cZero, a.E[u, c]);
								}
							// Now A[u, j] will be 0, since
							// A[u, j] - A[u, j] * A[i, j] = A[u, j] - 1 * A[u, j] = 0
							Calc.SetToNM(MultiCD.cZero, a.E[u, j]);
						}
					}
					i++;
				}
				j++;
			}
			MultiFD.expanded = saveExp;
			Calc.fMC = saveFMC;
			MultiID.mMax = saveMMax;
			Calc.SetToNM(a, z);
		} // RREF

		public static void RREF(MultiZ x, MultiZ z) // z = RREF of x
		// RREF = Reduced Row Echelon Form
		{
			RREF(x, 0, z);
		} // RREF

		public static void RREFN(MultiZ x, int zI) // z = RREF of x
		// RREF = Reduced Row Echelon Form
		{
			if (x.Rs == 0)
				MultiZD.ErrNotMatrix();
			else
			{
				MultiZ z = new MultiZ();
				RREF(x, z);
				Calc.SetToN(z, zI);
			}
		} // RREFN

		public static void RREFa(MultiZ x, MultiZ y, MultiZ z) // z = RREF of [x y]
		// RREF = Reduced Row Echelon Form
		{
			int tI;
			Common.InitNAdd(1, "RREFa-tI", out tI);
			ConcatN(x, y, tI);
			MultiZ zz = new MultiZ();
			RREF(Common.nList[tI], y.Cs, zz);
			if (x.Rs != x.Cs || zz.E[x.Rs - 1, x.Cs - 1] == MultiCD.cZero)
				MultiID.WriteDLn("Matrix is singular");
			MatrixN(y.Rs, y.Cs, tI);
			for (int r = 0; r < y.Rs; r++)
				for (int c = 0; c < y.Cs; c++)
					Calc.SetToNM(zz.E[r, c + x.Cs], Common.nList[tI].E[r, c]);
			Calc.SetToNM(Common.nList[tI], z);
			Common.LDelete(tI);
		} // RREFa

		public static void RREFaN(MultiZ x, MultiZ y, int zI) // z = RREF of [x y]
		// RREF = Reduced Row Echelon Form
		{
			if (x.Rs == 0 || y.Rs == 0)
				MultiZD.ErrNotMatrix();
			else if (x.Rs != y.Rs)
				MultiZD.ErrIncompatible();
			else
			{
				MultiZ z = new MultiZ();
				RREFa(x, y, z);
				Calc.SetToN(z, zI);
			}
		} // RREFaN

		public static void InvN(MultiZ x, int zI) // z = 1 / x
		// this code is adapted from Numerical Recipes in C
		{
				if (x.Rs == 0)
				XCCalLst.InvN(x, zI);
			else if (x.Cs != x.Rs)
				MultiZD.ErrNotSquare();
			else
			{
				bool saveExp = MultiFD.expanded;
				int saveFMC = Calc.fMC;
				int saveMMax = MultiID.mMax;
				XPCalLst.Expand();
				MultiZ z = new MultiZ(Calc.fMC, x.Rs, x.Rs);
				MultiZD.InvZ(x, z);
				Calc.SetToN(z, zI);
				MultiFD.expanded = saveExp;
				Calc.fMC = saveFMC;
				MultiID.mMax = saveMMax;
				Calc.NormN(zI);
			}
		} // InvN

		public static void RInvN(int zI) // z = 1 / z
		{
			InvN(Common.nList[zI], zI);
		}

		public static void SetFMBN(int zI) // set Modulo base FMB on Common.nList = value of zI
		{
			int fMBI;
			Common.LFindAdd("FMB", out fMBI);
			if (fMBI == zI) return;
			Calc.SetToII(zI, fMBI);
		}

		public static void ModFMBN(MultiZ x, int zI) // z = x % FMB
		{
			int fMBI;
			if (Common.FMBIok(zI, out fMBI) && (!Common.ZTestN(fMBI) || Common.nList[fMBI].Rs != 0))
				RModN(Common.nList[fMBI], zI);
		}

		public static void RModFMBN(int zI) // z = z % FMB
		{
			ModFMBN(Common.nList[zI], zI);
		}

		public static void SinN(MultiZ x, int zI)	// z = Sin(x)
		{
			if (x.Rs == 0)
				XCCalLst.SinN(x, zI);
			else
			{
				int tI;
				Common.InitNAdd(1, "SinN-tI", out tI);
				if (Common.nList[zI].Rs != x.Rs || Common.nList[zI].Cs != x.Cs)
					MatrixN(x.Rs, x.Cs, zI);
				for (int r = 0; r < x.Rs; r++)
					for (int c = 0; c < x.Cs; c++)
					{
						XCCalLst.SinN(x.E[r, c], tI);
						if (MultiID.muErr)
						{
							Common.LDelete(tI);
							return;
						}
						Calc.SetToNM(Common.nList[tI], Common.nList[zI].E[r, c]);
					}
				Common.LDelete(tI);
			}
		} // SinN

		public static void RSinN(int zI)	// z = Sin(z)
		{
			SinN(Common.nList[zI], zI);
		}

		public static void ConjN(MultiZ x, int zI) // z = Complex Conjugate(x)
		// x may have the same address -in memory as z
		// x is not changed unless it shares memory with z
		{
			Calc.SetToN(x, zI);
			if (x.Rs == 0)
				MultiFD.ChSF(Common.nList[zI].I);
			else
			{
				for (int r = 0; r < x.Rs; r++)
					for (int c = 0; c < x.Cs; c++)
					{
						MultiFD.ChSF(Common.nList[zI].E[r, c].I);
					}
			}
		} // ConjN

		public static void RConjN(int zI) // z = Complex Conjugate(z)
		{
			ConjN(Common.nList[zI], zI);
		}

		public static void SignN(MultiZ x, int zI)	// z = Sin(x)
		{
			if (x.Rs == 0)
				XCCalLst.SignN(x, zI);
			else
			{
				int tI;
				Common.InitNAdd(1, "SignN-tI", out tI);
				if (Common.nList[zI].Rs != x.Rs || Common.nList[zI].Cs != x.Cs)
					MatrixN(x.Rs, x.Cs, zI);
				for (int r = 0; r < x.Rs; r++)
					for (int c = 0; c < x.Cs; c++)
					{
						XCCalLst.SignN(x.E[r, c], tI);
						if (MultiID.muErr)
						{
							Common.LDelete(tI);
							return;
						}
						Calc.SetToNM(Common.nList[tI], Common.nList[zI].E[r, c]);
					}
				Common.LDelete(tI);
			}
		} // SignN

		public static void RSignN(int zI)	// z = Sin(z)
		{
			SignN(Common.nList[zI], zI);
		}

		public static void CosN(MultiZ x, int zI)	// z = Cos(x)
		// Cos(x) = Sin(x + pi/2)
		{
			if (x.Rs == 0)
				XCCalLst.CosN(x, zI);
			else
			{
				int tI;
				Common.InitNAdd(1, "CosN-tI", out tI);
				if (Common.nList[zI].Rs != x.Rs || Common.nList[zI].Cs != x.Cs)
					MatrixN(x.Rs, x.Cs, zI);
				for (int r = 0; r < x.Rs; r++)
					for (int c = 0; c < x.Cs; c++)
					{
						XCCalLst.CosN(x.E[r, c], tI);
						if (MultiID.muErr)
						{
							Common.LDelete(tI);
							return;
						}
						Calc.SetToNM(Common.nList[tI], Common.nList[zI].E[r, c]);
					}
				Common.LDelete(tI);
			}
		} // CosN

		public static void RCosN(int zI) // z = Cos(z)
		{
			CosN(Common.nList[zI], zI);
		}

		public static void TanN(MultiZ x, int zI)	// z = Tan(x)
		// Tan(x) = Sin(x) / SqRt(1 - Sq(Sin(x))
		// and change sign of Tan(x) if in 2nd or 3rd quadrant
		{
			bool forCot = false;
			TanN(x, zI, ref forCot);
		}

		public static void TanN(MultiZ x, int zI, ref bool forCot)	// z = Tan(x)
		// Tan(x) = Sin(x) / SqRt(1 - Sq(Sin(x))
		// and change sign of Tan(x) if in 2nd or 3rd quadrant
		{
			if (x.Rs == 0)
				XCCalLst.TanN(x, zI);
			else
			{
				int tI;
				Common.InitNAdd(1, "TanN-tI", out tI);
				if (Common.nList[zI].Rs != x.Rs || Common.nList[zI].Cs != x.Cs)
					MatrixN(x.Rs, x.Cs, zI);
				for (int r = 0; r < x.Rs; r++)
					for (int c = 0; c < x.Cs; c++)
					{
						XCCalLst.TanN(x.E[r, c], tI);
						if (MultiID.muErr)
						{
							Common.LDelete(tI);
							return;
						}
						Calc.SetToNM(Common.nList[tI], Common.nList[zI].E[r, c]);
					}
				Common.LDelete(tI);
			}
		} // TanN

		public static void RTanN(int zI) // z = Tan(z)
		{
			TanN(Common.nList[zI], zI);
		}

		public static void ATanN(MultiZ x, int zI)	// z = ATan(x)
		{
			if (x.Rs == 0)
				XCCalLst.ATanN(x, zI);
			else
			{
				int tI;
				Common.InitNAdd(1, "ATanN-tI", out tI);
				if (Common.nList[zI].Rs != x.Rs || Common.nList[zI].Cs != x.Cs)
					MatrixN(x.Rs, x.Cs, zI);
				for (int r = 0; r < x.Rs; r++)
					for (int c = 0; c < x.Cs; c++)
					{
						XCCalLst.ATanN(x.E[r, c], tI);
						if (MultiID.muErr)
						{
							Common.LDelete(tI);
							return;
						}
						Calc.SetToNM(Common.nList[tI], Common.nList[zI].E[r, c]);
					}
				Common.LDelete(tI);
			}
		} // ATanN

		public static void RATanN(int zI) // z = ATan(z)
		{
			ATanN(Common.nList[zI], zI);
		}

		public static void ATan2N(MultiZ y, MultiZ x, int zI)	// z = ATan(y / x)
		// Compute Arctan(y / x) from -pi to pi
		{
			if (x.Rs == 0 && y.Rs == 0)
				XCCalLst.ATan2N(y, x, zI);
			else if (x.Rs != y.Rs || x.Cs != y.Cs)
				MultiZD.ErrIncompatible();
			else
			{
				int tI;
				Common.InitNAdd(1, "ATan2N-tI", out tI);
				if (Common.nList[zI].Rs != x.Rs || Common.nList[zI].Cs != x.Cs)
					MatrixN(x.Rs, x.Cs, zI);
				for (int r = 0; r < x.Rs; r++)
					for (int c = 0; c < x.Cs; c++)
					{
						XCCalLst.ATan2N(y.E[r, c], x.E[r, c], tI);
						if (MultiID.muErr)
						{
							Common.LDelete(tI);
							return;
						}
						Calc.SetToNM(Common.nList[tI], Common.nList[zI].E[r, c]);
					}
				Common.LDelete(tI);
			}
		} // ATan2N

		public static void AbsN(MultiZ x, int zI) // z = |x|
		{
			Calc.SetToN(x, zI);
			if (x.Rs == 0)
				XCCalLst.AbsN(x, zI);
			else
			{
				int tI;
				Common.InitNAdd(1, "AbsN-tI", out tI);
				for (int r = 0; r < x.Rs; r++)
					for (int c = 0; c < x.Cs; c++)
					{
						XCCalLst.AbsN(x.E[r, c], tI);
						Calc.SetToNM(Common.nList[tI], Common.nList[zI].E[r, c]);
					}
				Common.LDelete(tI);
			}
		} // AbsN

		public static void MagN(MultiZ x, MultiZ y, int zI)	// z = SqRt(Sq(x) + Sq(y))
		{
			if (x.Rs == 0 && y.Rs == 0)
				XCCalLst.MagN(x, y, zI);
			else if (x.Rs != y.Rs || x.Cs != y.Cs)
				MultiZD.ErrIncompatible();
			else
			{
				int tI;
				Common.InitNAdd(1, "MagN-tI", out tI);
				if (Common.nList[zI].Rs != x.Rs || Common.nList[zI].Cs != x.Cs)
					MatrixN(x.Rs, x.Cs, zI);
				for (int r = 0; r < x.Rs; r++)
					for (int c = 0; c < x.Cs; c++)
					{
						XCCalLst.MagN(x.E[r, c], y.E[r, c], tI);
						if (MultiID.muErr)
						{
							Common.LDelete(tI);
							return;
						}
						Calc.SetToNM(Common.nList[tI], Common.nList[zI].E[r, c]);
					}
				Common.LDelete(tI);
			}
		} // MagN

		public static void MeanCN(MultiZ x, MultiZ y, int zI)	// z = common mean of x and y
		{
			if (x.Rs == 0 && y.Rs == 0)
				XCCalLst.MeanCN(x, y, zI);
			else if (x.Rs != y.Rs || x.Cs != y.Cs)
				MultiZD.ErrIncompatible();
			else
			{
				int tI;
				Common.InitNAdd(1, "MeanCN-tI", out tI);
				if (Common.nList[zI].Rs != x.Rs || Common.nList[zI].Cs != x.Cs)
					MatrixN(x.Rs, x.Cs, zI);
				for (int r = 0; r < x.Rs; r++)
					for (int c = 0; c < x.Cs; c++)
					{
						XCCalLst.MeanCN(x.E[r, c], y.E[r, c], tI);
						if (MultiID.muErr)
						{
							Common.LDelete(tI);
							return;
						}
						Calc.SetToNM(Common.nList[tI], Common.nList[zI].E[r, c]);
					}
				Common.LDelete(tI);
			}
		} //MeanCN

		public static void MaxN(MultiZ x, MultiZ y, int zI)	// z = max of x and y
		{
			if (x.Rs == 0 && y.Rs == 0)
				XCCalLst.MaxN(x, y, zI);
			else if (x.Rs != y.Rs || x.Cs != y.Cs)
				MultiZD.ErrIncompatible();
			else
			{
				int tI;
				Common.InitNAdd(1, "MaxN-tI", out tI);
				if (Common.nList[zI].Rs != x.Rs || Common.nList[zI].Cs != x.Cs)
					MatrixN(x.Rs, x.Cs, zI);
				for (int r = 0; r < x.Rs; r++)
					for (int c = 0; c < x.Cs; c++)
					{
						XCCalLst.MaxN(x.E[r, c], y.E[r, c], tI);
						if (MultiID.muErr)
						{
							Common.LDelete(tI);
							return;
						}
						Calc.SetToNM(Common.nList[tI], Common.nList[zI].E[r, c]);
					}
				Common.LDelete(tI);
			}
		} //MaxN

		public static void MinN(MultiZ x, MultiZ y, int zI)	// z = min of x and y
		{
			if (x.Rs == 0 && y.Rs == 0)
				XCCalLst.MinN(x, y, zI);
			else if (x.Rs != y.Rs || x.Cs != y.Cs)
				MultiZD.ErrIncompatible();
			else
			{
				int tI;
				Common.InitNAdd(1, "MinN-tI", out tI);
				if (Common.nList[zI].Rs != x.Rs || Common.nList[zI].Cs != x.Cs)
					MatrixN(x.Rs, x.Cs, zI);
				for (int r = 0; r < x.Rs; r++)
					for (int c = 0; c < x.Cs; c++)
					{
						XCCalLst.MinN(x.E[r, c], y.E[r, c], tI);
						if (MultiID.muErr)
						{
							Common.LDelete(tI);
							return;
						}
						Calc.SetToNM(Common.nList[tI], Common.nList[zI].E[r, c]);
					}
				Common.LDelete(tI);
			}
		} //MinN

		public static void ModN(MultiZ x, MultiZ y, int zI)	// z = Mod(x, y))
		// same as FORTRAN definition: z = Mod(x, y) = x - (Floor(x/y) * y)
		// sign(z) = sign(y) or z = 0
		{
			if (x.Rs == 0 && y.Rs == 0)
				XCCalLst.ModN(x, y, zI);
			else if (x.Rs != y.Rs || x.Cs != y.Cs)
				MultiZD.ErrIncompatible();
			else
			{
				int tI;
				Common.InitNAdd(1, "ModN-tI", out tI);
				if (Common.nList[zI].Rs != x.Rs || Common.nList[zI].Cs != x.Cs)
					MatrixN(x.Rs, x.Cs, zI);
				for (int r = 0; r < x.Rs; r++)
					for (int c = 0; c < x.Cs; c++)
					{
						XCCalLst.ModN(x.E[r, c], y.E[r, c], tI);
						if (MultiID.muErr)
						{
							Common.LDelete(tI);
							return;
						}
						Calc.SetToNM(Common.nList[tI], Common.nList[zI].E[r, c]);
					}
				Common.LDelete(tI);
			}
		} // ModN int

		public static void RModN(MultiZ x, int zI) // z = Mod(z, x))
		{
			ModN(Common.nList[zI], x, zI);
		}

		public static void GamPN(MultiZ x, MultiZ y, int zI) // z = GamP(x, y)
		// GamP(a, x) is the lower regularized incomplete gamma function with parameter a and argument x
		{
			if (x.Rs == 0 && y.Rs == 0)
				XCCalLst.GamPN(x, y, zI);
			else if (x.Rs != y.Rs || x.Cs != y.Cs)
				MultiZD.ErrIncompatible();
			else
			{
				int tI;
				Common.InitNAdd(1, "GamPN-tI", out tI);
				if (Common.nList[zI].Rs != x.Rs || Common.nList[zI].Cs != x.Cs)
					MatrixN(x.Rs, x.Cs, zI);
				for (int r = 0; r < x.Rs; r++)
					for (int c = 0; c < x.Cs; c++)
					{
						XCCalLst.GamPN(x.E[r, c], y.E[r, c], tI);
						if (MultiID.muErr)
						{
							Common.LDelete(tI);
							return;
						}
						Calc.SetToNM(Common.nList[tI], Common.nList[zI].E[r, c]);
					}
				Common.LDelete(tI);
			}		
		} // GamPN

		public static void GamQN(MultiZ x, MultiZ y, int zI) // z = GamQ(x, y)
		// GamQ(a, x) is the upper regularized incomplete gamma function with parameter a and argument x
		{
			if (x.Rs == 0 && y.Rs == 0)
				XCCalLst.GamQN(x, y, zI);
			else if (x.Rs != y.Rs || x.Cs != y.Cs)
				MultiZD.ErrIncompatible();
			else
			{
				int tI;
				Common.InitNAdd(1, "GamQN-tI", out tI);
				if (Common.nList[zI].Rs != x.Rs || Common.nList[zI].Cs != x.Cs)
					MatrixN(x.Rs, x.Cs, zI);
				for (int r = 0; r < x.Rs; r++)
					for (int c = 0; c < x.Cs; c++)
					{
						XCCalLst.GamQN(x.E[r, c], y.E[r, c], tI);
						if (MultiID.muErr)
						{
							Common.LDelete(tI);
							return;
						}
						Calc.SetToNM(Common.nList[tI], Common.nList[zI].E[r, c]);
					}
				Common.LDelete(tI);
			}		
		} // GamQN

		public static void GamLN(MultiZ x, MultiZ y, int zI) // z = GamL(x, y)
		// GamL(a, x) is the lower incomplete gamma function with parameter a and argument x
		{
			if (x.Rs == 0 && y.Rs == 0)
				XCCalLst.GamLN(x, y, zI);
			else if (x.Rs != y.Rs || x.Cs != y.Cs)
				MultiZD.ErrIncompatible();
			else
			{
				int tI;
				Common.InitNAdd(1, "GamLN", out tI);
				if (Common.nList[zI].Rs != x.Rs || Common.nList[zI].Cs != x.Cs)
					MatrixN(x.Rs, x.Cs, zI);
				for (int r = 0; r < x.Rs; r++)
					for (int c = 0; c < x.Cs; c++)
					{
						XCCalLst.GamLN(x.E[r, c], y.E[r, c], tI);
						if (MultiID.muErr)
						{
							Common.LDelete(tI);
							return;
						}
						Calc.SetToNM(Common.nList[tI], Common.nList[zI].E[r, c]);
					}
				Common.LDelete(tI);
			}		
		} // GamLN

		public static void GamUN(MultiZ x, MultiZ y, int zI) // z = GamU(x, y)
		// GamU(a, x) is the upper incomplete gamma function with parameter a and argument x
		{
			if (x.Rs == 0 && y.Rs == 0)
				XCCalLst.GamUN(x, y, zI);
			else if (x.Rs != y.Rs || x.Cs != y.Cs)
				MultiZD.ErrIncompatible();
			else
			{
				int tI;
				Common.InitNAdd(1, "GamUN-tI", out tI);
				if (Common.nList[zI].Rs != x.Rs || Common.nList[zI].Cs != x.Cs)
					MatrixN(x.Rs, x.Cs, zI);
				for (int r = 0; r < x.Rs; r++)
					for (int c = 0; c < x.Cs; c++)
					{
						XCCalLst.GamUN(x.E[r, c], y.E[r, c], tI);
						if (MultiID.muErr)
						{
							Common.LDelete(tI);
							return;
						}
						Calc.SetToNM(Common.nList[tI], Common.nList[zI].E[r, c]);
					}
				Common.LDelete(tI);
			}		
		} // GamUN

		public static void BetaCN(MultiZ x, MultiZ y, int zI)
		// z = the complete beta function B(x, y) = gamma(x) * gamma(y) / gamma(x+y)
		// also equal to 1 / (x * Bino(x+y-1, y-1)
		{
			if (x.Rs == 0 && y.Rs == 0)
				XCCalLst.BetaCN(x, y, zI);
			else if (x.Rs != y.Rs || x.Cs != y.Cs)
				MultiZD.ErrIncompatible();
			else
			{
				int tI;
				Common.InitNAdd(1, "BetaCN-tI", out tI);
				if (Common.nList[zI].Rs != x.Rs || Common.nList[zI].Cs != x.Cs)
					MatrixN(x.Rs, x.Cs, zI);
				for (int r = 0; r < x.Rs; r++)
					for (int c = 0; c < x.Cs; c++)
					{
						XCCalLst.BetaCN(x.E[r, c], y.E[r, c], tI);
						if (MultiID.muErr)
						{
							Common.LDelete(tI);
							return;
						}
						Calc.SetToNM(Common.nList[tI], Common.nList[zI].E[r, c]);
					}
				Common.LDelete(tI);
			}
		} // BetaCN

		public static void BinoGN(MultiZ x, MultiZ y, int zI)
		// z = Binomial coefficient (x, y) using the gamma function
		{
			if (x.Rs == 0 && y.Rs == 0)
				XCCalLst.BinoGN(x, y, zI);
			else if (x.Rs != y.Rs || x.Cs != y.Cs)
				MultiZD.ErrIncompatible();
			else
			{
				int tI;
				Common.InitNAdd(1, "BinoGN-tI", out tI);
				if (Common.nList[zI].Rs != x.Rs || Common.nList[zI].Cs != x.Cs)
					MatrixN(x.Rs, x.Cs, zI);
				for (int r = 0; r < x.Rs; r++)
					for (int c = 0; c < x.Cs; c++)
					{
						XCCalLst.BinoGN(x.E[r, c], y.E[r, c], tI);
						if (MultiID.muErr)
						{
							Common.LDelete(tI);
							return;
						}
						Calc.SetToNM(Common.nList[tI], Common.nList[zI].E[r, c]);
					}
				Common.LDelete(tI);
			}
		} // BinoGN

		public static void GCD(MultiZ a, MultiZ b, MultiZ z) // z = GCD(a, b))
		{
			if (a.Rs == 0 && b.Rs == 0)
				XCCalLst.GCD(a, b, z);
			else if (a.Rs != b.Rs || a.Cs != b.Cs)
				MultiZD.ErrIncompatible();
			else
			{
				int tI;
				Common.InitNAdd(1, "GCD-tI", out tI);
				if (z.Rs != a.Rs || z.Cs != a.Cs)
					Matrix(a.Rs, a.Cs, z);
				for (int r = 0; r < a.Rs; r++)
					for (int c = 0; c < a.Cs; c++)
					{
						XCCalLst.GCDN(a.E[r, c], b.E[r, c], tI);
						if (MultiID.muErr)
						{
							Common.LDelete(tI);
							return;
						}
						Calc.SetToNM(Common.nList[tI], z.E[r, c]);
					}
				Common.LDelete(tI);
			}
		} // GCD

		public static void GCDN(MultiZ a, MultiZ b, int zI) // z = GCD(a, b))
		{
			MultiZ zz = new MultiZ();
			GCD(a, b, zz);
			Calc.SetToN(zz, zI);
		} // GCDN

		public static void LCM(MultiZ a, MultiZ b, MultiZ z) // z = LCM(a, b))
		// LCM(a, b) = a * b / GCD(a, b)
		{
			if (a.Rs == 0 && b.Rs == 0)
				XCCalLst.LCM(a, b, z);
			else if (a.Rs != b.Rs || a.Cs != b.Cs)
				MultiZD.ErrIncompatible();
			else
			{
				int tI;
				Common.InitNAdd(1, "GCD-tI", out tI);
				if (z.Rs != a.Rs || z.Cs != a.Cs)
					Matrix(a.Rs, a.Cs, z);
				for (int r = 0; r < a.Rs; r++)
					for (int c = 0; c < a.Cs; c++)
					{
						XCCalLst.LCMN(a.E[r, c], b.E[r, c], tI);
						if (MultiID.muErr)
						{
							Common.LDelete(tI);
							return;
						}
						Calc.SetToNM(Common.nList[tI], z.E[r, c]);
					}
				Common.LDelete(tI);
			}
		} // LCM

		public static void LCMN(MultiZ a, MultiZ b, int zI) // z = LCM(a, b))
		// LCM(a, b) = a * b / GCD(a, b)
		{
			MultiZ zz = new MultiZ();
			LCM(a, b, zz);
			Calc.SetToN(zz, zI);
		} // LCMN

		public static void ASinN(MultiZ x, int zI)	// z = ASin(x)
		// ASin(S) = ATan2(S, SqRt(1 - Sq(S)), but if Sq(S) > 1 then Error
		{
			if (x.Rs == 0)
				XCCalLst.ASinN(x, zI);
			else
			{
				int tI;
				Common.InitNAdd(1, "ASinN-tI", out tI);
				if (Common.nList[zI].Rs != x.Rs || Common.nList[zI].Cs != x.Cs)
					MatrixN(x.Rs, x.Cs, zI);
				for (int r = 0; r < x.Rs; r++)
					for (int c = 0; c < x.Cs; c++)
					{
						XCCalLst.ASinN(x.E[r, c], tI);
						if (MultiID.muErr)
						{
							Common.LDelete(tI);
							return;
						}
						Calc.SetToNM(Common.nList[tI], Common.nList[zI].E[r, c]);
					}
				Common.LDelete(tI);
			}
		} // ASinN

		public static void RASinN(int zI) // z = ASin(z)
		{
			ASinN(Common.nList[zI], zI);
		}

		public static void ACosN(MultiZ x, int zI)	// z = ACos(x)
		// ACos(C) = ATan2(SqRt(1 - Sq(C), C), but if Sq(C) > 1 then Error
		{
			if (x.Rs == 0)
				XCCalLst.ACosN(x, zI);
			else
			{
				int tI;
				Common.InitNAdd(1, "ACosN-tI", out tI);
				if (Common.nList[zI].Rs != x.Rs || Common.nList[zI].Cs != x.Cs)
					MatrixN(x.Rs, x.Cs, zI);
				for (int r = 0; r < x.Rs; r++)
					for (int c = 0; c < x.Cs; c++)
					{
						XCCalLst.ACosN(x.E[r, c], tI);
						if (MultiID.muErr)
						{
							Common.LDelete(tI);
							return;
						}
						Calc.SetToNM(Common.nList[tI], Common.nList[zI].E[r, c]);
					}
				Common.LDelete(tI);
			}
		} // ACosN

		public static void RACosN(int zI) // z = ACos(z)
		{
			ACosN(Common.nList[zI], zI);
		}

		public static void ExpLN(MultiZ x, bool bL1, int zI)
		// z = (e^x) - 1, accurate for small x, (- 1 only if bL1 is true)
		// n = Math.Floor(x / Ln(10))
		// g = Frac(x / Ln(10)) * Ln(10)
		// answer = Exp(g) * 10^n
		{
			if (x.Rs == 0)
				XCCalLst.ExpLN(x, bL1, zI);
			else
			{
				int tI;
				Common.InitNAdd(1, "ExpLN-tI", out tI);
				if (Common.nList[zI].Rs != x.Rs || Common.nList[zI].Cs != x.Cs)
					MatrixN(x.Rs, x.Cs, zI);
				for (int r = 0; r < x.Rs; r++)
					for (int c = 0; c < x.Cs; c++)
					{
						XCCalLst.ExpLN(x.E[r, c], bL1, tI);
						if (MultiID.muErr)
						{
							Common.LDelete(tI);
							return;
						}
						Calc.SetToNM(Common.nList[tI], Common.nList[zI].E[r, c]);
					}
				Common.LDelete(tI);
			}
		} // ExpLN

		public static void RExpLN(int zI) // z = (e^z) - 1, accurate for small x
		{
			ExpLN(Common.nList[zI], true, zI);
		}

		public static void ExpN(MultiZ x, int zI)	// z = e^x
		{
			ExpLN(x, false, zI);
		}

		public static void RExpN(int zI) // z = e^z
		{
			ExpN(Common.nList[zI], zI);
		}

		public static void LnLN(MultiZ y, bool bL1, int zI)
		// z = Ln(y + 1), accurate for small y
		// x = y + 1 = Integer * MultiID.baseM^C
		// Multiply by 10 until MSD of i >= MultiID.baseM / 10, save count
		// change C of x to -N so x < 1, save original C+N
		// answer = Ln(x) + Ln(10) * NT, NT = MultiID.muDMax * C+N - count
		{
			if (y.Rs == 0)
				XCCalLst.LnLN(y, bL1, zI);
			else
			{
				int tI;
				Common.InitNAdd(1, "LnLN-tI", out tI);
				if (Common.nList[zI].Rs != y.Rs || Common.nList[zI].Cs != y.Cs)
					MatrixN(y.Rs, y.Cs, zI);
				for (int r = 0; r < y.Rs; r++)
					for (int c = 0; c < y.Cs; c++)
					{
						XCCalLst.LnLN(y.E[r, c], bL1, tI);
						if (MultiID.muErr)
						{
							Common.LDelete(tI);
							return;
						}
						Calc.SetToNM(Common.nList[tI], Common.nList[zI].E[r, c]);
					}
				Common.LDelete(tI);
			}
		} // LnLN

		public static void RLnLN(int zI) // z = Ln(z + 1), accurate for small z
		{
			LnLN(Common.nList[zI], true, zI);
		}

		public static void LnN(MultiZ x, int zI)	// z = Ln(x)
		{
			LnLN(x, false, zI);
		}

		public static void RLnN(int zI) // z = Ln(z)
		{
			LnN(Common.nList[zI], zI);
		}

		public static void LogN(MultiZ x, int zI)	// z = Log(x)
		// Log(x) = Ln(x) / Ln(10), Log base 10 of x
		{
			if (x.Rs == 0)
				XCCalLst.LogN(x, zI);
			else
			{
				int tI;
				Common.InitNAdd(1, "LogN-tI", out tI);
				if (Common.nList[zI].Rs != x.Rs || Common.nList[zI].Cs != x.Cs)
					MatrixN(x.Rs, x.Cs, zI);
				for (int r = 0; r < x.Rs; r++)
					for (int c = 0; c < x.Cs; c++)
					{
						XCCalLst.LogN(x.E[r, c], tI);
						if (MultiID.muErr)
						{
							Common.LDelete(tI);
							return;
						}
						Calc.SetToNM(Common.nList[tI], Common.nList[zI].E[r, c]);
					}
				Common.LDelete(tI);
			}
		}

		public static void RLogN(int zI) // z = Log(z)
		{
			LogN(Common.nList[zI], zI);
		}

		public static void SinHN(MultiZ x, int zI)	// z = SinH(x)
		// for (x >= 0.1 SinH(x) = (y - 1/y) / 2, where y = Exp(x)
		// for (x <  0.1 SinH(x) = y / (2 * SqRt(y + 1)), where y = Exp(2 * x) - 1
		// SinH(-x) = -SinH(x)
		{
			if (x.Rs == 0)
				XCCalLst.SinHN(x, zI);
			else
			{
				int tI;
				Common.InitNAdd(1, "SinHN-tI", out tI);
				if (Common.nList[zI].Rs != x.Rs || Common.nList[zI].Cs != x.Cs)
					MatrixN(x.Rs, x.Cs, zI);
				for (int r = 0; r < x.Rs; r++)
					for (int c = 0; c < x.Cs; c++)
					{
						XCCalLst.SinHN(x.E[r, c], tI);
						if (MultiID.muErr)
						{
							Common.LDelete(tI);
							return;
						}
						Calc.SetToNM(Common.nList[tI], Common.nList[zI].E[r, c]);
					}
				Common.LDelete(tI);
			}
		} // SinHN

		public static void RSinHN(int zI) // z = SinH(z)
		{
			SinHN(Common.nList[zI], zI);
		}

		public static void RiN(MultiZ x, int zI)	// z = Riemann prime counting function of x
		// Ri(x) = 1 + sum{n = 1, ...} ( Ln(x)^n / (n*n!*Zeta(n+1)) )
		{
			if (x.Rs == 0)
				XPCalLst.RiN(x, zI);
			else
			{
				int tI;
				Common.InitNAdd(1, "RiN-tI", out tI);
				if (Common.nList[zI].Rs != x.Rs || Common.nList[zI].Cs != x.Cs)
					MatrixN(x.Rs, x.Cs, zI);
				for (int r = 0; r < x.Rs; r++)
					for (int c = 0; c < x.Cs; c++)
					{
						XPCalLst.RiN(x.E[r, c], tI);
						if (MultiID.muErr)
						{
							Common.LDelete(tI);
							return;
						}
						Calc.SetToNM(Common.nList[tI], Common.nList[zI].E[r, c]);
					}
				Common.LDelete(tI);
			}
		} // RiN

		public static void LiN(MultiZ x, int zI)	// z = Logarithmic integral of x
		// Li(x) = Ei(Ln(x))
		{
			if (x.Rs == 0)
				XCCalLst.LiN(x, zI);
			else
			{
				int tI;
				Common.InitNAdd(1, "LiN-tI", out tI);
				if (Common.nList[zI].Rs != x.Rs || Common.nList[zI].Cs != x.Cs)
					MatrixN(x.Rs, x.Cs, zI);
				for (int r = 0; r < x.Rs; r++)
					for (int c = 0; c < x.Cs; c++)
					{
						XCCalLst.LiN(x.E[r, c], tI);
						if (MultiID.muErr)
						{
							Common.LDelete(tI);
							return;
						}
						Calc.SetToNM(Common.nList[tI], Common.nList[zI].E[r, c]);
					}
				Common.LDelete(tI);
			}
		} // LiN

		public static void EiN(MultiZ x, int zI)	// z = entire exponential integral of x
		// Ei(x) = -Ei(-x)
		{
			if (x.Rs == 0)
				XCCalLst.EiN(x, zI);
			else
			{
				int tI;
				Common.InitNAdd(1, "EiN-tI", out tI);
				if (Common.nList[zI].Rs != x.Rs || Common.nList[zI].Cs != x.Cs)
					MatrixN(x.Rs, x.Cs, zI);
				for (int r = 0; r < x.Rs; r++)
					for (int c = 0; c < x.Cs; c++)
					{
						XCCalLst.EiN(x.E[r, c], tI);
						if (MultiID.muErr)
						{
							Common.LDelete(tI);
							return;
						}
						Calc.SetToNM(Common.nList[tI], Common.nList[zI].E[r, c]);
					}
				Common.LDelete(tI);
			}
		} // EiN

		public static void E1N(MultiZ x, int zI)	// z = exponential integral 1 of x
		// E1(x) = integral{x to inf}(e^-t / t)dt = -Ei(-x)
		{
			if (x.Rs == 0)
				XCCalLst.E1N(x, zI);
			else
			{
				int tI;
				Common.InitNAdd(1, "E1N-tI", out tI);
				if (Common.nList[zI].Rs != x.Rs || Common.nList[zI].Cs != x.Cs)
					MatrixN(x.Rs, x.Cs, zI);
				for (int r = 0; r < x.Rs; r++)
					for (int c = 0; c < x.Cs; c++)
					{
						XCCalLst.E1N(x.E[r, c], tI);
						if (MultiID.muErr)
						{
							Common.LDelete(tI);
							return;
						}
						Calc.SetToNM(Common.nList[tI], Common.nList[zI].E[r, c]);
					}
				Common.LDelete(tI);
			}
		} // E1N

		public static void DilogN(MultiZ x, int zI)	// z = Dilog(x)
		// Dilog(x) = Polylog(2, x) = x * Lerch(x, 2, 1)
		{
			if (x.Rs == 0)
				PolylogN(MultiZD.mTwo, x, zI);
			else
			{
				int tI;
				Common.InitNAdd(1, "DilogN-tI", out tI);
				MatrixN(x.Rs, x.Cs, tI);
				for (int r = 0; r < x.Rs; r++)
					for (int c = 0; c < x.Cs; c++)
					{
						Calc.SetToNM(MultiCD.cTwo, Common.nList[tI].E[r, c]);
					}
				PolylogN(Common.nList[tI], x, zI);
				Common.LDelete(tI);
			}
		} // DilogN

		public static void PolylogN(MultiZ x, MultiZ y, int zI)	// z = Polylog(x, y)
		// Polylog(s, x) = x * Lerch(x, s, 1)
		{
			if (x.Rs == 0 && y.Rs == 0)
				XCCalLst.PolylogN(x, y, zI);
			else if (x.Rs != y.Rs || x.Cs != y.Cs)
				MultiZD.ErrIncompatible();
			else
			{
				int tI;
				Common.InitNAdd(1, "PolylogN-tI", out tI);
				if (Common.nList[zI].Rs != x.Rs || Common.nList[zI].Cs != x.Cs)
					MatrixN(x.Rs, x.Cs, zI);
				for (int r = 0; r < x.Rs; r++)
					for (int c = 0; c < x.Cs; c++)
					{
						XCCalLst.PolylogN(x.E[r, c], y.E[r, c], tI);
						if (MultiID.muErr)
						{
							Common.LDelete(tI);
							return;
						}
						Calc.SetToNM(Common.nList[tI], Common.nList[zI].E[r, c]);
					}
				Common.LDelete(tI);
			}		
		} // PolylogN

		public static void ZetaHN(MultiZ s, MultiZ a, int zI)	// z = ZetaH(s, a) Hurwitz Zeta Function
		// ZetaH(s, a) = Lerch(1, s, a)
		{
			if ((s.Rs == 0 && a.Rs == 0) || s.Rs != a.Rs || s.Cs != a.Cs)
				Lerch(MultiZD.mOne, s, a, false, zI);
			else
			{
				int tI;
				Common.InitNAdd(1, "ZetaHN-tI", out tI);
				MatrixN(s.Rs, s.Cs, tI);
				for (int r = 0; r < s.Rs; r++)
					for (int c = 0; c < s.Cs; c++)
					{
						Calc.SetToNM(MultiCD.cOne, Common.nList[tI].E[r, c]);
					}
				Lerch(Common.nList[tI], s, a, false, zI);
				Common.LDelete(tI);
			}
		} // ZetaHN

		public static void CosHN(MultiZ x, int zI)	// z = CosH(x)
		// CosH(x) = (y + 1/y) / 2, where y = Exp(|x|)
		{
			if (x.Rs == 0)
				XCCalLst.CosHN(x, zI);
			else
			{
				int tI;
				Common.InitNAdd(1, "CosHN-tI", out tI);
				if (Common.nList[zI].Rs != x.Rs || Common.nList[zI].Cs != x.Cs)
					MatrixN(x.Rs, x.Cs, zI);
				for (int r = 0; r < x.Rs; r++)
					for (int c = 0; c < x.Cs; c++)
					{
						XCCalLst.CosHN(x.E[r, c], tI);
						if (MultiID.muErr)
						{
							Common.LDelete(tI);
							return;
						}
						Calc.SetToNM(Common.nList[tI], Common.nList[zI].E[r, c]);
					}
				Common.LDelete(tI);
			}
		} // CosHN

		public static void RCosHN(int zI) // z = CosH(z)
		{
			CosHN(Common.nList[zI], zI);
		}

		public static void TanHN(MultiZ x, int zI)	// z = TanH(x)
		// TanH(x) = y / (y + 2), where y = Exp(2 * x) - 1
		// TanH(-x) = -TanH(x)
		{
			if (x.Rs == 0)
				XCCalLst.TanHN(x, zI);
			else
			{
				int tI;
				Common.InitNAdd(1, "TanHN-tI", out tI);
				if (Common.nList[zI].Rs != x.Rs || Common.nList[zI].Cs != x.Cs)
					MatrixN(x.Rs, x.Cs, zI);
				for (int r = 0; r < x.Rs; r++)
					for (int c = 0; c < x.Cs; c++)
					{
						XCCalLst.TanHN(x.E[r, c], tI);
						if (MultiID.muErr)
						{
							Common.LDelete(tI);
							return;
						}
						Calc.SetToNM(Common.nList[tI], Common.nList[zI].E[r, c]);
					}
				Common.LDelete(tI);
			}
		} // TanHN

		public static void RTanHN(int zI) // z = TanH(z)
		{
			TanHN(Common.nList[zI], zI);
		}

		public static void ASinHN(MultiZ x, int zI)	// z = ASinH(x)
		// for (x >= 0.1 ASinH(x) = Ln(x + SqRt(1 + Sq(x)))
		// for (x <  0.1 ASinH(x) = LnL(x + Sq(x) / SqRt(1 + Sq(x)))
		// ASinH(-x) = -ASinH(x)
		{
			if (x.Rs == 0)
				XCCalLst.ASinHN(x, zI);
			else
			{
				int tI;
				Common.InitNAdd(1, "ASinHN-tI", out tI);
				if (Common.nList[zI].Rs != x.Rs || Common.nList[zI].Cs != x.Cs)
					MatrixN(x.Rs, x.Cs, zI);
				for (int r = 0; r < x.Rs; r++)
					for (int c = 0; c < x.Cs; c++)
					{
						XCCalLst.ASinHN(x.E[r, c], tI);
						if (MultiID.muErr)
						{
							Common.LDelete(tI);
							return;
						}
						Calc.SetToNM(Common.nList[tI], Common.nList[zI].E[r, c]);
					}
				Common.LDelete(tI);
			}
		} // ASinHN

		public static void RASinHN(int zI) // z = ASinH(z)
		{
			ASinHN(Common.nList[zI], zI);
		}

		public static void ACosHN(MultiZ x, int zI)	// z = ACosH(x)
		// ACosH(x) = Ln(x + SqRt(Sq(x) - 1))
		// error if x < 1
		{
			if (x.Rs == 0)
				XCCalLst.ACosHN(x, zI);
			else
			{
				int tI;
				Common.InitNAdd(1, "ACosHN-tI", out tI);
				if (Common.nList[zI].Rs != x.Rs || Common.nList[zI].Cs != x.Cs)
					MatrixN(x.Rs, x.Cs, zI);
				for (int r = 0; r < x.Rs; r++)
					for (int c = 0; c < x.Cs; c++)
					{
						XCCalLst.ACosHN(x.E[r, c], tI);
						if (MultiID.muErr)
						{
							Common.LDelete(tI);
							return;
						}
						Calc.SetToNM(Common.nList[tI], Common.nList[zI].E[r, c]);
					}
				Common.LDelete(tI);
			}
		} // ACosHN

		public static void RACosHN(int zI) // z = ACosH(z)
		{
			ACosHN(Common.nList[zI], zI);
		}

		public static void RACotN(int zI) // z = ACot(z)
		// ACot(z) = ATan(1/z), error if z = 0
		{
			int rs = Common.nList[zI].Rs;
			if (rs == 0)
				XCCalLst.RACotN(zI);
			else
			{
				int cs = Common.nList[zI].Cs;
				int tI;
				Common.InitNAdd(1, "RACotN-tI", out tI);
				for (int r = 0; r < rs; r++)
					for (int c = 0; c < cs; c++)
					{
						Calc.SetToNM(Common.nList[zI].E[r, c], Common.nList[tI]);
						XCCalLst.RACotN(tI);
						if (MultiID.muErr)
						{
							Common.LDelete(tI);
							return;
						}
						Calc.SetToNM(Common.nList[tI], Common.nList[zI].E[r, c]);
					}
				Common.LDelete(tI);
			}
		} // RACotN

		public static void RACotHN(int zI) // z = ACotH(z)
		// ACotH(z) = ATanH(1/z), error if |z| <= 1
		{
			int rs = Common.nList[zI].Rs;
			if (rs == 0)
				XCCalLst.RACotHN(zI);
			else
			{
				int cs = Common.nList[zI].Cs;
				int tI;
				Common.InitNAdd(1, "RACotHN-tI", out tI);
				for (int r = 0; r < rs; r++)
					for (int c = 0; c < cs; c++)
					{
						Calc.SetToNM(Common.nList[zI].E[r, c], Common.nList[tI]);
						XCCalLst.RACotHN(tI);
						if (MultiID.muErr)
						{
							Common.LDelete(tI);
							return;
						}
						Calc.SetToNM(Common.nList[tI], Common.nList[zI].E[r, c]);
					}
				Common.LDelete(tI);
			}
		} // RACotHN

		public static void RACscN(int zI) // z = ACsc(z)
		// ACsc(z) = ASin(1/z), error if z = 0
		{
			int rs = Common.nList[zI].Rs;
			if (rs == 0)
				XCCalLst.RACscN(zI);
			else
			{
				int cs = Common.nList[zI].Cs;
				int tI;
				Common.InitNAdd(1, "RACscN-tI", out tI);
				for (int r = 0; r < rs; r++)
					for (int c = 0; c < cs; c++)
					{
						Calc.SetToNM(Common.nList[zI].E[r, c], Common.nList[tI]);
						XCCalLst.RACscN(tI);
						if (MultiID.muErr)
						{
							Common.LDelete(tI);
							return;
						}
						Calc.SetToNM(Common.nList[tI], Common.nList[zI].E[r, c]);
					}
				Common.LDelete(tI);
			}
		} // RACscN

		public static void RACscHN(int zI) // z = ACscH(z)
		// ACscH(z) = ASinH(1/z), error if z = 0
		{
			int rs = Common.nList[zI].Rs;
			if (rs == 0)
				XCCalLst.RACscHN(zI);
			else
			{
				int cs = Common.nList[zI].Cs;
				int tI;
				Common.InitNAdd(1, "RACscHN-tI", out tI);
				for (int r = 0; r < rs; r++)
					for (int c = 0; c < cs; c++)
					{
						Calc.SetToNM(Common.nList[zI].E[r, c], Common.nList[tI]);
						XCCalLst.RACscHN(tI);
						if (MultiID.muErr)
						{
							Common.LDelete(tI);
							return;
						}
						Calc.SetToNM(Common.nList[tI], Common.nList[zI].E[r, c]);
					}
				Common.LDelete(tI);
			}
		} // RACscHN

		public static void RASecN(int zI) // z = ASec(z)
		// ASec(z) = ACos(1/z), error if z = 0
		{
			int rs = Common.nList[zI].Rs;
			if (rs == 0)
				XCCalLst.RASecN(zI);
			else
			{
				int cs = Common.nList[zI].Cs;
				int tI;
				Common.InitNAdd(1, "RASecN-tI", out tI);
				for (int r = 0; r < rs; r++)
					for (int c = 0; c < cs; c++)
					{
						Calc.SetToNM(Common.nList[zI].E[r, c], Common.nList[tI]);
						XCCalLst.RASecN(tI);
						if (MultiID.muErr)
						{
							Common.LDelete(tI);
							return;
						}
						Calc.SetToNM(Common.nList[tI], Common.nList[zI].E[r, c]);
					}
				Common.LDelete(tI);
			}
		} // RASecN

		public static void RASecHN(int zI) // z = ASecH(z)
		// ASecH(z) = ACosH(1/z), error if z <= 0 or z > 1
		{
			int rs = Common.nList[zI].Rs;
			if (rs == 0)
				XCCalLst.RASecHN(zI);
			else
			{
				int cs = Common.nList[zI].Cs;
				int tI;
				Common.InitNAdd(1, "RASecHN-tI", out tI);
				for (int r = 0; r < rs; r++)
					for (int c = 0; c < cs; c++)
					{
						Calc.SetToNM(Common.nList[zI].E[r, c], Common.nList[tI]);
						XCCalLst.RASecHN(tI);
						if (MultiID.muErr)
						{
							Common.LDelete(tI);
							return;
						}
						Calc.SetToNM(Common.nList[tI], Common.nList[zI].E[r, c]);
					}
				Common.LDelete(tI);
			}
		} // RASecHN

		public static void RCotN(int zI) // z = Cot(z)
		// Cot(z) = 1/Tan(z), error if Tan(z) = 0
		{
			int rs = Common.nList[zI].Rs;
			if (rs == 0)
				XCCalLst.RCotN(zI);
			else
			{
				int cs = Common.nList[zI].Cs;
				int tI;
				Common.InitNAdd(1, "RCotN-tI", out tI);
				for (int r = 0; r < rs; r++)
					for (int c = 0; c < cs; c++)
					{
						Calc.SetToNM(Common.nList[zI].E[r, c], Common.nList[tI]);
						XCCalLst.RCotN(tI);
						if (MultiID.muErr)
						{
							Common.LDelete(tI);
							return;
						}
						Calc.SetToNM(Common.nList[tI], Common.nList[zI].E[r, c]);
					}
				Common.LDelete(tI);
			}
		} // RCotN

		public static void RCotHN(int zI) // z = CotH(z)
		// CotH(z) = 1/TanH(z), error if TanH(z) = 0
		{
			int rs = Common.nList[zI].Rs;
			if (rs == 0)
				XCCalLst.RCotHN(zI);
			else
			{
				int cs = Common.nList[zI].Cs;
				int tI;
				Common.InitNAdd(1, "RCotHN-tI", out tI);
				for (int r = 0; r < rs; r++)
					for (int c = 0; c < cs; c++)
					{
						Calc.SetToNM(Common.nList[zI].E[r, c], Common.nList[tI]);
						XCCalLst.RCotHN(tI);
						if (MultiID.muErr)
						{
							Common.LDelete(tI);
							return;
						}
						Calc.SetToNM(Common.nList[tI], Common.nList[zI].E[r, c]);
					}
				Common.LDelete(tI);
			}
		} // RCotHN

		public static void RCscN(int zI) // z = Csc(z)
		// Csc(z) = 1/Sin(z), error if Sin(z) = 0
		{
			int rs = Common.nList[zI].Rs;
			if (rs == 0)
				XCCalLst.RCscN(zI);
			else
			{
				int cs = Common.nList[zI].Cs;
				int tI;
				Common.InitNAdd(1, "RCscN-tI", out tI);
				for (int r = 0; r < rs; r++)
					for (int c = 0; c < cs; c++)
					{
						Calc.SetToNM(Common.nList[zI].E[r, c], Common.nList[tI]);
						XCCalLst.RCscN(tI);
						if (MultiID.muErr)
						{
							Common.LDelete(tI);
							return;
						}
						Calc.SetToNM(Common.nList[tI], Common.nList[zI].E[r, c]);
					}
				Common.LDelete(tI);
			}
		} // RCscN

		public static void RCscHN(int zI) // z = CscH(z)
		// CscH(z) = 1/SinH(z), error if SinH(z) = 0
		{
			int rs = Common.nList[zI].Rs;
			if (rs == 0)
				XCCalLst.RCscHN(zI);
			else
			{
				int cs = Common.nList[zI].Cs;
				int tI;
				Common.InitNAdd(1, "RCscHN-tI", out tI);
				for (int r = 0; r < rs; r++)
					for (int c = 0; c < cs; c++)
					{
						Calc.SetToNM(Common.nList[zI].E[r, c], Common.nList[tI]);
						XCCalLst.RCscHN(tI);
						if (MultiID.muErr)
						{
							Common.LDelete(tI);
							return;
						}
						Calc.SetToNM(Common.nList[tI], Common.nList[zI].E[r, c]);
					}
				Common.LDelete(tI);
			}
		} // RCscHN

		public static void RSecN(int zI) // z = Sec(z)
		// Sec(z) = 1/Cos(z), error if Cos(z) = 0
		{
			int rs = Common.nList[zI].Rs;
			if (rs == 0)
				XCCalLst.RSecN(zI);
			else
			{
				int cs = Common.nList[zI].Cs;
				int tI;
				Common.InitNAdd(1, "RSecN-tI", out tI);
				for (int r = 0; r < rs; r++)
					for (int c = 0; c < cs; c++)
					{
						Calc.SetToNM(Common.nList[zI].E[r, c], Common.nList[tI]);
						XCCalLst.RSecN(tI);
						if (MultiID.muErr)
						{
							Common.LDelete(tI);
							return;
						}
						Calc.SetToNM(Common.nList[tI], Common.nList[zI].E[r, c]);
					}
				Common.LDelete(tI);
			}
		} // RSecN

		public static void RSecHN(int zI) // z = SecH(z)
		// SecH(z) = 1/CosH(z)
		{
			int rs = Common.nList[zI].Rs;
			if (rs == 0)
				XCCalLst.RSecHN(zI);
			else
			{
				int cs = Common.nList[zI].Cs;
				int tI;
				Common.InitNAdd(1, "RSecHN-tI", out tI);
				for (int r = 0; r < rs; r++)
					for (int c = 0; c < cs; c++)
					{
						Calc.SetToNM(Common.nList[zI].E[r, c], Common.nList[tI]);
						XCCalLst.RSecHN(tI);
						if (MultiID.muErr)
						{
							Common.LDelete(tI);
							return;
						}
						Calc.SetToNM(Common.nList[tI], Common.nList[zI].E[r, c]);
					}
				Common.LDelete(tI);
			}
		} // RSecHN

		public static void ATanHN(MultiZ x, int zI)	// z = ATanH(x)
		// ATanH(x) = LnL(2 * x / (1 - x))
		// ATanH(-x) = -ATanH(x)
		// error if |x| >= 1
		{
			if (x.Rs == 0)
				XCCalLst.ATanHN(x, zI);
			else
			{
				int tI;
				Common.InitNAdd(1, "ATanHN-tI", out tI);
				if (Common.nList[zI].Rs != x.Rs || Common.nList[zI].Cs != x.Cs)
					MatrixN(x.Rs, x.Cs, zI);
				for (int r = 0; r < x.Rs; r++)
					for (int c = 0; c < x.Cs; c++)
					{
						XCCalLst.ATanHN(x.E[r, c], tI);
						if (MultiID.muErr)
						{
							Common.LDelete(tI);
							return;
						}
						Calc.SetToNM(Common.nList[tI], Common.nList[zI].E[r, c]);
					}
				Common.LDelete(tI);
			}
		} // ATanHN

		public static void RATanHN(int zI) // z = ATanH(z)
		{
			ATanHN(Common.nList[zI], zI);
		}

		public static void CeilingN(MultiZ x, int zI)	// z = smallest integer >= z
		{
			if (x.Rs == 0)
				XCCalLst.CeilingN(x, zI);
			else
			{
				int tI;
				Common.InitNAdd(1, "CeilingN-tI", out tI);
				if (Common.nList[zI].Rs != x.Rs || Common.nList[zI].Cs != x.Cs)
					MatrixN(x.Rs, x.Cs, zI);
				for (int r = 0; r < x.Rs; r++)
					for (int c = 0; c < x.Cs; c++)
					{
						XCCalLst.CeilingN(x.E[r, c], tI);
						if (MultiID.muErr)
						{
							Common.LDelete(tI);
							return;
						}
						Calc.SetToNM(Common.nList[tI], Common.nList[zI].E[r, c]);
					}
				Common.LDelete(tI);
			}
		} // CeilingN

		public static void RCeilingN(int zI) // z = smallest integer >= z
		{
			CeilingN(Common.nList[zI], zI);
		}

		public static void FloorN(MultiZ x, int zI)	// z = largest integer <= z
		{
			if (x.Rs == 0)
				XCCalLst.FloorN(x, zI);
			else
			{
				int tI;
				Common.InitNAdd(1, "FloorN-tI", out tI);
				if (Common.nList[zI].Rs != x.Rs || Common.nList[zI].Cs != x.Cs)
					MatrixN(x.Rs, x.Cs, zI);
				for (int r = 0; r < x.Rs; r++)
					for (int c = 0; c < x.Cs; c++)
					{
						XCCalLst.FloorN(x.E[r, c], tI);
						if (MultiID.muErr)
						{
							Common.LDelete(tI);
							return;
						}
						Calc.SetToNM(Common.nList[tI], Common.nList[zI].E[r, c]);
					}
				Common.LDelete(tI);
			}
		} // FloorN

		public static void RFloorN(int zI) // z = largest integer <= z
		{
			FloorN(Common.nList[zI], zI);
		}

		public static void RoundN(MultiZ x, int zI)	// z = integer nearest to x
		{
			if (x.Rs == 0)
				XCCalLst.RoundN(x, zI);
			else
			{
				int tI;
				Common.InitNAdd(1, "RoundN-tI", out tI);
				if (Common.nList[zI].Rs != x.Rs || Common.nList[zI].Cs != x.Cs)
					MatrixN(x.Rs, x.Cs, zI);
				for (int r = 0; r < x.Rs; r++)
					for (int c = 0; c < x.Cs; c++)
					{
						XCCalLst.RoundN(x.E[r, c], tI);
						if (MultiID.muErr)
						{
							Common.LDelete(tI);
							return;
						}
						Calc.SetToNM(Common.nList[tI], Common.nList[zI].E[r, c]);
					}
				Common.LDelete(tI);
			}
		} // RoundN

		public static void RRoundN(int zI) // z = integer nearest to z
		{
			RoundN(Common.nList[zI], zI);
		}

		public static void IntN(MultiZ x, int zI)	// z = Integer part of x
		{
			if (x.Rs == 0)
				XCCalLst.IntN(x, zI);
			else
			{
				int tI;
				Common.InitNAdd(1, "IntN-tI", out tI);
				if (Common.nList[zI].Rs != x.Rs || Common.nList[zI].Cs != x.Cs)
					MatrixN(x.Rs, x.Cs, zI);
				for (int r = 0; r < x.Rs; r++)
					for (int c = 0; c < x.Cs; c++)
					{
						XCCalLst.IntN(x.E[r, c], tI);
						if (MultiID.muErr)
						{
							Common.LDelete(tI);
							return;
						}
						Calc.SetToNM(Common.nList[tI], Common.nList[zI].E[r, c]);
					}
				Common.LDelete(tI);
			}
		} // IntN

		public static void RIntN(int zI) // z = Integer part of z
		{
			IntN(Common.nList[zI], zI);
		}

		public static void FracN(MultiZ x, int zI)	// z = Fractional part of x
		{
			if (x.Rs == 0)
				XCCalLst.FracN(x, zI);
			else
			{
				int tI;
				Common.InitNAdd(1, "FracN-tI", out tI);
				if (Common.nList[zI].Rs != x.Rs || Common.nList[zI].Cs != x.Cs)
					MatrixN(x.Rs, x.Cs, zI);
				for (int r = 0; r < x.Rs; r++)
					for (int c = 0; c < x.Cs; c++)
					{
						XCCalLst.FracN(x.E[r, c], tI);
						if (MultiID.muErr)
						{
							Common.LDelete(tI);
							return;
						}
						Calc.SetToNM(Common.nList[tI], Common.nList[zI].E[r, c]);
					}
				Common.LDelete(tI);
			}
		} // FracN

		public static void RFracN(int zI) // z = Fractional part of z
		{
			FracN(Common.nList[zI], zI);
		}

		public static void RLopN(int zI) // z = z with one less least significant "digit"
		{
			MultiZD.RLopZ(Common.nList[zI]);
			Calc.NormN(zI);
		}

		public static void ToDegN(MultiZ x, int zI) // z = x converted Rad to Degrees
		{
			if (x.Rs == 0)
				XCCalLst.ToDegN(x, zI);
			else
			{
				int tI;
				Common.InitNAdd(1, "ToDegN-tI", out tI);
				if (Common.nList[zI].Rs != x.Rs || Common.nList[zI].Cs != x.Cs)
					MatrixN(x.Rs, x.Cs, zI);
				for (int r = 0; r < x.Rs; r++)
					for (int c = 0; c < x.Cs; c++)
					{
						XCCalLst.ToDegN(x.E[r, c], tI);
						if (MultiID.muErr)
						{
							Common.LDelete(tI);
							return;
						}
						Calc.SetToNM(Common.nList[tI], Common.nList[zI].E[r, c]);
					}
				Common.LDelete(tI);
			}
		} // ToDegN

		public static void RToDegN(int zI) // z = z converted Radians to Degrees
		{
			ToDegN(Common.nList[zI], zI);
		}

		public static void ToRadN(MultiZ x, int zI)	// z = x converted Deg to Radians
		{
			if (x.Rs == 0)
				XCCalLst.ToRadN(x, zI);
			else
			{
				int tI;
				Common.InitNAdd(1, "ToRadN-tI", out tI);
				if (Common.nList[zI].Rs != x.Rs || Common.nList[zI].Cs != x.Cs)
					MatrixN(x.Rs, x.Cs, zI);
				for (int r = 0; r < x.Rs; r++)
					for (int c = 0; c < x.Cs; c++)
					{
						XCCalLst.ToRadN(x.E[r, c], tI);
						if (MultiID.muErr)
						{
							Common.LDelete(tI);
							return;
						}
						Calc.SetToNM(Common.nList[tI], Common.nList[zI].E[r, c]);
					}
				Common.LDelete(tI);
			}
		} // ToRadN

		public static void RToRadN(int zI) // z = z converted Deg to Radians
		{
			ToRadN(Common.nList[zI], zI);
		}

		public static void CalSFTn(int zI) // set number of digits in truncate in display
		{
			if (Common.nList[zI].Rs != 0)
				MultiZD.ErrNotScalar();
			else
				XCCalLst.CalSFTn(zI);
		}

		public static void NormC(MultiZ x, MultiZ m, MultiZ s, int zI)
		// z = Cumulative distribution function (cdf) of normal distribution
		// m = mean, s = standard deviation
		// x, m, s may have the same address in memory as z
		// x, m, s are not changed unless it shares memory with z
		// sets muErr true if overflow or error
		{
			if (x.Rs == 0 && m.Rs == 0 && s.Rs == 0)
				XCCalLst.NormC(x, m, s, zI);
			else if (x.Rs != m.Rs || x.Rs != s.Rs || x.Cs != m.Cs || x.Cs != s.Cs)
				MultiZD.ErrIncompatible();
			else
			{
				int tI;
				Common.InitNAdd(1, "NormC-tI", out tI);
				if (Common.nList[zI].Rs != x.Rs || Common.nList[zI].Cs != x.Cs)
					MatrixN(x.Rs, x.Cs, zI);
				for (int r = 0; r < x.Rs; r++)
					for (int c = 0; c < x.Cs; c++)
					{
						XCCalLst.NormC(x.E[r, c], m.E[r, c], s.E[r, c],tI);
						if (MultiID.muErr)
						{
							Common.LDelete(tI);
							return;
						}
						Calc.SetToNM(Common.nList[tI], Common.nList[zI].E[r, c]);
					}
				Common.LDelete(tI);
			}
		} // NormC

		public static void NormS(int zI)
		// z = Cumulative distribution function (cdf) of standard normal distribution
		// mean = 0, standard deviation = 1
		// sets muErr true if overflow or error
		{
			if (Common.nList[zI].Rs == 0)
				NormC(Common.nList[zI], MultiZD.mZero, MultiZD.mOne, zI);
			else
			{
				int t0I;
				int t1I;
				Common.InitNAdd(1, "NormS-t0I", out t0I);
				Common.InitNAdd(1, "NormS-t1I", out t1I);
				MatrixN(Common.nList[zI].Rs, Common.nList[zI].Cs, t0I);
				MatrixN(Common.nList[zI].Rs, Common.nList[zI].Cs, t1I);
				for (int r = 0; r < Common.nList[zI].Rs; r++)
					for (int c = 0; c < Common.nList[zI].Cs; c++)
					{
						Calc.SetToNM(MultiCD.cZero, Common.nList[t0I].E[r, c]);
						Calc.SetToNM(MultiCD.cOne, Common.nList[t1I].E[r, c]);
					}
				NormC(Common.nList[zI], Common.nList[t0I], Common.nList[t1I], zI);
				Common.LDelete(t1I);
				Common.LDelete(t0I);
			}
		} // NormS

		public static void QToMN(MultiZ x, int zI)	// z = rotation matrix from rotation quaternion
		{
			bool saveExp = MultiFD.expanded;
			int saveFMC = Calc.fMC;
			int saveMMax = MultiID.mMax;
			XPCalLst.Expand();
			MultiZ z = new MultiZ(Calc.fMC, 3, 3);
			MultiZD.QToZ(x, z);
			Calc.SetToN(z, zI);
			MultiFD.expanded = saveExp;
			Calc.fMC = saveFMC;
			MultiID.mMax = saveMMax;
			Calc.NormN(zI);
		} // QToMN

		public static void MToQN(MultiZ x, int zI)	// z = rotation quaternion from rotation matrix
		{
			bool saveExp = MultiFD.expanded;
			int saveFMC = Calc.fMC;
			int saveMMax = MultiID.mMax;
			XPCalLst.Expand();
			MultiZ z = new MultiZ(Calc.fMC, 4, 1);
			MultiZD.MToQ(x, z);
			Calc.SetToN(z, zI);
			MultiFD.expanded = saveExp;
			Calc.fMC = saveFMC;
			MultiID.mMax = saveMMax;
			Calc.NormN(zI);
		} // MToQN

		public static void MToRN(MultiZ x, int zI)	// z = rotation vector from rotation matrix
		{
			bool saveExp = MultiFD.expanded;
			int saveFMC = Calc.fMC;
			int saveMMax = MultiID.mMax;
			XPCalLst.Expand();
			MultiZ z = new MultiZ(Calc.fMC, 4, 1);
			MultiZD.MToR(x, z);
			Calc.SetToN(z, zI);
			MultiFD.expanded = saveExp;
			Calc.fMC = saveFMC;
			MultiID.mMax = saveMMax;
			Calc.NormN(zI);
		} // MToRN

		public static void RToMN(MultiZ x, int zI)	// z = rotation matrix from rotation vector
		{
			bool saveExp = MultiFD.expanded;
			int saveFMC = Calc.fMC;
			int saveMMax = MultiID.mMax;
			XPCalLst.Expand();
			MultiZ z = new MultiZ(Calc.fMC, 3, 3);
			MultiZD.RToZ(x, z);
			Calc.SetToN(z, zI);
			MultiFD.expanded = saveExp;
			Calc.fMC = saveFMC;
			MultiID.mMax = saveMMax;
			Calc.NormN(zI);
		} // RToMN

		public static void QToRN(MultiZ x, int zI)	// z = rotation vector from rotation quaternion
		{
			bool saveExp = MultiFD.expanded;
			int saveFMC = Calc.fMC;
			int saveMMax = MultiID.mMax;
			XPCalLst.Expand();
			MultiZ z = new MultiZ(Calc.fMC, 4, 1);
			MultiZD.QToR(x, z);
			Calc.SetToN(z, zI);
			MultiFD.expanded = saveExp;
			Calc.fMC = saveFMC;
			MultiID.mMax = saveMMax;
			Calc.NormN(zI);
		} // QToRN

		public static void RToQN(MultiZ x, int zI)	// z = rotation quaternion from rotation vector
		{
			bool saveExp = MultiFD.expanded;
			int saveFMC = Calc.fMC;
			int saveMMax = MultiID.mMax;
			XPCalLst.Expand();
			MultiZ z = new MultiZ(Calc.fMC, 4, 1);
			MultiZD.RToQ(x, z);
			Calc.SetToN(z, zI);
			MultiFD.expanded = saveExp;
			Calc.fMC = saveFMC;
			MultiID.mMax = saveMMax;
			Calc.NormN(zI);
		} // RToQN

		public static void PrinQN(MultiZ x, int zI)	// z = principal quaternion of x and -x
		{
			MultiZ z = new MultiZ(Calc.fMC, 4, 1);
			MultiZD.PrinQ(x, z);
			Calc.SetToN(z, zI);
			Calc.NormN(zI);
		} // PrinQN

		public static void RotateQN(MultiZ x, MultiZ y, int zI)	// z = quaternion rotation x*y
		{
			bool saveExp = MultiFD.expanded;
			int saveFMC = Calc.fMC;
			int saveMMax = MultiID.mMax;
			XPCalLst.Expand();
			MultiZ z = new MultiZ(Calc.fMC, 3, 1);
			MultiZD.RotateQ(x, y, z);
			Calc.SetToN(z, zI);
			MultiFD.expanded = saveExp;
			Calc.fMC = saveFMC;
			MultiID.mMax = saveMMax;
			Calc.NormN(zI);
		} // RotateQN

		public static void RotateMN(MultiZ x, MultiZ y, int zI)	// z = 3x3 matrix rotation x*y
		{
			bool saveExp = MultiFD.expanded;
			int saveFMC = Calc.fMC;
			int saveMMax = MultiID.mMax;
			XPCalLst.Expand();
			MultiZ z = new MultiZ(Calc.fMC, 3, 1);
			MultiZD.RotateZ(x, y, z);
			Calc.SetToN(z, zI);
			MultiFD.expanded = saveExp;
			Calc.fMC = saveFMC;
			MultiID.mMax = saveMMax;
			Calc.NormN(zI);
		} // RotateMN

		public static void EqQN(MultiZ x, int zI)	// z = quaternion z normalized
		{
			MultiZ z = new MultiZ(Calc.fMC, 4, 1);
			MultiZD.EqQ(x, z);
			Calc.SetToN(z, zI);
			Calc.NormN(zI);
		} // EqQN

		public static void InvQN(MultiZ x, int zI)	// z = quaternion z normalized
		{
			MultiZ z = new MultiZ(Calc.fMC, 4, 1);
			MultiZD.InvQ(x, z);
			Calc.SetToN(z, zI);
			Calc.NormN(zI);
		} // InvQN

		public static void NormQN(MultiZ x, int zI)	// z = quaternion z normalized
		{
			bool saveExp = MultiFD.expanded;
			int saveFMC = Calc.fMC;
			int saveMMax = MultiID.mMax;
			XPCalLst.Expand();
			MultiZ z = new MultiZ(Calc.fMC, 4, 1);
			MultiZD.NormQ(x, z);
			Calc.SetToN(z, zI);
			MultiFD.expanded = saveExp;
			Calc.fMC = saveFMC;
			MultiID.mMax = saveMMax;
			Calc.NormN(zI);
		} // NormQN

		public static void NormRN(MultiZ x, int zI)	// z = rotation vector z normalized
		{
			bool saveExp = MultiFD.expanded;
			int saveFMC = Calc.fMC;
			int saveMMax = MultiID.mMax;
			XPCalLst.Expand();
			MultiZ z = new MultiZ(Calc.fMC, 4, 1);
			MultiZD.NormR(x, z);
			Calc.SetToN(z, zI);
			MultiFD.expanded = saveExp;
			Calc.fMC = saveFMC;
			MultiID.mMax = saveMMax;
			Calc.NormN(zI);
		} // NormRN

		public static void NormVN(MultiZ x, int zI) // z = length of vector x
		{
			bool saveExp = MultiFD.expanded;
			int saveFMC = Calc.fMC;
			int saveMMax = MultiID.mMax;
			XPCalLst.Expand();
			MultiZ z = new MultiZ(Calc.fMC);
			MultiZD.NormV(x, z);
			Calc.SetToN(z, zI);
			MultiFD.expanded = saveExp;
			Calc.fMC = saveFMC;
			MultiID.mMax = saveMMax;
			Calc.NormN(zI);
		} // NormVN

		public static void VanderN(MultiZ x, int zI) // z = Vandermonde's matrix from vector x
		// zij = xi^(j-1)
		{

			if (x.Rs != 1 && x.Cs != 1)
				MultiZD.ErrNotVector();
			else
			{
				bool saveExp = MultiFD.expanded;
				int saveFMC = Calc.fMC;
				int saveMMax = MultiID.mMax;
				XPCalLst.Expand();
				int n = Math.Max(x.Rs, x.Cs);
				MultiZ z = new MultiZ(Calc.fMC, n, n);
				MultiC jC = new MultiC(1, 1);
				if (x.Rs == 1)
					for (int i = 0; i < n; i++)
						for (int j = 0; j < n; j++)
						{
							jC &= j;
							XCCalLst.PowNC(x.E[0, i], jC, z.E[i, j]);
						}
				else
					for (int i = 0; i < n; i++)
						for (int j = 0; j < n; j++)
						{
							jC &= j;
							XCCalLst.PowNC(x.E[i, 0], jC, z.E[i, j]);
						}
				Calc.SetToN(z, zI);
				MultiFD.expanded = saveExp;
				Calc.fMC = saveFMC;
				MultiID.mMax = saveMMax;
				Calc.NormN(zI);
			}
		} // VanderN

		public static void QuadRoot(MultiZ a, MultiZ b, MultiZ c, int zI)
		// z = a root of quadratic equation 0 = a*x^2 + b*x + c, roots also stored in X1 and X2
		// a, b, c may have the same address in memory as z
		// a, b, c are not changed unless it shares memory with z
		// sets muErr true if overflow or error
		{
			if (a.Rs == 0 && b.Rs == 0 && c.Rs == 0)
				XCCalLst.QuadRoot(a, b, c, zI);
			else if (a.Rs != b.Rs || a.Cs != b.Cs || a.Rs != c.Rs || a.Cs != c.Cs)
				MultiZD.ErrIncompatible();
			else
			{
				int tI;
				int x1MI;
				int x2MI;
				Common.InitNAdd(1, "QuadRoot-tI", out tI);
				Common.InitNAdd(1, "QuadRoot-x1MI", out x1MI);
				Common.InitNAdd(1, "QuadRoot-x2MI", out x2MI);
				MatrixN(a.Rs, a.Cs, x1MI);
				MatrixN(a.Rs, a.Cs, x2MI);
				if (Common.nList[zI].Rs != a.Rs || Common.nList[zI].Cs != a.Cs)
					MatrixN(a.Rs, a.Cs, zI);
				int x1I;
				int x2I;
				Common.LFindAdd("X1", out x1I);
				Common.LFindAdd("X2", out x2I);
				for (int r = 0; r < a.Rs; r++)
					for (int ci = 0; ci < a.Cs; ci++)
					{
						XCCalLst.QuadRoot(a.E[r, ci], b.E[r, ci], c.E[r, ci], tI);
						if (MultiID.muErr)
						{
							Common.LDelete(x2MI);
							Common.LDelete(x1MI);
							Common.LDelete(tI);
							return;
						}
						Calc.SetToNM(Common.nList[x1I], Common.nList[x1MI].E[r, ci]);
						Calc.SetToNM(Common.nList[x2I], Common.nList[x2MI].E[r, ci]);
					}
				Calc.SetToII(x1MI, x1I);
				Calc.SetToII(x2MI, x2I);
				Common.LDelete(x2MI);
				Common.LDelete(x1MI);
				Common.LDelete(tI);
			}
		} // QuadRoot

		public static void ChSZN(int zI) // change sign, n = - n
		{
			MultiZD.ChSZ(Common.nList[zI]);
		}

		public static bool ZTestN(int zI) // true if z is Zero
		{
			return MultiZD.ZTestZ(Common.nList[zI]);
		}

		public static void Matrix(int rs, int cs, MultiZ z)
			// z = zero matrix with rs rows and cs columns
		{
			MultiZ zz;
			MultiZD.Matrix(0, rs, cs, out zz);
			Calc.SetToNM(zz, z);
		} // Matrix

		public static void MatrixN(MultiZ rsF, MultiZ csF, int zI)
		// z = zero matrix with rs rows and cs columns
		{
			if (rsF.Rs != 0 || csF.Rs != 0)
				MultiZD.ErrNotScalar();
			else
			{
				int rs;
				int cs;
				MultiFD.Get1F(out rs, rsF);
				MultiFD.Get1F(out cs, csF);
				MatrixN(rs, cs, zI);
			}
		} // MatrixN

		public static void MatrixN(int rs, int cs, int zI)
		// z = zero matrix with rs rows and cs columns
		{
			MultiZ z;
			MultiZD.Matrix(0, rs, cs, out z);
			Calc.SetToN(z, zI);
		} // MatrixN

		public static void CharPoly(MultiZ x, MultiZ z)
		// z = Characteristic polynomial of square matrix or scalar x
		{
			int n = x.Rs;
			MultiZ zz;
			if (x.Cs != n)
				MultiZD.ErrNotSquare();
			else if (n == 0)					// if a scalar
			{
				if (x.I == MultiFD.fZero)	// if a real scalar
				{
					zz = new MultiZ(1, 2, 1);
					zz.E[0, 0] &= MultiFD.fOne;
					Calc.SetToNM(-x, zz.E[1, 0]);
				}
				else						// if an imaginary or complex scalar
				{
					zz = new MultiZ(1, 3, 1);
					zz.E[0, 0] &= MultiFD.fOne;
					XPCalLst.SetToNM(-((MultiF)x + x), zz.E[1, 0]);
					XPCalLst.SetToNM((MultiF)x * x + x.I * x.I, zz.E[2, 0]);
				}
				Calc.SetToNM(zz, z);
			}
			else
			{
				bool saveExp = MultiFD.expanded;
				int saveFMC = Calc.fMC;
				int saveMMax = MultiID.mMax;
				XPCalLst.Expand();
				int i, j, k, m;
				zz = new MultiZ(Calc.fMC, n + 1, 1);
				MultiZ s = new MultiZ(Calc.fMC);
				MultiZ t0 = new MultiZ(Calc.fMC);
				MultiZ b = new MultiZ(Calc.fMC, n, n);
				MultiZ c = new MultiZ(Calc.fMC, n, n);
				if (n % 2 != 0)
					zz.E[0, 0] &= MultiCD.cMOne;
				else
					zz.E[0, 0] &= MultiCD.cOne;
				for (m = 0; m < n; m++) 
				{
					if (m == 0) 
					{
						for (i = 0; i < n; i++)
							for (j = 0; j < n; j++)
								c.E[i, j] &= x.E[i, j];
					}
					else 
					{
						for (i = 0; i < n; i++) 
						{
							for (j = 0; j < n; j++) 
							{
								s &= MultiCD.cZero;
								for (k = 0; k < n; k++)
									s &= s + b.E[i, k] * x.E[k, j];
								c.E[i, j] &= s;
							}
						}
					}
					Trace(c, t0);
					Calc.SetToNM(t0 / (m + 1), t0);
					zz.E[m + 1, 0] &= -t0 * zz.E[0, 0];
					if (m < n - 1)
						for (i = 0; i < n; i++)
							for (j = 0; j < n; j++)
								if (j == i)
									b.E[i, j] &= c.E[i, j] - t0;
								else
									b.E[i, j] &= c.E[i, j];
				}
				if (n % 2 != 0)
					zz &= -zz;
				MultiFD.expanded = saveExp;
				Calc.fMC = saveFMC;
				MultiID.mMax = saveMMax;
				Calc.SetToNM(zz, z);
			}
		} // CharPoly

		public static void CharPolyN(MultiZ x, int zI)
		// z = Characteristic polynomial of square matrix or scalar x
		{
			if (x.Cs != x.Rs)
				MultiZD.ErrNotSquare();
			else
			{
				MultiZ z = new MultiZ();
				CharPoly(x, z);
				Calc.SetToN(z, zI);
			}
		} // CharPolyN

		public static void MinPoly(MultiZ x, MultiZ z)
		// z = Minimal polynomial of square matrix or scalar x
		{
			int n = x.Rs;
			if (n <= 1)
				CharPoly(x, z);
			else
			{
				bool saveExp = MultiFD.expanded;
				int saveFMC = Calc.fMC;
				int saveMMax = MultiID.mMax;
				XPCalLst.Expand();
				MultiZ p = new MultiZ(1, 1, 1);
				MultiZ r = new MultiZ(1, 1, 1);
				CharPoly(x, p);
				PolRoots(p, true, r);
				MultiFD.expanded = saveExp;
				Calc.fMC = saveFMC;
				MultiID.mMax = saveMMax;
				MultiZ po = new MultiZ(1, r.Rs, 1); // powers
				MultiZ rr = new MultiZ(1, r.Rs, 1); // reduced roots
				int i, j;
				bool equal = false;
				for (i = 0; i < r.Rs; i++)
				{
					po.E[i, 0] &= MultiCD.cOne;
					Calc.SetToNM(r.E[i, 0], rr.E[i, 0]);
					MultiCD.ClearSmall(rr.E[i, 0]);
				}
				XPCalLst.Expand();
				for (i = r.Rs - 1; i > 0; i--)
				{
					if (rr.E[i, 0] == rr.E[i - 1, 0])
					{
						equal = true;
						po.E[i - 1, 0] += po.E[i, 0];
						DelR(r, i, r);
						DelR(po, i, po);
					}
				}
				if (!equal || MultiID.muErr || MultiID.muAbort)
				{
					MultiFD.expanded = saveExp;
					Calc.fMC = saveFMC;
					MultiID.mMax = saveMMax;
					Calc.SetToNM(p, z);
					return;
				}
				MultiZD.DoneZ(ref rr);
				MultiZ m = new MultiZ(1, 1, 1);
				MultiZ mm = new MultiZ(1, 1, 1);
				Calc.SetToNM(x, m);
				MultiZ rank = new MultiZ();
				MultiZ nullity = new MultiZ();
				for (i = 0; i < po.Rs; i++)
				{
					if (po.E[i, 0] == MultiCD.cOne)
						continue;
					for (j = 0; j < n; j++)
						Calc.SetToNM(x.E[j, j] - r.E[i, 0], m.E[j, j]);
					Calc.SetToNM(m, mm);
					MultiF k = new MultiF();
					k.V[0] = 0;
					do
					{
						k++;
						if (k > MultiCD.cOne)
							MultiZD.MulZ(mm, m, mm);
						if (MultiID.muErr || MultiID.muAbort)
						{
							Calc.SetToNM(p, z);
							MultiFD.expanded = saveExp;
							Calc.fMC = saveFMC;
							MultiID.mMax = saveMMax;
							return;
						}
						Rank(mm, rank);
						nullity &= n - rank;
					}
					while (nullity < po.E[i, 0] && k < po.E[i, 0]);
					po.E[i, 0] &= k;
				}
				MultiF iF = new MultiF();
				MultiZ p1 = new MultiZ(1, 2, 1);
				p1.E[0, 0] &= MultiCD.cOne;
				p = new MultiZ(1, 1, 1);
				p.E[0, 0] &= MultiCD.cOne;
				for (i = 0; i < po.Rs; i++)
				{
					Calc.SetToNM(-r.E[i, 0], p1.E[1, 0]);
					for (iF &= MultiFD.fOne; iF <= (MultiF)po.E[i, 0]; iF++)
						PolMul(p1, p, p);
				}
				MultiFD.expanded = saveExp;
				Calc.fMC = saveFMC;
				MultiID.mMax = saveMMax;
				Calc.SetToNM(p, z);
			}
		} // MinPoly

		public static void MinPolyN(MultiZ x, int zI)
		// z = Minimal polynomial of square matrix or scalar x
		{
			if (x.Cs != x.Rs)
				MultiZD.ErrNotSquare();
			else
			{
				MultiZ z = new MultiZ();
				MinPoly(x, z);
				Calc.SetToN(z, zI);
			}
		} // MinPolyN

		public static void CompM(MultiZ p, MultiZ z)
		// z = companion matrix of polynomial p
		{
			if (p.Cs != 1)
				MultiZD.ErrNotColumn();
			else
			{
				PolMonic(p);
				int r;
				int n = p.Rs - 1;
				MultiZ M = new MultiZ(1, n, n);
				for (r = 0; r < n; r++)
					Calc.SetToNM(-p.E[n - r, 0], M.E[r, n - 1]);
				for (r = 1; r < n; r++)
					XPCalLst.SetToNM(MultiFD.fOne, M.E[r, r - 1]);
				Calc.SetToNM(M, z);
			}
		} // CompM

		public static void CompMN(MultiZ p, int zI)
		// z = companion matrix of polynomial p
		{
			if (p.Cs != 1)
				MultiZD.ErrNotColumn();
			else
			{
				MultiZ z = new MultiZ();
				CompM(p, z);
				Calc.SetToN(z, zI);
			}
		} // CompMN

		public static void PolMonic(MultiZ z)
		// z = Monic polynomial of z, leading coeff. = 1 or {0}
		{
			if (z.Cs != 1)
				MultiZD.ErrNotColumn();
			else
			{
				PolNorm(z);
				if (z.E[0, 0] == MultiCD.cZero || z.E[0, 0] == MultiCD.cOne)
					return;
				for (int r = 1; r < z.Rs; r++)
					Calc.SetToNM(z.E[r, 0] / z.E[0, 0], z.E[r, 0]);
				Calc.SetToNM(MultiCD.cOne, z.E[0, 0]);
			}
		} // PolMonic

		public static void PolMonicN(MultiZ x, int zI)
		// z = Monic polynomial of z, leading coeff. = 1 or {0}
		{
			if (x.Cs != 1)
				MultiZD.ErrNotColumn();
			else
			{
				PolMonic(x);
				Calc.SetToN(x, zI);
			}
		} // PolMonicN

		public static void PolNorm(MultiZ z)
		// z = norm of polynomial z, leading coeff. not zero or {0}
		{
			if (z.Cs != 1)
				MultiZD.ErrNotColumn();
			else
			{
				if (z.E[0, 0] != MultiCD.cZero)
					return;
				int nz = 0;
				int r;
				for (r = 0; r < z.Rs - 1; r++)
					if (z.E[r, 0] == MultiCD.cZero)
						nz++;
					else
						break;
				MultiZ zz = new MultiZ(1, z.Rs - nz, 1);
				for (r = nz; r < z.Rs; r++)
					Calc.SetToNM(z.E[r, 0], zz.E[r - nz, 0]);
				Calc.SetToNM(zz, z);
			}
		} // PolNorm

		public static void PolNormN(MultiZ x, int zI)
		// z = norm of polynomial x, leading coeff. not zero or {0}
		{
			if (x.Cs != 1)
				MultiZD.ErrNotColumn();
			else
			{
				PolNorm(x);
				Calc.SetToN(x, zI);
			}
		} // PolNormN

		public static void PolAdd(MultiZ u, MultiZ v, MultiZ z) // z = polynomial u + v
		{
			if (u.Cs != 1 || v.Cs != 1)
				MultiZD.ErrNotColumn();
			else
			{
				int m = u.Rs;
				int n = v.Rs;
				int p = Math.Max(m, n);
				MultiZ zz = new MultiZ(Calc.fMC, p, 1);
				int i, j, k;
				for (i = m - 1, j = n - 1, k = p - 1; i >= 0 && j >= 0; i--, j--, k--)
					Calc.SetToNM(u.E[i, 0] + v.E[j, 0], zz.E[k, 0]);
				if (m >= n)
					for (i = 0; i < m - n; i++)
						Calc.SetToNM(u.E[i, 0], zz.E[i, 0]);
				else
					for (i = 0; i < n - m; i++)
						Calc.SetToNM(v.E[i, 0], zz.E[i, 0]);
				PolNorm(zz);
				Calc.SetToNM(zz, z);
			}
		} // PolAdd

		public static void PolAddN(MultiZ u, MultiZ v, int zI) // z = polynomial u + v
		{
			if (u.Cs != 1 || v.Cs != 1)
				MultiZD.ErrNotColumn();
			else
			{
				MultiZ z = new MultiZ();
				PolAdd(u, v, z);
				Calc.SetToN(z, zI);
			}
		} // PolAddN

		public static void PolMod(MultiZ u, MultiZ v, MultiZ z)
		// if (div)  z = polynomial u/v, "Re" = remainder
		// if (!div) z = "Re" = remainder of u/v
		{
			PolDiv(u, v, false, z);
		} // PolMod

		public static void PolDisc(MultiZ p, MultiZ z) // z = discriminant of polynomial p
		{
			PolNorm(p);
			if (p.Cs != 1)
				MultiZD.ErrNotColumn();
			else if (p.Rs == 1)
				Calc.SetToNM(MultiCD.cZero, z);
			else if (p.Rs == 2)
				Calc.SetToNM(MultiCD.cOne, z);
			else
			{
				int r;
				int c;
				int n = p.Rs - 1;
				int m = n + n - 1;
				MultiZ M = new MultiZ(1, m, m);
				for (r = 0; r < n - 1; r++)
					for (c = r; c <= r + n; c++)
						Calc.SetToNM(p.E[c - r, 0], M.E[r, c]);
				for (r = n - 1; r < m; r++)
					for (c = r - n + 1; c <= r; c++)
						Calc.SetToNM((r - c + 1) * p.E[c - r + n - 1, 0], M.E[r, c]);
				Determinant(M, z);
				if (n/2 % 2 == 1)
					MultiCD.ChSC(z);
				bool saveExp = MultiFD.expanded;
				int saveFMC = Calc.fMC;
				int saveMMax = MultiID.mMax;
				XPCalLst.Expand();
				Calc.SetToNM(z / p.E[0, 0], z);
				MultiFD.expanded = saveExp;
				Calc.fMC = saveFMC;
				MultiID.mMax = saveMMax;
				Calc.NormNM(z);
			}
		} // PolDisc

		public static void PolDiscN(MultiZ p, int zI) // z = discriminant of polynomial p
		{
			if (p.Cs != 1)
				MultiZD.ErrNotColumn();
			else
			{
				MultiZ z = new MultiZ();
				PolDisc(p, z);
				Calc.SetToN(z, zI);
			}
		} // PolDiscN

		public static void PolDiv(MultiZ u, MultiZ v, bool div, MultiZ z)
		// if (div)  z = polynomial u/v, "Re" = remainder
		// if (!div) z = "Re" = remainder of u/v
		{
			if (u.Cs != 1 || v.Cs != 1)
				MultiZD.ErrNotColumn();
			else
			{

				int reI;
				Common.LFindAdd("Re", out reI);
				PolNorm(u);
				PolNorm(v);
				if (u.Rs < v.Rs)
				{
					Calc.SetToN(u, reI);
					if (div)
						Calc.SetToNM(MultiZD.vZero, z);
					else
						Calc.SetToNM(u, z);
				}
				else
				{
					bool saveExp = MultiFD.expanded;
					int saveFMC = Calc.fMC;
					int saveMMax = MultiID.mMax;
					XPCalLst.Expand();
					int m = u.Rs;
					int n = v.Rs;
					int p = m - n + 1;
					MultiZ q = new MultiZ(Calc.fMC, p, 1);
					MultiZ r = new MultiZ(Calc.fMC, m, 1);
					int k, j;
					for (j = 0; j < m; j++) 
						r.E[j, 0] &= u.E[j, 0];
					for (j = 0; j < p; j++) 
						q.E[j, 0] &= MultiCD.cZero;
					for (k = 0; k <= m - n; k++) 
					{
						q.E[k, 0] &= r.E[k, 0] / v.E[0, 0];
						if (MultiID.muErr || MultiID.muAbort)
						{
							MultiFD.expanded = saveExp;
							Calc.fMC = saveFMC;
							MultiID.mMax = saveMMax;
							return;
						}
						for (j = k + 1; j < n + k; j++)
							r.E[j, 0] -= q.E[k, 0] * v.E[j - k, 0];
					}
					for (j = 0; j < m - n + 1; j++)
						r.E[j, 0] &= MultiCD.cZero;
					PolNorm(r);
					MultiFD.expanded = saveExp;
					Calc.fMC = saveFMC;
					MultiID.mMax = saveMMax;
					Calc.SetToN(r, reI);
					if (div)
						Calc.SetToNM(q, z);
					else
						Calc.SetToNM(r, z);
				}
			}
		} // PolDiv

		public static void PolDivN(MultiZ u, MultiZ v, bool div, int zI)
		// if (div)  z = polynomial u/v, "Re" = remainder
		// if (!div) z = "Re" = remainder of u/v
		{
			if (u.Cs != 1 || v.Cs != 1)
				MultiZD.ErrNotColumn();
			else
			{
				MultiZ z = new MultiZ();
				PolDiv(u, v, div, z);
				Calc.SetToN(z, zI);
			}
		} // PolDivN

		public static void PolEval(MultiZ p, MultiC x, MultiC z)	// z = polynomial evaluated at x
		{
			if (p.Cs != 1)
				MultiZD.ErrNotColumn();
			else
			{
				bool saveExp = MultiFD.expanded;
				int saveFMC = Calc.fMC;
				int saveMMax = MultiID.mMax;
				XPCalLst.Expand();
				int n = p.Rs;
				MultiZ zz = new MultiZ(Calc.fMC);
				zz &= p.E[0, 0];
				for (int i = 1; i < n; i++)
					zz &= zz * x + p.E[i, 0];
				MultiFD.expanded = saveExp;
				Calc.fMC = saveFMC;
				MultiID.mMax = saveMMax;
				Calc.SetToNM(zz, z);
			}
		} // PolEval

		public static void PolEvalN(MultiZ p, MultiZ x, int zI) // z = polynomial evaluated at x
		{
			if (p.Cs != 1)
				MultiZD.ErrNotColumn();
			else if (x.Rs != 0)
				MultiZD.ErrNotScalar();
			else
			{
				MultiC z = new MultiC();
				PolEval(p, x, z);
				Calc.SetToN(z, zI);
			}
		} // PolEvalN

		public static void PolMul(MultiZ u, MultiZ v, MultiZ z)	// z = polynomial u*v
		{
			if ((u.Cs == 0 && v.Cs == 1) || (u.Cs == 1 && v.Cs == 0))
			{
				MultiZ zz = new MultiZ(Calc.fMC, Math.Max(u.Rs, v.Rs), 1);
				MultiZD.MulZ(u, v, zz);	// polynomial * a scalar
				Calc.SetToNM(zz, z);
				return;
			}
			if (u.Cs != 1 || v.Cs != 1)
				MultiZD.ErrNotColumn();
			else
			{
				bool saveExp = MultiFD.expanded;
				int saveFMC = Calc.fMC;
				int saveMMax = MultiID.mMax;
				XPCalLst.Expand();
				int m = u.Rs;
				int n = v.Rs;
				MultiZ zz = new MultiZ(Calc.fMC, n + m - 1, 1);
				for (int i = 0; i < m; i++)
					for (int j = 0; j < n; j++)
						zz.E[i + j, 0] += u.E[i, 0] * v.E[j, 0];
				MultiFD.expanded = saveExp;
				Calc.fMC = saveFMC;
				MultiID.mMax = saveMMax;
				Calc.SetToNM(zz, z);
			}
		} // PolMul

		public static void PolMulN(MultiZ u, MultiZ v, int zI)	// z = polynomial u*v
		{
			MultiZ z = new MultiZ();
			PolMul(u, v, z);
			Calc.SetToN(z, zI);
		} // PolMulN

		public static void PolRecip(MultiZ p, MultiZ z) // z = reciprocal polynomial of p
		// p is a polynomial, z = x^deg(p) * p(1/x), the coefficients are reversed
		// and the Complex Conjugate of the coefficients are taken
		{
			if (p.Cs != 1)
				MultiZD.ErrNotColumn();
			else
			{
				int n = p.Rs - 1;
				MultiZ zz = new MultiZ(1, p.Rs, 1);
				for (int r = 0; r <= n; r++)
				{
					Calc.SetToNM(p.E[r, 0], zz.E[n - r, 0]);
					MultiFD.ChSF(zz.E[n - r, 0].I);
				}
				Calc.SetToNM(zz, z);
			}
		} // PolRecip

		public static void PolRecipN(MultiZ p, int zI) // z = reciprocal polynomial of p
		// p is a polynomial, z = x^deg(p) * p(1/x), the coefficients are reversed
		// and the Complex Conjugate of the coefficients are taken
		{
			if (p.Cs != 1)
				MultiZD.ErrNotColumn();
			else
			{
				MultiZ z = new MultiZ();
				PolRecip(p, z);
				Calc.SetToN(z, zI);
			}
		} // PolRecipN

		public static void Laguer(MultiZ a, out int its, MultiC x, bool polish) // x = a root of a
		// The following routine implements the Laguerre method to find one root of a
		// given polynomial of degree m, whose coefficients can be complex. As usual,
		// the first coefficient a(1) is the constant term, while a(m+1) is the
		// coefficient of the highest power of x. The routine implements a simplified
		// version of an elegant stopping criterion due to Adams [5], which neatly
		// balances the desire to achieve full machine accuracy, on the one hand, with
		// the danger of iterating forever in the presence of roundoff error, on the
		// other.
			// Given the degree m and the complex coefficients a(1:m+1) of the
			// polynomial SUM{i=1, m+1}[a(i)*x^(i-1)], and given a complex value x,
			// this routine improves x by Laguerre's method until it converges, within
			// the achievable roundoff limit, to a root of the given polynomial. The
			// number of iterations taken is returned as its. Parameters: EPSS is the
			// estimated fractional roundoff error. We try to break(r are) limit
			// cycles with MR different fractional values, once every MT steps, for
			// MAXIT total allowed iterations.
		{
			its = 0;
			if (a.Cs != 1)
				MultiZD.ErrNotColumn();
			else
			{
				bool saveExp = MultiFD.expanded;
				int saveFMC = Calc.fMC;
				int saveMMax = MultiID.mMax;
				if (polish)
					XPCalLst.ExpandMore(Calc.fMC);
				int n = a.Rs; 
				int m = n - 1; 
				int mR = 8;
				int mT = 10;
				int iter, itToGo = 0, j;
				MultiF abb = new MultiF(Calc.fMC);
				MultiF abx = new MultiF(Calc.fMC);
				MultiF abp = new MultiF(Calc.fMC);
				MultiF abm = new MultiF(Calc.fMC);
				MultiF err = new MultiF(Calc.fMC);
				MultiC[] frac = new MultiC[mR];		// Fractions used to break limit cycle.
				for (j = 0; j < mR; j++)
					frac[j] = new MultiC();
				MultiC dx = new MultiC(Calc.fMC);
				MultiC xx = new MultiC(Calc.fMC);
				MultiC x1 = new MultiC(Calc.fMC);
				MultiC b = new MultiC(Calc.fMC);
				MultiC d = new MultiC(Calc.fMC);
				MultiC f = new MultiC(Calc.fMC);
				MultiC g = new MultiC(Calc.fMC);
				MultiC h = new MultiC(Calc.fMC);
				MultiC sq = new MultiC(Calc.fMC);
				MultiC gp = new MultiC(Calc.fMC);
				MultiC gm = new MultiC(Calc.fMC);
				MultiC g2 = new MultiC(Calc.fMC);
				MultiFD.ValueF10("0.50", out j, frac[0]);
				MultiFD.ValueF10("0.25", out j, frac[1]);
				MultiFD.ValueF10("0.75", out j, frac[2]);
				MultiFD.ValueF10("0.13", out j, frac[3]);
				MultiFD.ValueF10("0.38", out j, frac[4]);
				MultiFD.ValueF10("0.62", out j, frac[5]);
				MultiFD.ValueF10("0.88", out j, frac[6]);
				MultiFD.ValueF10("1.00", out j, frac[7]);
				xx &= x;
				int tI;
				Common.InitNAdd(1, "Laguer-tI", out tI);
				// Loop over iterations up to allowed max
				for (iter = 1; ; iter++)
				{
					if (MultiID.muErr || MultiID.muAbort)
						break;
					its = iter;
					b &= a.E[0, 0];
					err &= Abs(b);
					d &= MultiCD.cZero;
					f &= MultiCD.cZero;
					abx &= Abs(xx);
					// Efficient computation of the polynomial and its first
					for (j = 1; j < n; j++)
					{
						f &= xx * f + d;	// two derivatives. f stores P''/2.
						d &= xx * d + b;
						b &= xx * b + a.E[j, 0];
						err &= Abs(b) + abx * err;
					}
					err.C -= Calc.fMC - MultiFD.fGuard; // Estimate of roundoff error in evaluating polynomial
					abb &= Abs(b);
					 if (abb <= err && itToGo == 0)
						itToGo = 2 * iter;
					if (b == MultiCD.cZero || (abb <= err && iter >= itToGo))	// We are on the root
					{
						int qq = 0;		// for testing
						qq++;
						break;
					}
					else // The generic case: use Laguerres formula
					{
						g &= d / b;
						g2 &= g * g;
						h &= g2 - MultiCD.cTwo * f / b;
						MultiCD.SqRootC((m - 1) * (m * h - g2), sq);
						gp &= g + sq;
						gm &= g - sq;
						abp &= Abs(gp);
						abm &= Abs(gm);
						if(abp < abm)
							gp &= gm;
						XPCalLst.MaxN(abp, abm, tI);
						if (Common.nList[tI] > MultiCD.cZero)
							dx &= m / gp;
						else
						{
							XPCalLst.LnN(MultiFD.fOne + abx, tI);
							dx &= Common.nList[tI] + MultiCD.ci * iter;
							XCCalLst.ExpN(dx, tI);
							dx &= Common.nList[tI];
						}
					}
					x1 &= xx - dx;
					Math.DivRem(iter, mT, out j);
					if (j != 0)
						xx &= x1;
					else	// Every so often we take a fractional step, to break any
						// limit cycle (itself a rare occurrence)
					{
						Math.DivRem(-1 + iter / mT, mR, out j);
						xx &= xx - dx * frac[j];
					}
				}
				Common.LDelete(tI);
				MultiFD.expanded = saveExp;
				Calc.fMC = saveFMC;
				MultiID.mMax = saveMMax;
				Calc.SetToNM(xx, x);
			}
		} // Laguer

		public static void PolRoots(MultiZ a, bool polish, MultiZ roots) // roots = roots of a
		// a is a polynomial, roots is its roots
		// Adapted from http://www.library.cornell.edu/nr/bookfpdf/f9-5.pdf
		// Here is a driver routine that calls Laguer in succession for each root,
		// Performs the deflation, optionally polishes the roots by the same Laguerre
		// method -- if you are not going to polish in some other way -- and finally
		// sorts the roots by their real parts.
			// Given the degree m and the complex coefficients a(1:m+1) of the
			// polynomial SUM{i=1, m+1}[a(i)*x^(i-1)], this routine successively calls
			// Laguer and finds all m complex roots. The logical variable polish
			// should be input as .true. if polishing (also by Laguerres method) is
			// desired, .false. if the roots will be subsequently polished by other
			// means.
		{
			if (a.Cs != 1)
				MultiZD.ErrNotColumn();
			else
			{
				bool saveExp = MultiFD.expanded;
				int saveFMC = Calc.fMC;
				int saveMMax = MultiID.mMax;
				XPCalLst.Expand();
				if (saveExp)	// if already expanded, expand again so complex roots
					XPCalLst.ExpandMore(MultiFD.fGuard); // will be sorted correctly
				MultiZ aa = new MultiZ(Calc.fMC);
				Calc.SetToNM(a, aa);
				PolNorm(aa);
				int n = aa.Rs;			// size of polynomial
				int m = n - 1;			// degree of polynomial
				int MaxM = n;			// maximum anticipated value of m+1
				int i, j, jj, its;
				MultiZ ad = new MultiZ(Calc.fMC, MaxM, 1);
				MultiZ rr = new MultiZ(Calc.fMC, m, 1);
				MultiC x = new MultiC(Calc.fMC);
				MultiC b = new MultiC(Calc.fMC);
				MultiC c = new MultiC(Calc.fMC);
				Calc.SetToNM(aa, ad);
				for (j = 0; j < m; j++)					// Loop over each root to be found.
				{
					// Start at zero to favor convergence to smallest remaining root
					Calc.SetToNM(MultiCD.cZero, x);
					Laguer(ad, out its, x, false);		// Find the root
					MultiCD.ClearSmall(x);				// clear small real or imaginary part
					Calc.SetToNM(x, rr.E[j, 0]);		// save the root
					Calc.SetToNM(ad.E[0, 0], b);		// Forward deflation.
					for (jj = 1; jj <= m - j; jj++)
					{
						Calc.SetToNM(ad.E[jj, 0], c);
						Calc.SetToNM(b, ad.E[jj - 1, 0]);
						Calc.SetToNM(x * b + c, b);
					}
					ad.Rs--; // fix ad to have one less row
				}
				if (polish)
				{
					// Polish the roots using the undeflated coefficients
					for (j = 0; j < m; j++)
						Laguer(a, out its, rr.E[j, 0], polish);
					Common.LFind("Pi", out j);
					if (j != 0)
						j = XPCalLst.GetPiPtr();	// reduce precision of Pi
					Common.LFind("Ln10", out j);
					if (j != 0)
						j = XPCalLst.GetLn10Ptr();	// reduce precision of Ln10
				}
				MultiFD.expanded = saveExp;
				Calc.fMC = saveFMC;
				MultiID.mMax = saveMMax;
				Calc.SetToNM(rr, roots);
				for (j = 1; j < m; j++)	// Sort roots by straight insertion
				{
					Calc.SetToNM(roots.E[j, 0], x);
					for (i = j - 1; i >= 0; i--)
					{
						if (roots.E[i, 0] <= x)
							break;
						Calc.SetToNM(roots.E[i, 0], roots.E[i + 1, 0]);
					}
					Calc.SetToNM(x, roots.E[i + 1, 0]);
				}
			}
		} // PolRoots

		public static void PolRootsN(MultiZ p, bool polish, int zI) // z = roots of p
		{
			if (p.Cs != 1)
				MultiZD.ErrNotColumn();
			else
			{
				MultiZ z = new MultiZ();
				PolRoots(p, polish, z);
				Calc.SetToN(z, zI);
			}
		} // PolRootsN

		public static void PolSturm(MultiZ p, MultiZ z) // z = number of real roots in p
		// p is a polynomial with real coeffs. should be square free
		{
			if (p.Cs != 1)
				MultiZD.ErrNotColumn();
			else
			{
				PolMonic(p);
				MultiZ a = new MultiZ(Calc.fMC);
				a &= MultiZD.mOne;
				for (int r = 1; r < p.Rs; r++)
					a &= a + Abs(p.E[r, 0]);
				Calc.NormNM(a);
				PolSturmI(p, -a, a, z);
			}
		} // PolSturm

		public static void PolSturmN(MultiZ p, int zI) // z = number of real roots in p
		// p is a polynomial with real coeffs. should be square free
		{
			if (p.Cs != 1)
				MultiZD.ErrNotColumn();
			else
			{
				MultiZ z = new MultiZ();
				PolSturm(p, z);
				Calc.SetToN(z, zI);
			}
		} // PolSturmN

		public static void PolSturmI(MultiZ p, MultiZ a, MultiZ b, MultiZ z) // z = number of real roots in p
		// in interval (a, b). roots at a or b may or may not be counted
		// p is a polynomial with real coeffs. should be square free
		{
			if (p.Cs != 1)
				MultiZD.ErrNotColumn();
			else if (a.Rs != 0 || b.Rs != 0)
				MultiZD.ErrNotScalar();
			else
			{
				int r;
				for (r = 0; r < p.Rs; r++)
					if (p.E[r, 0].I != MultiFD.fZero)
					{
						MultiID.MuWriteErr("PolSturm: Polynomial not real");
						return;
					}
				if (a.I != MultiFD.fZero || b.I != MultiFD.fZero)
				{
					MultiID.MuWriteErr("PolSturm: Interval not real");
					return;
				}
				PolMonic(p);
				if (p.Rs == 1)
					z &= MultiZD.mZero;
				else
				{
					int d = 0;
					MultiZ f0 = new MultiZ(1, p.Rs, 1);
					MultiZ f1 = new MultiZ(1, p.Rs - 1, 1);
					MultiZ f2 = new MultiZ(1, p.Rs - 2, 1);
					MultiZ v0 = new MultiZ(2);
					MultiZ v1 = new MultiZ(2);
					Calc.SetToNM(p, f0);
					for (r = 0; r < p.Rs - 1; r++)
						Calc.SetToNM((p.Rs - 1 - r) * f0.E[r, 0], f1.E[r, 0]);
					PolEval(f0, a, v0);
					if(v0 == MultiFD.fZero)
						MultiID.WriteLn("PolSturmI: PolEval(p, a) = 0.0");
					PolEval(f1, a, v1);
					if (v0.S != v1.S)
						d++;
					while (f1.Rs > 1)
					{
						PolMod(-f0, f1, f2);
						Calc.SetToNM(f1, f0);
						Calc.SetToNM(v1, v0);
						Calc.SetToNM(f2, f1);
						PolEval(f1, a, v1);
						if (v0.S != v1.S)
							d++;			
					}
					if (f1.E[0, 0] == MultiCD.cZero)
					{
						MultiID.MuWriteErr("PolSturm failed: Polynomial is not square free");
						return;
					}
					Calc.SetToNM(p, f0);
					f1 = new MultiZ(1, p.Rs - 1, 1);
					for (r = 0; r < p.Rs - 1; r++)
						 Calc.SetToNM((p.Rs - 1 - r) * f0.E[r, 0], f1.E[r, 0]);
					PolEval(f0, b, v0);
					if(v0 == MultiFD.fZero)
						MultiID.WriteLn("PolSturmI: PolEval(p, b) = 0.0");
					PolEval(f1, b, v1);
					if (v0.S != v1.S)
						d--;
					while (f1.Rs > 1)
					{
						PolMod(-f0, f1, f2);
						Calc.SetToNM(f1, f0);
						Calc.SetToNM(v1, v0);
						Calc.SetToNM(f2, f1);
						PolEval(f1, b, v1);
						if (v0.S != v1.S)
							d--;			
					}
					z &= Math.Abs(d);
				}
				Calc.NormNM(z);
			}
		} // PolSturmI

		public static void PolSturmIN(MultiZ p, MultiZ a, MultiZ b, int zI) // z = number of real roots in p
		// in interval (a, b). roots at a or b may or may not be counted
		// p is a polynomial with real coeffs. should be square free
		{
			if (p.Cs != 1)
				MultiZD.ErrNotColumn();
			else if (a.Rs != 0 || b.Rs != 0)
				MultiZD.ErrNotScalar();
			else
			{
				MultiZ z = new MultiZ();
				PolSturmI(p, a, b, z);
				Calc.SetToN(z, zI);
			}
		} // PolSturmIN

		public static void SetAll(MultiZ m, MultiZ x, int zI)	// z = m, z[r, c] = x for all z[r, c]
		{
			if (x.Rs != 0)
				MultiZD.ErrNotScalar();
			else if (m.Rs == 0)
				MultiZD.ErrElementOfScalar();
			else
			{
				for (int r = 0; r < m.Rs; r++)
					for (int c = 0; c < m.Cs; c++)
					{
						Calc.SetToNM(x, m.E[r, c]);
					}
				Calc.SetToN(m, zI);
			}
		} // SetAll

		public static void SetElement(MultiZ m, MultiZ rF, MultiZ cF, MultiZ x, int zI)	// z = m, z[r, c] = x
		{
			if (rF.Rs != 0 || cF.Rs != 0 || x.Rs != 0)
				MultiZD.ErrNotScalar();
			else if (m.Rs == 0)
				MultiZD.ErrElementOfScalar();
			else
			{
				int r;
				int c;
				MultiFD.Get1F(out r, rF);
				MultiFD.Get1F(out c, cF);
				r -= uFirst;
				c -= uFirst;
				Calc.SetToN(m, zI);
				if (r >= 0 && c >= 0 && r < Common.nList[zI].Rs && c < Common.nList[zI].Cs)
					Calc.SetToNM(x, Common.nList[zI].E[r, c]);
				else
					MultiZD.ErrOutside();
			}
		} // SetElement

		public static void GetElement(MultiZ x, MultiZ rF, MultiZ cF, int zI)	// z = x[r, c]
		{
			if (rF.Rs != 0 || cF.Rs != 0)
				MultiZD.ErrNotScalar();
			else
			{
				int r;
				int c;
				MultiFD.Get1F(out r, rF);
				MultiFD.Get1F(out c, cF);
				GetElement(x, r, c, zI);
			}
		} // GetElement

		public static void GetElement(MultiZ x, int ri, int ci, int zI)	// z = x[r, c]
		{
			if (x.Rs == 0)
				MultiZD.ErrNotMatrix();
			else
			{
				int r = ri - uFirst;
				int c = ci - uFirst;
				if (r >= 0 && c >= 0 && r < x.Rs && c < x.Cs)
				{
					MultiZ z = new MultiZ();
					Calc.SetToNM(x.E[r, c], z);
					Calc.SetToN(z, zI);
				}
				else
					MultiZD.ErrOutside();
			}
		} // GetElement

		public static MultiZ Transpose(MultiZ x)	// return x'
		{
			if (x.Rs == 0)
			{
				MultiZD.ErrNotMatrix();
				return MultiZD.mZero;
			}
			else
			{
				MultiZ z = new MultiZ(0, x.Cs, x.Rs);
				for (int r = 0; r < x.Rs; r++)
					for (int c = 0; c < x.Cs; c++)
					{
						Calc.SetToNM(x.E[r, c], z.E[c, r]);
					}
				return z;
			}
		} // Transpose

		public static void TransposeN(int zI)	// z = z'
		{
			Calc.SetToN(Transpose(Common.nList[zI]), zI);
		} // TransposeN

		public static MultiZ ConjTran(MultiZ x)	// return x'
		{
			if (x.Rs == 0)
			{
				MultiZD.ErrNotMatrix();
				return MultiZD.mZero;
			}
			else
			{
				MultiZ z = new MultiZ(0, x.Cs, x.Rs);
				for (int r = 0; r < x.Rs; r++)
					for (int c = 0; c < x.Cs; c++)
					{
						Calc.SetToNM(x.E[r, c], z.E[c, r]);
						MultiFD.ChSF(z.E[c, r].I);
					}
				return z;
			}
		} // ConjTran

		public static void ConjTranN(int zI)	// z = z'
		{
			Calc.SetToN(ConjTran(Common.nList[zI]), zI);
		} // ConjTranN

		public static MultiZ Adjoint(MultiZ x)	// return Adjoint matrix of square matrix x
		{
			if (x.Rs == 0 || x.Rs != x.Cs)
			{
				MultiZD.ErrNotSquare();
				return MultiZD.mZero;
			}
			else if (x.Rs == 1)
			{
				MultiZ z = new MultiZ(1, 1, 1); // 1 by 1 identity matrix
				z.E[0, 0] &= MultiCD.cOne;
				return z;
			}
			else
			{
				int n = x.Rs;
				int nm = n - 1;
				int mI;
				Common.InitNAdd(1, "Adjoint-mI", out mI);
				MultiZ z = new MultiZ(0, n, n);
				for (int h = 0; h < n; h++)
					for (int j = 0; j < n; j++)
					{
						MatrixN(nm, nm, mI);
						for (int r = 0; r < nm; r++)
							for (int c = 0; c < nm; c++)
							{
								int r2 = r;
								int c2 = c;
								if (r2 >= h) r2++;
								if (c2 >= j) c2++;
								Calc.SetToNM(x.E[r2, c2], Common.nList[mI].E[r, c]);
							}
						DeterminantN(Common.nList[mI], mI);
						Calc.SetToNM(Common.nList[mI], z.E[j, h]);
						int odd;
						Math.DivRem(h + j, 2, out odd);
						if (odd == 1)
							MultiCD.ChSC(z.E[j, h]);
					}
				return z;
			}
		} // Adjoint

		public static void AdjointN(int zI)	// z = Adjoint matrix of square matrix z
		{
			Calc.SetToN(Adjoint(Common.nList[zI]), zI);
		} // AdjointN

		public static MultiZ Balance(MultiZ x)	// return a balanced matrix with identical eigenvalues as x
		// Given a matrix x[1..n][1..n], this routine replaces it by a balanced matrix with identical
		// eigenvalues. A symmetric matrix is already balanced and is unaffected by this procedure. The
		// parameter radix should be the machine's floating-point radix.
		{
			if (x.Rs == 0 || x.Rs != x.Cs)
			{
				MultiZD.ErrNotSquare();
				return MultiZD.mZero;
			}
			else if (x.Rs == 1)
			{
				MultiZ z = new MultiZ(1, 1, 1); // 1 by 1 matrix
				Calc.SetToNM(x, z);
				return z;
			}
			else
			{
				bool saveExp = MultiFD.expanded;
				int saveFMC = Calc.fMC;
				int saveMMax = MultiID.mMax;
				XPCalLst.Expand();
				int n = x.Rs;
				int i, j, last = 0;
				MultiZ z = new MultiZ(Calc.fMC, n, n);
				MultiC s = new MultiC(Calc.fMC);
				MultiC r = new MultiC(Calc.fMC);
				MultiC g = new MultiC(Calc.fMC);
				MultiC f = new MultiC(Calc.fMC);
				MultiC c = new MultiC(Calc.fMC);
				MultiC radix = new MultiC(1);
				MultiC sqrdx = new MultiC(1);
				MultiCD.SetTo1C(10, 0, radix);
				sqrdx &= radix * radix;
				MultiC cP95 = new MultiC(1);
				MultiFD.ValueF10("0.95", out i, cP95);
				z &= x;
				while (last == 0) 
				{
					last = 1;
					for (i = 0; i < n; i++)	// Calculate row and column norms.
					{
						r &= c &= MultiCD.cZero;
						for (j = 0; j < n; j++)
							if (j != i)
							{
								c += Abs(z.E[j, i]);
								r += Abs(z.E[i, j]);
							}
						if (c != MultiCD.cZero && r != MultiCD.cZero)	// if both are nonzero
						{
							g &= r / radix;
							f &= MultiCD.cOne;
							s &= c + r;
							while (c < g)	// find the integer power of the machine radix that
							{				// comes closest to balancing the matrix.
								f *= radix;
								c *= sqrdx;
							}
							g &= r * radix;
							while (c > g) 
							{
								f /= radix;
								c /= sqrdx;
							}
							if ((c + r) / f < cP95 * s) 
							{
								last = 0;
								g &= MultiCD.cOne / f;
								for (j = 0; j < n; j++) z.E[i, j] *= g;	// Apply similarity transformation.
								for (j = 0; j < n; j++) z.E[j, i] *= f;
							}
						}
					}
				}
				MultiFD.expanded = saveExp;
				Calc.fMC = saveFMC;
				MultiID.mMax = saveMMax;
				Calc.NormNM(z);
				return z;
			}
		} // Balance

		public static void BalanceN(int zI)	// replace z by a balanced matrix with identical eigenvalues
		{
			Calc.SetToN(Balance(Common.nList[zI]), zI);
		} // BalanceN

		public static void SwapC(ref MultiC x, ref MultiC y)	// interchange x and y (pointers only)
		{
			MultiC t;
			t = x;
			x = y;
			y = t;
		} // SwapC

		public static MultiZ Hess(MultiZ a)	// return upper Hessenberg matrix with identical eigenvalues as a
		// Reduction to Hessenberg form by the elimination method
		// Recommended, but not required, is that this routine be preceded by Balance
		{
			if (a.Rs == 0 || a.Rs != a.Cs)
			{
				MultiZD.ErrNotSquare();
				return MultiZD.mZero;
			}
			else if (a.Rs <= 2)
			{
				MultiZ z = new MultiZ(); // 1 by 1 or 2 by 2 matrix
				Calc.SetToNM(a, z);
				return z;
			}
			else
			{
				bool saveExp = MultiFD.expanded;
				int saveFMC = Calc.fMC;
				int saveMMax = MultiID.mMax;
				XPCalLst.Expand();
				int n = a.Rs;
				int m, j, i;
				MultiZ z = new MultiZ(Calc.fMC, n, n);
				MultiC y = new MultiC(Calc.fMC);
				MultiC x = new MultiC(Calc.fMC);
				z &= a;
				for (m = 1; m < n - 1; m++)	// m is called r + 1 in the text.
				{
					x &= MultiCD.cZero;
					i = m;
					for (j = m; j < n; j++)	// Find the pivot.
					{
						if (Abs(z.E[j, m - 1]) > Abs(x))
						{
							x &= z.E[j, m - 1];
							i = j;
						}
					}
					if (i != m)	// Interchange rows and columns.
					{
						for (j = m - 1; j < n; j++) SwapC(ref z.E[i, j], ref z.E[m, j]);
						for (j = 0; j < n; j++) SwapC(ref z.E[j, i], ref z.E[j, m]);
					}
					if (x != MultiCD.cZero)	// Carry out the elimination.
					{
						for (i = m + 1; i < n; i++) 
						{
							if ((y &= z.E[i, m - 1]) != MultiCD.cZero) 
							{
								y /= x;
								z.E[i, m - 1] &= y;
								for (j = m; j < n; j++)
									z.E[i, j] -= y * z.E[m, j];
								for (j = 0; j < n; j++)
									z.E[j, m] += y * z.E[j, i];
							}
						}
					}
				}
				for (j = 0; j < n - 2; j++)
					for (i = j + 2; i < n; i++)
						z.E[i, j] &= MultiCD.cZero;
				MultiFD.expanded = saveExp;
				Calc.fMC = saveFMC;
				MultiID.mMax = saveMMax;
				Calc.NormNM(z);
				return z;
			}
		} // Hess

		public static void HessN(int zI)	// replace z by a balanced matrix with identical eigenvalues
		{
			Calc.SetToN(Hess(Common.nList[zI]), zI);
		} // HessN

		public static void Deriv(MultiZ p, MultiZ z)	// z = derivative of polynomial p
		{
			if (p.Cs != 1)
				MultiZD.ErrNotColumn();
			else
			{
				int n = p.Rs;
				MultiZ zz = new MultiZ(1, p.Rs - 1, 1);
				for (int r = 0; r < n - 1; r++)
					Calc.SetToNM((p.Rs - 1 - r) * p.E[r, 0], zz.E[r, 0]);
				Calc.SetToNM(zz, z);
			}
		} // Deriv

		public static void DerivN(MultiZ p, int zI)	// z = derivative of polynomial p
		{
			if (p.Cs != 1)
				MultiZD.ErrNotColumn();
			else
			{
				MultiZ z = new MultiZ();
				Deriv(p, z);
				Calc.SetToN(z, zI);
			}
		} // DerivN

		public static void Integ(MultiZ p, MultiZ z)	// z = formal integral of polynomial p
		{
			if (p.Cs != 1)
				MultiZD.ErrNotColumn();
			else
			{
				int n = p.Rs;
				MultiZ zz = new MultiZ(1, p.Rs + 1, 1);
				for (int r = 0; r < n; r++)
					Calc.SetToNM(p.E[r, 0] / (n - r), zz.E[r, 0]);
				Calc.SetToNM(zz, z);
			}
		} // Integ

		public static void IntegN(MultiZ p, int zI)	// z = formal integral of polynomial p
		{
			if (p.Cs != 1)
				MultiZD.ErrNotColumn();
			else
			{
				MultiZ z = new MultiZ();
				Integ(p, z);
				Calc.SetToN(z, zI);
			}
		} // IntegN

		public static void Determinant(MultiZ x, MultiZ z)	// z = |x|
		{
			int rs = x.Rs;
			int cs = x.Cs;
			if (rs == 0 || rs != cs)
				MultiZD.ErrNotSquare();
			else if (rs == 1)
			{
				Calc.SetToNM(x.E[0, 0], z);
			}
			else
			{
				bool saveExp = MultiFD.expanded;
				int saveFMC = Calc.fMC;
				int saveMMax = MultiID.mMax;
				XPCalLst.Expand();
				MultiC d = new MultiC(Calc.fMC);
				if (rs == 2)
				{
					d &= x.E[0, 0] * x.E[1, 1] - x.E[0, 1] * x.E[1, 0];
				}
				else
				{
					int n = rs;
					MultiZ a = new MultiZ(Calc.fMC, n, n);
					int[] index;
					index = new int[n];
					a &= x;
					bool sing;
					MultiZD.LUdecomp(a, index, d, out sing); // decompose the matrix
					if (sing)
						d &= MultiCD.cZero;
					else
						for (int i = 0; i < n; i++)
							d *= a.E[i, i];
				}
				MultiFD.expanded = saveExp;
				Calc.fMC = saveFMC;
				MultiID.mMax = saveMMax;
				Calc.SetToNM(d, z);
			}
		} // Determinant

		public static void DeterminantN(MultiZ x, int zI)	// z = |x|
		{
			int rs = x.Rs;
			int cs = x.Cs;
			if (rs == 0 || rs != cs)
				MultiZD.ErrNotSquare();
			else
			{
				MultiZ z = new MultiZ();
				Determinant(x, z);
				Calc.SetToN(z, zI);
			}
		} // DeterminantN

		public static void EigenVa(MultiZ x, MultiZ z)	// z = eigenvalues of x
		{
			int n = x.Rs;
			if (n == 0 || n != x.Cs)
				MultiZD.ErrNotSquare();
			else
			{
				bool saveExp = MultiFD.expanded;
				int saveFMC = Calc.fMC;
				int saveMMax = MultiID.mMax;
				XPCalLst.Expand();
				MultiZ p = new MultiZ(1, 1, 1);
				MultiZ r = new MultiZ(1, 1, 1);
				CharPoly(x, p);
				PolRoots(p, true, r);
				MultiFD.expanded = saveExp;
				Calc.fMC = saveFMC;
				MultiID.mMax = saveMMax;
				Calc.SetToNM(r, z);
				if (!saveExp)
				{
					int i, j;
					for (j = 1; j < z.Rs; j++)	// resort roots by straight insertion
					{
						Calc.SetToNM(z.E[j, 0], x);
						for (i = j - 1; i >= 0; i--)
						{
							if (z.E[i, 0] <= x)
								break;
							Calc.SetToNM(z.E[i, 0], z.E[i + 1, 0]);
						}
						Calc.SetToNM(x, z.E[i + 1, 0]);
					}
				}
			}
		} // EigenVa

		public static void EigenVaN(MultiZ x, int zI)	// z = eigenvalues of x
		{
			int rs = x.Rs;
			int cs = x.Cs;
			if (rs == 0 || rs != cs)
				MultiZD.ErrNotSquare();
			else
			{
				MultiZ z = new MultiZ();
				EigenVa(x, z);
				Calc.SetToN(z, zI);
			}
		} // EigenVaN

		public static void EigenVe(MultiZ x, MultiZ z)	// z = eigenvectors of x
		{
			int n = x.Rs;
			if (n == 0 || n != x.Cs)
				MultiZD.ErrNotSquare();
			else
			{
				int ri, ci;
				bool saveExp = MultiFD.expanded;
				int saveFMC = Calc.fMC;
				int saveMMax = MultiID.mMax;
				XPCalLst.Expand();
				MultiZ p = new MultiZ(1, 1, 1);
				MultiZ r = new MultiZ(1, 1, 1);
				MultiZ h = new MultiZ(1, n, 1);
				MultiZ H = new MultiZ(1, n, n);
				MultiZ P = new MultiZ(1, n, n);
				MultiZ v = new MultiZ(1, n, n);
				MultiC d = new MultiC();
				CharPoly(x, p);
				PolRoots(p, true, r);
				int hI;
				Common.InitNAdd(1, "EigenVe-hI", out hI);
				MatrixN(n, 1, hI);
				Calc.RRanMN(hI);
				Calc.SetToNM(Common.nList[hI], h);
				Common.LDelete(hI);
				for (ci = 0; ci < n; ci++)
				{
					if (ci != 0)
						Calc.SetToNM(x * h, h);
					for (ri = 0; ri < n; ri++)
						Calc.SetToNM(h.E[ri, 0], H.E[ri, ci]);
				}
				for (ri = 0; ri < n; ri++)
				{
					p.Rs--;
					for (ci = 0; ci < n; ci++)
						PolEval(p, r.E[ci, 0], P.E[ri, ci]);
				}
				Calc.SetToNM(H * P, v);
				for (ci = 0; ci < n; ci++)
				{
					Calc.SetToNM(MultiCD.cOne, d);
					for (ri = n - 1; ri >= 0; ri--)
						if (v.E[ri, ci] != MultiCD.cZero)
						{
							Calc.SetToNM(v.E[ri, ci], d);
							break;
						}
					for (ri = 0; ri < n; ri++)
						Calc.SetToNM(v.E[ri, ci] / d, v.E[ri, ci]);
				}
				MultiFD.expanded = saveExp;
				Calc.fMC = saveFMC;
				MultiID.mMax = saveMMax;
				Calc.NormNM(v);
				for (ci = n - 2; ci >= 0; ci--)
				{
					int m = 0;
					for (ri = 0; ri < n; ri++)
					{
						if (v.E[ri, ci] == v.E[ri, ci + 1])
							m++;
						else
							break;
					}
					if (m == n)
						Del(v, ci, v);
				}
				bool err = false;
				for (ci = v.Cs - 1; ci >= 0; ci--)
				{
					int m = 0;
					for (ri = 0; ri < n; ri++)
					{
						if (v.E[ri, ci] == MultiCD.cZero)
							m++;
						else
							break;
					}
					if (m == n)
					{
						Del(v, ci, v);
						err = true;
					}
				}
				Calc.SetToNM(v, z);
				if (err)
					MultiID.WriteLn("Not all eigenvectors found");
			}
		} // EigenVe

		public static void EigenVeN(MultiZ x, int zI)	// z = eigenvectors of x
		{
			int rs = x.Rs;
			int cs = x.Cs;
			if (rs == 0 || rs != cs)
				MultiZD.ErrNotSquare();
			else
			{
				MultiZ z = new MultiZ();
				EigenVe(x, z);
				Calc.SetToN(z, zI);
			}
		} // EigenVeN

		public static void EigenVe2(MultiZ x, MultiZ z)	// z = eigenvectors of square matrix x by SVD
		{
			int m = x.Rs;
			int n = x.Cs;
			if (m == 0 || m != n)
				MultiZD.ErrNotSquare();
			else
			{
				bool saveExp = MultiFD.expanded;
				int saveFMC = Calc.fMC;
				int saveMMax = MultiID.mMax;
				XPCalLst.Expand();
				MultiZ p = new MultiZ(1, 1, 1);
				MultiZ r = new MultiZ(1, 1, 1);
				MultiZ zz = new MultiZ(1, n, n);
				CharPoly(x, p);						// p = characteristic polynomial of square matrix x
				MultiFD.expanded = saveExp;
				Calc.fMC = saveFMC;
				MultiID.mMax = saveMMax;
				PolRoots(p, true, r);				// r = All of the roots of polynomial p
				int i, j, ri, ci, rm, f = 0;
				for (i = n - 2; i >= 0; i--)		// delete roots that are duplicates
					if (r.E[i, 0] == r.E[i + 1, 0])
						DelR(r, i, r);
				XPCalLst.Expand();
				rm = r.Rs;
				MultiZ u;
				MultiZ w;
				MultiZ v;
				MultiC di = new MultiC();
				for (i = 0; i < rm; i++)
				{
					u = new MultiZ(Calc.fMC, n, n);
					w = new MultiZ(Calc.fMC, 1, n);
					v = new MultiZ(Calc.fMC, n, n);
					Calc.SetToNM(x, u);
					for (j = 0; j < n; j++)				// subtract i-th root from all diagonal elements of u
						Calc.SetToNM(u.E[j, j] - r.E[i, 0], u.E[j, j]);
					SVD(u, w, v);						// Decompose u to (u, w, v)
					MultiFD.expanded = saveExp;
					Calc.fMC = saveFMC;
					MultiID.mMax = saveMMax;
					Calc.NormNM(w);
					XPCalLst.Expand();
					for (j = n - 1; j >= 0; j--)		// Delete columns of v for w != 0
					{
						if (w.E[0, j] != MultiFD.fZero)	// w is real
							Del(v, j, v);
					}
					for (ci = 0; ci < v.Cs; ci++)		// concat columns of v to zz	
					{
						if (f < zz.Cs)	// for safety
						{
							for (ri = 0; ri < n; ri++)
							{
								Calc.SetToNM(v.E[ri, ci], zz.E[ri, f]);
							}
							f++;
						}
						else
							break;
					}
				}
				Calc.fMC -= MultiFD.fGuard;
				Calc.NormNMCol(zz);
				Calc.fMC += MultiFD.fGuard;
				for (ci = n - 1; ci >= 0; ci--)			// for all columns of zz
				{
					Calc.SetToNM(MultiCD.cZero, di);
					for (ri = n - 1; ri >= 0; ri--)		// find last non-zero element in this column
						if (zz.E[ri, ci] != MultiCD.cZero)
						{
							Calc.SetToNM(zz.E[ri, ci], di);
							break;
						}
					if (di == MultiCD.cZero)			// if column is all zeros
					{
						Del(zz, ci, zz);				// Delete columns of zz with all zeros
					}
					else
					{
						for (ri = 0; ri < n; ri++)		// Make last non-zero element equal to one
							Calc.SetToNM(zz.E[ri, ci] / di, zz.E[ri, ci]);
					}
				}
				MultiFD.expanded = saveExp;
				Calc.fMC = saveFMC;
				MultiID.mMax = saveMMax;
				Calc.NormNM(zz);
				bool err = false;
				for (ci = zz.Cs - 1; ci > 0; ci--)
				{
					int d = 0;
					for (ri = 0; ri < n; ri++)
					{
						if (zz.E[ri, ci] == zz.E[ri, ci - 1])
							d++;
						else
							break;
					}
					if (d == n)
					{
						Del(zz, ci, zz);
						err = true;
					}
				}
				Calc.SetToNM(zz, z);
				if (err)
					MultiID.WriteLn("Not all eigenvectors found");
			}
		} // EigenVe2

		public static void EigenVe2N(MultiZ x, int zI)	// z = eigenvectors of x
		{
			int rs = x.Rs;
			int cs = x.Cs;
			if (rs == 0 || rs != cs)
				MultiZD.ErrNotSquare();
			else
			{
				MultiZ z = new MultiZ();
				EigenVe2(x, z);
				Calc.SetToN(z, zI);
			}
		} // EigenVe2N

		public static void NotN(MultiZ x, int zI)	// z = !x
		{
			if (x.Rs == 0)
				XCCalLst.NotN(x, zI);
			else
			{
				int tI;
				Common.InitNAdd(1, "NotN-tI", out tI);
				if (Common.nList[zI].Rs != x.Rs || Common.nList[zI].Cs != x.Cs)
					MatrixN(x.Rs, x.Cs, zI);
				for (int r = 0; r < x.Rs; r++)
					for (int c = 0; c < x.Cs; c++)
					{
						XCCalLst.NotN(x.E[r, c], tI);
						if (MultiID.muErr)
						{
							Common.LDelete(tI);
							return;
						}
						Calc.SetToNM(Common.nList[tI], Common.nList[zI].E[r, c]);
					}
				Common.LDelete(tI);
			}		
		} // NotN

		public static void RNotN(int zI) // z = !z
		{
			NotN(Common.nList[zI], zI);
		}

		public static void Primorial(int zI) // change sign, z = - z
		{
			if (Common.nList[zI].Rs == 0)
			{
				MultiFD.ClearF(Common.nList[zI].I);
				XPCalLst.RIntSIN(zI);
				XICommon.Primorial(zI);
				Calc.NormN(zI);
			}
			else
			{
				int tI;
				Common.InitNAdd(1, "Primorial-tI", out tI);
				for (int r = 0; r < Common.nList[zI].Rs; r++)
					for (int c = 0; c < Common.nList[zI].Cs; c++)
					{
						Calc.SetToN(Common.nList[zI].E[r, c], tI);
						XPCalLst.RIntSIN(tI);
						XICommon.Primorial(tI);
						Calc.SetToNM(Common.nList[tI], Common.nList[zI].E[r, c]);
						XPCalLst.NormNM(Common.nList[zI].E[r, c]);
					}
				Common.LDelete(tI);
			}
		} // Primorial

		public static void BinoS(int xI, int zI)
		// z = Binomial coefficient (x, z) slow method
		// a, b are not modified unless they share memory with z
		// will write to standard output if diagOn is true		
		{
			int xrs = Common.nList[xI].Rs;
			int zrs = Common.nList[zI].Rs;
			int xcs = Common.nList[xI].Cs;
			int zcs = Common.nList[zI].Cs;
			if (xrs == 0 && zrs == 0)
			{
				XPCalLst.RIntSIN(zI);
				XPCalLst.RIntSIN(xI);
				XICommon.BinoS(Common.nList[zI], Common.nList[xI], Common.nList[zI]);
				Calc.NormN(zI);
			}
			else if (xrs != zrs || xcs != zcs)
				MultiZD.ErrIncompatible();
			else
			{
				int tI, t2I;
				Common.InitNAdd(1, "BinoS-tI", out tI);
				Common.InitNAdd(1, "BinoS-t2I", out t2I);
				for (int r = 0; r < zrs; r++)
					for (int c = 0; c < zcs; c++)
					{
						Calc.SetToN(Common.nList[zI].E[r, c], tI);
						Calc.SetToN(Common.nList[xI].E[r, c], t2I);
						XPCalLst.RIntSIN(tI);
						XPCalLst.RIntSIN(t2I);
						Common.nList[zI].E[r, c].C = 0;
						XICommon.BinoS(Common.nList[tI], Common.nList[t2I], Common.nList[zI].E[r, c]);
						XPCalLst.NormNM(Common.nList[zI].E[r, c]);
					}
				Common.LDelete(t2I);
				Common.LDelete(tI);
			}
		} // BinoS

		public static bool swapped;		// true if SortC or SortR swapped elements
		public static int uFirst;		// the user index of the first element in a matrix
	}
}
